package Server;

use strict;
use warnings;

use IO::Socket qw(AF_INET SOCK_STREAM);
use Xormon;

# Server class to manage socket creation and client connections
# Attributes:
#   port        - The port number on which the server will listen (default: 3999, dynamically checked via Xormon::checkPrefixPort).
#   host        - The host address on which the server will bind (default: '0.0.0.0').
#   reuse_port  - Boolean indicating whether the server should reuse the port (default: 1).
#   listen      - The maximum number of pending connections (default: 5).
#   domain      - The socket domain (optional, e.g., AF_INET for IPv4).
#   type        - The socket type (optional, e.g., SOCK_STREAM for TCP).
#   server      - The server socket object.

sub new {
    my ($class, %args) = @_;
    # Default port number
    my $default_port = $args{port} || 3999;
    my $port = Xormon::checkPrefixPort($default_port);

    my $self = {
        port => $port,  # Final port number after checking
        host => $args{host} || '0.0.0.0',
        reuse_port => $args{reuse_port} || 1,
        listen => $args{listen} || 5,
        domain => $args{domain} || AF_INET,  # Default to AF_INET (IPv4)
        type   => $args{type} || SOCK_STREAM, # Default to SOCK_STREAM (TCP)
        server => undef,
    };
    bless $self, $class;
    return $self;
}

# Method to start the server by creating and binding the socket
sub start {

 my ($self) = @_;
    $self->{server} = IO::Socket->new(
        Domain    => $self->{domain},
        Type      => $self->{type},
        LocalHost => $self->{host},
        LocalPort => $self->{port},
        Proto     => 'tcp',
        Listen    => $self->{listen},
        Reuse     => $self->{reuse_port},
    ) or die "Could not create socket: $!";
    Xormon::log("Service listening on port $self->{port}");
    return $self->{server};
}

# Method to accept a client connection
# Returns a socket object representing the client connection
sub accept_client {
    my ($self) = @_;       
    return $self->{server}->accept();
}

# Method to close the server socket
sub close {
    my ($self) = @_;
    $self->{server}->close() if $self->{server};
}

1;