package Xormon;

use strict;
use warnings;

use LWP;
use Data::Dumper;
use JSON;
use POSIX qw(strftime ceil);
use Date::Parse;
use Time::Local;
use Scalar::Util qw(looks_like_number);
use File::Temp qw/tempdir/;
use File::Basename;
use Exporter;

our @ISA       = qw(Exporter);
our @EXPORT_OK = qw(VrDump);

my $BACKEND_PORT        = defined $ENV{'APP_PORT'} ? $ENV{'APP_PORT'} : 8443;
my $METRICS_SERVER_PORT = defined $ENV{'METRICS_SERVER_PORT'} ? $ENV{'METRICS_SERVER_PORT'} : 3030;
my $DEFAULT_TIMEOUT     = 30;

my $protocol_https = 'https://';
#my $protocol_http = 'http://';

my $PERF_JSON_API        = "/api/metrics/v3/data/store";
my $PERF_CSV_API         = "/api/metrics/v2/data/store";
my $HEALTH_STATUS_API_MS = "/api/health_status/v1/store";
my $MS_VARIABLES_API     = "/api/configuration/v1/microservices/variables";
my $CONNTEST_API_MS      = "/api/health_status/v1/conntest";
my $DATATEST_API_MS      = "/api/health_status/v1/datatest";
my $ARCHITECTURE_API     = "/api/configuration/v1/architecture";

#my $PERF_JSON_API = "/perf/json";
#my $PERF_CSV_API = "/perf/csv";

my $splice_data_by = 100;

my $md5module                                       = 1;
my $password                                        = "";
eval "use Digest::MD5 qw(md5_hex); 1" or $md5module = 0;

if( !$md5module ) { # binary MD5 module not found, use bundled pure Perl one
  #use Digest::Perl::MD5 qw(md5_hex);
  require Digest::Perl::MD5;

  # print STDERR "Bundled MD5\n";
}

sub new {
  my ( $self, $url, $token, $microservice, $hostcfg_id, $hostalias ) = @_;

  # if the url contains a port
  my @host = split(/:/, $url);

  my $o                  = {};
  $o->{'url'}            = $host[0] . ":" . $BACKEND_PORT;
  $o->{'metrics-server'} = $host[0] . ":" . $METRICS_SERVER_PORT;
  $o->{'token'}          = $token;
  $o->{'microservice'}   = $microservice;
  if( defined $hostcfg_id && $hostcfg_id ne '' ) {
    $o->{'hostcfg_id'} = $hostcfg_id;
  }
  if( defined $hostalias && $hostalias ne '' ) {
    $o->{'hostalias'} = $hostalias;
  }
  $o->{'destinations'} = ();
  bless $o, $self;

  return $o;
}

sub setDestinations {
  my ( $self, $destinations ) = @_;

  if( defined $self->{'hostcfg_id'} ) {

    $self->{'destinations'} = $destinations;

    for my $destination ( @{$destinations} ) {
      registerDestination($destination->{'destination'},
                          $destination->{'port'},
                          $destination->{'token'},
                          $self->{'hostcfg_id'},
                          $self->{'hostalias'},
                          $self->{'microservice'},
                          $destination->{'data'});
    }

  } else {
    Xormon::error("hostcfg_id is not defined");
  }
}

sub validateJSON {
  my ( $self, $data ) = @_;

}

sub validateCSV {
  my ( $self, $data ) = @_;

}

sub JSONtoCSV {
  my ( $self, $data ) = @_;

  my %metrics;

  foreach my $uuid ( keys %{$data} ) {
    foreach my $time ( keys %{$data->{$uuid}} ) {
      foreach my $metric ( keys %{$data->{$uuid}{$time}} ) {
        if( !defined $metrics{$metric} ) {
          $metrics{$metric} = 1;
        }
      }
    }
  }
  my $csv = "uuid;timestamp;";
  foreach my $metric ( keys %metrics ) {
    $csv .= "$metric;";
  }
  foreach my $uuid ( keys %{$data} ) {
    foreach my $time ( keys %{$data->{$uuid}} ) {
      $csv .= "\n$uuid;$time;";
      foreach my $metric ( keys %metrics ) {
        if( defined $data->{$uuid}{$time}{$metric} ) {
          $csv .= $data->{$uuid}{$time}{$metric} . ";";
        } else {
          $csv .= "U;";
        }
      }
    }
  }

  return $csv;
}

sub CSVtoJSON {
  my ( $self, $data ) = @_;
  my %data;
  my @rows    = split(/\n/, $data);
  my @metrics = split(/;/, $rows[0]);
  my $i       = 0;
  foreach my $row ( @rows ) {
    if( $i eq "0" ) {
      $i++;
      next;
    }
    my @values = split(/;/, $row);
    my $c      = 0;
    foreach my $value ( @values ) {
      if( $c <= 1 || $values[$c] eq "U" ) {
        $c++;
        next;
      }
      $data{$values[0]}{$values[1]}{$metrics[$c]} = $values[$c];
      $c++;
    }
    $i++;
  }

  return \%data;
}

sub split_hash {
  my ( $hash, $val ) = @_;
  $val               = defined $val ? $val : $splice_data_by;
  my @keys           = keys %{$hash};
  my @hashes;

  while( my @subset = splice(@keys, 0, $val) ) {
    push @hashes, {map { $_ => $hash->{$_} } @subset};
  }
  return \@hashes;
}

sub keepLastResponse {
  my $data       = shift;
  my $hw_type    = shift;
  my $hostcfg_id = shift;
  my $filename  = shift;
  
  my $logdir   = dirname( $ENV{pm_err_log_path} );
  my $fileout  = "$logdir/$hw_type/$hostcfg_id/$filename.json";

  if ( !-d "$logdir/$hw_type" ) {
    mkdir( "$logdir/$hw_type", 0755 ) || Xormon::error( "Cannot mkdir $logdir/$hw_type $!" . __FILE__ . ":" . __LINE__ ) && return 0;
  }
  if ( !-d "$logdir/$hw_type/$hostcfg_id" ) {
    mkdir( "$logdir/$hw_type/$hostcfg_id", 0755 ) || Xormon::error( "Cannot mkdir $logdir/$hw_type/$hostcfg_id $!" . __FILE__ . ":" . __LINE__ ) && return 0;
  }

  save_as_json($fileout,$data);

  return 1;
}

sub keepLastFiles {
  my $type       = shift;
  my $format     = shift;
  my $data       = shift;
  my $hw_type    = shift;
  my $hostcfg_id = shift;

  use Time::HiRes; # current timestamp in ms

  my $logdir   = dirname($ENV{'pm_err_log_path'});
  my $filename = $type . "-" . sprintf('%.6f', Time::HiRes::time()) . "." . $format;
  my $fileout  = "$logdir/$hw_type/$hostcfg_id/$filename";

  if( !-d "$logdir/$hw_type" ) {
    mkdir("$logdir/$hw_type",
          0755) || Xormon::error("Cannot mkdir $logdir/$hw_type $!" . __FILE__ . ":" . __LINE__) && return 0;
  }
  if( !-d "$logdir/$hw_type/$hostcfg_id" ) {
    mkdir("$logdir/$hw_type/$hostcfg_id",
          0755) || Xormon::error("Cannot mkdir $logdir/$hw_type/$hostcfg_id $!" . __FILE__ . ":" . __LINE__) && return 0;
  }

  if( $format eq "json" ) {
    save_as_json($fileout, $data);
  } else {
    open(my $file, '>', $fileout) || error("Couldn't open file $fileout $!" . __FILE__ . ":" . __LINE__) && return 0;
    print $file $data;
    close($file);
  }

  # once per 15 minutes remove files older than 15 minutes
  my $last_check_file = "$logdir/$hw_type/$hostcfg_id/last_check";
  if( -f $last_check_file ) {
    if( file_time_diff($last_check_file) > 900 ) {
      open(my $file,
           '>',
           $last_check_file) || error("Couldn't open file $last_check_file $!" . __FILE__ . ":" . __LINE__) && return 0;
      close($file);

      removeFiles("$logdir/$hw_type/$hostcfg_id/");
    }
  } else {
    open(my $file,
         '>',
         $last_check_file) || error("Couldn't open file $last_check_file $!" . __FILE__ . ":" . __LINE__) && return 0;
    close($file);

    removeFiles("$logdir/$hw_type/$hostcfg_id/");
  }

  return 1;
}

sub removeFiles {
  my $path = shift;

  opendir(my $dir, "$path") || Xormon::error("Cannot open directory '$path' " . __FILE__ . ":" . __LINE__) && return 0;
  my @files = grep !/^\.\.?$/, readdir($dir);
  chomp @files;
  closedir($dir);

  my @perf_files = grep /^perf-/, @files;
  my @conf_files = grep /^conf-/, @files;
  my @arch_files = grep /^arch-/, @files;
  my @hs_files   = grep /^hs-/, @files;

  # keep perf files only for the last 15 minutes
  foreach my $filename ( @perf_files ) {
    my $file = $path . $filename;
    if( file_time_diff($file) > 900 ) {
      # remove older files than 15 minutes
      unlink($file);
    }
  }

  # keep last conf file
  my $idx     = 0;
  @conf_files = sort @conf_files;
  @conf_files = reverse @conf_files;
  foreach my $filename ( @conf_files ) {
    my $file = $path . $filename;
    if( $idx > 0 ) {
      unlink($file);
    }
    $idx++;
  }

  # keep last arch file
  $idx        = 0;
  @arch_files = sort @arch_files;
  @arch_files = reverse @arch_files;
  foreach my $filename ( @arch_files ) {
    my $file = $path . $filename;
    if( $idx > 0 ) {
      unlink($file);
    }
    $idx++;
  }

  # keep last hs file
  $idx      = 0;
  @hs_files = sort @hs_files;
  @hs_files = reverse @hs_files;
  foreach my $filename ( @hs_files ) {
    my $file = $path . $filename;
    if( $idx > 0 ) {
      unlink($file);
    }
    $idx++;
  }

  return 1;
}

sub saveJSON2 {
  my ( $self, $data ) = @_;

  if( exists $self->{'microservice'} && $self->{'microservice'} ne '' && exists $self->{'hostcfg_id'} && $self->{'hostcfg_id'} ne '' && defined $data && $data ne "" ) {
    keepLastFiles("perf", "json", $data, $self->{'microservice'}, $self->{'hostcfg_id'});
  }

  my $resp_all = 0;

  for( keys %{$data} ) {
    my $hw_type = $_;
    for( keys %{$data->{$hw_type}} ) {
      my $subsystem     = $_;
      my $item_id       = $_;
      my $hash_to_split = $data->{$hw_type}{$subsystem};
      my $hashes        = split_hash($hash_to_split);

      for( @{$hashes} ) {
        my $out->{$hw_type}{$subsystem} = ( scalar keys %{$_} > 0 ) ? $_ : {};

        my $json = JSON->new;

        my $req = HTTP::Request->new('POST' => $protocol_https . $self->{'metrics-server'} . $PERF_JSON_API);
        $req->content_type('application/json');
        $req->content(encode_json($out));
        my $ua = LWP::UserAgent->new(
          'timeout'  => $DEFAULT_TIMEOUT,
          'ssl_opts' => {'SSL_cipher_list' => 'DEFAULT:!DH', 'verify_hostname' => 0, 'SSL_verify_mode' => 0},
        );

        $ua->default_header('Authorization' => 'Bearer ' . $self->{'token'});

        my $resp = ();
        eval {
          $resp = $ua->request($req);
        };

        if( $@ ) {
          my $error = $@;
          print STDERR $error;
        }

        if( $resp->code == 500 || $resp->code == 400 ) {
          Xormon:
          error("[savePerfJSON2] [" . $self->{'hostalias'} . "] Error: " . $resp->content);
        }

        $resp_all = $resp->code > $resp_all ? $resp->code : $resp_all;

        # target destinations
        if( defined $self->{'destinations'} ) {
          for my $destination ( @{$self->{'destinations'}} ) {
            # debug print
            print "[DestinationTarget] [Perf]: " . $destination->{'destination'};

            my $destUrl = $protocol_https . $destination->{'destination'} . ":" . $destination->{'port'} . $PERF_JSON_API;
            sendPostData($destUrl, $destination->{'token'}, $out);
          }
        }
      }
    }
  }
  return $resp_all;
}

sub getMicroserviceVariables {
  my ( $self, $hw_type, $hostcfg_id ) = @_;

  my $json = JSON->new;

  my $req_url = "$protocol_https" . "$self->{'url'}" . "$MS_VARIABLES_API?";
  $req_url    .= defined $hw_type ? "hw_type=$hw_type" : "";
  $req_url    .= defined $hostcfg_id ? "hw_type=$hostcfg_id" : "";

  my $req = HTTP::Request->new('GET' => $req_url);

  $req->content_type('application/json');
  my $ua = LWP::UserAgent->new(
    'timeout'  => $DEFAULT_TIMEOUT,
    'ssl_opts' => {'SSL_cipher_list' => 'DEFAULT:!DH', 'verify_hostname' => 0, 'SSL_verify_mode' => 0},
  );

  $ua->default_header('Authorization' => 'Bearer ' . $self->{'token'});

  my $response = ();
  eval {
    $response = $ua->request($req);
  };
  if( $@ ) {
    my $error = $@;
    error("[getMicroserviceVariables] error response from API Error: $error");
  }

  my $resp;
  eval {
    $resp = $json->decode($response->content);
  };

  if( defined $resp->{'status'} ) {
    if( $resp->{'status'} eq "Failure" ) {
      error("[getData] API Request Failure: " . $resp->{'message'} . " (bad token?)");
    }
  }

  if( $resp->{'statusCode'} eq "404" || !defined $resp->{'data'} ) {
    return ();
  }

  return $resp->{'data'};
}

sub saveJSON {
  my ( $self, $data, $hw_type, $subsystem ) = @_;

  my $json = JSON->new;

  my $param = '';
  if( defined $hw_type && defined $subsystem ) {
    $param = "?hw_type=$hw_type&subsystem=$subsystem";
  }
  my $req = HTTP::Request->new('POST' =>
                               $protocol_https . $self->{'metrics-server'} . "/api/metrics/v1/data/store" . $param);
  $req->content_type('application/json');
  $req->content(encode_json($data));
  my $ua = LWP::UserAgent->new(
    'timeout'  => $DEFAULT_TIMEOUT,
    'ssl_opts' => {'SSL_cipher_list' => 'DEFAULT:!DH', 'verify_hostname' => 0, 'SSL_verify_mode' => 0},
  );

  $ua->default_header('Authorization' => 'Bearer ' . $self->{'token'});

  my $resp = ();

  eval {
    $resp = $ua->request($req);
  };

  if( $@ ) {
    my $error = $@;
    print STDERR $error;
  }

  if( $resp->code == 500 || $resp->code == 400 ) {
    Xormon:
    error("[savePerfJSON] [" . $self->{'hostalias'} . "] Error: " . $resp->content);
  }

  # target destinations
  if( defined $self->{'destinations'} ) {
    for my $destination ( @{$self->{'destinations'}} ) {
      # debug print
      print "[DestinationTarget] [Perf]: " . $destination->{'destination'};

      my $destUrl = $protocol_https . $destination->{'destination'} . ":" . $destination->{'port'} . "/api/metrics/v1/data/store" . $param;
      sendPostData($destUrl, $destination->{'token'}, $data);
    }
  }

  return $resp->code;
}

sub saveCSV {
  my ( $self, $data, $hw_type, $subsystem ) = @_;

  if( exists $self->{'microservice'} && $self->{'microservice'} ne '' && exists $self->{'hostcfg_id'} && $self->{'hostcfg_id'} ne '' ) {
    keepLastFiles("perf", "csv", $data, $self->{'microservice'}, $self->{'hostcfg_id'});
  }

  my $json = JSON->new;

  my $param = '';
  if( defined $hw_type && defined $subsystem ) {
    $param = "?hw_type=$hw_type&subsystem=$subsystem";
  }

  my $req = HTTP::Request->new('POST' => $protocol_https . $self->{'metrics-server'} . $PERF_CSV_API . $param);
  $req->content_type('text/plain');
  $req->content($data);

  my $ua = LWP::UserAgent->new(
    'timeout'  => $DEFAULT_TIMEOUT,
    'ssl_opts' => {'SSL_cipher_list' => 'DEFAULT:!DH', 'verify_hostname' => 0, 'SSL_verify_mode' => 0},
  );

  $ua->default_header('Authorization' => 'Bearer ' . $self->{'token'});

  my $resp = ();

  eval {
    $resp = $ua->request($req);
  };

  if( $@ ) {
    my $error = $@;
    print STDERR $error;
  }

  if( $resp->code == 500 || $resp->code == 400 ) {
    Xormon:
    error("[savePerfCSV] [" . $self->{'hostalias'} . "] Error: " . $resp->content);
  }

  # target destinations
  if( defined $self->{'destinations'} ) {
    for my $destination ( @{$self->{'destinations'}} ) {
      # debug print
      print "[DestinationTarget] [Perf]: " . $destination->{'destination'};

      my $destUrl = $protocol_https . $destination->{'destination'} . ":" . $destination->{'port'} . $PERF_CSV_API . $param;
      sendPostData($destUrl, $destination->{'token'}, $data, 1);
    }
  }

  return $resp->code;
}

sub saveConfJSON {
  my ( $self, $data, $hs_data ) = @_;

  if( exists $self->{'microservice'} && $self->{'microservice'} ne '' && exists $self->{'hostcfg_id'} && $self->{'hostcfg_id'} ne '' ) {
    if( defined $hs_data && $hs_data eq "1" ) {
      keepLastFiles("hs", "json", $data, $self->{'microservice'}, $self->{'hostcfg_id'});
    } else {
      keepLastFiles("conf", "json", $data, $self->{'microservice'}, $self->{'hostcfg_id'});
    }
  }

  my $resp_all = 0;
  my $hashes   = split_hash($data);
  for( @{$hashes} ) {
    my $json = JSON->new;
    my $req  = HTTP::Request->new('POST' =>
                                  $protocol_https . $self->{'metrics-server'} . "/api/configuration/v1/properties");
    $req->content_type('application/json');
    $req->content(encode_json($_));

    my $ua = LWP::UserAgent->new(
      'timeout'  => $DEFAULT_TIMEOUT,
      'ssl_opts' => {'SSL_cipher_list' => 'DEFAULT:!DH', 'verify_hostname' => 0, 'SSL_verify_mode' => 0},
    );

    $ua->default_header('Authorization' => 'Bearer ' . $self->{'token'});

    my $resp = ();

    eval {
      $resp = $ua->request($req);
    };

    if( $@ ) {
      my $error = $@;
      print STDERR $error;
    }
    if( $resp->code == 500 || $resp->code == 400 ) {
      Xormon:
      error("[saveConfJSON] [" . $self->{'hostalias'} . "] Error: " . $resp->content);
    }

    $resp_all = $resp->code > $resp_all ? $resp->code : $resp_all;

    # target destinations
    if( defined $self->{'destinations'} ) {
      for my $destination ( @{$self->{'destinations'}} ) {
        # debug print
        print "[DestinationTarget] [Conf]: " . $destination->{'destination'};

        my $destUrl = $protocol_https . $destination->{'destination'} . ":" . $destination->{'port'} . "/api/configuration/v1/properties";
        sendPostData($destUrl, $destination->{'token'}, $data);
      }
    }
  }
  return $resp_all;
}

sub deleteArchitecture {
  my ( $self, $hostcfg_id ) = @_;

  my $json = JSON->new;
  my $req  = HTTP::Request->new('DELETE' => $protocol_https . $self->{'metrics-server'} . $ARCHITECTURE_API);
  $req->content_type('application/json');

  my %data = (
    "hostcfg_id" => $hostcfg_id,
  );

  $req->content(encode_json(\%data));

  my $ua = LWP::UserAgent->new(
    'timeout'  => $DEFAULT_TIMEOUT,
    'ssl_opts' => {'SSL_cipher_list' => 'DEFAULT:!DH', 'verify_hostname' => 0, 'SSL_verify_mode' => 0},
  );

  $ua->default_header('Authorization' => 'Bearer ' . $self->{'token'});

  my $resp = ();

  eval {
    $resp = $ua->request($req);
  };

  if( $@ ) {
    my $error = $@;
    print STDERR $error;
  }

  return $resp->code;
}

sub saveArchitecture {
  my ( $self, $data ) = @_;

  if( exists $self->{'microservice'} && $self->{'microservice'} ne '' && exists $self->{'hostcfg_id'} && $self->{'hostcfg_id'} ne '' && ( defined $data && $data ne "" ) ) {
    keepLastFiles("arch", "json", $data, $self->{'microservice'}, $self->{'hostcfg_id'});
  }

  my $resp_all = 0;

  while( my @next_n = splice @{$data}, 0, $splice_data_by ) {
    my $json        = JSON->new;
    my $req         = HTTP::Request->new('POST' => $protocol_https . $self->{'metrics-server'} . $ARCHITECTURE_API);
    $req->content_type('application/json');
    $req->content(encode_json(\@next_n));

    my $ua = LWP::UserAgent->new(
      'timeout'  =>
      60, #$DEFAULT_TIMEOUT, change has been made to reduce number of timeout errs as architectures tend to be quite large
      'ssl_opts' =>
      {'SSL_cipher_list' => 'DEFAULT:!DH', 'verify_hostname' => 0, 'SSL_verify_mode' => 0},
    );

    $ua->default_header('Authorization' => 'Bearer ' . $self->{'token'});

    my $resp = ();
    eval {
      $resp = $ua->request($req);
    };
    if( $@ ) {
      my $error = $@;
      print STDERR $error;
    }

    if( $resp->code == 500 || $resp->code == 400 ) {
      Xormon:
      error("[saveArchitecture] [" . $self->{'hostalias'} . "] Error: " . $resp->content);
    }

    $resp_all = $resp->code > $resp_all ? $resp->code : $resp_all;

    # target destinations
    if( defined $self->{'destinations'} ) {
      for my $destination ( @{$self->{'destinations'}} ) {
        # debug print
        print "[DestinationTarget] [Arch]: " . $destination->{'destination'};

        my $destUrl = $protocol_https . $destination->{'destination'} . ":" . $destination->{'port'} . $ARCHITECTURE_API;
        sendPostData($destUrl, $destination->{'token'}, \@next_n);
      }
    }
    #print "save partial architecture: ".$resp->code."\n";
  }

  return $resp_all;
}

sub saveArchitectureAgent {
  my ( $self, $data ) = @_;

  if( exists $self->{'microservice'} && $self->{'microservice'} ne '' && exists $self->{'hostcfg_id'} && $self->{'hostcfg_id'} ne '' ) {
    keepLastFiles("arch", "json", $data, $self->{'microservice'}, $self->{'hostcfg_id'});
  }

  my $json = JSON->new;

  my $req = HTTP::Request->new('POST' => $protocol_https . $self->{'metrics-server'} . $ARCHITECTURE_API . "?agent=1");
  $req->content_type('application/json');
  $req->content(encode_json($data));

  my $ua = LWP::UserAgent->new(
    'timeout'  => $DEFAULT_TIMEOUT,
    'ssl_opts' => {'SSL_cipher_list' => 'DEFAULT:!DH', 'verify_hostname' => 0, 'SSL_verify_mode' => 0},
  );

  $ua->default_header('Authorization' => 'Bearer ' . $self->{'token'});

  my $resp = ();

  eval {
    $resp = $ua->request($req);
  };

  if( $@ ) {
    my $error = $@;
    print STDERR $error;
  }

  # target destinations
  if( defined $self->{'destinations'} ) {
    for my $destination ( @{$self->{'destinations'}} ) {
      # debug print
      print "[DestinationTarget] [AgentArch]: " . $destination->{'destination'};

      my $destUrl = $protocol_https . $destination->{'destination'} . ":" . $destination->{'port'} . $ARCHITECTURE_API . "?agent=1";
      sendPostData($destUrl, $destination->{'token'}, $data);
    }
  }

  return $resp->code;
}

sub linuxAgentMapping {
  my ( $self, $uuid ) = @_;

  my $json = JSON->new;

  my $req = HTTP::Request->new('GET' => $protocol_https . $self->{'metrics-server'} . "/api/mapping/v1/linux/$uuid");
  $req->content_type('application/json');

  my $ua = LWP::UserAgent->new(
    'timeout'  => $DEFAULT_TIMEOUT,
    'ssl_opts' => {'SSL_cipher_list' => 'DEFAULT:!DH', 'verify_hostname' => 0, 'SSL_verify_mode' => 0},
  );

  $ua->default_header('Authorization' => 'Bearer ' . $self->{'token'});

  my $resp = ();
  my $response;

  eval {
    $response = $ua->request($req);
  };

  if( $@ ) {
    my $error = $@;
    print STDERR $error;
  }

  eval {
    $resp = $json->decode($response->content);
  };

  if( $@ ) {
    my $error = $@;
    error("JSON decode response from API Error: $error");
    print Dumper($response->content);
  }

  return $resp->{'data'};
}

sub powerAgentMapping {
  my ( $self, $serial, $label ) = @_;

  my $json = JSON->new;

  my $req = HTTP::Request->new('GET' =>
                               $protocol_https . $self->{'metrics-server'} . "/api/mapping/v1/power/$serial/$label");
  $req->content_type('application/json');

  my $ua = LWP::UserAgent->new(
    'timeout'  => $DEFAULT_TIMEOUT,
    'ssl_opts' => {'SSL_cipher_list' => 'DEFAULT:!DH', 'verify_hostname' => 0, 'SSL_verify_mode' => 0},
  );

  $ua->default_header('Authorization' => 'Bearer ' . $self->{'token'});

  my $resp = ();
  my $response;

  eval {
    $response = $ua->request($req);
  };

  if( $@ ) {
    my $error = $@;
    print STDERR $error;
  }

  eval {
    $resp = $json->decode($response->content);
  };

  if( $@ ) {
    my $error = $@;
    error("JSON decode response from API Error: $error");
    print Dumper($response->content);
  }

  return $resp->{'data'};
}

sub saveStatus {
  my ( $self, $data ) = @_;

  if ( exists $self->{microservice} && $self->{microservice} ne '' && exists $self->{hostcfg_id} && $self->{hostcfg_id} ne '' && defined $data && $data ne "") {
    keepLastFiles("hs","json",$data,$self->{microservice},$self->{hostcfg_id});
  }

  my $json = JSON->new;

  my $req = HTTP::Request->new('POST' =>
                               $protocol_https . $self->{'metrics-server'} . ":" . $METRICS_SERVER_PORT . $HEALTH_STATUS_API_MS);
  $req->content_type('application/json');
  $req->content(encode_json($data));

  my $ua = LWP::UserAgent->new(
    'timeout'  => $DEFAULT_TIMEOUT,
    'ssl_opts' => {'SSL_cipher_list' => 'DEFAULT:!DH', 'verify_hostname' => 0, 'SSL_verify_mode' => 0},
  );

  $ua->default_header('Authorization' => 'Bearer ' . $self->{'token'});

  my $resp = ();

  eval {
    $resp = $ua->request($req);
  };

  if( $@ ) {
    my $error = $@;
    print STDERR $error;
  }

  # target destinations
  if( defined $self->{'destinations'} ) {
    for my $destination ( @{$self->{'destinations'}} ) {
      # debug print
      print "[DestinationTarget] [HealthStatus]: " . $destination->{'destination'};

      my $destUrl = $protocol_https . $destination->{'destination'} . ":" . $destination->{'port'} . "/api/health_status/v1/store";
      sendPostData($destUrl, $destination->{'token'}, $data);
    }
  }

  return $resp->code;
}

sub isConnected {
  my ( $self, $status, $message ) = @_;

  my %data;
  $data{'hostcfg_id'} = $self->{'hostcfg_id'};
  $data{'connection'} = $status;

  if( defined $message ) {
    $data{'message'} = $message;
  }

  my $json = JSON->new;
  my $req  = HTTP::Request->new('POST' =>
                                $protocol_https . $self->{'metrics-server'} . ":" . $METRICS_SERVER_PORT . $CONNTEST_API_MS);
  $req->content_type('application/json');
  $req->content(encode_json(\%data));

  my $ua = LWP::UserAgent->new(
    'timeout'  => $DEFAULT_TIMEOUT,
    'ssl_opts' => {'SSL_cipher_list' => 'DEFAULT:!DH', 'verify_hostname' => 0, 'SSL_verify_mode' => 0},
  );

  $ua->default_header('Authorization' => 'Bearer ' . $self->{'token'});

  my $resp = ();
  eval {
    $resp = $ua->request($req);
  };

  if( $@ ) {
    my $error = $@;
    print STDERR $error;
  }

  return $resp->code;
}

sub isConnectedForHostcfgId {
  # Used for Power to send Status for servers connected to inactive HMC.
  # Sends hostcfg_id, code is same as isConnected => combine later (?)
  # Possible extension: send subsystem
  my ( $self, $data ) = @_;

  my $json = JSON->new;
  my $req  = HTTP::Request->new('POST' =>
                                $protocol_https . $self->{'metrics-server'} . ":" . $METRICS_SERVER_PORT . $CONNTEST_API_MS);
  $req->content_type('application/json');
  $req->content(encode_json($data));

  my $ua = LWP::UserAgent->new(
    'timeout'  => $DEFAULT_TIMEOUT,
    'ssl_opts' => {'SSL_cipher_list' => 'DEFAULT:!DH', 'verify_hostname' => 0, 'SSL_verify_mode' => 0},
  );

  $ua->default_header('Authorization' => 'Bearer ' . $self->{'token'});

  my $resp = ();
  eval {
    $resp = $ua->request($req);
  };

  if( $@ ) {
    my $error = $@;
    print STDERR $error;
  }

  return $resp->code;
}

sub hasData {
  my ( $self, $status, $message ) = @_;

  my %data;
  $data{'hostcfg_id'} = $self->{'hostcfg_id'};
  $data{'data'}       = $status;

  if( defined $message ) {
    $data{'message'} = $message;
  }

  my $json = JSON->new;
  my $req  = HTTP::Request->new('POST' =>
                                $protocol_https . $self->{'metrics-server'} . ":" . $METRICS_SERVER_PORT . $DATATEST_API_MS);
  $req->content_type('application/json');
  $req->content(encode_json(\%data));

  my $ua = LWP::UserAgent->new(
    'timeout'  => $DEFAULT_TIMEOUT,
    'ssl_opts' => {'SSL_cipher_list' => 'DEFAULT:!DH', 'verify_hostname' => 0, 'SSL_verify_mode' => 0},
  );

  $ua->default_header('Authorization' => 'Bearer ' . $self->{'token'});

  my $resp = ();

  eval {
    $resp = $ua->request($req);
  };

  if( $@ ) {
    my $error = $@;
    print STDERR $error;
  }

  return $resp->code;
}

sub getData {
  my ( $self ) = @_;

  if( !defined $self->{'hostcfg_id'} ) {
    Xormon::error("hostcfg_id is not defined");
    return ();
  }

  my $json = JSON->new;

  my $req = HTTP::Request->new('GET' =>
                               $protocol_https . $self->{'url'} . "/api/microservices/" . $self->{'microservice'} . "/data/" . $self->{'hostcfg_id'});
  $req->content_type('application/json');

  my $ua = LWP::UserAgent->new(
    'timeout'  => $DEFAULT_TIMEOUT,
    'ssl_opts' => {'SSL_cipher_list' => 'DEFAULT:!DH', 'verify_hostname' => 0, 'SSL_verify_mode' => 0},
  );

  $ua->default_header('Authorization' => 'Bearer ' . $self->{'token'});

  my $resp = ();
  my $response;

  eval {
    $response = $ua->request($req);
  };

  if( $@ ) {
    my $error = $@;
    print STDERR $error;
  }

  eval {
    $resp = $json->decode($response->content);
  };

  if( $@ ) {
    my $error = $@;
    error("[getData] JSON decode response from API Error: $error");
    print Dumper($response->content);
  }

  if( defined $resp->{'status'} ) {
    if( $resp->{'status'} eq "Failure" ) {
      error("[getData] API Request Failure: " . $resp->{'message'} . " (bad token?)");
    }
  }

  if( $resp->{'statusCode'} eq "404" || !defined $resp->{'data'} ) {
    return ();
  }
  return $resp->{'data'};
}

sub saveData {
  my ( $self, $data ) = @_;

  if( !defined $self->{'hostcfg_id'} ) {
    Xormon::error("hostcfg_id is not defined");
    return ();
  }

  my $json = JSON->new;

  my %body;
  $body{'microservice'} = $self->{'microservice'};
  $body{'hostcfg_id'}   = $self->{'hostcfg_id'};
  $body{'value'}        = $data;

  my $req = HTTP::Request->new('PUT' =>
                               $protocol_https . $self->{'url'} . "/api/microservices/" . $self->{'microservice'} . "/data");
  $req->content_type('application/json');
  $req->content(encode_json(\%body));

  my $ua = LWP::UserAgent->new(
    'timeout'  => $DEFAULT_TIMEOUT,
    'ssl_opts' => {'SSL_cipher_list' => 'DEFAULT:!DH', 'verify_hostname' => 0, 'SSL_verify_mode' => 0},
  );

  $ua->default_header('Authorization' => 'Bearer ' . $self->{'token'});

  my $resp = ();

  eval {
    $resp = $ua->request($req);
  };

  if( $@ ) {
    my $error = $@;
    print STDERR $error;
  }

  return $resp->code;
}

sub startJob {
  my ( $self, $job, $identification, $pid ) = @_;

  my $json = JSON->new;

  my %data;
  $data{'microservice'}   = $self->{'microservice'};
  $data{'job'}            = $job;
  $data{'identification'} = $identification;
  $data{'start'}          = time();
  $data{'pid'}            = $pid ? $pid : 0;

  my $req = HTTP::Request->new('POST' =>
                               $protocol_https . $self->{'url'} . "/api/configuration/v1/microservices/job/status");
  $req->content_type('application/json');
  $req->content(encode_json(\%data));

  my $ua = LWP::UserAgent->new(
    'timeout'  => $DEFAULT_TIMEOUT,
    'ssl_opts' => {'SSL_cipher_list' => 'DEFAULT:!DH', 'verify_hostname' => 0, 'SSL_verify_mode' => 0},
  );

  $ua->default_header('Authorization' => 'Bearer ' . $self->{'token'});

  my $resp = ();

  eval {
    $resp = $ua->request($req);
  };

  if( $@ ) {
    my $error = $@;
    print STDERR $error;
  }

  return $resp->code;
}

sub endJob {
  my ( $self, $job, $identification, $status, $pid ) = @_;

  my $json = JSON->new;

  my %data;
  $data{'microservice'}   = $self->{'microservice'};
  $data{'job'}            = $job;
  $data{'identification'} = $identification;
  $data{'end'}            = time();
  $data{'status'}         = $status;
  $data{'pid'}            = $pid ? $pid : 0;

  my $req = HTTP::Request->new('POST' =>
                               $protocol_https . $self->{'url'} . "/api/configuration/v1/microservices/job/status");
  $req->content_type('application/json');
  $req->content(encode_json(\%data));

  my $ua = LWP::UserAgent->new(
    'timeout'  => $DEFAULT_TIMEOUT,
    'ssl_opts' => {'SSL_cipher_list' => 'DEFAULT:!DH', 'verify_hostname' => 0, 'SSL_verify_mode' => 0},
  );

  $ua->default_header('Authorization' => 'Bearer ' . $self->{'token'});

  my $resp = ();

  eval {
    $resp = $ua->request($req);
  };

  if( $@ ) {
    my $error = $@;
    print STDERR $error;
  }

  return $resp->code;
}

sub registerMetrics {
  my ( $self, $data ) = @_;

  my $json = JSON->new;

  my $req = HTTP::Request->new('POST' => $protocol_https . $self->{'url'} . "/api/metrics/v1/definitions");
  $req->content_type('application/json');
  $req->content(encode_json($data));

  my $ua = LWP::UserAgent->new(
    'timeout'  => $DEFAULT_TIMEOUT,
    'ssl_opts' => {'SSL_cipher_list' => 'DEFAULT:!DH', 'verify_hostname' => 0, 'SSL_verify_mode' => 0},
  );

  $ua->default_header('Authorization' => 'Bearer ' . $self->{'token'});

  my $resp = ();

  eval {
    $resp = $ua->request($req);
  };

  if( $@ ) {
    my $error = $@;
    print STDERR $error;
  }

  return $resp->code;
}

sub log {
  my $text = shift;
  # my $act_time = localtime();
  chomp($text);

  $| = 1;
  print STDOUT "$text\n";
  return 1;
}

sub error {
  my $text = shift;
  # my $act_time = localtime();
  chomp($text);

  $| = 1;
  print STDERR "$text\n";
  return 1;
}

sub conn_test {
  my $success = shift;
  my $data    = shift;

  my %conn;
  $conn{'message'} = $data;
  if( $success eq "1" ) {
    $conn{'connected'} = \1;
  } else {
    $conn{'connected'} = \0;
  }

  return encode_json(\%conn);
}

sub unobscure_password {
  my $string    = shift;
  my $unobscure = DecodeBase64($string);
  $unobscure    = unpack(chr(ord("a") + 19 + print ""), $unobscure);
  return $unobscure;
}

sub DecodeBase64 {
  my $d = shift;
  $d =~ tr!A-Za-z0-9+/!!cd;
  $d =~ s/=+$//;
  $d =~ tr!A-Za-z0-9+/! -_!;
  my $r = '';
  while( $d =~ /(.{1,60})/gs ) {
    my $len = chr(32 + length($1) * 3 / 4);
    $r      .= unpack("u", $len . $1);
  }
  $r;
}


sub round {
  my $number         = shift;
  my $decimal_plates = shift;
  if ( !defined $decimal_plates ) {
    $decimal_plates = 3;
  }
  my $num = $number;
  if ( $number =~ /\./ ) {
    $number =~ s/.*\.//g;
    my $length = length($number);
    if ( $length > $decimal_plates ) {
      my $rounded = sprintf( "%.$decimal_plates" . "f", $num );
      return $rounded;
    }
    else {
      return $num;
    }
  }
  else {
    return $num;
  }
}

sub isdigit {
  my $digit = shift;

  if( !defined($digit) ) {
    return 0;
  }
  if( $digit eq '' ) {
    return 0;
  }

  if( $digit eq '-' || $digit eq '--' ) {
    return 0;
  }
  
  if ( $digit eq 'U' ) { # Solaris zone return "U" sometimes
    return 1;
  }

  my $digit_work = $digit;
  $digit_work =~ s/[0-9]//g;
  $digit_work =~ s/\.//;
  $digit_work =~ s/^-//;
  $digit_work =~ s/e|E//;
  $digit_work =~ s/\+//;
  $digit_work =~ s/\-//;

  if( length($digit_work) == 0 ) {

    # is a number
    return 1;
  }

  # NOT a number
  return 0;
} ## end sub isdigit

sub isAscii {
  my $string = shift;

  if( !defined $string || $string eq '' ) {
    return 1;
  }

  if( $string =~ /[^!-~\s]/g ) {
    return 0; #Non-ASCII character found
  } else {
    return 1;
  }
}

sub save_as_json {
  my $path        = shift;
  my $hash_ref    = shift;
  my $path_suffix = "$path-tmp";
  my $json        = JSON->new->utf8;

  if( ref($hash_ref) ne "HASH" && ref($hash_ref) ne "ARRAY" ) {
    warn("Hash ref or array ref expected, got: \"$hash_ref\" (Ref:" . ref($hash_ref) . ") in " . __FILE__ . ":" . __LINE__ . "\n");
    return 0;
  }

  $json->pretty([1]);
  $json->canonical([1]);

  open(FILE,
       ">",
       "$path_suffix") || error("Couldn't open file $path_suffix $!" . __FILE__ . ":" . __LINE__) && return 0;
  print FILE $json->encode($hash_ref);
  #print to_json($hash_ref, {utf8 => 1, pretty => 1});
  close(FILE);
  rename($path_suffix, $path);

  return 1;
}

sub read_json {
  my $file   = shift;
  my $encode = shift;
  my $json   = JSON->new;
  my $string;
  my $output;

  {
    local $/ = undef;
    if( !defined $encode ) {
      open(FILE, '<', $file) or error("Couldn't open file $file $!" . __FILE__ . ":" . __LINE__) && return {};
    } else {
      open(FILE,
           '<:encoding(UTF-8)',
           $file) or error("Couldn't open file $file $!" . __FILE__ . ":" . __LINE__) && return {};
    }
    $string = <FILE>;
    close FILE;
  }

  if( $string ) {
    eval {
      $output = $json->decode($string);
      1;
    } or do {
      error("Couldn't parse json file $file $@ " . __FILE__ . ":" . __LINE__) && return {};
    };
  }

  return $output;
}

sub create_hash_key {
  my $string = shift;
  my $hash   = substr(md5_hex($string), 0, 10);
  #print "$string : $hash\n";

  return $hash;
}

sub create_hash_key_long {
  my $string = shift;
  my $hash   = md5_hex($string);
  #print "$string : $hash\n";

  return $hash;
}

sub file_time_diff {
  my $file = shift;

  my $act_time  = time();
  my $file_time = $act_time;
  my $time_diff = 0;

  if( -f $file ) {
    $file_time = ( stat($file) )[9];
    $time_diff = $act_time - $file_time;
  }

  return ( $time_diff );
}

sub parse_CSV_line_by_header {
  my $separator   = shift;
  my $header_line = shift;
  my $data_line   = shift;

  my $data_out;

  #parse header
  my @header = split(/$separator/, $header_line);

  # parse data lines
  my @values    = split(/$separator/, $data_line);
  my $value_idx = 0;
  foreach my $value ( @values ) {
    if( defined $header[$value_idx] && $header[$value_idx] ne '' ) {
      if( $header[$value_idx] ne "" && $value ne '' ) {
        $data_out->{ $header[$value_idx] } = $value;
      }
    }
    $value_idx++;
  }

  return $data_out;
}

sub calculate_weighted_avg {
  my $data = shift;
  my $deb  = shift; # is not mandatory

  # data must be in this format (array of hashes):
  #$VAR1 = [
  #          {
  #            'count' => '3.413',
  #            'value' => '3.660'
  #          },
  #          ...
  #          {
  #            'value' => '0.902',
  #            'count' => '663.150'
  #          }
  #        ];

  my $sum         = "";
  my $value_total = "";
  my $value_idx   = 0;

  if( ref($data) eq "ARRAY" ) {
    foreach( @{$data} ) {
      #if ( exists $_->{count} && isdigit($_->{count}) && exists $_->{value} && isdigit($_->{value}) ) {
      if( exists $_->{'count'} && isdigit($_->{'count'}) && $_->{'count'} != 0 && exists $_->{'value'} && isdigit($_->{'value'}) ) {
        if( isdigit($sum) ) {
          $sum += $_->{'count'};
        } else {
          $sum = $_->{'count'};
        }

        if( isdigit($value_total) ) {
          $value_total += $_->{'value'} * $_->{'count'};
        } else {
          $value_total = $_->{'value'} * $_->{'count'};
        }
        $value_idx++;
      }
    }
  }
  if( defined $sum && $sum eq "0" ) {
    # when IO is zero, then latency is also zero
    return 0;
  }
  if( $value_idx > 0 ) {
    my $weighted_avg = sprintf('%.4f', $value_total / $sum);

    # debug
    #$deb = 1;
    if( $deb ) {
      print Dumper $data;
      print "$weighted_avg = $value_total / $sum\n";
    }

    return $weighted_avg;
  } else {
    return undef;
  }
}

sub calculate_avg {
  my $data = shift;
  my $deb  = shift; # is not mandatory

  # data must be an array
  #$VAR1 = [
  #          '30.434',
  #          '1.744',
  #          '0.000'
  #        ];

  my $sum       = "";
  my $value_idx = 0;

  if( ref($data) eq "ARRAY" ) {
    foreach my $value_act ( @{$data} ) {
      if( isdigit($value_act) ) {
        if( isdigit($sum) ) {
          $sum += $value_act;
        } else {
          $sum = $value_act;
        }
        $value_idx++;
      }
    }
  }
  if( $value_idx > 0 ) {
    my $avg = sprintf('%.4f', $sum / $value_idx);

    # debug example:
    #$VAR1 = [
    #          '30.434',
    #          '1.744',
    #          '0.000'
    #        ];
    # 10.7260 = 32.178 / 3
    #$deb = 1;
    if( $deb ) {
      print Dumper $data;
      print "$avg = $sum / $value_idx\n";
    }

    return $avg;
  } else {
    return undef;
  }
}

# accepts two arguments, e.g. returnBytes (100, "GiB"), returnBytes (10, "MB"), returnBytes (10, "kB"), returnBytes (10, "KB")
# and returns number of bytes
# 1 KB = 1 KiB = 1024 B #weird but seems valid
# 1 kB = 1000 B
# 1 KiB = 1024 B #weird but seems valid
# 1 kiB (= 1 KiB?) = 1024 B
sub returnBytes {
  my $number = shift;
  my $suffix = shift;

  unless( defined $number && isdigit($number) && defined $suffix && $suffix ne '' ) { return ""; }

  my $last_char = chop($suffix);

  my ( $power, $base ) = get_power_and_base($suffix);
  if( $power == -1 || $base == -1 ) {
    #warn "Wrong intput number: $number, suffix: $suffix\n";
    return "";
  }
  if( $last_char eq "b" ) {
    $number /= 8; #bits to bytes
  }
  return $number * ( $base ** $power );
}

#accepts the suffix e.g. "MB", "GiB" etc and returns base and power to calculate the bytes in returnBytes()
sub get_power_and_base {
  my $suffix     = shift;
  my @arr1       = ( "", "k", "M", "G", "T", "P", "E", "Z", "Y", "R", "Q" );
  my @arr2       = ( "", "ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi", "Yi", "Ri", "Qi" );
  my @target_arr = ();
  if( grep(/^$suffix$/, @arr1) ) {
    for( 0 .. $#arr1 ) {
      return ( $_, 1000 ) if( $suffix eq $arr1[$_] );
    }
  } elsif( grep(/^$suffix$/, @arr2) ) {
    for( 0 .. $#arr2 ) {
      return ( $_, 1024 ) if( $suffix eq $arr2[$_] );
    }
  } elsif( $suffix eq "K" ) {
    return ( 1, 1024 );
  } elsif( $suffix eq "Ki" ) {
    return ( 1, 1024 );
  } else {
    #warn "Error, cannot continue with $suffix\n";
    return ( -1, -1 );
  }
}

sub formatNumberByType {
  my $type  = shift;
  my $value = shift;

  if( !defined $value || $value eq '' || !isdigit($value) || !defined $type || $type eq '' ) {
    return "";
  }

  if( $type eq "int" || $type eq "int2" || $type eq "int4" || $type eq "int8" ) {
    return sprintf('%.0f', $value);
  } elsif( $type eq "float" || $type eq "float4" || $type eq "float8" ) {
    return sprintf('%.4f', $value);
  } else {
    # unsupported types
    return "";
  }
}

sub formatWWN {
  my $wwn = shift;

  unless( defined $wwn ) { return ""; }

  $wwn =~ s/://g;
  $wwn =~ s/\s//g;
  $wwn = uc($wwn);

  return $wwn;
}

sub openssh_new {
  my ( $host, $port, $user, $timeout, $passwd, $ctl_dir ) = @_;
  my $ssh                                                 = "";
  if( !defined $ctl_dir ) {
    $ctl_dir = "";
  }
  my $error = "";
  require Net::OpenSSH;

  eval {
    #Set alarm
    my $act_time       = localtime();
    local $SIG{'ALRM'} = sub { die "$act_time: died in SIG ALRM"; };
    alarm($timeout);
    $password = $passwd;
    if( !defined($ctl_dir) || $ctl_dir eq '' ) {
      $ctl_dir = tempdir('CLEANUP' => 1); # It must be there because of : too long for Unix domain socket
    }
    $ssh = Net::OpenSSH->new(
      $host, 'user'   => $user, 'strict_mode' => 0, 'ctl_dir' => $ctl_dir,
      'master_opts'   => [
        '-q', '-o' => 'NumberOfPasswordPrompts=1',
        '-p'       => $port,
        '-o'       => 'StrictHostKeyChecking=no',
        '-o'       => 'ConnectTimeout=80',
        '-o'       => 'PreferredAuthentications=keyboard-interactive,password',
      ],
      'login_handler' => \&mi_login_handler,
    );
    if( $ssh->error ) {

      #print STDERR ("Error: Unable to connect to remote machine. ".$ssh->error."\n");
      #error("Error: Unable to connect to remote machine " . $ssh->error);
      my $error_local                            = $ssh->error;
      if( !defined $error_local ) { $error_local = ""; }
      $error                                     = "Error: Unable to connect to remote machine $error_local";
    }
    alarm(0);
  };
  if( $@ ) {
    if( $@ =~ /died in SIG ALRM/ ) {
      return "Error command: ssh login timed out after : $timeout seconds";
    }
    return "Error command: ssh login failed $ssh $@";
  }
  if( defined $error && $error ne "" ) {
    return $error;
  }
  return $ssh;
}

sub mi_login_handler {
  my ( $ssh, $pty, $data ) = @_;

  # print "custom login handler called!";
  my $read = sysread($pty, ${$data}, 1024, length ${$data});
  if( $read ) {

    # print "buffer: >$$data<\n";
    if( ${$data} =~ s/.*://s ) {
      print $pty "$password\n";
      return 1;
    }
  }
  return 0;
}

sub openssh_disconnect {
  my ( $ssh, $timeout ) = @_;
  require Net::OpenSSH;

  my $async = 0;
  my $rc    = 0;
  eval {
    #Set alarm
    my $act_time       = localtime();
    local $SIG{'ALRM'} = sub { die "$act_time: died in SIG ALRM"; };
    alarm($timeout);
    my $pid;
    $ssh->disconnect($async);
    if( $ssh->error ) {
      if( $ssh->error =~ /aborted/ ) { # async==0 -> aborted
        #print STDERR ("openssh_disconnect: ".$ssh->error."\n");
        $rc = 1;
      } elsif( $ssh->error =~ /exit/ ) { # async==0 -> exit
        #print STDERR ("openssh_disconnect: ".$ssh->error."\n");
        $rc = 1;
      } else {
        #print STDERR ("Error: openssh_disconnect failed. ".$ssh->error."\n");
        $rc = 0;
      }
    }
    alarm(0);
  };
  if( $@ ) {
    if( $@ =~ /died in SIG ALRM/ ) {
      my $act_time = localtime();

      #print STDERR "Error: Session clossing on timed out after : $timeout seconds\n";
      return 0;
    }

    #print STDERR "Error: ".$@."\n";
    return 0;
  }
  return $rc;
}

sub openssh_runcmd {
  my ( $ssh, $cmd, $timeout, $type_data, $st_type ) = ( shift, shift, shift, shift, shift );
  require Net::OpenSSH;

  my $data = {};
  my @data = ();
  eval {
    #Set alarm
    my $act_time       = localtime();
    local $SIG{'ALRM'} = sub { die "$act_time: died in SIG ALRM"; };
    alarm($timeout);
    if( $cmd =~ /^scp,/ ) {
      my %param;
      my ( undef, $remote_file, $local_file ) = split(",", $cmd);
      $data                                   = $ssh->scp_get(\%param, $remote_file, $local_file);
    } else {
      if( defined $type_data && $type_data eq "ARRAY" ) {

        #{stderr_discard => 0},
        @data = $ssh->capture({'stderr_discard' => 1}, $cmd);
      } else {
        if( !defined $st_type ) { $data = $ssh->capture({'stderr_discard' => 1},
                                                        $cmd); } elsif( $st_type eq "MACROSAN" ) { $data = $ssh->capture({
          'stderr_discard' =>
          1,
          'stdin_data'     =>
          $cmd}); }
      }
    }
    alarm(0);
  };
  if( $@ ) {
    if( $@ =~ /died in SIG ALRM/ ) {
      my $act_time = localtime();
      if( defined $type_data && $type_data eq "ARRAY" ) {
        return "(Error command: $cmd timed out after : $timeout seconds)";
      } else {
        return "Error command: $cmd timed out after : $timeout seconds";
      }
    }
    if( defined $type_data && $type_data eq "ARRAY" ) {
      my $scal = join("", @data);
      return "(Error command: $cmd : $scal\n)";
    } else {
      my @output = %{$data};
      my $scal   = join("", @output);
      return "Error command: $cmd : $scal\n";
    }
  }
  if( defined $type_data && $type_data eq "ARRAY" ) {
    return @data;
  }
  return $data;
}

sub conntest_tcp {
  my $host    = shift;
  my $port    = shift;
  my $timeout = shift;
  if( !Xormon::isdigit($timeout) ) {
    $timeout = 5;
  }
  my $sock;

  use IO::Socket::INET;
  eval {
    local $SIG{'ALRM'} = sub { die 'Timed Out'; };
    alarm($timeout + 5);
    $sock = new IO::Socket::INET(
      'PeerAddr' => $host,
      'PeerPort' => $port,
      'Proto'    => 'tcp',
      'Timeout'  => $timeout,
    );
    alarm 0;
  };
  if( $@ && $@ =~ /Timed Out/ ) {
    return ( 0, "TCP connection to $host:$port timed out after 10 seconds!" )
  } elsif( $sock ) {
    return ( 1, "TCP connection to $host:$port is OK" );
  } else {
    return ( 0, "TCP connection to $host:$port has failed! Open it on the firewall." );
  }

}

sub conntest_udp {
  my $host    = shift;
  my $port    = shift;
  my $timeout = shift;
  if( !Xormon::isdigit($timeout) ) {
    $timeout = 5;
  }
  my $socket;
  my @sock;

  use IO::Socket::INET;
  use IO::Select;

  my $ret = scanUDP($host, $port);

  if( $ret ) {
    return ( 1, "UDP connection to $host:$port is OK" );
  } else {
    return ( 0, "UDP connection to $host:$port has failed! Open it on the firewall." );
  }
}

sub scanUDP {
  my $address  = shift;
  my $port     = shift;
  my $socket   = new IO::Socket::INET(
    'PeerAddr' => $address,
    'PeerPort' => $port,
    'Proto'    => 'udp',
  ) or return 0;
  $socket->send('Hello', 0);
  my $select = new IO::Select();
  $select->add($socket);
  my @socket = $select->can_read(1);
  if( @socket == 1 ) {
    $socket->recv(my $temp, 1, 0) or return 0;
    return 1;
  }
  return 1;
}

sub checkPrefixPort {
  my $default_port = shift;
  my $prefix_port  = defined $ENV{'MS_PORT_PREFIX'} ? $ENV{'MS_PORT_PREFIX'} : ();
  return defined $prefix_port ? ( $prefix_port * 1000 ) + substr($default_port, 1) : $default_port;
}

sub isPortOpen {
  my $port = shift;
  #my $netstat = `lsof -nP -i4TCP:$port | grep LISTEN`; # MacOS
  my $netstat = `netstat -an | egrep " *LISTEN"| egrep ":$port *"`;
  if( $netstat =~ m/$port/ ) {
    return 0;
  } else {
    return 1;
  }
}

sub registerDestination {
  my $destination   = shift;
  my $port          = shift;
  my $token         = shift;
  my $hostcfg_id    = shift;
  my $hostcfg_label = shift;
  my $hw_type       = shift;
  my $data          = shift;

  my %hostcfg = (
    "hostcfg_id" => $hostcfg_id,
    "label"      => $hostcfg_label,
    "hw_type"    => $hw_type,
    "data"       => $data,
  );

  sendPostData($protocol_https . $destination . ':' . $port . '/api/configuration/v1/hostcfg', $token, \%hostcfg);

}

sub sendPostData {
  my $url       = shift;
  my $token     = shift;
  my $data      = shift;
  my $plainData = shift;

  my $req = HTTP::Request->new('POST' => $url);

  if( defined $plainData && $plainData eq "1" ) {
    $req->content_type('text/plain');
    $req->content($data);
  } else {
    $req->content_type('application/json');
    $req->content(encode_json($data));
  }

  my $ua = LWP::UserAgent->new(
    'timeout'  => $DEFAULT_TIMEOUT,
    'ssl_opts' => {'SSL_cipher_list' => 'DEFAULT:!DH', 'verify_hostname' => 0, 'SSL_verify_mode' => 0},
  );

  $ua->default_header('Authorization' => 'Bearer ' . $token);

  my $resp = ();

  eval {
    $resp = $ua->request($req);
  };

  if( $@ ) {
    my $error = $@;
    print STDERR "[SendPostData] error: " . $error;
  }

  return $resp->code;
}

# improved Data::Dumper provides more information
# @param
#  @var mixed $data - the variable I want to dump
#  @var string $title - title of the dump
#  @var bool $trace - if true print trace information
#  @var bool $callerLevel - function call level default=1
#  @var bool $noPrint - if true the text is not sent to the output but is returned in the return value
sub VrDump {
  my $data        = shift;
  my $title       = shift;
  my $trace       = shift;
  my $callerLevel = shift;
  my $noPrint     = shift;
  $callerLevel = $callerLevel // 1;

  my ( $package, $filename, $line )                                                                                 = caller;
  my ( $package1, $filename1, $line1, $subroutine, $hasargs, $wantarray, $evaltext, $is_require, $hints, $bitmask ) = caller($callerLevel);

  my $hr = '-' x 78;
  my $txt;
  if(!defined($subroutine)){
    $subroutine = '';
  }

  $txt = '[' . $callerLevel . '] ' . $subroutine . ' | ' . $filename . '#' . $line . "\n";
  $txt .= "$hr\n";
  if( length($title) ) {
    $txt .= $title . "\n";
    $txt .= "$hr\n";
  }
  $txt .=  Dumper($data);
  if( $trace ) {
    $txt  .= "$hr\n";
    my $i = 0;
    while( 1 ) {
      my ( $package1, $filename1, $line1, $subroutine, $hasargs, $wantarray, $evaltext, $is_require, $hints, $bitmask ) = caller($i);
      if( $subroutine ) {
        $txt .= sprintf("[%s] %-30s | %s#%s\n", $i, $subroutine, $filename1, $line1);
      } else {
        last;
      }
      $i++;
    }
  }
  $txt =~ s/^/| /gm;
  $txt = "\n/-$hr\n" . $txt . "\\-$hr\n\n";

  if( !$noPrint ) {
    print($txt);
  }
  return $txt;
}

1;
