"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const config_1 = require("@nestjs/config");
const core_1 = require("@nestjs/core");
const platform_ws_1 = require("@nestjs/platform-ws");
const swagger_1 = require("@nestjs/swagger");
const compression = require("compression");
const express_1 = require("express");
const session = require("express-session");
const pg = require("pg");
const app_module_1 = require("./app.module");
const dataSource_1 = require("./config/dataSource");
const global_exception_1 = require("./exception/global.exception");
const App_1 = require("./service/configuration/App");
const Settings_1 = require("./service/configuration/Settings");
const typeorm_transactional_1 = require("typeorm-transactional");
const PM2_1 = require("./service/configuration/PM2");
async function bootstrap() {
    await (0, dataSource_1.createDataSource)();
    (0, typeorm_transactional_1.initializeTransactionalContext)({ storageDriver: typeorm_transactional_1.StorageDriver.AUTO });
    const fs = require('fs');
    let httpsOptions;
    let useDefaultCert = true;
    const settings = await Settings_1.Settings.get({
        prefix: 'CERT'
    });
    if (settings['CERT_KEY'] && settings['CERT_KEY'].length >= 1) {
        const keyFile = settings['CERT_KEY'];
        const crtFile = settings['CERT_CRT'];
        if (fs.existsSync(keyFile) && fs.existsSync(crtFile)) {
            httpsOptions = {
                key: fs.readFileSync(keyFile),
                cert: fs.readFileSync(crtFile),
            };
            useDefaultCert = false;
        }
    }
    if (useDefaultCert) {
        httpsOptions = {
            key: fs.readFileSync('ssl/xormon.key'),
            cert: fs.readFileSync('ssl/xormon.crt'),
        };
    }
    const app = await core_1.NestFactory.create(app_module_1.AppModule, httpsOptions ? {
        httpsOptions,
    } : undefined);
    const configService = app.get(config_1.ConfigService);
    app.enableCors({
        origin: ['http://localhost:3003', '*'], credentials: true,
        allowedHeaders: ['Accept', 'Accept-Language', 'Content-Language', 'Content-Type', 'Cookie', 'Content-Length', 'Connection', 'Accept-Encoding',
            'Sec-Fetch-Mode', 'Referer', 'Origin', 'Host', 'Authorization', 'Cache-Control', 'If-None-Match', '*'],
        methods: ['GET', 'PUT', 'POST', 'DELETE', 'PATCH', 'HEAD', 'OPTIONS']
    });
    app.useWebSocketAdapter(new platform_ws_1.WsAdapter(app));
    app.use((0, express_1.json)({ limit: '50mb' }));
    app.use((0, express_1.urlencoded)({ extended: true, limit: '50mb' }));
    app.use(compression());
    const { httpAdapter } = app.get(core_1.HttpAdapterHost);
    app.useGlobalFilters(new global_exception_1.AllExceptionsFilter(httpAdapter));
    const cookieParser = require('cookie-parser');
    app.use(cookieParser());
    const isProduction = configService.get('NODE_ENV') && configService.get('NODE_ENV') === 'production';
    const dbConfig = await (0, dataSource_1.databaseConfiguration)();
    const pgPool = new pg.Pool({
        database: dbConfig.database,
        user: dbConfig.username,
        host: dbConfig.host,
        password: dbConfig.password,
        port: dbConfig.port,
    });
    const pgSession = require('connect-pg-simple')(session);
    app.use(session({
        store: new pgSession({
            pool: pgPool,
            tableName: 'sessions',
            schemaName: 'public'
        }),
        secret: App_1.App.getSecret(),
        resave: false,
        saveUninitialized: false,
        cookie: {
            secure: true,
            httpOnly: true,
            sameSite: isProduction || 'none'
        }
    }));
    if (isProduction) {
        app.set('trust proxy', 1);
    }
    if (fs.existsSync('./client/')) {
        if (isProduction) {
            if (PM2_1.PM2.isMasterBackend()) {
                if (fs.existsSync('./client/openapi.html') && !fs.existsSync('./client/docs')) {
                    if (!fs.existsSync('./client/docs')) {
                        fs.mkdirSync('./client/docs');
                        fs.copyFileSync('./client/openapi.html', './client/docs/index.html');
                    }
                }
            }
        }
        else {
            const config = new swagger_1.DocumentBuilder().setTitle('NextGen').setDescription('The NextGen API description').addBearerAuth().build();
            const document = swagger_1.SwaggerModule.createDocument(app, config);
            swagger_1.SwaggerModule.setup('docs', app, document);
            fs.writeFileSync('./client/swagger-spec.json', JSON.stringify(document));
        }
    }
    if (PM2_1.PM2.isMasterBackend()) {
        if (!fs.existsSync('./files/')) {
            fs.mkdirSync('./files/');
        }
    }
    app.enableShutdownHooks();
    const port = +configService.get('APP_PORT') || 8443;
    await app.listen(port);
}
bootstrap();
//# sourceMappingURL=main.js.map