"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.CapacityTable = exports.GenerateRow = void 0;
const pageResponse_constants_1 = require("../../../common/constants/pageResponse.constants");
const dataSource_1 = require("../../../config/dataSource");
const objectItem_entity_1 = require("../../../entity/objectItem.entity");
const subsystem_entity_1 = require("../../../entity/subsystem.entity");
const MetricsWrapper_1 = require("../../metrics/MetricsWrapper");
const typeorm_1 = require("typeorm");
class GenerateRow {
}
exports.GenerateRow = GenerateRow;
class CapacityTable {
    constructor() {
        this.header = [];
        this.rows = [];
        this.generatedRows = [];
        this.lastIndex = 0;
        this.itemIds = [];
        this.metrics = [];
    }
    addHeader(header) {
        this.header.push({
            order: this.lastIndex,
            value: header.value,
            prefix: header.prefix ? header.prefix : undefined,
            thermometer: typeof header.thermometer !== 'undefined' ? header.thermometer : undefined,
            number: header.number ? header.number : undefined,
            decimal: typeof header.decimal !== 'undefined' ? header.decimal : undefined,
            shortcut: header.shortcut ? header.shortcut : undefined
        });
        this.lastIndex++;
        return this;
    }
    addHeaders(headers) {
        for (const header of headers) {
            this.header.push({
                order: this.lastIndex,
                value: header.value,
                prefix: header.prefix ? header.prefix : undefined,
                thermometer: typeof header.thermometer !== 'undefined' ? header.thermometer : undefined,
                number: header.number ? header.number : undefined,
                decimal: typeof header.decimal !== 'undefined' ? header.decimal : undefined,
                shortcut: header.shortcut ? header.shortcut : undefined
            });
            this.lastIndex++;
        }
        return this;
    }
    addRow(columns) {
        if (columns.length === this.header.length) {
            this.rows.push(columns);
        }
        return this;
    }
    addRows(rows) {
        for (const columns of rows) {
            if (columns.length === this.header.length) {
                this.rows.push(columns);
            }
        }
        return this;
    }
    addGenerateRows(columns) {
        const tmpColumns = [];
        for (const column of columns) {
            if (column.itemId) {
                this.itemIds.push(column.itemId);
            }
            if (column.metric) {
                this.metrics.push(column.metric);
            }
            tmpColumns.push(column);
        }
        if (tmpColumns.length === this.header.length) {
            this.generatedRows.push(tmpColumns);
        }
        return this;
    }
    addSummaryRow(row) {
        this.summaryRow = row;
    }
    async generateRows() {
        const lastValues = await new MetricsWrapper_1.MetricsWrapper(this.metrics, this.itemIds).getLastValuesFromTable();
        for (const generatedRow of this.generatedRows) {
            let row = [];
            for (const generatedColumn of generatedRow) {
                if (generatedColumn.metric && generatedColumn.itemId) {
                    row.push({
                        order: generatedColumn.order,
                        value: lastValues[generatedColumn.itemId] ? lastValues[generatedColumn.itemId][generatedColumn.metric] : null
                    });
                }
                else if (generatedColumn.value) {
                    row.push({
                        order: generatedColumn.order,
                        value: generatedColumn.value,
                        url: generatedColumn.url ? generatedColumn.url : undefined
                    });
                }
            }
            if (row.length === this.header.length) {
                this.rows.push(row);
            }
        }
        return this;
    }
    sortByIndex(i, rev = false, isString = false) {
        if (rev) {
            if (isString) {
                this.rows.sort((a, b) => b[i].value.localeCompare(a[i].value, undefined, { numeric: true, sensitivity: 'base' }));
            }
            else {
                this.rows.sort((a, b) => (a[i].value > b[i].value) ? 1 : ((b[i].value > a[i].value) ? -1 : 0));
            }
        }
        else {
            if (isString) {
                this.rows.sort((a, b) => a[i].value.localeCompare(b[i].value, undefined, { numeric: true, sensitivity: 'base' }));
            }
            else {
                this.rows.sort((a, b) => (a[i].value < b[i].value) ? 1 : ((b[i].value < a[i].value) ? -1 : 0));
            }
        }
    }
    generateBox() {
        let exportBody = {
            headers: this.header.map(h => `${h.value}${h.shortcut ? " [" + h.shortcut + "]" : ""}`),
            values: this.rows.map(r => r.map(c => c.value ? c.value.toString() : ""))
        };
        return {
            type: pageResponse_constants_1.BOX_TYPE.CapacityTableBox,
            header: this.header,
            rows: this.rows,
            export: {
                fileName: 'Capacity',
                method: 'POST',
                url: '/api/exporter/v1/values',
                body: exportBody
            },
            summaryRow: this.summaryRow
        };
    }
    static async createSimpleCapacityTable(itemIds, metrics, hwType, subsystem) {
        const items = await (0, dataSource_1.getDataSource)().getRepository(objectItem_entity_1.ObjectItem).findBy({ item_id: (0, typeorm_1.In)(itemIds) });
        const subsystemDefinition = await (0, dataSource_1.getDataSource)().getRepository(subsystem_entity_1.Subsystem).findOneBy({ hw_type: hwType, subsystem: subsystem });
        if (!subsystemDefinition) {
            throw new Error("[createSimpleCapacityTable] Error: subsystem " + subsystem + " not found");
        }
        const metricDefinitions = await MetricsWrapper_1.MetricsWrapper.getMetrics(hwType, subsystem);
        let capacityTable = new CapacityTable()
            .addHeader({ value: subsystemDefinition.label });
        for (const metric of metrics) {
            const definition = metricDefinitions.find(m => m.metric === metric);
            if (definition) {
                capacityTable.addHeader({ value: definition.label, number: true, prefix: definition.defaultPrefix, decimal: definition.defaultPrecision });
            }
            else {
                capacityTable.addHeader({ value: metric, number: true });
            }
        }
        const lastMetrics = await new MetricsWrapper_1.MetricsWrapper(metrics, itemIds).getLastValuesFromTable();
        for (const item of items) {
            let row = [{ order: 0, value: item.label }];
            let i = 1;
            for (const metric of metrics) {
                row.push({ order: i, value: lastMetrics[item.item_id] ? lastMetrics[item.item_id][metric] : null });
                i++;
            }
            capacityTable.addRow(row);
        }
        capacityTable.sortByIndex(0, false, true);
        let totalRow = [{ order: 0, value: 'Total' }];
        let i = 1;
        for (const metric of metrics) {
            totalRow.push({
                order: i, value: Object.keys(lastMetrics).map(id => lastMetrics[id][metric]).reduce((previous, current) => previous + current, 0)
            });
        }
        capacityTable.addRow(totalRow);
        return capacityTable.generateBox();
    }
}
exports.CapacityTable = CapacityTable;
//# sourceMappingURL=CapacityTable.js.map