"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TaskMonitor = void 0;
const dataSource_1 = require("../config/dataSource");
const objectItem_entity_1 = require("../entity/objectItem.entity");
const App_1 = require("../service/configuration/App");
const TableGenerator_1 = require("../service/metrics/TableGenerator");
const uuid_1 = require("uuid");
class TaskMonitor {
    constructor(task, label, logLabel = "TaskMonitor") {
        this.task = task;
        this.label = label;
        this.logLabel = logLabel;
        this.namespace = App_1.App.DEFAULT_NAMESPACE;
        this.workerStarted = false;
    }
    start() {
        if (!this.valid()) {
            return;
        }
        this.log("task started");
        this.started = new Date().getTime() / 1000;
    }
    setTask(task) {
        this.task = task;
    }
    setLabel(label) {
        this.label = label;
        this.workerStarted = true;
    }
    getWorkerStarted() {
        return this.workerStarted;
    }
    end() {
        this.workerStarted = false;
        if (!this.valid()) {
            return;
        }
        const seconds = (new Date().getTime() / 1000) - this.started;
        this.log(`task finished in ${Math.ceil(seconds)} sec.`);
        this.saveArchitecture().then(item => {
            this.savePerformance(item.item_id, seconds).catch(e => {
                console.error(`[${this.logLabel}] [${this.task}] save perf error: ${e}`);
            });
        }).catch(e => {
            console.error(`[${this.logLabel}] [${this.task}] save arch error: ${e}`);
        });
    }
    log(text) {
        console.log(`[${this.logLabel}] [${this.task}]: ${text}`);
    }
    valid() {
        if (this.task === undefined) {
            return false;
        }
        else {
            return true;
        }
    }
    async saveArchitecture() {
        const item = {
            item_id: (0, uuid_1.v5)(this.task, this.namespace),
            label: this.label ? this.label : this.task,
            object_id: 'xormon',
            hw_type: 'application',
            subsystem: 'task',
            updatedAt: new Date()
        };
        return await (0, dataSource_1.getDataSource)().getRepository(objectItem_entity_1.ObjectItem).save(item);
    }
    async savePerformance(itemId, seconds) {
        const repository = (0, dataSource_1.getDataSource)().getRepository('metrics.metrics_application_task');
        await repository.createQueryBuilder().insert().values({
            item_id: itemId,
            time: new Date().toISOString(),
            duration: seconds
        }).orIgnore().execute();
    }
    static async checkInternalMonitoringTable() {
        const tables = [
            'metrics_application_task',
            'metrics_database_schema',
            'metrics_database_hw_type',
            'metrics_database_subsystem',
            'metrics_database_table'
        ];
        for (const table of tables) {
            const exists = await (0, dataSource_1.getDataSource)().query("SELECT to_regclass('metrics." + table + "')");
            if (exists[0]['to_regclass'] === null) {
                const entity = (0, dataSource_1.getDataSource)().entityMetadatas.find(ent => ent.tableName === table);
                if (entity) {
                    await TableGenerator_1.TableGenerator.createPerformanceTableFromEntity(entity);
                }
            }
        }
    }
}
exports.TaskMonitor = TaskMonitor;
//# sourceMappingURL=TaskMonitor.js.map