"""
Microservice interface. Main communication point is controller.

READY:
process POST request with specified action in query

Lower priority:
TODO: async requests handling / switch server package

"""

from http.server import BaseHTTPRequestHandler, HTTPServer
from urllib.parse import urlparse, parse_qs
import socket

import measure.timing as timing
import sys
import json

import controller

HOST, PORT = '0.0.0.0', 3020


class GetHandler(BaseHTTPRequestHandler):

    def do_GET(self):
        message = 'Analyzer service'
        self.send_response(200)
        self.end_headers()
        self.wfile.write(message)
        return

    def do_POST(self):
        print("-----------------------------------------------------------------------")
        print("                              NEW REQUEST                              ")
        print("-----------------------------------------------------------------------")

        t = timing.Timer("TOTAL SERVICE TIME")
        t.start()
        print(f"Received request: ")
        if (len(self.requestline)<=500):
            print(self.requestline)
        else:
            print("Large => head:")
            print(self.requestline[:500])

        print(f"Received path:    {self.path}")
        print()
        resp_code = 200

        content_len = int(self.headers.get('Content-Length'))
        post_body = self.rfile.read(content_len).decode()

        query  = urlparse(self.path).query
        fields = parse_qs(query)

        action = ""
        if 'action' in fields.keys():
            action = fields['action'][0]
        else:
            action = ""
            resp_code = 422

        # TODO: proper returns
        print(f'Received action: {action}')
        print(f"Post body - size: {sys.getsizeof(post_body)} B")
        print("Post body: ")
        if (len(post_body)<=500):
            print(post_body)
        else:
            print("Large => head:")
            print(post_body[:500])

        post_data: dict = {}
        try:
            post_data = json.loads(post_body)
        except:
            post_data = {}
            resp_code = 400

        if 'request' in post_data.keys():
            if type(post_data["request"]) is dict:
                print("Request is dict!")
            elif type(post_data["request"]) is list:
                print("Request is list!")
                number_of_requests = len(post_data["request"])
                print(f"Individual requests: {number_of_requests}")
            else:
                print("Request structure is unexpected!")
                print(type(post_data["request"]))
                action="invalid"
                resp_code = 400
        else:
            print("Unexpected request structure!")
            resp_code = 400

        response  = {}

        print()
        print("-----------------------------------------------------------------------")
        print()

        """
        action:
            function: specifies function in controller
            input: decides to pass/not pass POST data as an argument
            description: Externally available info about action.
        """
        action_tree = {
            "info":{
                "description": "Information call. Returns list of all available actions."
            },
            "add_requests_and_calculate_all": {
                "function": controller.add_requests_and_calculate_all,
                "input": True,
                "description": "Adds received requests from POST body to prediction table DB."
            },
            "calculate_all": {
                "function": controller.act,
                "input": False,
                "description": "Runs all calculations for saved request."
            },
            "load_prediction": {
                "function": controller.load_data,
                "input": True,
                "description": "Preferred way to call prediction. Takes request, returns prediction JSON."
            },
            "load_times_to_fill": {
                "function": controller.load_data_for_capacity_table,
                "input": True,
                "description": "Load saved times from DB."
            },
            #"requests_to_database": {
            #    "function": controller.more_requests_to_database,
            #    "input": True,
            #    "description": "Saves requests to DB without computations."
            #},
            #
            # Rest is for testing and development
            # Possibility: callable only in development mode
            #
            "calculate_now": {
                "function": controller.process_one_request,
                "input": True,
                "description": "Calculates and returns prediction for one request."
            },
            "do_one_and_save": {
                "function": controller.process_one_and_save,
                "input": True,
                "description": "Calculates one prediction and saves appropriate data to prediction table."
            },
            "request_to_database": {
                "function": controller.request_to_database,
                "input": True,
                "description": "Saves one request to prediction table without computations."
            },
            "return_data_for_request": {
                "function": controller.data_used_for_request,
                "input": True,
                "description": "Returns performance data that are used for computations."
            },
            "drop_prediction_table": {
                "function": controller.drop_prediction_table,
                "input": False,
                "description": "Drops prediction table."
            },
            "process_series": {
                "function": controller.run_all_processings,
                "input": True,
                "description": "Processes input series. Returns prediction. Expects 5m frequency."
            },
        }


        if action in action_tree.keys() and action != "info":

            takes_input = action_tree[action]["input"]

            if takes_input:
                # only expected input is in request format => only one check, should be here
                response = action_tree[action]["function"](post_data)
            else:
                response = action_tree[action]["function"]()

        else:
            if action in action_tree.keys():
                print("INFOCALL")
            else:
                print("UNKNOWN ACTION")
                resp_code = 501

            response = {}
            response["available_actions"] = {}

            for key in action_tree.keys():
                description = action_tree[key]["description"]
                response["available_actions"][key] = description

        t.stop(show=True)
        t.start("REST")

        json_response = json.dumps(response)

        self.send_response(resp_code)
        self.send_header("Content-type", "application/json")
        self.end_headers()
        self.wfile.write(json_response.encode())
        #self.wfile.close()

        t.stop(show=True)
        print("\n\n")
        return


if __name__ == '__main__':

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(2)

    result = sock.connect_ex((HOST,PORT))

    if result == 0:
        print(f'Address {HOST}:{PORT} is already in use.')
    else:
        server_address = (HOST, PORT)
        server = HTTPServer(server_address, GetHandler)
        print (f'Starting server at http://{HOST}:{PORT}')
        server.serve_forever()

"""

        if action == "add_requests_and_calculate_all":
            response = controller.add_requests_and_calculate_all(post_data)

        elif action == "calculate_all":
            # MAIN: run all calculations
            response = controller.act()

        elif action == "load_prediction":
            # MAIN: look into prediction table and return prepared response
            response = controller.load_data(post_data)

        elif action == "load_times_to_fill":
            # MAIN: look into prediction table and return times to fill 90/100
            response = controller.load_data_for_capacity_table(post_data)

        elif action == "requests_to_database":
            # MAIN
            if type(post_data["request"]) == list:
                response = controller.more_requests_to_database(post_data["request"])
            else:
                response = {"message": "check post body structure"}
        #
        # Rest is for testing and development
        #
        elif action == "calculate_now":
            # for testing: calculates for one request and returns prediction
            response = controller.process_one_request(post_data)

        elif action == "do_one_and_save":
            # for testing: make one calculation
            controller.process_one_and_save(post_data)

        elif action == "request_to_database":
            response = controller.request_to_database(post_data)

        elif action == "return_data_for_request":
            # for testing: data used for calculation
            response = controller.data_used_for_request(post_data)

        elif action == "drop_prediction_table":
            response = controller.drop_prediction_table()

        elif action == "process_series":
            response = controller.run_all_processings(post_data["request"])

        else:
            exception_message = "Exception! Unknown action!"
            print(exception_message)
            response = {"message": exception_message}
            resp_code = 501
"""


"""
        CALCULATE + RETURN
        calculate_now   -> calculate_and_return + request

        CALCULATE + SAVE
        calculate_all   -> calculate_all
        do_one_and_save -> calculate_one + request

        RETURN
        return_data_for_request -> load_used_data       + request
        times_to_fill           -> load_times_to_fill   + request
        load_data               -> load_prediction      + request

        TABLE CONTROL
        delete_predictions
        show_state

        structure: 
        action:
            expected request structure
            respone function
            description fo info


            "action": {
                "response": controller.function,
                "input": True
                "description": ""
            },

        action_tree = {
            "add_requests_and_calculate_all": {
                "response": controller.add_requests_and_calculate_all,
                "description": "Main call. Uses provided list of requests"
            },
            "calculate_all": {
                "response": controller.act, # this could be a problem: no request given
                "description": ""
            },
            "load_prediction": {
                "response": controller.load_data,
                "description": ""
            },
        }
"""
