"""
USAGE:
- call only function execute_command(...), returns prediction dictionary

READY:
- execute command + test args
- specific command run

TODO: check conditions on data size

Low priority:
TODO: external methods: add parallel timer to exit
TODO: note: percentage growth over last n days - average vs cumulative (CPU)
TODO: standard model object
TODO: with standard model object: try out computations from params + save params
TODO: different trend computation
TODO: steps_in_day: from external module
TODO: command decision: ifs to dict

"""
import numpy as np
import pandas as pd

import warnings
import re
from dateutil import parser
import datetime


import model.tools as tools
import model.models as models
import model.seasons as seasons

import utility.requestwork as rwk
import utility.postprocess as postprocess
import measure.timing as timing

from utility.logger import log


steps_in_day = {
    "1m":  int(24*60),
    "5m":  int(24*12),
    "1h":  int(24),
    "6h":  int(4),
    "1d":  int(1)
}


def execute_command(command,
                    df,
                    maximal_value,
                    frequency="5m"):
    """
    command: decision on model set selection
    df: dataframe[index = date, Values = values]
    maximal_value: value used for postprocessing - find days to fill
    frequency: step in data in time

    """
    log("Execute command")

    data = df["Value"].to_numpy(dtype=float)

    try:
        # large gap handling
        valid_map = df["valid_map"].to_numpy()
    except:
        valid_map = []

    #
    # TIME
    #
    ## TODO: compare with now
    last_time = str(df.index.max())

    date = parser.parse(last_time)

    unix_time = date.timestamp()

    log("Last time: DATE:     ", last_time)
    log("Last time: UNIX:     ", unix_time)

    #
    # CHECKS
    #
    if not rwk.command_is_implemented(command):
        return {}
    if not rwk.frequency_is_implemented(command, frequency):
        return {}

    if len(data) == 0:
        log("NO DATA!")
        log("Return NONE!")
        return None

    #
    # DECISION TREE
    #
    t = timing.Timer(f"ANALYZER - prediction")
    t.start("Compute")

    if ( command == "predict_capacity" ):
        log("Predict Capacity")
        prediction = predict_capacity(data, unix_time, frequency, valid_map)
        t.restart("postprocess")
        prediction = postprocess.postprocessing_for_capacity(prediction, maximal_value)
    elif ( command == "predict_cpu" ):
        log("Predict CPU")
        prediction = predict_cpu(data, unix_time, frequency, valid_map)
        t.restart("postprocess")
        prediction = postprocess.postprocessing_for_cpu(prediction, maximal_value, frequency)
    elif ( command == "predict_vm" ):
        log("Predict VM")
        prediction = predict_vm(data, unix_time, frequency, valid_map)
        t.restart("postprocess")
        prediction = postprocess.postprocessing_for_cpu(prediction, maximal_value, frequency)
    
    else:
        warnings.warn("UNKNOWN COMMAND")
        prediction = {}
        return prediction

    t.restart("Final")

    rwk.check_prediction_response_structure(prediction)

    t.stop()

    t.summary()

    return prediction

def prepare_weekly_extrapolation(
        input_data,
        period,
        shifting,
        method,
        steps_ahead=8,
        interpolation_alpha=0.9,
        extrapolation_alpha=0.9,
        interpolation_percentile=80,
        extrapolation_percentile=80):

    weekly_splits = seasons.prepare_splits(input_data, period)

    weekly_interpolation = seasons.get_fast_robust_seasonal_interpolation(
        input_data,
        period,
        shifting=shifting,
        method=method,
        percentile=interpolation_percentile,
        alpha=interpolation_alpha)

    weekly_extrapolation = seasons.prepare_extrapolation(
        weekly_interpolation,
        weekly_splits,
        steps_ahead=steps_ahead,
        method=method,
        shifting=shifting,
        percentile=extrapolation_percentile,
        alpha=extrapolation_alpha)

    return weekly_extrapolation


def prepare_weekly_extrapolation_wrap(
        name,
        data,
        day_steps,
        prediction_steps,
        start_time):

    interpolation_alpha = 0
    extrapolation_alpha = 0
    
    if name == "robust":
        shifting = "median"
        method = "median"
    elif name == "reactive-robust":
        shifting = "median"
        method = "exponential"
        interpolation_alpha = 0.5
        extrapolation_alpha = 0.9
    else:
        shifting = "median"
        method = "median"

    week_steps = 7 * day_steps
    #prediction_steps = 31*day_steps
    steps_ahead = int(prediction_steps/week_steps)+1
    weekly_extrapolation = prepare_weekly_extrapolation(
        data,
        week_steps,
        shifting=shifting,
        method=method,
        steps_ahead=steps_ahead,
        interpolation_alpha=interpolation_alpha,
        extrapolation_alpha=extrapolation_alpha,
        interpolation_percentile=80,
        extrapolation_percentile=80)

    prediction_sample = {}

    prediction_sample[f'{name}_weekly_extrapolation'] = {"prediction": weekly_extrapolation.tolist()[:prediction_steps]}
    prediction_sample[f'{name}_weekly_extrapolation']["name"] = f"Weekly extrapolation: {name} model"
    prediction_sample[f'{name}_weekly_extrapolation']['start_time'] = start_time
    prediction_sample[f'{name}_weekly_extrapolation']['start_steps'] = 0
    prediction_sample[f'{name}_weekly_extrapolation']['description'] = f'Extrapolation of typical weekly behaviour'

    return prediction_sample


def prepare_daily_extrapolation(
        name,
        input_data,
        day_steps,
        prediction_steps,
        start_time):

    interpolation_alpha = 0
    extrapolation_alpha = 0

    if name == "robust":
        shifting = "median"
        method = "median"
    elif name == "reactive-robust":
        shifting = "median"
        method = "exponential"
        interpolation_alpha = 0.5
        extrapolation_alpha = 0.5
    elif name == "highly-reactive-robust":
        shifting = "median"
        method = "exponential"
        interpolation_alpha = 0.8
        extrapolation_alpha = 0.8
    elif name == "short":
        shifting = "mean"
        method = "mean"
    elif name == "long":
        shifting = "mean"
        method = "mean"
    else:
        shifting = "median"
        method = "median"

    daily_splits = seasons.prepare_splits(input_data, day_steps)

    daily_interpolation = seasons.get_fast_robust_seasonal_interpolation(
        input_data,
        day_steps,
        shifting=shifting,
        method=method,
        percentile=80,
        alpha=interpolation_alpha)

    daily_extrapolation = seasons.prepare_extrapolation(
        daily_interpolation,
        daily_splits,
        steps_ahead=40,
        method=method,
        shifting=shifting,
        percentile=80,
        alpha=extrapolation_alpha)

    return daily_extrapolation


def prepare_daily_extrapolation_wrap(
        name,
        data,
        day_steps,
        prediction_steps,
        start_time):

    daily_extrapolation = prepare_daily_extrapolation(
        name,
        data,
        day_steps,
        prediction_steps,
        start_time)
    prediction_steps = 31*day_steps

    prediction_sample = {}

    prediction_sample[f'{name}_ext_daily_averages'] = {"prediction": daily_extrapolation.tolist()[:prediction_steps]}
    prediction_sample[f'{name}_ext_daily_averages']["name"] = f"Daily extrapolation: {name} model"
    prediction_sample[f'{name}_ext_daily_averages']['start_time'] = start_time
    prediction_sample[f'{name}_ext_daily_averages']['start_steps'] = 0
    prediction_sample[f'{name}_ext_daily_averages']['description'] = f'Extrapolation of typical daily behaviour'

    return prediction_sample


def prepare_trend(name, predictor_days, frequency, last_time_in_data, input_data, prediction_steps, prediction_label_name="", valid_map=[]):

    _series = models.Series(input_data)
    if len(valid_map) > 0:
        _series.extend_valid_map(valid_map)
    _series.make_trendline(steps_ahead = prediction_steps)

    trend = _series.trendline

    prediction_sample = {}

    prediction_sample[name] = {"prediction": trend.tolist()}
    prediction_sample[name]['name'] = f'{prediction_label_name} daily trend'
    prediction_sample[name]['start_time'] = int( last_time_in_data - len(input_data) / steps_in_day[frequency] * 86400 )#predictor_days * 86400
    prediction_sample[name]['start_steps'] = len(input_data)
    prediction_sample[name]['description'] = f'Extrapolation of {predictor_days}-day trend'

    #std_upper = np.ones_like(trend) * trend_std
    #std_lower = np.ones_like(trend) * trend_std


    #prediction_sample[name]['sigmas_upper'] = std_upper.tolist()
    #prediction_sample[name]['sigmas_lower'] = std_lower.tolist()

    prediction_sample[name]['start_value'] = _series.trendline_start_y
    #prediction_sample[name]['slope'] = _series.trendline_slope
    #prediction_sample[name]['std'] = trend_std
    #prediction_sample[name]['max_dif'] = trend_max_dif
    #prediction_sample[name]['min_dif'] = trend_min_dif

    return prediction_sample


def get_naive(data, prediction_steps, start_time):
    """Useful for testing"""

    prediction_sample = {}

    naive = np.full((len(data.tolist()) + prediction_steps), data[-1])

    prediction_sample['naive'] = {"series": naive.tolist()}
    prediction_sample['naive']["prediction"] = naive.tolist()[-prediction_steps:]
    prediction_sample['naive']["name"] = "Naive Testing"
    prediction_sample['naive']["description"] = "Naive Testing"
    prediction_sample['naive']["start_time"] = start_time
    prediction_sample['naive']["start_steps"] = 0

    return prediction_sample


def predict_cpu(data, start_time, multiply_key="5m", valid_map=[]):
    """
    Implemented:
    Trends: short, medium, long
    Extrapolations: daily, weekly, weekly robust-reactive

    """
    cpu_timer = timing.Timer("CPU")
    cpu_timer.start("Preparations")

    prediction_data = {}

    # short/medium/long intervals by frequency
    if multiply_key == '5m':
        short_interval_days     = 7
        medium_interval_days    = 14
        long_interval_days      = 30

    elif multiply_key == '6h':
        short_interval_days     = 30
        medium_interval_days    = 60
        long_interval_days      = 365
    
    # future steps
    prediction_steps        = long_interval_days    * steps_in_day[multiply_key]
    # history steps
    short_interval_steps    = short_interval_days   * steps_in_day[multiply_key]
    medium_interval_steps   = medium_interval_days  * steps_in_day[multiply_key]
    long_interval_steps     = long_interval_days    * steps_in_day[multiply_key]

    # start with daily period
    period = steps_in_day[multiply_key]

    #
    # NAIVE for tests
    #
    #prediction_sample = get_naive(data,prediction_steps,start_time)
    #prediction_data.update(prediction_sample)

    #
    # PEAKS
    # TODO: think about more robust trend e.g. trend of 95 percentiles
    #

    day_steps = int(steps_in_day[multiply_key])


    if len(data) > steps_in_day[multiply_key] and multiply_key == "5m":
        cpu_timer.restart("Peaks")

        day_splits = seasons.prepare_splits(data, day_steps)
        daily_peaked_data = np.ravel(np.array([np.full_like(_split, np.nanmax(_split)) for _split in day_splits]))

        # TODO: better fix
        peak_series = models.Series(daily_peaked_data)
        peak_series.make_trendline(steps_ahead = prediction_steps)

        trend_peaks = peak_series.trendline

        cpu_timer.restart("PREPARE DICT")

        prediction_data['trend_peaks'] = {"prediction":trend_peaks.tolist()[-prediction_steps:]}
        prediction_data['trend_peaks']['name'] = "Trend of peaks"
        prediction_data['trend_peaks']['start_steps'] = 0
        prediction_data['trend_peaks']['start_time'] = start_time
        prediction_data['trend_peaks']['description'] = 'Extrapolation of daily peaks trend'

    #
    # TRENDS
    #
    # Default option: prediction year-year and month-month
    # TODO: add combined prediction = short term development
    #       takes into account long term trend expectation
    cpu_timer.stop()
    cpu_timer.start("Trends")


    if (len(data)>prediction_steps):
        # capacity graph: year-year
        data = data[-prediction_steps:]

    series = models.Series(data)
    series.make_trendline(steps_ahead = prediction_steps)
    all_time_slope = series.trendline_slope

    all_time_steps = len(data)
    all_time_interval = int(all_time_steps / steps_in_day[multiply_key])

    if (len(data) >= short_interval_steps):

        prediction_sample = prepare_trend(
            name='short_trend',
            predictor_days=short_interval_days,
            frequency=multiply_key,
            last_time_in_data=start_time,
            input_data=data[-short_interval_steps:],
            prediction_steps=prediction_steps,
            prediction_label_name="Short",
            valid_map=valid_map)

        prediction_data.update(prediction_sample)

    if (len(data) >= medium_interval_steps):

        prediction_sample = prepare_trend(
            name='medium_trend',
            predictor_days=medium_interval_days,
            frequency=multiply_key,
            last_time_in_data=start_time,
            input_data=data[-medium_interval_steps:],
            prediction_steps=prediction_steps,
            prediction_label_name="Medium",
            valid_map=valid_map)

        prediction_data.update(prediction_sample)

    if (len(data) >= long_interval_steps):

        prediction_sample = prepare_trend(
            name='long_trend',
            predictor_days=long_interval_days,
            frequency=multiply_key,
            last_time_in_data=start_time,
            input_data=data[-long_interval_steps:],
            prediction_steps=prediction_steps,
            prediction_label_name="Long",
            valid_map=valid_map)

        prediction_data.update(prediction_sample)

    if (len(data) < short_interval_steps
        or len(data) < medium_interval_steps
        or len(data) < long_interval_steps):

        prediction_sample = prepare_trend(
            name='trend',
            predictor_days=all_time_interval,
            frequency=multiply_key,
            last_time_in_data=start_time,
            input_data=data,
            prediction_steps=prediction_steps,
            prediction_label_name="Longest available",
            valid_map=valid_map)

        prediction_data.update(prediction_sample)


    #
    # EXTRAPOLATIONS: DAILY/WEEKLY MODELS
    #
    if multiply_key == "6h":
        cpu_timer.restart("Weekly extrapolation - reactive-robust")

        prediction_sample = {}
        if len(data) > day_steps*15: # at least 2 weeks of data
            prediction_sample = prepare_weekly_extrapolation_wrap(
                name="reactive-robust",
                data=data,
                day_steps=period,
                prediction_steps=365*day_steps,
                start_time=start_time)

            prediction_data.update(prediction_sample)

    if ((multiply_key == "5m") and (len(data) > day_steps*4)):
        # TODO: change structure
        datas = data[int(len(data)/2):]

        cpu_timer.restart("Daily extrapolation - short")

        prediction_sample = {}

        prediction_sample = prepare_daily_extrapolation_wrap(
            name="short",
            data=datas,
            day_steps=period,
            prediction_steps=31*day_steps,
            start_time=start_time)

        prediction_data.update(prediction_sample)

        cpu_timer.restart("Daily extrapolation - long")

        prediction_sample = {}

        prediction_sample = prepare_daily_extrapolation_wrap(
            name="long",
            data=data,
            day_steps=period,
            prediction_steps=31*day_steps,
            start_time=start_time)

        prediction_data.update(prediction_sample)

        cpu_timer.restart("Daily extrapolation - robust")

        prediction_sample = {}
        prediction_sample = prepare_daily_extrapolation_wrap(
            name="robust",
            data=data,
            day_steps=period,
            prediction_steps=31*day_steps,
            start_time=start_time)

        prediction_data.update(prediction_sample)

        cpu_timer.restart("Daily extrapolation - reactive-robust")

        prediction_sample = {}
        prediction_sample = prepare_daily_extrapolation_wrap(
            name="reactive-robust",
            data=data,
            day_steps=period,
            prediction_steps=31*day_steps,
            start_time=start_time)

        prediction_data.update(prediction_sample)

        cpu_timer.restart("Weekly extrapolation - robust")

        prediction_sample = {}
        if len(data) > day_steps*15:
            prediction_sample = prepare_weekly_extrapolation_wrap(
                name="robust",
                data=data,
                day_steps=period,
                prediction_steps=31*day_steps,
                start_time=start_time)

            prediction_data.update(prediction_sample)

            prediction_sample = prepare_weekly_extrapolation_wrap(
                name="reactive-robust",
                data=data,
                day_steps=period,
                prediction_steps=31*day_steps,
                start_time=start_time)
    
            prediction_data.update(prediction_sample)

    cpu_timer.stop()
    cpu_timer.summary()

    return prediction_data


def predict_capacity(data, start_time, multiply_key="1d", valid_map=[]):
    # Implemented:
    # Trends: short, medium, long, clean
    #         monthly extrapolation
    # TODO: cleaned extrapolation
    cap_timer = timing.Timer("CAPACITY")
    cap_timer.start("Preparations")

    short_interval_days     = 30
    medium_interval_days    = 60
    long_interval_days      = 365
    prediction_steps        = 365

    short_interval_steps    = short_interval_days   * steps_in_day[multiply_key]
    medium_interval_steps   = medium_interval_days  * steps_in_day[multiply_key]
    long_interval_steps     = long_interval_days    * steps_in_day[multiply_key]

    # start with daily period
    period              = steps_in_day[multiply_key]
    prediction_steps   *= steps_in_day[multiply_key]

    prediction_data = {}


    #
    # NAIVE for tests
    #
    #prediction_sample = get_naive(data,prediction_steps,start_time)
    #prediction_data.update(prediction_sample)


    #
    # TRENDS
    #
    cap_timer.restart("Trend")

    if (len(data) >= short_interval_steps):

        prediction_sample = prepare_trend(
            name='short_trend',
            predictor_days=short_interval_days,
            frequency=multiply_key,
            last_time_in_data=start_time,
            input_data=data[-short_interval_steps:],
            prediction_steps=prediction_steps,
            prediction_label_name="Short",
            valid_map=valid_map)

        prediction_data.update(prediction_sample)

    if (len(data) >= medium_interval_steps):

        prediction_sample = prepare_trend(
            name='medium_trend',
            predictor_days=medium_interval_days,
            frequency=multiply_key,
            last_time_in_data=start_time,
            input_data=data[-medium_interval_steps:],
            prediction_steps=prediction_steps,
            prediction_label_name="Medium",
            valid_map=valid_map)

        prediction_data.update(prediction_sample)

    if (len(data) >= long_interval_steps):

        prediction_sample = prepare_trend(
            name='long_trend',
            predictor_days=long_interval_days,
            frequency=multiply_key,
            last_time_in_data=start_time,
            input_data=data[-long_interval_steps:],
            prediction_steps=prediction_steps,
            prediction_label_name="Long",
            valid_map=valid_map)

        prediction_data.update(prediction_sample)

    if (len(data) < short_interval_steps
        or len(data) < medium_interval_steps
        or len(data) < long_interval_steps):

        all_time_steps = len(data)
        all_time_interval = int(all_time_steps / steps_in_day[multiply_key])

        prediction_sample = prepare_trend(
            name='trend',
            predictor_days=all_time_interval,
            frequency=multiply_key,
            last_time_in_data=start_time,
            input_data=data,
            prediction_steps=prediction_steps,
            prediction_label_name="Longest available",
            valid_map=valid_map)

        prediction_data.update(prediction_sample)


    #
    # CLEAN TRENDLINE
    #
    cap_timer.restart("Clean trend")

    series = models.Series(data)
    series.make_differenced_series()

    series.differenced.make_differenced_series()
    series.differenced.make_clean()

    series.differenced.clean.make_integrated_series()
    series.differenced.clean.make_clean()

    series.clean = series.differenced.clean.integrated

    series.differenced.clean.integrated.make_trendline(steps_ahead=prediction_steps)

    cleaned_trendline = series.differenced.clean.integrated.trendline
    cleaned_trendline +=  data[0] - cleaned_trendline[len(series.series)]
    cleaned_trendline = tools.continuous_concatenation(data, cleaned_trendline)

    cleaned_trendline_pred = series.differenced.clean.integrated.trendline
    cleaned_trendline_pred +=  data[-1] - cleaned_trendline_pred[0]

    prediction_data['once_cleaned_trend'] = {"prediction": cleaned_trendline_pred.tolist()[:prediction_steps]}
    prediction_data['once_cleaned_trend']['start_time'] = start_time
    prediction_data['once_cleaned_trend']['start_steps'] = 0
    prediction_data['once_cleaned_trend']['name'] = "Clean trend"
    prediction_data['once_cleaned_trend']['description'] = "Trend of data without statistical extremes"

    #prediction_data['once_cleaned_trend']['std'] = cleaned_trend_std
    #prediction_data['once_cleaned_trend']['max_dif'] = cleaned_trend_max_dif
    #prediction_data['once_cleaned_trend']['min_dif'] = cleaned_trend_min_dif

    #
    # SEASONAL DAILY AVERAGES
    #
    # TODO: add robust and reactive option / make one composed (more options + more periods) model
    #
    period = 30

    if len(data) > period + 1:
        # TODO: try to evaluate PACF of detrended data
        #try:
        #    period = seasons.get_expected_period(data)
        #except:
        #    period = 30
        #
        #if period < 10:
        #    period = 30
        cap_timer.restart("Prepare seasons")

        series = models.Series(data)
        series.make_differenced_series()
        series.differenced.make_clean()
        series.differenced.clean.make_integrated_series()
        series.clean = series.differenced.clean.integrated

        clean_data = series.differenced.clean.integrated.series

        #period = seasons.get_expected_period(clean_data)

        cycle_position = len(clean_data) % period
        data_ready_to_split = np.array(clean_data[cycle_position:])
        split_points = np.arange(1, int(data_ready_to_split.size / period +1 ))*int(period)
        splits = np.split(data_ready_to_split, np.array(split_points))

        #splits_centered = seasons.center_seasons_by_start(splits)
        splits_centered = splits
        # each split should be centered around common mean
        mean_splits, sigmas, sigmas_upper, sigmas_lower = seasons.get_seasonal_means_and_sigmas(splits_centered, period)

        num_of_concatenations = int((prediction_steps/period)+2)

        concat_result = mean_splits

        for con_num in range(num_of_concatenations):
            concat_result = tools.continuous_concatenation(concat_result, mean_splits)

        mean_splits_prediction = concat_result.copy()
        #mean_splits_prediction = mean_splits_prediction - np.mean(clean_data) + np.mean(data)
        # Shift for now:
        mean_splits_prediction = mean_splits_prediction - mean_splits_prediction[0] + data[-1]

        prediction_data['mean_splits'] = {"prediction": mean_splits_prediction.tolist()[:prediction_steps]}
        prediction_data['mean_splits']['name'] = f'Expected monthly behaviour'
        prediction_data['mean_splits']['start_time'] = start_time
        prediction_data['mean_splits']['start_steps'] = 0
        prediction_data['mean_splits']['description'] = f'Extrapolation of monthly behaviour'

    else:
        prediction_data['mean_splits'] = {"prediction": [data[-1]]}
        prediction_data['mean_splits']['name'] = f'Expected monthly behaviour'
        prediction_data['mean_splits']['start_time'] = start_time
        prediction_data['mean_splits']['start_steps'] = 0
        prediction_data['mean_splits']['description'] = f'Extrapolation of monthly behaviour not computed: requires at least 1 month of data'


    cap_timer.stop()
    cap_timer.summary()

    return prediction_data


def predict_vm(data, start_time, multiply_key="5m", valid_map=[]):
    """
    Implemented:
    Trends: short, medium, long
    """
    cpu_timer = timing.Timer("VM")
    cpu_timer.start("Preparations")

    prediction_data = {}

    # short/medium/long intervals by frequency
    if multiply_key == '5m':
        short_interval_days     = 7
        medium_interval_days    = 14
        long_interval_days      = 30

    elif multiply_key == '6h':
        short_interval_days     = 30
        medium_interval_days    = 60
        long_interval_days      = 365

    # m-m/year-year
    prediction_steps        = long_interval_days * steps_in_day[multiply_key]

    short_interval_steps    = short_interval_days   * steps_in_day[multiply_key]
    medium_interval_steps   = medium_interval_days  * steps_in_day[multiply_key]
    long_interval_steps     = long_interval_days    * steps_in_day[multiply_key]

    period    = steps_in_day[multiply_key]

    #
    # TRENDS
    #
    # Default option: prediction year-year and month-month
    # TODO: add combined prediction = short term development
    #       takes into account long term trend expectation
    cpu_timer.stop()
    cpu_timer.start("Trends")


    if (len(data)>prediction_steps):
        # capacity graph: year-year
        data = data[-prediction_steps:]

    series = models.Series(data)
    series.make_trendline(steps_ahead = prediction_steps)
    all_time_slope = series.trendline_slope

    all_time_steps = len(data)
    all_time_interval = int(all_time_steps / steps_in_day[multiply_key])

    if (len(data) >= short_interval_steps):

        prediction_sample = prepare_trend(
            name='short_trend',
            predictor_days=short_interval_days,
            frequency=multiply_key,
            last_time_in_data=start_time,
            input_data=data[-short_interval_steps:],
            prediction_steps=prediction_steps,
            prediction_label_name="Short",
            valid_map=valid_map)

        prediction_data.update(prediction_sample)

    if (len(data) >= medium_interval_steps):

        prediction_sample = prepare_trend(
            name='medium_trend',
            predictor_days=medium_interval_days,
            frequency=multiply_key,
            last_time_in_data=start_time,
            input_data=data[-medium_interval_steps:],
            prediction_steps=prediction_steps,
            prediction_label_name="Medium",
            valid_map=valid_map)

        prediction_data.update(prediction_sample)

    if (len(data) >= long_interval_steps):

        prediction_sample = prepare_trend(
            name='long_trend',
            predictor_days=long_interval_days,
            frequency=multiply_key,
            last_time_in_data=start_time,
            input_data=data[-long_interval_steps:],
            prediction_steps=prediction_steps,
            prediction_label_name="Long",
            valid_map=valid_map)

        prediction_data.update(prediction_sample)

    if (len(data) < short_interval_steps
        or len(data) < medium_interval_steps
        or len(data) < long_interval_steps):

        prediction_sample = prepare_trend(
            name='trend',
            predictor_days=all_time_interval,
            frequency=multiply_key,
            last_time_in_data=start_time,
            input_data=data,
            prediction_steps=prediction_steps,
            prediction_label_name="Longest available",
            valid_map=valid_map)

        prediction_data.update(prediction_sample)

    return prediction_data

"""
def prepare_arima(input_data, prediction_steps, start_time, arima_order=(3,1,0)):
    import statsmodels.api as sm

    model = sm.tsa.arima.ARIMA(input_data, order=arima_order)
    results = model.fit()
    arima_prediction = results.predict(len(input_data), len(input_data)+prediction_steps).tolist()
    arima_prediction = results.forecast(prediction_steps).tolist()

    log("TYPE ARIMA: ", type(arima_prediction))

    order_name = f"_{arima_order[0]}_{arima_order[1]}_{arima_order[2]}"

    prediction_sample = {}
    prediction_sample[f'arima{order_name}'] = {"prediction": arima_prediction}
    prediction_sample[f'arima{order_name}']["name"] = f"ARIMA {arima_order}"
    prediction_sample[f'arima{order_name}']["start_steps"] = 0
    prediction_sample[f'arima{order_name}']["start_time"] = start_time
    prediction_sample[f'arima{order_name}']['description'] = f'Prediction based on statistical ARIMA{arima_order} model'

    return prediction_sample

def prepare_sarima(input_data, prediction_steps, start_time, pdq=(1,0,0), PDQS=(0,0,1,120)):
    from statsmodels.tsa.statespace.sarimax import SARIMAX

    sarimax_pqd = pdq
    sarimax_PQDS = PDQS

    sarimax_data = input_data
    model = SARIMAX(sarimax_data,
                    order=sarimax_pqd,
                    seasonal_order=sarimax_PQDS,
                    method="Newton-CG",
                    enforce_stationarity=False )

    in_sample_prediction_steps = int(len(sarimax_data)/3)
    in_sample_prediction_steps = 0

    model_fit = model.fit(disp=False)
    #sarimax_prediction = model_fit.get_prediction(len(sarimax_data)-in_sample_prediction_steps, len(sarimax_data)+prediction_steps).predicted_mean.tolist()#predict(len(sarimax_data)-50, len(sarimax_data)+prediction_steps).tolist()
    sarimax_prediction = model_fit.forecast(prediction_steps).tolist()

    prediction_sample = {}

    prediction_sample['sarimax'] = {"prediction": sarimax_prediction}
    prediction_sample['sarimax']["name"] = f"SARIMAX{sarimax_pqd}{sarimax_PQDS}"
    prediction_sample['sarimax']["start_steps"] = in_sample_prediction_steps
    prediction_sample['sarimax']["start_time"] = start_time

    return prediction_sample
"""

"""
#
# CAPACITY: ARIMA
#
try:
    # TODO: low priority: add parallel timer to exit
    prediction_sample = prepare_arima(data, prediction_steps, start_time, arima_order=(3,1,1))
    prediction_data.update(prediction_sample)
except:
    log("ARIMA prediction failed")
try:
    # TODO: low priority: add parallel timer to exit
    prediction_sample = prepare_arima(data, prediction_steps, start_time, arima_order=(3,0,1))
    prediction_data.update(prediction_sample)
except:
    log("ARIMA prediction failed")
"""

"""
#
# CPU: ARIMA
#
if multiply_key == "6h":
    try:
        prediction_sample = prepare_arima(data, prediction_steps, start_time, arima_order=(4,2,2))
        prediction_data.update(prediction_sample)
    except:
        log("ARIMA prediction failed")

    try:
        prediction_sample = prepare_arima(data, prediction_steps, start_time, arima_order=(21,1,1))
        prediction_data.update(prediction_sample)
    except:
        log("ARIMA prediction failed")

    #try:
    #    prediction_sample = prepare_sarima(data, prediction_steps, start_time, pdq=(3,1,0), PDQS=(1,0,1,21))
    #    prediction_data.update(prediction_sample)
    #except:
    #    log("SARIMA prediction failed")
"""


"""
def prepare_auto_arima(data, prediction_steps, start_time):
    import pmdarima as pm

    y = np.array(data)

    model = pm.auto_arima(y, seasonal=True, m=30)

    forecasts = model.predict(prediction_steps)

    prediction_sample = {}

    prediction_sample['sarimax'] = {"prediction": forecasts}
    prediction_sample['sarimax']["name"] = f"Auto ARIMA"
    prediction_sample['sarimax']["start_steps"] = 0
    prediction_sample['sarimax']["start_time"] = start_time

    return prediction_sample
"""
