
"""
dbwork includes all operations needed to interact with database.

Basic operations
    - metrics: 
        - load data
    - prediction table:
        - load data
        - load times to fill 90/100
        - save data

TODO: check login validity

TODO: check: are saves ok?
TODO: show sizes: table/row...
TODO: on upgrade (?) prediction table autodelete instead of migrations
TODO: use saved time to compute times_to_fill
TODO: do not remove response when adding new request to table
TODO: add tests

Lower priority
TODO: request-response table interface - object? - CREATE UPDATE DELETE READ
TODO: optimize database times

"""
import pandas as pd
import re
import json
import os

from dateutil import parser
import time
import datetime

import sqlalchemy as db
from sqlalchemy.sql import select, insert, update, exists, and_
from sqlalchemy.orm import Session
from sqlalchemy import Table, Column, Integer, String, DateTime
from sqlalchemy import JSON

import utility.requestwork as rwk
import utility.postprocess as postprocess

import measure.timing as timing
from utility.logger import log

import urllib.parse

prediction_schema = "public"
prediction_table_name = 'prediction'

def get_db_login():

    db_name = os.getenv("DB_DATABASE")
    db_user = os.getenv("DB_USERNAME")
    db_host = os.getenv("DB_HOST")
    db_pwd  = os.getenv("DB_PASSWORD")
    db_port = os.getenv("DB_PORT")

    db_pwd = urllib.parse.quote_plus(db_pwd) # encode

    return (db_name, db_user, db_host, db_pwd, db_port)


def prepare_engine():
    log("Preparing engine")
    (db_name, db_user, db_host, db_pwd, db_port) = get_db_login()
    #timescaledb
    connection_string = f'timescaledb://{db_user}:{db_pwd}@{db_host}:{db_port}/{db_name}'

    # isolation_level="AUTOCOMMIT"
    engine = db.create_engine(connection_string)#, echo=True)
    # hint: use echo for logs
    #engine = db.create_engine(connection_string, echo=True)

    return engine


def start_engine_and_metadata():
    engine      = prepare_engine()
    meta_data   = db.MetaData(schema=prediction_schema, bind=engine)

    log("Loading sqlalchemy metadata")
    meta_data.reflect(extend_existing=True)

    return engine, meta_data


def prediction_table_loading_exception(engine):
    log("Problem during table load...")
    inspector = db.inspect(engine)

    log(f"Checking existence of {prediction_table_name}...")
    if not inspector.has_table(prediction_table_name, schema="public"):
        log(f"Table {prediction_table_name} does not exist.")
    else:
        log(f"Table {prediction_table_name} exists.")

def create_prediction_table(meta_data, engine):
    PredictionTable = Table(
        prediction_table_name,
        meta_data,
        Column('command', String),
        Column('frequency', String),
        Column('schema', String),
        Column('table', String),
        Column('column', String),
        Column('id', String),
        Column('max', String),
        Column('date', DateTime),
        Column('time_to_full', String),
        Column('time_to_less', String),
        Column('max', String),
        Column('response', JSON),
    )
    try:
        meta_data.create_all(engine)
    except:
        log("Cannot create {prediction_table_name}")

def update_prediction_table(request, response):
    # IN: request dictionary
    # response dictionary
    # 

    # if non existent => create
    # if non existent props => insert props
    # update response (json) where props=props
    # return ? state message
    log("--- UPDATE PREDICTION TABLE ---")
    time_now = datetime.datetime.now()

    time_to_less, time_to_full = postprocess.times_to_fill(response)

    (command, schema, table, metric,
     time_name, item_id, frequency, maximal_value) = rwk.read_request(request, show_read=True)

    maximal_value = str(maximal_value)
    #response_json = json.dumps(response)

    if command == "predict_capacity":
        response_json = response
    elif command == "predict_cpu":
        response_json = {}

    #engine, meta_data = start_engine_and_metadata()
    engine = prepare_engine()
    meta_data = db.MetaData()

    try:
        log("Load existing table from name")
        with engine.connect() as conn:
            PredictionTable = Table(f"{prediction_table_name}",
                                       meta_data,
                                       autoload_with=conn,
                                       schema=prediction_schema,
                                       extend_existing=True)
        #PredictionTable = meta_data.tables[f"{prediction_schema}.{prediction_table_name}"]
    except:
        log(f"Table {prediction_table_name} does not exist.")
        log("Creating prediction table...")
        create_prediction_table(meta_data, engine)

    with engine.connect() as conn:
        PredictionTable = Table(f"{prediction_table_name}",
                                   meta_data,
                                   autoload_with=conn,
                                   schema='public',
                                   extend_existing=True)


    #
    # INSERT STATEMENT
    #
    insert_statement = insert(PredictionTable).values(
        command   = command,
        frequency = frequency,
        schema    = schema,
        table     = table,
        column    = metric,
        id        = item_id,
        max       = maximal_value,
        response  = response_json,
        date      = time_now,
        time_to_full = time_to_full,
        time_to_less = time_to_less
    )

    #
    # UPDATE
    #
    update_statement = update(PredictionTable).values(
        response      = response_json,
        time_to_full  = time_to_full,
        time_to_less  = time_to_less,
        date          = time_now,
        ).where(
        and_(
            PredictionTable.c.command       == command,
            PredictionTable.c.frequency     == frequency,
            PredictionTable.c.schema        == schema,
            PredictionTable.c.table         == table,
            PredictionTable.c.column        == metric,
            PredictionTable.c.id            == item_id,
            PredictionTable.c.max           == maximal_value
        )
    )

    #
    # EXECUTE
    #
    row_exists = does_row_exist(request)
    
    with engine.connect() as conn:
        if row_exists:
            log("Updating prediction table")
            result = conn.execute(update_statement)
        else:
            log("Inserting into prediction table")
            result = conn.execute(insert_statement)

        log(result)

    engine.dispose()
    pass

def load_all_prediction_times_to_fill(id_metric):
    # global capacity table

    t = timing.Timer("DB")
    t.start("Engine")

    engine = prepare_engine()
    meta_data = db.MetaData()

    try:
        log("Loading table...")
        with engine.connect() as conn:
            PredictionTable = Table(f"{prediction_table_name}",
                                       meta_data,
                                       autoload_with=conn,
                                       schema='public',
                                       extend_existing=True)
    except:
        prediction_table_loading_exception(engine)
        return {}

    t.stop()
    t.start("Statement")

    select_statement = select(
            PredictionTable.c.id,
            PredictionTable.c.column,
            PredictionTable.c.time_to_less,
            PredictionTable.c.time_to_full
        ).where(
        and_(
            PredictionTable.c.command   == "predict_capacity",
        )
    )

    t.stop()
    t.start("Connection execution")

    print_counter = 0

    times_data = {}
    with engine.connect() as conn:
        try:
            log("Selecting prediction table")
            result = conn.execute(select_statement).all()
            for row in result:
                item_id      = row["id"]
                column       = row["column"]
                time_to_less = row["time_to_less"]
                time_to_full = row["time_to_full"]

                # TODO: better: filter during select
                if (item_id, column) in id_metric:

                    if print_counter < 5:
                        log(f"{print_counter} time_to_less {time_to_less}")
                        log(f"{print_counter} time_to_full {time_to_full}")
                        print_counter += 1

                    times_data[item_id] = {
                        "time_to_less": time_to_less,
                        "time_to_full": time_to_full,
                    }
        except:
            log("Problem during row selection...")
            log(select_statement)
            return 0, 0

    t.stop()
    t.summary()

    engine.dispose()

    return(times_data)


def load_prediction_times_to_fill(request):
    """
    Function reads prediction table rows that correspond to request.
    Return: saved times: time_to_less, time_to_full
    """
    engine = prepare_engine()
    meta_data = db.MetaData()

    try:
        log("Loading table...")
        with engine.connect() as conn:
            PredictionTable = Table(f"{prediction_table_name}",
                                       meta_data,
                                       autoload_with=conn,
                                       schema='public',
                                       extend_existing=True)
    except:
        prediction_table_loading_exception(engine)
        return {}

    (command, schema, table, metric,
     time_name, item_id, frequency, maximal_value) = rwk.read_request(request, show_read=True)

    maximal_value = str(maximal_value)

    select_statement = select(
            PredictionTable.c.time_to_less,
            PredictionTable.c.time_to_full
        ).where(
        and_(
            PredictionTable.c.command   == command,
            PredictionTable.c.frequency == frequency,
            PredictionTable.c.schema    == schema,
            PredictionTable.c.table     == table,
            PredictionTable.c.column    == metric,
            PredictionTable.c.id        == item_id,
            PredictionTable.c.max       == maximal_value,
        )
    )

    with engine.connect() as conn:
        try:
            log("Selecting prediction table")
            result = conn.execute(select_statement).all()[0]
        except:
            log("Problem during row selection...")
            print(select_statement)
            result = [0, 0]

    time_to_less, time_to_full = result[0], result[1]

    engine.dispose()

    return time_to_less, time_to_full

def load_prediction_response(request):
    engine = prepare_engine()
    meta_data = db.MetaData()

    try:
        log("Loading table...")
        with engine.connect() as conn:
            PredictionTable = Table(f"{prediction_table_name}",
                                       meta_data,
                                       autoload_with=conn,
                                       schema='public',
                                       extend_existing=True)
    except:
        prediction_table_loading_exception(engine)
        return {}

    (command, schema, table, metric,
     time_name, item_id, frequency, maximal_value) = rwk.read_request(request, show_read=True)

    maximal_value = str(maximal_value)

    select_statement = select(PredictionTable.c.response).where(
        and_(
            PredictionTable.c.command   == command,
            PredictionTable.c.frequency == frequency,
            PredictionTable.c.schema    == schema,
            PredictionTable.c.table     == table,
            PredictionTable.c.column    == metric,
            PredictionTable.c.id        == item_id,
            PredictionTable.c.max       == maximal_value,
        )
    )

    with engine.connect() as conn:
        try:
            log("Selecting prediction table")
            result = conn.execute(select_statement).all()[0][0]
            #log(type(result))
            #result = json.loads(result)
        except:
            log("Problem during row selection...")
            print(select_statement)
            result = {}

    engine.dispose()

    return result

def load_all_prediction_table_requests():
    """
    Function reads all prediction table requests and returns them as
    a list of requests.
    
    """
    requests = []
    
    engine = prepare_engine()
    meta_data = db.MetaData()

    try:
        log("Loading table...")
        with engine.connect() as conn:
            PredictionTable = Table(f"{prediction_table_name}",
                                       meta_data,
                                       autoload_with=conn,
                                       schema='public',
                                       extend_existing=True)
    except:
        prediction_table_loading_exception(engine)

        return []

    select_statement = select(
        PredictionTable.c.command,
        PredictionTable.c.frequency,
        PredictionTable.c.schema,
        PredictionTable.c.table,
        PredictionTable.c.column,
        PredictionTable.c.id,
        PredictionTable.c.max,
    )
    log("Selecting prediction table")

    with engine.connect() as conn:
        try:
            log("Selecting prediction table")
            result = conn.execute(select_statement).all()
        except:
            log("Problem during row selection...")
            print(select_statement)
            result = {}

    engine.dispose()

    # TODO: use requestwork
    # TODO: move out, not dbwork
    # the following might not be necessary..
    all_keys = ['command', 'frequency', 'schema', 'table', 'column', 'id', 'max']
    for res in result:
        request = {}
        for used_key in all_keys:
            request[used_key] = str(res[used_key])
        request['schema'] = 'metrics'
        requests.append(request)

    return requests

def delete_prediction_table():
    #engine = prepare_engine()
    #meta_data = db.MetaData()

    engine, meta_data = start_engine_and_metadata() # use for now, only for delete, takes a long time TODO

    with engine.connect() as conn:
        PredictionTable = Table(f"{prediction_table_name}",
                                   meta_data,
                                   autoload_with=conn,
                                   schema='public',
                                   extend_existing=True)
        PredictionTable.drop()
        meta_data.reflect(extend_existing=True)

    try:
        with engine.connect() as conn:
            PredictionTable = Table(f"{prediction_table_name}",
                                       meta_data,
                                       autoload_with=conn,
                                       schema='public',
                                       extend_existing=True)
        PredictionTable.drop()
        meta_data.reflect(extend_existing=True)
    except:
        log("No table found")

    engine.dispose()


#
# METRICS TIMESERIES
#
def metrics_data(table, metric, item_id, time_name="time"):
    """
    OUT: ts in 2xN matrix
    """
    data_timer = timing.Timer(f"Load metric - {table}")
    data_timer.start("Engine")

    engine = prepare_engine()
    meta_data = db.MetaData()

    data_timer.restart("ORM: loading table")

    try:
        log("Loading table...")
        with engine.connect() as conn:
            # NOTE: 23.69% - Timer: 0.2527 s - Load metric - metrics_vmware_cluster_5m - ORM: loading table
            #ViewTable = Table(f"{table}",
            #                  meta_data,
            #                  autoload_with=conn,
            #                  schema='metrics',
            #                  extend_existing=True)
            # NOTE: 10.71% - Timer: 0.0827 s - Load metric - metrics_vmware_cluster_5m - ORM: loading table
            ViewTable = Table(
                f"{table}", meta_data,
                Column(f'{time_name}', DateTime),
                Column(f'{metric}', String),
                Column('item_id', String),
                schema='metrics',
            )
    except:
        log("Problem during table load...")
        return {}

    data_timer.restart("Select - make query")

    #now = datetime.datetime.now()
    #start_date = now-datetime.timedelta(days=62)

    try:
        # sqlalchemy < 2.0
        query = select([ViewTable.c[time_name], ViewTable.c[metric]]).where(
            ViewTable.c.item_id   == item_id,
        )
    except:
        query = select(ViewTable.c[time_name, metric]).where(
            ViewTable.c.item_id   == item_id,
        )

    data_timer.restart("Select - exec or else")

    with engine.connect() as conn:
        try:
            log("Selecting prediction table")
            # TODO: optimize: first call takes long
            result = conn.execute(query).all()

            #data_timer.restart("Select - #2 exec or else")
            #result = conn.execute(query).all()

            # alternative:
            #result = psycopg2_execute_query(query)
            data_timer.restart("Prepare response")

            result = prepare_ts_from_response(result)
        except:
            log("Problem during row selection...")
            print(query)
            result = {}

    data_timer.restart("Engine disposal")

    engine.dispose()

    data_timer.stop()
    data_timer.summary()

    return result

#
# CHECKS
#
def does_row_exist(request):
    (command, schema, table, metric,
     time_name, item_id, frequency, maximal_value) = rwk.read_request(request, show_read=True)

    engine = prepare_engine()
    meta_data = db.MetaData()

    # TODO: if table exists

    with engine.connect() as conn:
        try:
            PredictionTable = Table(f"{prediction_table_name}",
                                    meta_data,
                                    autoload_with=conn,
                                    schema='public',
                                    extend_existing=True)
        except:
            prediction_table_loading_exception(engine)
            return False

    row_exists = False

    with Session(engine) as session:
        try:
            row_exists = session.query(exists().where(
                and_(
                    PredictionTable.c.command   == command,
                    PredictionTable.c.frequency == frequency,
                    PredictionTable.c.schema    == schema,
                    PredictionTable.c.table     == table,
                    PredictionTable.c.column    == metric,
                    PredictionTable.c.id        == item_id,
                )
            )).scalar()
        except:
            row_exists = False

    engine.dispose()

    return row_exists

def prepare_ts_from_response(res):
    # TODO: find more elegant solution
    ts = [[0]*2 for i in range(len(res))]

    for i,row in enumerate(res):

        date_format = f"{row[0]}"
        date = parser.parse(date_format)

        # 2023-09-11 05:00:00+00:00
        #date = re.sub("[\+\-]\d\d\:\d\d$", "", date)
        # 2023-09-11 05:00:00

        #unix_time = datetime.datetime.timestamp(date_format)

        ts[i][0] = date
        ts[i][1] = row[1]

        if (row[1]):
            pass
            #log("{} {}".format(row[0], row[1]))

    return ts

"""

def db_data_sqlalchemy2dataframe(schema, table, metric, item_id, time_name="time"):
    (db_name, db_user, db_host, db_pwd, db_port) = get_db_login()

    connection_string = f'timescaledb://{db_user}:{db_pwd}@{db_host}:{db_port}/{db_name}'

    #engine = create_engine(connection_string)
    #query = make_query(time_name, schema, table, metric, item_id)

    result = pd.read_sql_table(table_name=table, con=connection_string, schema=schema, columns=(time_name, 'item_id', metric)) 

    #TODO: change to use where
    # inefficient
    result = result.loc[result['item_id'] == item_id, [time_name, metric]]

    log(result)
    return result
"""

#
# Below: old psycopg2
#
"""
import psycopg2
import psycopg2.extras


def db_data(schema, table, metric, item_id, time_name="time"):
    #OUT: ts in 2xN matrix
    # Uses fully qualified name params to access 
    # column data in the database. 

    (db_name, db_user, db_host, db_pwd, db_port) = get_db_login()

    cursor = connect_to_db(db_name, db_user, db_host, db_pwd, db_port)

    query = make_query(time_name, schema, table, metric, item_id)

    cursor.execute(query)

    res = cursor.fetchall()

    ts = prepare_ts_from_response(res)

    return ts
def psycopg2_execute_query(query):
    (db_name, db_user, db_host, db_pwd, db_port) = get_db_login()

    cursor = psycopg2_connect_to_db(db_name, db_user, db_host, db_pwd, db_port)

    cursor.execute(query)

    res = cursor.fetchall()
    return res

def psycopg2_connect_to_db(db_name, db_user, db_host, db_pwd, db_port):
    conn = psycopg2.connect(database = db_name, 
                            user = db_user, 
                            host = db_host,
                            password = db_pwd,
                            port = db_port)


    cur = conn.cursor()
    return cur

def make_query( time_metric, schema, table, metric, item_id ):
    # !!! DO NOT COMMIT THIS TO WORKING PROJECT !!!
    # USE ORM !!!
    query = ''
    query += f"SELECT {time_metric}, \"{metric}\" "
    query += f"FROM {schema}.{table} "
    query += f"WHERE item_id='{item_id}'"

    return query


"""
