package Azure;

use strict;
use warnings;

use XML::Parser;
use XML::LibXML;
use HTTP::Request::Common;
use LWP;
use Data::Dumper;
use JSON;
use POSIX qw(strftime ceil);
use Date::Parse;
use Time::Local;
use Digest::MD5 qw(md5_hex);
use feature qw(switch);
no warnings qw( experimental::smartmatch );

sub new {
  my($self, $tenant, $client, $secret, $uuid) = @_;
  my $o = {};

  my $token = &getToken($tenant, $client, $secret);
  $o->{token} = $token;
  $o->{uuid} = $uuid;
  bless $o, $self;
  return $o;
}

sub getToken {
  my $tenant = shift;
  my $client = shift;
  my $secret = shift;

  my $json = JSON->new;

  my $ua = LWP::UserAgent->new(
    timeout => 30,
    ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 },
  );

  my %data;
  $data{grant_type} = "client_credentials";
  $data{client_id} = $client;
  $data{client_secret} = $secret;
  $data{resource} = "https://management.azure.com/";

  my $response = $ua->post( "https://login.microsoftonline.com/$tenant/oauth2/token", \%data );

  if ($response->is_success) {
    my $resp = $json->decode($response->content);
    #print Dumper($resp);

    return $resp->{access_token};
  }
  else {
    error_die("ERROR: Can't handle login request, bad credentials?");
    return 1;
  }

  return 0;
}

sub testToken {
  my $tenant = shift;
  my $client = shift;
  my $secret = shift;

  my $json = JSON->new;

  my $ua = LWP::UserAgent->new(
    timeout => 30,
    ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 },
  );

  my %data;
  $data{grant_type} = "client_credentials";
  $data{client_id} = $client;
  $data{client_secret} = $secret;
  $data{resource} = "https://management.azure.com/";

  my $response = $ua->post( "https://login.microsoftonline.com/$tenant/oauth2/token", \%data );

  if ($response->is_success) {
    my $resp = $json->decode($response->content);
    #print Dumper($resp);

    return $resp->{access_token};
  }
  else {
    return 0;
  }

}

sub getInstances {
  my($self, $resource, $subscription) = @_;

  my $json = JSON->new;

  my $ua = LWP::UserAgent->new(
    timeout => 30,
    ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 },
  );
  $ua->default_header(Authorization => 'Bearer '.$self->{token});

  my $response = $ua->get( "https://management.azure.com/subscriptions/$subscription/resourceGroups/$resource/providers/Microsoft.Compute/virtualMachines?api-version=2019-12-01");

  my $resp = $json->decode($response->content);

  #print Dumper($resp);

  return $resp;
}

sub getStorageInstances {
  my($self, $resource, $subscription) = @_;

  my $json = JSON->new;

  my $ua = LWP::UserAgent->new(
    timeout => 30,
    ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 },
  );
  $ua->default_header(Authorization => 'Bearer '.$self->{token});

  my $response = $ua->get( "https://management.azure.com/subscriptions/$subscription/resourceGroups/$resource/providers/Microsoft.Storage/storageAccounts?api-version=2022-09-01");

  my $resp = $json->decode($response->content);

  return $resp;
}

sub getSubscription {
  my ($self, $subscription) = @_;

  my $json = JSON->new;

  my $ua = LWP::UserAgent->new(
    timeout => 30,
    ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 },
  );
  $ua->default_header(Authorization => 'Bearer '.$self->{token});

  my $response = $ua->get( "https://management.azure.com/subscriptions/$subscription?api-version=2016-06-01");

  my $resp = $json->decode($response->content);

  return $resp->{displayName};

}

sub getResourceGroups {
  my ($self, $subscription) = @_;

  my $json = JSON->new;

  my $ua = LWP::UserAgent->new(
    timeout => 30,
    ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 },
  );
  $ua->default_header(Authorization => 'Bearer '.$self->{token});

  my $response = $ua->get( "https://management.azure.com/subscriptions/$subscription/resourcegroups?api-version=2020-06-01");

  my $resp = $json->decode($response->content);

  my $resources;
  for(@{$resp->{value}}) {
    my $res = $_;
    push(@{$resources}, $res->{name});
  }

  return $resources;
}

sub getDatabaseServers {
  my($self, $resource, $subscription) = @_;

  my $json = JSON->new;

  my $ua = LWP::UserAgent->new(
    timeout => 30,
    ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 },
  );
  $ua->default_header(Authorization => 'Bearer '.$self->{token});

  my $response = $ua->get( "https://management.azure.com/subscriptions/$subscription/resourceGroups/$resource/providers/Microsoft.Sql/servers?api-version=2019-06-01-preview");

  my $resp = $json->decode($response->content);

  #print Dumper($resp);

  return $resp;

}

sub getDatabases {
  my($self, $resource, $subscription, $server) = @_;

  my $json = JSON->new;

  my $ua = LWP::UserAgent->new(
    timeout => 30,
    ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 },
  );
  $ua->default_header(Authorization => 'Bearer '.$self->{token});

  my $response = $ua->get( "https://management.azure.com/subscriptions/$subscription/resourceGroups/$resource/providers/Microsoft.Sql/servers/$server/databases?api-version=2017-10-01-preview");

  my $resp = $json->decode($response->content);

  #print Dumper($resp);

  return $resp;

}

sub getVmList {

  my $self = shift;
  my $subscription = shift;

  my $json = JSON->new;

  my $ua = LWP::UserAgent->new(
    timeout => 30,
    ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 },
  );
  $ua->default_header(Authorization => 'Bearer '.$self->{token});

  my $response = $ua->get( "https://management.azure.com/subscriptions/$subscription/providers/Microsoft.Compute/skus?api-version=2021-07-01");

  my $vms = ();
  foreach(@{$json->decode($response->content)->{value}}){
    my $value = $_;
    my $location = $value->{locations}[0];

    $vms->{$location}->{$value->{name}}->{name} = $value->{name};
    $vms->{$location}->{$value->{name}}->{locations} = $value->{locations};
    $vms->{$location}->{$value->{name}}->{locationInfo} = $value->{locationInfo};
    $vms->{$location}->{$value->{name}}->{size} = $value->{size};
    $vms->{$location}->{$value->{name}}->{tier} = $value->{tier};
    $vms->{$location}->{$value->{name}}->{resourceType} = $value->{resourceType};
    $vms->{$location}->{$value->{name}}->{restrictions} = $value->{restrictions};
    $vms->{$location}->{$value->{name}}->{family} = $value->{family};

    foreach(@{$value->{capabilities}}){
      my $capability = $_;
      $vms->{$location}->{$value->{name}}->{capabilities}->{$capability->{name}} = $capability->{value};
    }
    
  }

  return $vms;
}

sub getAgentMetrics {
  my($self, $resource, $subscription, $vm) = @_;

  my $json = JSON->new;

  my $ua = LWP::UserAgent->new(
    timeout => 30,
    ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 },
  );
  $ua->default_header(Authorization => 'Bearer '.$self->{token});

  my $response = $ua->get( "https://management.azure.com/subscriptions/$subscription/resourceGroups/$resource/providers/Microsoft.Compute/virtualMachines/$vm/metricDefinitions?api-version=2014-04-01");

  my $resp = $json->decode($response->content);

  if (!defined $resp->{value} || scalar @{$resp->{value}} == 0) {
    if (defined $resp->{code}){
        warn $resp->{code};
        warn $resp->{message};
    }
    return 0;
  }

  my %data;

  for (@{$resp->{value}}) {
    my $metric = $_;
    my $time_grain_idx = 0;

    for(0 .. scalar $metric->{metricAvailabilities}){
      if($metric->{metricAvailabilities}->[$time_grain_idx]->{timeGrain} eq "PT1M"){
        last;
      }
      $time_grain_idx = $time_grain_idx + 1;
    }
    if (!defined $metric->{metricAvailabilities}->[$time_grain_idx]->{location}->{tableInfo}->[0]->{sasToken}) {
      next;
    }
    if ($metric->{name}->{value} ne "/builtin/memory/usedmemory" && $metric->{name}->{value} ne "\\Memory\\Available Bytes" && $metric->{name}->{value} ne "/builtin/memory/availablememory") {
      next;
    }

    my $i = 0;
    my $act_time = time();
    my $check_time = str2time($metric->{metricAvailabilities}->[0]->{location}->{tableInfo}->[$i]->{endTime});

    while ($check_time <  $act_time) {
      $i = $i+1;
      $check_time = str2time($metric->{metricAvailabilities}->[$time_grain_idx]->{location}->{tableInfo}->[$i]->{endTime});
    }

    #print "\nURL: $metric->{metricAvailabilities}->[0]->{location}->{tableEndpoint}";
    #print "\nTable: $metric->{metricAvailabilities}->[0]->{location}->{tableInfo}->[0]->{tableName}";
    #print "\nToken: $metric->{metricAvailabilities}->[0]->{location}->{tableInfo}->[0]->{sasToken}";

    if ($metric->{name}->{value} eq "/builtin/memory/usedmemory") {
      $data{usedMemory} = getAgentMetricRequest($metric->{metricAvailabilities}->[$time_grain_idx]->{location}->{tableEndpoint}, $metric->{metricAvailabilities}->[$time_grain_idx]->{location}->{tableInfo}->[$i]->{tableName}, $metric->{metricAvailabilities}->[$time_grain_idx]->{location}->{tableInfo}->[$i]->{sasToken}, "/builtin/memory/usedmemory", "usedMemory");
    } elsif ($metric->{name}->{value} eq "/builtin/memory/availablememory") {
      $data{freeMemory} = getAgentMetricRequest($metric->{metricAvailabilities}->[$time_grain_idx]->{location}->{tableEndpoint}, $metric->{metricAvailabilities}->[$time_grain_idx]->{location}->{tableInfo}->[$i]->{tableName}, $metric->{metricAvailabilities}->[$time_grain_idx]->{location}->{tableInfo}->[$i]->{sasToken}, "/builtin/memory/availablememory", "freeMemory");
    } elsif ($metric->{name}->{value} eq "\\Memory\\Available Bytes"){
      $data{freeMemory} = getAgentMetricRequest($metric->{metricAvailabilities}->[$time_grain_idx]->{location}->{tableEndpoint}, $metric->{metricAvailabilities}->[$time_grain_idx]->{location}->{tableInfo}->[$i]->{tableName}, $metric->{metricAvailabilities}->[$time_grain_idx]->{location}->{tableInfo}->[$i]->{sasToken}, "\\Memory\\Available Bytes", "freeMemory");
    }

  }

  return \%data;

}

sub getAgentMetricRequest {
  my ($url, $table, $sasToken, $metricName, $prettyName) = @_;
  my $json = JSON->new;

  my $ua = LWP::UserAgent->new(
    timeout => 30,
    ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 },
  );

  my %data;

  my $response = $ua->get( $url.$table.$sasToken."&\$filter=CounterName%20eq%20'".$metricName."'&\$top=25");

  my $dom = XML::LibXML->load_xml(string => $response->content);
  my $xc = XML::LibXML::XPathContext->new($dom);
  my @nodesValues = $xc->findnodes('//m:properties/d:Average');
  my @nodesTimestamps = $xc->findnodes('//m:properties/d:TIMESTAMP');

  my $actual = 10;
  my $time;
  my $value;
  for my $i (0..scalar @nodesTimestamps - 1){
    $time = $nodesTimestamps[$i]->to_literal;
    $time = str2time($time);
    $time = int($time-($time%60));
    $value = $nodesValues[$i]->to_literal;
    $data{$time} = $value;
    $actual = $actual+2;
    
  }

  #warn Dumper(\%data);

  return \%data;
}

sub getNetwork {
  my($self, $url) = @_;

  my $json = JSON->new;

  my $ua = LWP::UserAgent->new(
    timeout => 30,
    ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 },
  );
  $ua->default_header(Authorization => 'Bearer '.$self->{token});

  my $response = $ua->get( "https://management.azure.com$url/?api-version=2020-05-01");

  my $resp = $json->decode($response->content);

  #print Dumper($resp);

  return $resp;
}

sub getIpAdress {
  my($self, $url) = @_;

  my $json = JSON->new;

  my $ua = LWP::UserAgent->new(
    timeout => 30,
    ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 },
  );
  $ua->default_header(Authorization => 'Bearer '.$self->{token});

  my $response = $ua->get( "https://management.azure.com$url/?api-version=2020-05-01");

  my $resp = $json->decode($response->content);

  #print Dumper($resp);

  return $resp;
}

sub getInstanceView {
  my($self, $resource, $subscription, $vm) = @_;

  my $json = JSON->new;

  my $ua = LWP::UserAgent->new(
    timeout => 20,
    ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 },
  );
  $ua->default_header(Authorization => 'Bearer '.$self->{token});

  my $again = 0;
  my $response;
  my $resp; 

  eval {

    $response = $ua->get( "https://management.azure.com/subscriptions/$subscription/resourceGroups/$resource/providers/Microsoft.Compute/virtualMachines/$vm/instanceView?api-version=2019-12-01");
    $resp = $json->decode($response->content);
  }; if ($@) {
    
    $again = 1;

  }

  if ($again eq "1") {
    sleep(15);
    $response = $ua->get( "https://management.azure.com/subscriptions/$subscription/resourceGroups/$resource/providers/Microsoft.Compute/virtualMachines/$vm/instanceView?api-version=2019-12-01");
    $resp = $json->decode($response->content);
  }

  #print Dumper($resp);

  return $resp;
}

sub getMetrics {
  my($self, $resource, $vm, $subscription, $timestamp_from, $timestamp) = @_;

  my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = gmtime($timestamp_from);
  $year += 1900;
  $mon += 1;
  my $reformated = reformatTime({'month' => $mon, 'day' => $mday, 'hour' => $hour, 'min' => $min, 'sec' => $sec});
  my $start_time = "$year-$reformated->{month}-$reformated->{day}T$reformated->{hour}:$reformated->{min}:$reformated->{sec}";

  ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = gmtime($timestamp);
  $year += 1900;
  $mon += 1;
  $reformated = reformatTime({'month' => $mon, 'day' => $mday, 'hour' => $hour, 'min' => $min, 'sec' => $sec});
  my $end_time = "$year-$reformated->{month}-$reformated->{day}T$reformated->{hour}:$reformated->{min}:$reformated->{sec}";

  my $json = JSON->new;

  my $ua = LWP::UserAgent->new(
    timeout => 30,
    ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 },
  );
  $ua->default_header(Authorization => 'Bearer '.$self->{token});

  my $metrics = "Percentage CPU,Disk Read Bytes,Disk Write Bytes,Disk Read Operations/Sec,Disk Write Operations/Sec,Network In,Network Out";

  my $response = $ua->get( "https://management.azure.com/subscriptions/$subscription/resourceGroups/$resource/providers/Microsoft.Compute/virtualMachines/$vm/providers/microsoft.insights/metrics?metricnames=$metrics&timespan=".$start_time."Z/".$end_time."Z&api-version=2018-01-01");

  my $resp = $json->decode($response->content);
  
  return $resp;

}

sub getMetricsStorage {
  my($self, $resource, $storage, $subscription, $timestamp_from, $timestamp, $namespace) = @_;

    given($namespace){
        when ("blob")     {$namespace = "blobServices/default/";}
        when ("file")     {$namespace = "fileServices/default/";}
        when ("table")    {$namespace = "tableServices/default/";}
        when ("queue")    {$namespace = "queueServices/default/";}
        when ("account")  {$namespace = "";}
        default           {$namespace = "";}
    }

  my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = gmtime($timestamp_from);
  $year += 1900;
  $mon += 1;
  my $reformated = reformatTime({'month' => $mon, 'day' => $mday, 'hour' => $hour, 'min' => $min, 'sec' => $sec});
  my $start_time = "$year-$reformated->{month}-$reformated->{day}T$reformated->{hour}:$reformated->{min}:$reformated->{sec}";

  ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = gmtime($timestamp);
  $year += 1900;
  $mon += 1;
  $reformated = reformatTime({'month' => $mon, 'day' => $mday, 'hour' => $hour, 'min' => $min, 'sec' => $sec});
  my $end_time = "$year-$reformated->{month}-$reformated->{day}T$reformated->{hour}:$reformated->{min}:$reformated->{sec}";

  my $json = JSON->new;

  my $ua = LWP::UserAgent->new(
    timeout => 30,
    ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 },
  );
  $ua->default_header(Authorization => 'Bearer '.$self->{token});

  my $metrics = "Transactions,Ingress,Egress,SuccessServerLatency,SuccessE2ELatency,Availability";

  my $response = $ua->get( "https://management.azure.com/subscriptions/$subscription/resourceGroups/$resource/providers/Microsoft.Storage/storageAccounts/$storage/".$namespace."providers/microsoft.insights/metrics?api-version=2018-01-01&timespan=".$start_time."Z/".$end_time."&interval=PT1M&metricnames=$metrics");

  my $resp = $json->decode($response->content);
  
  return $resp;
}

sub getMetricsStorageHourGrain {
  my($self, $resource, $storage, $subscription, $timestamp_from, $timestamp, $namespace) = @_;


    my $metrics = "UsedCapacity";
    given($namespace){
        when ("blob")   {
                            $namespace = "blobServices/default/";
                            $metrics = "BlobCapacity,BlobCount,ContainerCount";
                        }
        when ("file")   {
                            $namespace = "fileServices/default/";
                            $metrics = "FileCapacity,FileCount,FileShareCount,FileShareSnapshotCount,FileShareSnapshotSize,FileShareCapacityQuota";
                        }
        when ("table")  {
                            $namespace = "tableServices/default/";
                            $metrics = "TableCapacity,TableCount,TableEntityCount";
                        }
        when ("queue")  {
                            $namespace = "queueServices/default/";
                            $metrics = "QueueCapacity,QueueCount,QueueMessageCount";
                        }
        when ("account"){
                            $namespace = "";
                            $metrics = "UsedCapacity";
                        }
        default         {
                            $namespace = "";
                            $metrics = "UsedCapacity";
                        }
    }
  $timestamp_from = $timestamp_from - 4800;
  my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = gmtime($timestamp_from);
  $year += 1900;
  $mon += 1;
  my $reformated = reformatTime({'month' => $mon, 'day' => $mday, 'hour' => $hour, 'min' => $min, 'sec' => $sec});
  my $start_time = "$year-$reformated->{month}-$reformated->{day}T$reformated->{hour}:$reformated->{min}:$reformated->{sec}";

  ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = gmtime($timestamp);
  $year += 1900;
  $mon += 1;
  $reformated = reformatTime({'month' => $mon, 'day' => $mday, 'hour' => $hour, 'min' => $min, 'sec' => $sec});
  my $end_time = "$year-$reformated->{month}-$reformated->{day}T$reformated->{hour}:$reformated->{min}:$reformated->{sec}";

  my $json = JSON->new;

  my $ua = LWP::UserAgent->new(
    timeout => 30,
    ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 },
  );
  $ua->default_header(Authorization => 'Bearer '.$self->{token});

  

  my $response = $ua->get( "https://management.azure.com/subscriptions/$subscription/resourceGroups/$resource/providers/Microsoft.Storage/storageAccounts/$storage/".$namespace."providers/microsoft.insights/metrics?api-version=2018-01-01&timespan=".$start_time."Z/".$end_time."&interval=PT1H&metricnames=$metrics");

  my $resp = $json->decode($response->content);
  
  return $resp;
}

sub getConfArch {
  my $self = shift;
  my $data = shift;

  my %conf;
  my $sub_count = 1;
  for (@{$data->{subscriptions}}) {
    my $subscription = $_;

    $conf{architecture}{subscriptions}{$subscription} = "Subscription $sub_count";
    

    my $resources = $self->getResourceGroups($subscription);
  
    for(@{$resources}){
      my $resource = $_;
      my $resource_id = md5_hex($resource.$self->{uuid});
      $conf{architecture}{resources}{$resource_id}{label} = $resource;

      my $instances = $self->getInstances($resource, $subscription);

      for(@{$instances->{value}}){

        my $instance = $_;
        my $instance_id = $instance->{properties}->{vmId};

        $conf{configuration}{subscriptions}{$subscription}{$resource_id}{vm}{$instance_id} = $instance->{name};

        $conf{architecture}{resources}{$resource_id}{vm}{$instance_id}{label} = $instance->{name};
        $conf{properties}{vm}{$instance_id}{location}       = $instance->{location};
        $conf{properties}{vm}{$instance_id}{instance_size}  = $instance->{properties}->{hardwareProfile}->{vmSize};
        $conf{properties}{vm}{$instance_id}{disk_size}      = $instance->{properties}->{storageProfile}->{osDisk}->{diskSizeGB};

        my $ip_object = $self->getIpAdress("/subscriptions/$subscription/resourceGroups/$resource/providers/Microsoft.Network/publicIPAddresses/$instance->{name}-ip");
        $conf{properties}{vm}{$instance_id}{ip_adress}      = $ip_object->{properties}->{ipAddress};
        
        my $statuses = $self->getInstanceView($resource, $subscription, $instance->{name});
        $conf{properties}{vm}{$instance_id}{os}             = "$statuses->{osName} ($statuses->{osVersion})";

        for(@{$statuses->{statuses}}){
          my $status = $_;
          
          if ($status->{code} eq 'PowerState/running'){
            $conf{properties}{vm}{$instance_id}{state}  = $status->{displayStatus};
            $conf{properties}{vm}{$instance_id}{powerStateCode}  = $status->{code};
            $conf{properties}{vm}{$instance_id}{powerStateLevel}  = $status->{level};
            $conf{properties}{vm}{$instance_id}{powerStateMessage} = $status->{displayStatus};
          }
          if ($status->{code} eq 'ProvisioningState/succeeded'){
            $conf{properties}{vm}{$instance_id}{provisioningStateCode}  = $status->{code};
            $conf{properties}{vm}{$instance_id}{provisioningStateLevel}  = $status->{level};
            $conf{properties}{vm}{$instance_id}{provisioningStateMessage}  = $status->{displayStatus};
          }
        }

      }

      my $storages = $self->getStorageInstances($resource, $subscription);

      for(@{$storages->{value}}){

        my $instance = $_;
        my $instance_id = md5_hex($instance->{id});

        $conf{configuration}{subscriptions}{$subscription}{$resource_id}{account}{$instance_id} = $instance->{name};
        $conf{architecture}{resources}{$resource_id}{storage}{$instance_id}{label} = $instance->{name};

        $conf{properties}{storage}{$instance_id}{location}            = $instance->{location};
        $conf{properties}{storage}{$instance_id}{publicNetworkAccess} = $instance->{properties}->{publicNetworkAccess};
        $conf{properties}{storage}{$instance_id}{minimumTlsVersion}   = $instance->{properties}->{minimumTlsVersion};
        $conf{properties}{storage}{$instance_id}{provisioningState}   = $instance->{properties}->{provisioningState};
        $conf{properties}{storage}{$instance_id}{performance}         = $instance->{sku}->{name};
        $conf{properties}{storage}{$instance_id}{kind}                = $instance->{kind};

      }

    }

  }

  return \%conf;

}

sub reformatTime {
  my ($time) = @_;
  my %data;
  foreach my $time_key (keys %{$time}) {
    $data{$time_key} = sprintf("%02d",$time->{$time_key});
  }
  return \%data;
}

sub error_die {
  my $message  = shift;
  my $act_time = localtime();
  print STDERR "$act_time: $message : $!\n";
  exit(1);
} 

1;
