use LWP;
use JSON;
use strict;
use warnings;
use Data::Dumper;
use FindBin;                 # locate this script
use lib "$FindBin::Bin/";
use Azure;
use Xormon;
use Date::Parse;
use Digest::MD5 qw(md5_hex);
use feature qw(switch);
no warnings qw( experimental::smartmatch );


$SIG{CHLD} = 'IGNORE';

use IO::Socket qw(AF_INET AF_UNIX SOCK_STREAM SHUT_WR);

my $DEFAULT_PORT = 3102;
my $port = Xormon::checkPrefixPort($DEFAULT_PORT);

my $server = IO::Socket->new(
    Domain => AF_INET,
    Type => SOCK_STREAM,
    Proto => 'tcp',
    LocalHost => '0.0.0.0',
    LocalPort => $port,
    ReusePort => 1,
    Listen => 5,
) || die "Can't open socket: $@";
Xormon::log("Service listening on port $port");

unless ( -d "./data") {
  mkdir( "./data", 0755 ) || warn( localtime() . ": Cannot mkdir ./data : $!" . __FILE__ . ':' . __LINE__ );
}

my $client;
while ($client = $server->accept()) {
    
    my $pid;
    unless ( defined( $pid = fork() ) ) {
        Xormon::error("Wait for fork");
        sleep 5;
    }

    if ($pid == 0) {

      # get information about a newly connected client
      my $client_address = $client->peerhost();
      my $client_port = $client->peerport();
      Xormon::log("Connection from $client_address:$client_port");

      # received data template 
      #
      # { 
      #   "jwt"  : "<backendToken>",
      #   "task" : "<taskName>"
      #   "data" : {
      #     <configuration_from_hostcfg>
      #   }
      # }

      # read up to 2048 characters from the connected client
      my $data = "";
      $client->recv($data, 2048);

      my $decoded = from_json($data);
      my $cfg = $decoded->{data};

      
      Xormon::log("Received task: ".$decoded->{task});
      my $response = '';
      if ($decoded->{task} eq "fetch") {

        $client->send("Task successfuly started");
        $client->shutdown(SHUT_WR);

        my $xormon = Xormon->new("$client_address", $decoded->{jwt}, 'azure', $cfg->{hostcfg_id}, $cfg->{hostalias});

        # target destinations check
        if (defined $decoded->{'destinations'}) {
            $xormon->setDestinations($decoded->{'destinations'});
        }

        $xormon->startJob($decoded->{task}, $cfg->{hostalias}, $$);


        # get metrics data and save cfg from technology
        my $status = process($xormon, $cfg);

        $xormon->endJob($decoded->{task}, $cfg->{hostalias}, $status, $$);

        Xormon::log("Task for ".$cfg->{hostalias}." is done");

       } elsif ($decoded->{task} eq "test") {
          my $cfg = $decoded->{data};
          $response = apitest($cfg);

          $client->send($response);
          $client->shutdown(SHUT_WR);
          

      } else {
          $response = "undefined task";

          $client->send($response);
          $client->shutdown(SHUT_WR);
      }

      exit();

    } else {
        $client->autoflush(1);    
    }
}

sub process {
    my $xormon  = shift;
    my $cfg     = shift;

    $cfg->{interval} = 60;
    my $perfData = api2json($xormon, $cfg);

    my $rc = $xormon->saveJSON2($perfData);

    my $status = "successful";
    if ($rc eq "201") {
        $status = "successful";
    }elsif ($rc eq "401") {
        $status = "unauthorized";
    } else {
        $status = "error";
    }
    return $status;
}

sub apitest {
  my $data = shift;

  my $azure = Azure->new($data->{tenant}, $data->{client}, $data->{secret});

  my @rows;
  for (@{$data->{subscriptions}}) {
    my $subscription = $_;
    my $resources = $azure->getResourceGroups($subscription);

    if (defined $resources) {
       my $res_sum = scalar @{$resources};
       push(@rows, "Number of resources under $subscription: $res_sum");
    }

  }

  if (scalar @rows >= 1) {
    return Xormon::conn_test(1, \@rows);
  } else {
    push(@rows, "Can not find any resources");
    return Xormon::conn_test(0, \@rows);
  }
}

sub saveConf {
  my $xormon = shift;
  my $conf = shift;
  my $cfg = shift;


  my @architecture;
  my $root_id = md5_hex($cfg->{hostalias});

    my %mgroup_architecture = (
        "item_id"    => $root_id,
        "label"      => $cfg->{hostalias},
        "hw_type"    => "azure",
        "subsystem"  => "mgroup",
        "parents"    => [],
        "hostcfg_id" => $cfg->{hostcfg_id}
    ); 

    push(@architecture, \%mgroup_architecture);

    foreach my $subscription (keys %{$conf->{configuration}->{subscriptions}}) {

        #subscription architecture
        my $subscription_id = $subscription;

        my %subscription_architecture = (
            "item_id"    => $subscription_id,
            "label"      => $conf->{architecture}->{subscriptions}->{$subscription},
            "hw_type"    => "azure",
            "subsystem"  => "subscription",
            "parents"    => [$root_id],
            "hostcfg_id" => $cfg->{hostcfg_id}
        );

        push(@architecture, \%subscription_architecture);

        foreach my $resource (keys %{$conf->{configuration}->{subscriptions}->{$subscription}}){
            my $resource_id = $resource;

            my %rgroup_architecture = (
                "item_id"    => $resource_id,
                "label"      => $conf->{architecture}->{resources}->{$resource_id}->{label},
                "hw_type"    => "azure",
                "subsystem"  => "rgroup",
                "parents"    => [$root_id, $subscription_id],
                "hostcfg_id" => $cfg->{hostcfg_id}
            );

            push(@architecture, \%rgroup_architecture);

            foreach my $type (keys %{$conf->{configuration}->{subscriptions}->{$subscription}->{$resource}}){
                foreach my $instance (keys %{$conf->{configuration}->{subscriptions}->{$subscription}->{$resource}->{$type}}){
                    my $instance_id = $instance;

                    my %sub_architecture = (
                        "item_id"    => $instance_id,
                        "label"      => $conf->{configuration}->{subscriptions}->{$subscription}->{$resource}->{$type}->{$instance},
                        "hw_type"    => "azure",
                        "subsystem"  => $type,
                        "parents"    => [$root_id, $subscription_id, $resource_id],
                        "hostcfg_id" => $cfg->{hostcfg_id}
                    );

                    push(@architecture, \%sub_architecture);
                }

            }
        }
    }

    $xormon->saveArchitecture(\@architecture);

    if(defined $conf->{properties}->{vm}){
      $xormon->saveConfJSON(\%{$conf->{properties}->{vm}});
    }
    if(defined $conf->{properties}->{storage}){
      $xormon->saveConfJSON(\%{$conf->{properties}->{storage}});
    }

}

sub fillTimestamps {

  my $values = shift;
  my $timeNow = shift;
  my @timeKeys = sort keys %{$values};
  my %newValues;
  my $size = scalar @timeKeys;
  for (my $i = 0; $i < $size; $i++){

    my $currentTmstmp = $timeKeys[$i];

    if($i + 1 >= $size){
      while($currentTmstmp +60 < $timeNow){
        $newValues{$currentTmstmp} = $values->{$timeKeys[$i]};
        $currentTmstmp = $currentTmstmp + 60;
      } 
    }else{
      while($currentTmstmp + 60 < $timeKeys[$i+1]){
        $newValues{$currentTmstmp} = $values->{$timeKeys[$i]};
        $currentTmstmp = $currentTmstmp + 60;
      } 
    }
  }

  return \%newValues;
}

sub api2json {
  my $xormon = shift;
  my $data = shift;

  my @storageServices = ("account", "blob", "file", "queue", "table");

  my $azure = Azure->new($data->{tenant}, $data->{client}, $data->{secret}, $data->{hostcfg_id});

  saveConf($xormon, $azure->getConfArch($data), $data);

  my %metricData;
  my %service_data_new;
  my $service_data = $xormon->getData();

  for (@{$data->{subscriptions}}) {
    my $subscription = $_;

    my $vmList = $azure->getVmList($subscription);

    my $resources = $azure->getResourceGroups($subscription);

    my $timestamp = time();

    for (@{$resources}) {
      my $resource = $_;

      my $last = (defined $service_data->{$subscription.'_'.$resource}) ? $service_data->{$subscription.'_'.$resource} : time()-1200;

      my $timestamp = time();
      my $time_end = $timestamp-180;
      my $time_start = $last;


      my $instances = $azure->getInstances($resource, $subscription);

      for (@{$instances->{value}}) {
        my $instance = $_;
        my $location = $instance->{location};
        my $vmSize = $instance->{properties}->{hardwareProfile}->{vmSize};
        #metrics data
        my $customMetrics = $azure->getAgentMetrics($resource, $subscription, $instance->{name});

        #print Dumper($customMetrics);

        my $metrics = $azure->getMetrics($resource, $instance->{name}, $subscription, $time_start, $time_end);

        my $type;

        for (@{$metrics->{value}}) {
	        my $metricValue = $_;

          if ($metricValue->{name}->{value} eq "Percentage CPU") {
            $type = "cpu_util";
          } elsif ($metricValue->{name}->{value} eq "Disk Read Bytes") {
            $type = "read_bytes";
          } elsif ($metricValue->{name}->{value} eq "Disk Write Bytes") {
            $type = "write_bytes";
          } elsif ($metricValue->{name}->{value} eq "Disk Read Operations/Sec") {
            $type = "read_ops";
          } elsif ($metricValue->{name}->{value} eq "Disk Write Operations/Sec") {
            $type = "write_ops";
          } elsif ($metricValue->{name}->{value} eq "Network In") {
            $type = "received_bytes";
          } elsif ($metricValue->{name}->{value} eq "Network Out") {
            $type = "sent_bytes";
          } else {
            $type = "undef";
          }

          for (@{$metricValue->{timeseries}}) {
            my $timeSerieData = $_;

            for (@{$timeSerieData->{data}}) {
              my $timeData = $_;
              #TODO
              my $pretty_time = str2time($timeData->{timeStamp});
              if(defined $timeData->{average}){
                $metricData{azure}{vm}{$instance->{properties}->{vmId}}{$pretty_time}{$type} = $timeData->{average};
              }
              else {
                $metricData{azure}{vm}{$instance->{properties}->{vmId}}{$pretty_time}{$type} = $timeData->{total};
              }
            }
          }

	        #custom metrics to global metrics
	        foreach my $timeKey (keys %{$metricData{azure}{vm}{$instance->{properties}->{vmId}}} ) {
	          if (ref($customMetrics) eq 'HASH' && defined $customMetrics->{usedMemory}->{$timeKey}) {
                $metricData{azure}{vm}{$instance->{properties}->{vmId}}{$timeKey}{usedMemory} = $customMetrics->{usedMemory}->{$timeKey};
              } else {
                $metricData{azure}{vm}{$instance->{properties}->{vmId}}{$timeKey}{usedMemory} = 0;
              }

              if (ref($customMetrics) eq 'HASH' && defined $customMetrics->{freeMemory}->{$timeKey}) {
                $metricData{azure}{vm}{$instance->{properties}->{vmId}}{$timeKey}{freeMemory} = $customMetrics->{freeMemory}->{$timeKey};
                if(defined $vmList->{$location}->{$vmSize}->{capabilities}->{MemoryGB} && !(defined$customMetrics->{usedMemory}->{$timeKey})){
                  $metricData{azure}{vm}{$instance->{properties}->{vmId}}{$timeKey}{usedMemory} = $vmList->{$location}->{$vmSize}->{capabilities}->{MemoryGB} * 1024 * 1024 * 1024 - $customMetrics->{freeMemory}->{$timeKey};
                }
              } else {
                $metricData{azure}{vm}{$instance->{properties}->{vmId}}{$timeKey}{freeMemory} = 0;
              }
	        }
        }
      }

      my $storages = $azure->getStorageInstances($resource, $subscription);

      for (@{$storages->{value}}) {
        my $instance = $_;
        for my $serviceType (@storageServices){
        
            my $metrics = $azure->getMetricsStorage($resource, $instance->{name}, $subscription, $time_start, $time_end, $serviceType);

            my $type;

            for (@{$metrics->{value}}) {
                my $metricValue = $_;

                if ($metricValue->{name}->{value} eq "Transactions") {
                    $type = $serviceType."_transactions";
                } elsif ($metricValue->{name}->{value} eq "Ingress") {
                    $type = $serviceType."_ingress";
                } elsif ($metricValue->{name}->{value} eq "Egress") {
                    $type = $serviceType."_egress";
                } elsif ($metricValue->{name}->{value} eq "SuccessServerLatency") {
                    $type = $serviceType."_suc_server_lat";
                } elsif ($metricValue->{name}->{value} eq "SuccessE2ELatency") {
                    $type = $serviceType."_suc_e2e_lat";
                } elsif ($metricValue->{name}->{value} eq "Availability") {
                    $type = $serviceType."_availibility";
                } else {
                    $type = "undef";
                }

                for (@{$metricValue->{timeseries}}) {
                    my $timeSerieData = $_;

                    for (@{$timeSerieData->{data}}) {
                        my $timeData = $_;
                        #TODO
                        my $pretty_time = str2time($timeData->{timeStamp});
                        if(defined $timeData->{average}){
                            $metricData{azure}{account}{md5_hex($instance->{id})}{$pretty_time}{$type} = $timeData->{average};
                        }
                        else {
                            if(defined $timeData->{total}){
                                $metricData{azure}{account}{md5_hex($instance->{id})}{$pretty_time}{$type} = $timeData->{total};
                            }
                        }
                    }
                }
            }
            
            $metrics = $azure->getMetricsStorageHourGrain($resource, $instance->{name}, $subscription, $time_start, $time_end, $serviceType);

            for (@{$metrics->{value}}) {
                my $metricValue = $_;
                my %hourly_data;
                given($serviceType){
                    when ("blob")   {
                                        if ($metricValue->{name}->{value} eq "BlobCapacity") {
                                            $type = "blob_capacity";
                                        } elsif ($metricValue->{name}->{value} eq "BlobCount") {
                                            $type = "blob_count";
                                        } elsif ($metricValue->{name}->{value} eq "ContainerCount") {
                                            $type = "container_count";
                                        } else {
                                            $type = "undef";
                                        }
                                    }
                    when ("file")   {
                                        if ($metricValue->{name}->{value} eq "FileCapacity") {
                                            $type = "file_capacity";
                                        } elsif ($metricValue->{name}->{value} eq "FileCount") {
                                            $type = "file_count";
                                        } elsif ($metricValue->{name}->{value} eq "FileShareCount") {
                                            $type = "file_share_count";
                                        } elsif ($metricValue->{name}->{value} eq "FileShareSnapshotCount") {
                                            $type = "file_share_snapshot_count";
                                        } elsif ($metricValue->{name}->{value} eq "FileShareSnapshotSize") {
                                            $type = "file_share_snapshot_size";
                                        } elsif ($metricValue->{name}->{value} eq "FileShareCapacityQuota") {
                                            $type = "file_share_capacity_quota";
                                        } else {
                                            $type = "undef";
                                        }
                                    }
                    when ("table")  {
                                        if ($metricValue->{name}->{value} eq "TableCapacity") {
                                            $type = "table_capacity";
                                        } elsif ($metricValue->{name}->{value} eq "TableCount") {
                                            $type = "table_count";
                                        } elsif ($metricValue->{name}->{value} eq "TableEntityCount") {
                                            $type = "table_entity_count";
                                        } else {
                                            $type = "undef";
                                        }
                                    }
                    when ("queue")  {
                                        if ($metricValue->{name}->{value} eq "QueueCapacity") {
                                            $type = "queue_capacity";
                                        } elsif ($metricValue->{name}->{value} eq "QueueCount") {
                                            $type = "queue_count";
                                        } elsif ($metricValue->{name}->{value} eq "QueueMessageCount") {
                                            $type = "queue_message_count";
                                        } else {
                                            $type = "undef";
                                        }
                                    }
                    when ("account"){
                                        if ($metricValue->{name}->{value} eq "UsedCapacity") {
                                            $type = "used_capacity";
                                        } else {
                                            $type = "undef";
                                        }
                                    }
                    default         {
                                        $type = "undef";
                                    }
                }

                for (@{$metricValue->{timeseries}}) {
                    my $timeSerieData = $_;

                    for (@{$timeSerieData->{data}}) {
                        my $timeData = $_;
                        #TODO
                        my $pretty_time = str2time($timeData->{timeStamp});
                        if(defined $timeData->{average}){
                            $metricData{azure}{account}{md5_hex($instance->{id})}{$pretty_time}{$type} = $timeData->{average};
                            $hourly_data{$pretty_time} = $timeData->{average};
                        }
                        else {
                            if(defined $timeData->{total}){
                                $metricData{azure}{account}{md5_hex($instance->{id})}{$pretty_time}{$type} = $timeData->{total};
                                $hourly_data{$pretty_time} = $timeData->{total};
                            }
                        }
                    }
                }
                my $newData = fillTimestamps(\%hourly_data, $time_end);

                for my $timeKey(keys %{$newData}){
                  $metricData{azure}{account}{md5_hex($instance->{id})}{$timeKey}{$type} = $newData->{$timeKey};
                }
            }
        }
      }
      $service_data_new{$subscription.'_'.$resource} = time()-600;
      

    }
  }

  $xormon->saveData(\%service_data_new);

  return \%metricData;

}

sub json2backend {
    my ($url, $data, $token) = @_;

    my $response = backendRequest($url, $data, $token);
}

sub backendRequest {
  my ($url, $data, $token) = @_;

  my $protocol = "http://";
  my $json = JSON->new;

  my $req = HTTP::Request->new(POST => $protocol.$url."/api/metrics/v1/save");
  $req->content_type('application/json');
  $req->content(to_json($data));

  my $ua = LWP::UserAgent->new(
    timeout => 30,
    ssl_opts => { SSL_cipher_list => 'DEFAULT:!DH', verify_hostname => 0, SSL_verify_mode => 0 },
  );

  $ua->default_header(Authorization => 'Bearer '.$token);

  my $resp = ();

  eval {
    $resp = $ua->request($req);
  };

  if ($@) {
    my $error = $@;
    print STDERR $error;
  }

  return $resp->code;

}

$server->close();