use LWP;
use JSON;
use strict;
use warnings;
use Data::Dumper;
use FindBin;                
use lib "$FindBin::Bin/"; # use lib in this folder
use lib "../../../perl-modules";
use GoogleCloud;
use Xormon;
use Digest::MD5 qw(md5_hex);
use Test;


$SIG{CHLD} = 'IGNORE';

use IO::Socket qw(AF_INET AF_UNIX SOCK_STREAM SHUT_WR);

# microservice port
my $DEFAULT_PORT = 3502;
my $port = Xormon::checkPrefixPort($DEFAULT_PORT);

my $server = IO::Socket->new(
    Domain => AF_INET,
    Type => SOCK_STREAM,
    Proto => 'tcp',
    LocalHost => '0.0.0.0',
    LocalPort => $port,
    ReusePort => 1,
    Listen => 5,
) || die "Can't open socket: $@";
Xormon::log("Service listening on port $port");

my $client;

 # waiting for a new client connection
while ($client = $server->accept()) {

    my $pid;
    unless ( defined( $pid = fork() ) ) {
        Xormon::error("Wait for fork");
        sleep 5;
    }

    if ($pid == 0) {
   
        # get information about a newly connected client
        my $client_address = $client->peerhost();
        my $client_port = $client->peerport();
        Xormon::log("Connection from $client_address:$client_port");

        # received data template 
        #
        # { 
        #   "jwt"  : "<backendToken>",
        #   "task" : "<taskName>"
        #   "data" : {
        #     <configuration_from_hostcfg>
        #   }
        # }

        # read up to 2048 characters from the connected client
        my $data = "";
        $client->recv($data, 1048576);

        my $decoded = from_json($data);
        my $cfg = $decoded->{data};
        
        Xormon::log("Received task: ".$decoded->{task});
        my $response = '';
        if ($decoded->{task} eq "fetch") {

            $client->send("Task successfuly started");
            $client->shutdown(SHUT_WR);

            my $xormon = Xormon->new("$client_address", $decoded->{jwt}, 'gcloud', $cfg->{hostcfg_id}, $cfg->{hostalias});

            # target destinations check
            if (defined $decoded->{'destinations'}) {
                $xormon->setDestinations($decoded->{'destinations'});
            }


            $xormon->startJob($decoded->{task}, $cfg->{hostalias}, $$);
            my $status = process($xormon, $cfg);

            $xormon->endJob($decoded->{task}, $cfg->{hostalias}, $status, $$);
            Xormon::log("Task [fetch] for ".$cfg->{hostalias}." is done");

        } elsif ($decoded->{task} eq "test") {

            $response = apitest($client, $cfg);

            $client->send($response);
            $client->shutdown(SHUT_WR);

        } else {
            $response = "undefined task";

            $client->send($response);
            $client->shutdown(SHUT_WR);

        }

        exit();

    } else {
        $client->autoflush(1);    
    }
}

sub process {
    my $xormon  = shift;
    my $cfg     = shift;

    $cfg->{interval} = 60;
    my $perfData = api2json($xormon, $cfg);

    my $rc = $xormon->saveJSON2($perfData);

    my $status;
    if ($rc eq "201") {
        $status = "successful";
    }elsif ($rc eq "401") {
        $status = "unauthorized";
    } else {
        $status = "error";
    }
    $status = "successful";
    return $status;
}

sub api2json {
    my $xormon  = shift;
    my $cfg     = shift;

    if(ref($cfg->{credentials}) ne "ARRAY"){
        $cfg->{credentials} = [$cfg->{credentials}];
    }

    my $service_data = $xormon->getData();

    my $timestamp_start = (defined $service_data->{timestamp}) ? $service_data->{timestamp} : time()-1200;
    my $timestamp_end = time()-180;

    $timestamp_start = $timestamp_start - ($timestamp_start % 60) - 180;
    $timestamp_end = $timestamp_end - ($timestamp_end % 60);

    my @architecture;
    my %perf;
    my $gcloud = GoogleCloud->new($cfg->{hostalias}, 1, $cfg->{hostcfg_id},  $cfg->{credentials});
    my $instances = $gcloud->listInstances(1);

    my $root_id = md5_hex($cfg->{hostalias});

    my %engine_architecture = (
        "item_id"    => $root_id,
        "label"      => $cfg->{hostalias},
        "hw_type"    => "gcloud",
        "subsystem"  => "project",
        "parents"    => [],
        "hostcfg_id" => $cfg->{hostcfg_id}
    ); 

    push(@architecture, \%engine_architecture);

    foreach my $region (keys %{$instances->{architecture}->{region_compute}}) {

        #zone architecture
        my $item_id = $region;

        my %region_architecture = (
            "item_id"    => $item_id,
            "label"      => $instances->{label}->{compute_region}->{$item_id},
            "hw_type"    => "gcloud",
            "subsystem"  => "region",
            "parents"    => [$root_id],
            "hostcfg_id" => $cfg->{hostcfg_id}
        );

        push(@architecture, \%region_architecture);

        for (@{$instances->{architecture}->{region_compute}->{$region}}){
            my $compute = $_;

            my %compute_architecture = (
                "item_id"    => $compute,
                "label"      => $instances->{specification}->{compute}->{$compute}->{name},
                "hw_type"    => "gcloud",
                "subsystem"  => "compute",
                "parents"    => [$root_id, $item_id],
                "hostcfg_id" => $cfg->{hostcfg_id}
            );

            push(@architecture, \%compute_architecture);

        }
    }

    foreach my $region (keys %{$instances->{architecture}->{region_database}}) {

        #zone architecture
        my $item_id = $region;

            my %region_architecture = (
                "item_id"    => $item_id,
                "label"      => $instances->{label}->{database_region}->{$item_id},
                "hw_type"    => "gcloud",
                "subsystem"  => "region",
                "parents"    => [$root_id],
                "hostcfg_id" => $cfg->{hostcfg_id}
            );

            push(@architecture, \%region_architecture);

        for (@{$instances->{architecture}->{region_database}->{$region}}){
            my $database = $_;
            my $database_id = $database;

            my %database_architecture = (
                "item_id"    => $database_id,
                "label"      => $instances->{specification}->{database}->{$database}->{name},
                "hw_type"    => "gcloud",
                "subsystem"  => "database",
                "parents"    => [$root_id, $item_id],
                "hostcfg_id" => $cfg->{hostcfg_id}
            );

            push(@architecture, \%database_architecture);

        }
    }
    
    #save architecture
    $xormon->saveArchitecture(\@architecture);

    if(defined $instances->{specification}->{compute}){
        $xormon->saveConfJSON(\%{$instances->{specification}->{compute}});
    }
    if(defined $instances->{specification}->{database}){
        $xormon->saveConfJSON(\%{$instances->{specification}->{database}});
    }
    if(defined $instances->{specification}->{region}){
        $xormon->saveConfJSON(\%{$instances->{specification}->{region}});
    }

    #save last read
    my %service_data_new = ('timestamp' => $timestamp_end);
    $xormon->saveData(\%service_data_new);

    return $gcloud->getMetrics($timestamp_start, $timestamp_end);
}

sub api2csv {
	my $data = shift;

    my $time1 = time();
    my $time2 = $time1-15;
    my $time3 = $time2-30;
    my $time4 = $time3-45;

    return "uuid;timestamp;data1;data2;data3;data4;data5;data6;data7;data8;
    f6969d67-453d-4337-970c-ccc72f5d8c43;$time1;1;2;3;4;5;6;7;8;
    f6969d67-453d-4337-970c-ccc72f5d8c43;$time2;2;3;2;4;5;6;7;8;
    f6969d67-453d-4337-970c-ccc72f5d8c43;$time3;2;3;3;4;5;6;7;8;
    f6969d67-453d-4337-970c-ccc72f5d8c43;$time4;5;4;1;4;5;6;7;8;"

}

sub apitest {
    my $client = shift;
    my $cfg = shift;
    my $gcloud = GoogleCloud->new($cfg->{hostalias}, 1, $cfg->{hostcfg_id},  $cfg->{credentials});
    my $testModule = Test->new($client);
    
    $gcloud->resetServiceAccounts();
    my $index = 0;
    my $version =  `gcloud version`;

    if($version eq ""){
        $testModule->addSection("GCloud CLI not found.", 0, "GCloud CLI test failed");
        return $testModule->genResult();
    }else{
        $testModule->addSection("GCloud CLI found", 1, "GCloud CLI test successful");
    }

    while(!$gcloud->setNextServiceAccount()){
        
        my $token = $gcloud->testToken();
        if($token) {
            $testModule->addSection("Token Test for service account $gcloud->{credentials}[$index]->{client_email}", 1, "Token test successful");
        }else {
            $testModule->addSection("Token Test for service account $gcloud->{credentials}[$index]->{client_email}", 0, "Token test failed");
        }
        my $instanceTest = $gcloud->testInstances();
        
        if($instanceTest){
            $testModule->addSection("Google Cloud Instances for service account $gcloud->{credentials}[$index++]->{client_email}", 1, "GCloud instances found");
        }else{
            $testModule->addSection("Google Cloud Instances for service account $gcloud->{credentials}[$index++]->{client_email}", 0, "Getting GCloud instances failed");
        }
    }

    return $testModule->genResult();

}

$server->close();