#!/usr/bin/env perl

use strict;
use warnings;
use Data::Dumper;
use Date::Parse;
use FindBin;
use lib "$FindBin::Bin/";    # use lib in this folder
use lib "$FindBin::Bin/../../../perl-modules";
use Netapp;
use Xormon;
use POSIX;
use Digest::MD5 qw(md5_hex);
use Sort::Naturally;
use Test;
use XormonLogger;

$SIG{CHLD} = 'IGNORE';

use IO::Socket qw(AF_INET AF_UNIX SOCK_STREAM SHUT_WR);

# microservice port
my $port = 3207;

my $server = IO::Socket->new(
  Domain    => AF_INET,
  Type      => SOCK_STREAM,
  Proto     => 'tcp',
  LocalHost => '0.0.0.0',
  LocalPort => $port,
  ReusePort => 1,
  Listen    => 5,
) || die "Can't open socket: $@";
Xormon::log("Service listening on port $port");

my $currentState;
my $currentCounters;
my $lastCounters = {};
my $allocated;          # use this to store allocated capacity per Tier (pool)
my $committed;          # use this to store committed capacity per Tier (pool)


my $debug = 0;
my $logger;

my $msdir = "$FindBin::Bin";

if ( !-d "$msdir/data" ) {
  mkdir "$msdir/data" || Xormon::log( "Cannot mkdir $msdir/data: $!" . __FILE__ . ":" . __LINE__ ) && return 0;
}

# waiting for a new client connection
my $client;
while ( $client = $server->accept() ) {

  my $pid;
  unless ( defined( $pid = fork() ) ) {
    Xormon::error("Wait for fork");
    sleep 5;
  }

  if ( $pid == 0 ) {

    # get information about a newly connected client
    my $client_address = $client->peerhost();
    my $client_port    = $client->peerport();
    Xormon::log("Connection from $client_address:$client_port");

    # received data template
    #
    # {
    #   "jwt"  : "<backendToken>",
    #   "task" : "<taskName>",
    #   "data" : {
    #     <configuration_from_hostcfg>
    #   }
    # }

    # read up to 2048 characters from the connected client
    my $data = "";
    $client->recv( $data, 2048 );

    my $decoded = Xormon::from_json($data);

    Xormon::log( "Received task: " . $decoded->{task} );
    my $cfg = $decoded->{data};
    $logger = XormonLogger->new( 'netapp', $cfg->{hostcfg_id}, $cfg->{hostalias} );

    # Netapp::file_write( "/tmp/naenv.dump", Dumper \%ENV );
    my $response = '';
    if ( $decoded->{task} eq "fetch" ) {

      $client->send("Task successfuly started");
      $client->shutdown(SHUT_WR);

      my $xormon = Xormon->new("$client_address", $decoded->{jwt}, 'netapp', $cfg->{hostcfg_id}, $cfg->{hostalias});

      # target destinations check
      if (defined $decoded->{'destinations'}) {
          $xormon->setDestinations($decoded->{'destinations'});
      }

      process( $xormon, $cfg );

    }
    elsif ( $decoded->{task} eq "test" ) {

      # Xormon::log( "Decoded data: " . Dumper $cfg );
      $response = apitest( $client, $cfg );

      $client->send($response);
      $client->shutdown(SHUT_WR);

    }
    else {
      $response = "undefined task";

      $client->send($response);
      $client->shutdown(SHUT_WR);
    }

    exit();

  }
  else {
    $client->autoflush(1);
  }
}

sub process {
  my $xormon = shift;
  my $cfg    = shift;
  my @architecture;
  my $config;
  my $last_perf_timestamp;
  my $nodes;

  my $service_data_in  = $xormon->getData();
  my $service_data_out = {};
  if ( exists $service_data_in->{last_conf_timestamp} ) {
    $service_data_out->{last_conf_timestamp} = $service_data_in->{last_conf_timestamp};
  }

  $xormon->startJob( 'fetch', $cfg->{hostalias} );

  my $netapp = Netapp->new( { host => $cfg->{apiHost}, username => $cfg->{apiUsername}, password => $cfg->{apiPassword} } );
  my $device = $netapp->readRestApi( { path => "/cluster", fields => "name,uuid" } );
  if ( !$device->{status} || !$device->{content}{uuid} ) {
    my $status = "Got no cluster UUID, skip stats processing.";
    $xormon->endJob( 'fetch', $cfg->{hostalias}, $status );
    $logger->error($status);
    Xormon::log( "Task [fetch] for " . $cfg->{hostalias} . " is done" );

    my $connected = Test::connTest($cfg->{apiHost}, $cfg->{apiPort});
    if ($connected == 1) {
      $xormon->isConnected(1);
    }
    else {
      $xormon->isConnected(0, "Connection failed!") && return "error";
    }
    exit;
  }
  else {
    $xormon->isConnected(1);
  }

  # device (cluster)
  my $device_arch = {
    "item_id"    => $cfg->{hostcfg_id},
    "cluster_id" => $device->{content}{uuid},
    "label"      => $cfg->{hostalias},
    "hw_type"    => "netapp",
    "subsystem"  => "device",
    "parents"    => [],
    "hostcfg_id" => $cfg->{hostcfg_id}
  };
  push( @architecture, $device_arch );

  use Storable;

  # Load old conters from file
  my $lastCountersFile = "$msdir/data/$cfg->{hostcfg_id}_lastCounters.file";
  if ( -r $lastCountersFile && -s $lastCountersFile ) {
    $lastCounters = retrieve($lastCountersFile);
  }

  my $readCounters;

  # cluster
  $readCounters->{device} = [];
  my $cluster_api = $netapp->readRestApi( { path => "/cluster", fields => "name,uuid,statistics,version,location,contact" } );
  if ( $cluster_api->{status} ) {

    # cluster space (merge keys with $currentState->{cluster} )
    my $cluster_space_api = $netapp->readRestApi( { path => "/storage/cluster", fields => "block_storage" } );
    if ( $cluster_space_api->{status} ) {
      @{ $cluster_api->{content} }{ keys %{ $cluster_space_api->{content} } } = values %{ $cluster_space_api->{content} };
    }
    push @{ $readCounters->{device} }, $cluster_api->{content};
    my $cluster_identity = $netapp->readRestApi( { path => "/private/cli/cluster/identity", fields => "" } );
    if ( $cluster_identity->{status} ) {
      $readCounters->{identity} = $cluster_identity->{content};
    }
  }
  my $apiResponse;

  # nodes
  $apiResponse = $netapp->readRestApi( { path => "/cluster/nodes", fields => "name,uuid,statistics,version,state,serial_number,model,system_id" } );
  if ( $apiResponse->{status} && $apiResponse->{records} ) {
    $readCounters->{system} = $apiResponse->{content};
  }

  # aggregates
  $apiResponse = $netapp->readRestApi( { path => "/storage/aggregates", fields => "statistics,node,space,state" } );
  if ( $apiResponse->{status} && $apiResponse->{records} ) {
    $readCounters->{pool} = $apiResponse->{content};
  }

  # volumes
  $apiResponse = $netapp->readRestApi( { path => "/storage/volumes", fields => "statistics,space,aggregates,svm,autosize" } );
  if ( $apiResponse->{status} && $apiResponse->{records} ) {
    $readCounters->{volume} = $apiResponse->{content};
  }

  # qtrees
  $apiResponse = $netapp->readRestApi( { path => "/storage/qtrees", fields => "statistics,volume,svm,nas,path&name=!\"\"" } );
  if ( $apiResponse->{status} && $apiResponse->{records} ) {
    $readCounters->{qtree} = $apiResponse->{content};
  }

  # luns
  $apiResponse = $netapp->readRestApi( { path => "/storage/luns", fields => "statistics,space,location" } );
  if ( $apiResponse->{status} && $apiResponse->{records} ) {
    $readCounters->{lun} = $apiResponse->{content};
  }

  # eth ports
  $apiResponse = $netapp->readRestApi( { path => "/network/ethernet/ports", fields => "statistics,state,enabled,node" } );
  if ( $apiResponse->{status} && $apiResponse->{records} ) {
    $readCounters->{ethport} = $apiResponse->{content};
  }

  # fc ports
  $apiResponse = $netapp->readRestApi( { path => "/network/fc/ports", fields => "statistics,state,enabled" } );
  if ( $apiResponse->{status} && $apiResponse->{records} ) {
    $readCounters->{fcport} = $apiResponse->{content};
  }

  # drives
  $apiResponse = $netapp->readRestApi( { path => "/cluster/counter/tables/disk/rows", fields => "*", counters => "capacity|read_ops|write_ops|total_transfer_count|read_data|write_data|total_data|operation_latency" } );
  if ( $apiResponse->{status} && $apiResponse->{records} ) {
    $readCounters->{drive} = $apiResponse->{content};
  }

  # node stats
  $apiResponse = $netapp->readRestApi( { path => "/cluster/counter/tables/system:node/rows", fields => "*", counters => "time|cpu_elapsed_time|average_processor_busy_percent|read_ops|write_ops|other_ops|total_ops|iscsi_ops|fcp_ops|nfs_ops|cifs_ops|read_data|write_data|total_data|iscsi_data_sent|iscsi_data_received|network_data_sent|network_data_received|fcp_data_sent|fcp_data_received|read_latency|write_latency|other_latency|total_latency" } );
  if ( $apiResponse->{status} && $apiResponse->{records} ) {
    $readCounters->{node} = $apiResponse->{content};
  }

  Netapp::file_write( "debug/$device->{content}{uuid}.dump", Dumper $readCounters ) if $debug;

  my $status = "successful";

  my @healthStatus;
  my $globHS = "ok";
  foreach my $node ( @{ $readCounters->{system} } ) {
    my $state;
    if ( $node->{state} && $node->{state} eq "up" ) {
      $state = "ok";
      if ( $node->{model} ) {
        $nodes->{models}{$node->{model}} = 1;
      }
      if ( $node->{name} ) {
        $nodes->{names}{$node->{name}} = 1;
      }
    }
    else {
      $state  = "error";
      $globHS = "error";
    }

    my $node_status = {
      "item_id" => $node->{uuid},
      "status"  => $state
    };
    push( @healthStatus, $node_status );

    my $arch = {
      item_id   => $node->{uuid},
      label     => $node->{name},
      hw_type   => "netapp",
      subsystem => "node",
      parents   => [ $cfg->{hostcfg_id} ],
      "hostcfg_id" => $cfg->{'hostcfg_id'}
    };
    push( @architecture, $arch );
  }

  foreach my $subsys (qw/device volume lun pool ethport fcport drive node/) {
    next unless $readCounters->{$subsys};

    foreach my $item ( @{ $readCounters->{$subsys} } ) {
      if ( $subsys eq "drive" ) {
        $item->{uuid} = $item->{id};
        $item->{name} = $item->{properties}{name};
      }
      elsif ( $subsys eq "node" ) {
        $item->{uuid} = $item->{properties}{uuid};
        $item->{name} = $item->{properties}{name};
      }
      if ( $item->{uuid} ) {
        my @parents = ( $cfg->{hostcfg_id} );
        if ( $subsys eq "device" ) {
          my $props;
          $props->{name}     = $item->{name}          if defined $item->{name};
          $props->{location} = $item->{location}      if defined $item->{location};
          $props->{contact}  = $item->{contact}       if defined $item->{contact};
          $props->{serial}   = $readCounters->{identity}{serialnumber} if defined $readCounters->{identity}{serialnumber};
          if ( defined $item->{version}{full} ) {
            # ( $props->{version} ) = split ":", $item->{version}{full};
            $props->{version} = "$item->{version}{generation}.$item->{version}{major}.$item->{version}{minor}";
          }
          if ( exists $nodes->{models} ) {
            $props->{model} = join ", ", sort keys %{ $nodes->{models} };
          }
          if ( exists $nodes->{names} ) {
            $props->{node_name} = join ", ", sort keys %{ $nodes->{names} };
          }
          $config->{ $cfg->{hostcfg_id} } = $props;
        }
        elsif ( $subsys eq "system" ) {
          my $props;
          $props->{serial}  = $item->{serial_number};
          if ($item->{model}) {
            $props->{model} = $item->{model};
          }
          if ( defined $item->{version}{full} ) {
            # ( $props->{version} ) = split ":", $item->{version}{full};
            $props->{version} = "$item->{version}{generation}.$item->{version}{major}.$item->{version}{minor}";
          }
          # $config->{ $item->{uuid} } = $props;
        }
        elsif ( $subsys eq "pool" ) {
          my $props;
          $props->{state}            = $item->{state}                               if defined $item->{state};
          $props->{physical_used}    = $item->{space}{block_storage}{physical_used} if defined $item->{space}{block_storage}{physical_used};
          $props->{efficiency_ratio} = sprintf( "%.3g", $item->{space}{efficiency}{ratio} ) if defined $item->{space}{efficiency}{ratio};
          $props->{efficiency_logical_used} = $item->{space}{efficiency}{logical_used}     if defined $item->{space}{efficiency}{logical_used};
          $props->{committed}        = $committed->{ $item->{uuid} }                 if defined $committed->{ $item->{uuid} };
          $props->{allocated}        = $allocated->{ $item->{uuid} }                 if defined $allocated->{ $item->{uuid} };

          $config->{ $item->{uuid} } = $props;
        }
        elsif ( $subsys eq "volume" ) {
          if ( defined $item->{aggregates} ) {
            foreach my $aggr ( @{ $item->{aggregates} } ) {
              if ( $aggr->{uuid} ) {
                push @parents, $aggr->{uuid};
                if ( defined $item->{autosize} )  {
                  if ( $item->{autosize}{mode} ne "off" ) {
                    $committed->{ $aggr->{uuid} } += $item->{autosize}{maximum};
                  }
                  else {
                    $committed->{ $aggr->{uuid} } += $item->{space}{afs_total};
                  }
                }
                $allocated->{ $aggr->{uuid} } += $item->{space}{afs_total};
              }
            }
          }
          if ( defined $item->{svm} ) {

            #foreach my $svm ( @{ $item->{svm} } ) {
            # push @parents, $item->{svm}{uuid};

            #}
          }
        }
        elsif ( $subsys eq "fcport" ) {
          next unless ( exists $item->{statistics} && exists $item->{statistics}{status} && $item->{statistics}{status} eq 'ok' );
          my $props;
          $props->{enabled} = $item->{enabled} if defined $item->{enabled};
          $props->{state}   = $item->{state} if defined $item->{state};
          $props->{wwnn}    = $item->{wwnn} if defined $item->{wwnn};
          if ( defined $item->{node} ) {
            push @parents, $item->{node}{uuid};
          }

          $config->{ $item->{uuid} } = $props;
        }
        elsif ( $subsys eq "ethport" ) {
          my $props;
          $props->{enabled} = $item->{enabled} if defined $item->{enabled};
          $props->{state}   = $item->{state}   if defined $item->{state};
          if ( defined $item->{node} ) {
            push @parents, $item->{node}{uuid};
          }

          $config->{ $item->{uuid} } = $props;
        }
        elsif ( $subsys eq "drive" && $item->{uuid} !~ "multiple-values" ) {
          my $props;
          $props->{speed} = $item->{properties}{speed} if defined $item->{properties}{speed};
          if ( defined $item->{properties}{'node.name'} ) {

            # push @parents, $item->{node}{uuid};
          }

          $config->{ $item->{uuid} } = $props;
        }

        if ( $subsys ne "device" && $subsys ne "node" && $subsys ne "system") {
          my $arch = {
            item_id   => $item->{uuid},
            label     => $item->{name},
            hw_type   => "netapp",
            subsystem => $subsys,
            "hostcfg_id" => $cfg->{'hostcfg_id'},
            parents   => \@parents
          };
          push( @architecture, $arch );
        }
      }
    }

    if ( $subsys ne "system" ) {
      my $csv = countDeltas( $subsys, $readCounters->{$subsys}, $cfg->{hostcfg_id} );
      if ($csv) {
        my $timestamp = time;
        $last_perf_timestamp = $timestamp;
        Netapp::file_write( "debug/$device->{content}{uuid}_${timestamp}_$subsys.csv", $csv ) if $debug;
        my $rc = $xormon->saveCSV( $csv, "netapp", $subsys );
        $logger->info("Save $subsys perf: $rc");
        if ( $rc eq "401" ) {
          $status = "unauthorized";
        }
        elsif ( $rc ne "201" ) {
          $status = "error";
        }
      }
    }

    # Netapp::file_write( "netapp.perf", "\n$subsys\n" . $csv );
  }

  my $gStatus = {
    "item_id" => $cfg->{hostcfg_id},
    "status"  => $globHS
  };
  push( @healthStatus, $gStatus );

  #save health status
  $xormon->saveStatus( \@healthStatus );


  if ( $last_perf_timestamp ) {
    $service_data_out->{last_perf_timestamp} = $last_perf_timestamp;
    $xormon->hasData(1);
  }
  else {
    # check current data
    if ( $service_data_out->{last_perf_timestamp} && ( ( time() - $service_data_out->{last_perf_timestamp} ) > 10800 ) ) {
      $xormon->hasData(0, "last available stats are older than 3 hours: $service_data_out->{last_perf_timestamp}");
    }
    else {
      $xormon->hasData(0, "no available stats");
    }
  }

  # send architecture and conf data (only twice per hour)
  #
  if ( ! $service_data_in->{last_conf_timestamp} || ( $last_perf_timestamp - $service_data_in->{last_conf_timestamp} > 1800 ) ) {
    $service_data_out->{last_conf_timestamp} = $last_perf_timestamp;

    #save architecture
    my $rc_arch = $xormon->saveArchitecture( \@architecture );
    $logger->info("save architecture: $rc_arch");

    #save configuration
    my $rc_conf = $xormon->saveConfJSON($config);
    $logger->info("save configuration: $rc_conf");
  }

  #
  # save microservice data
  #
  $xormon->saveData( $service_data_out );

  store( $currentCounters, $lastCountersFile );

  # Netapp::file_write( "data/$cfg->{hostcfg_id}_lastCounters.dump", Dumper $currentCounters );

  $xormon->endJob( 'fetch', $cfg->{hostalias}, $status );

  Xormon::log( "Task [fetch] for " . $cfg->{hostalias} . " is done" );
}

sub apitest {
  my $client    = shift;
  my $cfg       = shift;
  my $test      = Test->new($client);
  my @variables = ( "hostcfg_id", "hostalias", "apiHost", "apiPort", "apiUsername", "apiPassword" );
  foreach my $variable (@variables) {
    unless ( exists $cfg->{$variable} ) {
      $test->addSection( "Configuration check", 0, "Cannot continue, $variable is not configured!" );
      return test->stop();
    }
  }
  $test->addSection( "Configuration check", 1, "" );

  my $connected = Test::connTest($cfg->{apiHost}, $cfg->{apiPort});
  if ($connected == 1) {
    $test->addSection("Connection test", 1, "TCP connection to " . $cfg->{apiHost} . ":" . $cfg->{apiPort} . " is OK");
  }
  else {
    $test->addSection("Connection test", 0, "TCP connection to " . $cfg->{apiHost} . ":" . $cfg->{apiPort} . " has failed!");
    $test->stop();
  }

  my $netapp = Netapp->new( { host => $cfg->{apiHost}, username => $cfg->{apiUsername}, password => $cfg->{apiPassword} } );

  my $cluster = $netapp->readRestApi( { path => "/cluster", fields => "version" } );

  #warn Dumper $cluster;
  if ( !$cluster->{status} ) {
    $test->addSection( "Cluster version", 0, $cluster->{content} );
    return $test->stop();
  }
  else {
    $test->addSection( "Cluster version", 1, $cluster->{content}{version}{full} );
  }

  my $nodes = $netapp->readRestApi( { path => "/cluster/nodes", fields => "name,serial-number,uptime" } );
  my @rows;
  if ( $nodes->{records} ) {
    foreach my $node ( sort { ncmp( $a->{name}, $b->{name} ) } @{ $nodes->{content} } ) {
      my @parts  = gmtime( $node->{uptime} );
      my $uptime = sprintf( "%d day(s), %d:%02d:%02d", @parts[ 7, 2, 1, 0 ] );
      push( @rows, "$node->{name}: uptime $uptime\n" );
    }
    $test->addSection( "Cluster nodes", 1, \@rows );

  }
  else {
    $test->addSection( "Cluster nodes", 0, "Cannot get node info" );
    return test->stop();
  }
  return $test->stop();
}

sub countDeltas {
  my ( $subsys, $itemList, $cfg_id ) = @_;
  my $lines;

  if ( grep { $_ eq $subsys } qw/device volume lun pool fcport/ ) {
    if ( $subsys eq 'pool' ) {
      push @$lines, "item_id;time;io_rate;read_io;write_io;data_rate;read;write;resp_t;resp_t_r;resp_t_w;capacity_total;capacity_used;capacity_free;virtual_capacity_total;capacity_oversubscr;capacity_oversubscr_percent;capacity_used_percent";
    }
    elsif ( $subsys eq 'fcport' ) {
      push @$lines, "item_id;time;io_rate;read_io;write_io;data_rate;read;write;resp_t;resp_t_r;resp_t_w";
    }
    else {
      push @$lines, "item_id;time;io_rate;read_io;write_io;data_rate;read;write;resp_t;resp_t_r;resp_t_w;capacity_total;capacity_used;capacity_free;capacity_used_percent";
    }
    foreach my $item ( @{$itemList} ) {
      if ( $item->{uuid} ) {
        if ( $item->{statistics} && $item->{statistics}{status} eq "ok" ) {
          if ( $subsys eq "device" ) {
            $item->{uuid} = $cfg_id;
            # $logger->info("####### DEVICE UUID: $item->{uuid}");
          }
          $currentCounters->{$subsys}{ $item->{uuid} } = $item;
          if ( exists $lastCounters->{$subsys}{ $item->{uuid} } ) {
            my $currstats = $item->{statistics};
            my $laststats = $lastCounters->{$subsys}{ $item->{uuid} }{statistics};
            my $calculated;
            my $currtime = str2time( $currstats->{timestamp} );
            my $lasttime = str2time( $laststats->{timestamp} );
            my $delta    = $currtime - $lasttime;

            # warn "$currtime, $lasttime, $delta";
            my $line = "$item->{uuid};$currtime";

            foreach my $category (qw/iops_raw throughput_raw latency_raw/) {
              foreach my $metric (qw/total read write/) {
                if ( !( ( defined $currstats->{$category}{$metric} && $currstats->{$category}{$metric} >= 0 ) && ( defined $laststats->{$category}{$metric} && $laststats->{$category}{$metric} >= 0 ) ) ) {
                  next;
                }
                if ( $currstats->{$category}{$metric} == $laststats->{$category}{$metric} ) {
                  $calculated->{$category}{$metric} = 0;
                }
                elsif ( $currstats->{$category}{$metric} > $laststats->{$category}{$metric} ) {
                  if ( $category eq 'latency_raw' ) {
                    if ( $calculated->{iops_raw}{$metric} > 0 ) {
                      $calculated->{$category}{$metric} = sprintf( "%.4g", ( $currstats->{$category}{$metric} - $laststats->{$category}{$metric} ) / $delta / $calculated->{iops_raw}{$metric} / 1e3 );
                    }
                    else {
                      $calculated->{$category}{$metric} = "";
                    }
                  }

                  # elsif ( $category eq 'throughput_raw' ) {
                  #   $calculated->{$category}{$metric} = sprintf( "%.2f", ( $currstats->{$category}{$metric} - $laststats->{$category}{$metric} ) / $delta / 1024 );
                  # }
                  else {
                    $calculated->{$category}{$metric} = sprintf( "%.4g", ( $currstats->{$category}{$metric} - $laststats->{$category}{$metric} ) / $delta );
                  }
                }
                else {
                  $calculated->{$category}{$metric} = "";
                }
                $line .= ";$calculated->{$category}{$metric}";
              }
            }
            if ( $subsys eq "device" ) {
              if ( $item->{block_storage} ) {
                foreach my $metric (qw/size used available/) {
                  if ( defined $item->{block_storage}{$metric} ) {
                    $line .= ";" . sprintf( "%.4g", $item->{block_storage}{$metric} );
                  }
                  else {
                    $line .= ";";
                  }
                }
                if ( $item->{block_storage}{size} > 0 ) {
                  $line .= ";" . sprintf( "%.5g", ( $item->{block_storage}{used} / $item->{block_storage}{size} ) * 100 );
                }
                else {
                  $line .= ";";
                }
              }
            }
            elsif ( $subsys eq "pool" ) {
              if ( $item->{space} && $item->{space}{block_storage} ) {
                foreach my $metric (qw/size used available/) {
                  if ( defined $item->{space}{block_storage}{$metric} ) {
                    $line .= ";" . sprintf( "%.4g", $item->{space}{block_storage}{$metric} );
                  }
                  else {
                    $line .= ";";
                  }
                }
                if ( defined $allocated->{ $item->{uuid} } && $item->{space}{block_storage}{size} > 0 && $allocated->{ $item->{uuid} } > $item->{space}{block_storage}{size} ) {
                  $line .= ";" . sprintf( "%.5g", $allocated->{ $item->{uuid} } );
                  $line .= ";" . sprintf( "%.5g", $allocated->{ $item->{uuid} } - $item->{space}{block_storage}{size} );
                  $line .= ";" . sprintf( "%.5g", ( $allocated->{ $item->{uuid} } / $item->{space}{block_storage}{size} ) * 100 );
                }
                else {
                  $line .= ";;;";
                }
                if ( $item->{space}{block_storage}{size} > 0 ) {
                  $line .= ";" . sprintf( "%.5g", ( $item->{space}{block_storage}{used} / $item->{space}{block_storage}{size} ) * 100 );
                }
                else {
                  $line .= ";";
                }
              }
            }
            elsif ( $subsys eq "volume" ) {
              if ( $item->{space} ) {
                my $rsize;
                foreach my $metric (qw/size used available/) {
                  $rsize->{$metric} = $item->{space}{$metric};
                }
                $rsize->{used} = $rsize->{size} - $rsize->{available};

                foreach my $metric (qw/size used available/) {
                  if ( defined $rsize->{$metric} ) {
                    $line .= ";" . sprintf( "%.4g", $rsize->{$metric} );
                  }
                  else {
                    $line .= ";";
                  }
                }
                if ( $item->{space}{size} > 0 ) {
                  $line .= ";" . sprintf( "%.5g", ( $rsize->{used} / $rsize->{size} ) * 100 );
                }
                else {
                  $line .= ";";
                }
              }
            }
            elsif ( $subsys eq "lun" ) {
              if ( $item->{space} ) {
                if ( defined $item->{space}{size} && defined $item->{space}{used} ) {
                  $line .= ";" . sprintf( "%.4g;", $item->{space}{size} ) . sprintf( "%.4g;", $item->{space}{used} ) . sprintf( "%.4g", $item->{space}{size} - $item->{space}{used} );
                }
              }
            }
            push @$lines, $line;
          }
        }
      }
    }
  }
  elsif ( $subsys eq "ethport" ) {
    push @$lines, "item_id;time;data_rate;read;write;read_packet;read_error;write_packet;write_error;packet_rate";
    foreach my $item ( @{$itemList} ) {
      if ( $item->{uuid} ) {
        if ( $item->{statistics} && $item->{statistics}{status} eq "ok" ) {
          $currentCounters->{$subsys}{ $item->{uuid} } = $item;
          if ( exists $lastCounters->{$subsys}{ $item->{uuid} } ) {
            my $currstats = $item->{statistics};
            my $laststats = $lastCounters->{$subsys}{ $item->{uuid} }{statistics};
            my $calculated;
            my $currtime = str2time( $currstats->{timestamp} );
            my $lasttime = str2time( $laststats->{timestamp} );
            my $delta    = $currtime - $lasttime;

            # warn "$currtime, $lasttime, $delta";
            my $line = "$item->{uuid};$currtime";

            foreach my $category (qw/throughput_raw device/) {
              if ( $category eq "throughput_raw" ) {
                foreach my $metric (qw/total read write/) {
                  if ( !( ( defined $currstats->{$category}{$metric} && $currstats->{$category}{$metric} >= 0 ) && ( defined $laststats->{$category}{$metric} && $laststats->{$category}{$metric} >= 0 ) ) ) {
                    $line .= ";";
                    next;
                  }
                  if ( $currstats->{$category}{$metric} == $laststats->{$category}{$metric} ) {
                    $calculated->{$category}{$metric} = 0;
                  }
                  elsif ( $currstats->{$category}{$metric} > $laststats->{$category}{$metric} ) {
                    $calculated->{$category}{$metric} = sprintf( "%.4g", ( $currstats->{$category}{$metric} - $laststats->{$category}{$metric} ) / $delta );
                  }
                  else {
                    $calculated->{$category}{$metric} = "";
                  }
                  $line .= ";$calculated->{$category}{$metric}";
                }
              }
              elsif ( $category eq "device" ) {
                foreach my $direction (qw/receive_raw transmit_raw/) {
                  foreach my $metric (qw/packets errors/) {
                    if ( !( ( defined $currstats->{$category}{$direction}{$metric} && $currstats->{$category}{$direction}{$metric} >= 0 ) && ( defined $laststats->{$category}{$direction}{$metric} && $laststats->{$category}{$direction}{$metric} >= 0 ) ) ) {
                      $line .= ";";
                      next;
                    }
                    if ( $currstats->{$category}{$direction}{$metric} == $laststats->{$category}{$direction}{$metric} ) {
                      $calculated->{$category}{$direction}{$metric} = 0;
                    }
                    elsif ( $currstats->{$category}{$direction}{$metric} > $laststats->{$category}{$direction}{$metric} ) {
                      $calculated->{$category}{$direction}{$metric} = sprintf( "%.4g", ( $currstats->{$category}{$direction}{$metric} - $laststats->{$category}{$direction}{$metric} ) / $delta );
                    }
                    else {
                      $calculated->{$category}{$direction}{$metric} = "";
                    }
                    $line .= ";$calculated->{$category}{$direction}{$metric}";
                  }
                }
                {
                  no warnings 'uninitialized';
                  $line .= ";" . ( $calculated->{$category}{receive_raw}{packets} + $calculated->{$category}{transmit_raw}{packets} );
                }
              }
            }
            push @$lines, $line;
          }
        }
      }
    }
  }
  elsif ( $subsys eq 'node' ) {
    push @$lines, "item_id;time;io_rate;read_io;write_io;data_rate;read;write;resp_t;resp_t_r;resp_t_w;iscsi_ops;fcp_ops;nfs_ops;cifs_ops;iscsi_data_sent;iscsi_data_received;network_data_sent;network_data_received;fcp_data_sent;fcp_data_received;cpu_busy";

    foreach my $item ( @{$itemList} ) {
      if ( defined $item->{properties} && defined $item->{properties}{uuid} ) {
        my $uuid = $item->{properties}{uuid};
        $currentCounters->{$subsys}{$uuid}       = $item;
        $currentCounters->{$subsys}{$uuid}{uuid} = $uuid;
        $currentCounters->{$subsys}{$uuid}{name} = $item->{properties}{name};
        if ( exists $lastCounters->{$subsys}{$uuid} && ref $lastCounters->{$subsys}{$uuid} eq "HASH" ) {
          my $currstats = $item->{counters};
          my $laststats = $lastCounters->{$subsys}{$uuid}{counters};
          my $calculated;
          my $currtime = $item->{counters}{time};
          my $lasttime = $lastCounters->{$subsys}{$uuid}{counters}{time};
          my $delta    = $currtime - $lasttime;

          # warn "$currtime, $lasttime, $delta";
          my $line = "$uuid;$currtime";

          foreach my $metric (qw/total_ops read_ops write_ops total_data read_data write_data total_latency read_latency write_latency iscsi_ops fcp_ops nfs_ops cifs_ops iscsi_data_sent iscsi_data_received network_data_sent network_data_received fcp_data_sent fcp_data_received average_processor_busy_percent cpu_elapsed_time/) {
            if ( !( ( defined $currstats->{$metric} && $currstats->{$metric} >= 0 ) && ( defined $laststats->{$metric} && $laststats->{$metric} >= 0 ) ) ) {
              next;
            }
            if ( $currstats->{$metric} == $laststats->{$metric} ) {
              $calculated->{$metric} = 0;
            }
            elsif ( $currstats->{$metric} > $laststats->{$metric} ) {
              if ( $metric eq 'read_latency' ) {
                if ( $calculated->{read_ops} > 0 ) {
                  $calculated->{$metric} = sprintf( "%.4g", ( $currstats->{$metric} - $laststats->{$metric} ) / $delta / $calculated->{read_ops} / 1e3 );
                }
                else {
                  $calculated->{$metric} = "";
                }
              }
              elsif ( $metric eq 'write_latency' ) {
                if ( $calculated->{write_ops} > 0 ) {
                  $calculated->{$metric} = sprintf( "%.4g", ( $currstats->{$metric} - $laststats->{$metric} ) / $delta / $calculated->{write_ops} / 1e3 );
                }
                else {
                  $calculated->{$metric} = "";
                }
              }
              elsif ( $metric eq 'other_latency' ) {
                if ( $calculated->{other_ops} > 0 ) {
                  $calculated->{$metric} = sprintf( "%.4g", ( $currstats->{$metric} - $laststats->{$metric} ) / $delta / $calculated->{other_ops} / 1e3 );
                }
                else {
                  $calculated->{$metric} = "";
                }
              }
              elsif ( $metric eq 'total_latency' ) {
                if ( $calculated->{total_ops} > 0 ) {
                  $calculated->{$metric} = sprintf( "%.4g", ( $currstats->{$metric} - $laststats->{$metric} ) / $delta / $calculated->{total_ops} / 1e3 );
                }
                else {
                  $calculated->{$metric} = "";
                }
              }
              else {
                $calculated->{$metric} = sprintf( "%.4g", ( $currstats->{$metric} - $laststats->{$metric} ) / $delta );
              }
            }
            else {
              $calculated->{$metric} = "";
            }
            if ( $metric eq 'average_processor_busy_percent' ) {
              next;
            }
            if ( $metric eq 'cpu_elapsed_time' && $calculated->{$metric} > 0 ) {
              $line .= ";" . sprintf( "%.4g", ( $calculated->{average_processor_busy_percent} / $calculated->{$metric} * 100 ) );
            }
            else {
              $line .= ";$calculated->{$metric}";
            }
          }

          #if ( $item->{capacity} ) {
          #  if ( defined $item->{space}{$metric} ) {
          #    $line .= ";" . sprintf( "%.4g", $item->{space}{$metric} );
          #  }
          #  else {
          #    $line .= ";";
          #  }
          #}
          push @$lines, $line;
        }
      }
    }
  }
  elsif ( $subsys eq "drive" ) {
    push @$lines, "item_id;time;read_io;write_io;io_rate;read;write;data_rate;resp_t";
    foreach my $item ( @{$itemList} ) {
      if ( defined $item->{id} && $item->{id} !~ "multiple-values" ) {
        $currentCounters->{$subsys}{ $item->{id} } = $item;
        my $currtime = time();
        $currentCounters->{$subsys}{ $item->{id} }{timestamp} = $currtime;
        $currentCounters->{$subsys}{ $item->{id} }{uuid}      = $item->{id};
        $currentCounters->{$subsys}{ $item->{id} }{name}      = $item->{properties}{name};
        if ( exists $lastCounters->{$subsys}{ $item->{id} } && ref $lastCounters->{$subsys}{ $item->{id} } eq "HASH" ) {
          my $currstats = $item->{counters};
          my $laststats = $lastCounters->{$subsys}{ $item->{id} }{counters};
          my $calculated;
          my $lasttime = $lastCounters->{$subsys}{ $item->{id} }{timestamp};
          my $delta    = $currtime - $lasttime;

          # warn "$currtime, $lasttime, $delta";
          my $line = "$item->{id};$currtime";

          foreach my $metric (qw/read_ops write_ops total_transfer_count read_data write_data total_data operation_latency/) {
            if ( !( ( defined $currstats->{$metric} && $currstats->{$metric} >= 0 ) && ( defined $laststats->{$metric} && $laststats->{$metric} >= 0 ) ) ) {
              next;
            }
            if ( $currstats->{$metric} == $laststats->{$metric} ) {
              $calculated->{$metric} = 0;
            }
            elsif ( $currstats->{$metric} > $laststats->{$metric} ) {
              if ( $metric eq 'operation_latency' ) {
                if ( $calculated->{total_transfer_count} > 0 ) {
                  $calculated->{$metric} = sprintf( "%.4g", ( $currstats->{$metric} - $laststats->{$metric} ) / $delta / $calculated->{total_transfer_count} / 1e3 );
                }
                else {
                  $calculated->{$metric} = "";
                }
              }

              # elsif ( $category eq 'throughput_raw' ) {
              #   $calculated->{$category}{$metric} = sprintf( "%.2f", ( $currstats->{$category}{$metric} - $laststats->{$category}{$metric} ) / $delta / 1024 );
              # }
              else {
                $calculated->{$metric} = sprintf( "%.4g", ( $currstats->{$metric} - $laststats->{$metric} ) / $delta );
              }
            }
            else {
              $calculated->{$metric} = "";
            }
            $line .= ";$calculated->{$metric}";
          }

          #if ( $item->{capacity} ) {
          #  if ( defined $item->{space}{$metric} ) {
          #    $line .= ";" . sprintf( "%.4g", $item->{space}{$metric} );
          #  }
          #  else {
          #    $line .= ";";
          #  }
          #}
          push @$lines, $line;
        }
      }
    }
  }
  if ( defined $lines && scalar @$lines > 1 ) {
    return join( "\n", @$lines );
  }
  else {
    return "";
  }
}

# $currentState->{volumes} = $netapp->readRestApi( { path => "/storage/volumes", fields => "statistics,space,aggregates,svm" } );

$server->close();
