use LWP;
use JSON;
use strict;
use warnings;
use Data::Dumper;
use FindBin;                
use lib "$FindBin::Bin/"; # use lib in this folder
use lib "../../../perl-modules";
use Test;
use Nutanix;
use Xormon;
use Digest::MD5 qw(md5_hex);
use POSIX;
use HTTP::Async;
use Net::HTTPS::NB;

$SIG{CHLD} = 'IGNORE';

use IO::Socket qw(AF_INET AF_UNIX SOCK_STREAM SHUT_WR);

my $DEFAULT_PORT = 3105;
my $port = Xormon::checkPrefixPort($DEFAULT_PORT);

my $server = IO::Socket->new(
    Domain => AF_INET,
    Type => SOCK_STREAM,
    Proto => 'tcp',
    LocalHost => '0.0.0.0',
    LocalPort => $port,
    ReusePort => 1,
    Listen => 5,
) || die "Can't open socket: $@";
Xormon::log("Service listening on port $port");

# api version per type
my %api = (
  "central" => {
    "cluster" => "v3",
    "host" => "v3",
    "vm" => "v3",
    "disk" => "v2",
    "container" => "v2",
    "pool" => "v1",
    "vdisk" => "v2"
  },
  "element" => {
    "cluster" => "v2",
    "host" => "v2",
    "vm" => "v2",
    "disk" => "v2",
    "container" => "v2",
    "pool" => "v1",
    "vdisk" => "v2"
  },
  "element_old" => {
    "cluster" => "v1",
    "host" => "v1",
    "vm" => "v1",
    "disk" => "v1",
    "container" => "v1",
    "pool" => "v1",
    "vdisk" => "v1"
  }
);

my %dictionary = (
  "hypervisor_cpu_usage_ppm" => {
    "label" => "cpu_utilization_percent",
    "division" => 10000
  },
  "hypervisor_memory_usage_ppm" => {
    "label" => "memory_usage",
    "division" => 10000
  },
  "controller_read_io_bandwidth_kBps" => {
    "label" => "read_io_bandwidth",
    "multiplication" => 1024
  },
  "read_io_bandwidth_kBps" => {
    "label" => "read_io_bandwidth",
    "multiplication" => 1024
  },
  "controller_write_io_bandwidth_kBps" => {
    "label" => "write_io_bandwidth",
    "multiplication" => 1024
  },
  "write_io_bandwidth_kBps" => {
    "label" => "write_io_bandwidth",
    "multiplication" => 1024
  },
  "controller_io_bandwidth_kBps" => {
    "label" => "total_io_bandwidth",
    "multiplication" => 1024
  },
  "controller_avg_write_io_latency_usecs" => {
    "label" => "write_io_latency",
    "division" => 1000
  },
  "controller_avg_read_io_latency_usecs" => {
    "label" => "read_io_latency",
    "division" => 1000
  },
  "controller_avg_io_latency_usecs" => {
    "label" => "total_io_latency",
    "division" => 1000
  },
  "avg_io_latency_usecs" => {
    "label" => "total_io_latency",
    "division" => 1000
  },
  "memory_usage_ppm" => {
    "label" => "memory_usage",
    "division" => 10000
  },
  "controller_num_read_io" => {
    "label" => "read_io"
  },
  "num_read_io" => {
    "label" => "read_io"
  },
  "controller_num_write_io" => {
    "label" => "write_io"
  },
  "num_write_io" => {
    "label" => "write_io"
  },
  "controller_num_io" => {
    "label" => "total_io"
  },
  "num_io" => {
    "label" => "total_io"
  },
  "hypervisor_num_transmitted_bytes" => {
    "label" => "transmitted"
  },
  "hypervisor_num_received_bytes" => {
    "label" => "received"
  },
  "memory_free_bytes" => {
    "label" => "memory_free"
  },
  "memory_usage_bytes" => {
    "label" => "memory_used"
  }
);

my $page_max = 200; # items per page

# waiting for a new client connection
my $client;
while ($client = $server->accept()) {
    
    my $pid;
    unless ( defined( $pid = fork() ) ) {
        Xormon::error("Wait for fork");
        sleep 5;
    }

    if ($pid == 0) {

        # get information about a newly connected client
        my $client_address = $client->peerhost();
        my $client_port = $client->peerport();
        Xormon::log("Connection from $client_address:$client_port");

        # received data template 
        #
        # { 
        #   "jwt"  : "<backendToken>",
        #   "task" : "<taskName>",
        #   "data" : {
        #     <configuration_from_hostcfg>
        #   }
        # }

        # read up to 2048 characters from the connected client
        my $data = "";
        $client->recv($data, 2048);

        my $decoded = from_json($data);
        my $cfg = $decoded->{data};

        Xormon::log("Received task: ".$decoded->{task});
        my $response = '';
        if ($decoded->{task} eq "fetch") {

            $client->send("Task successfuly started");
            $client->shutdown(SHUT_WR);

            my $xormon = Xormon->new("$client_address", $decoded->{jwt}, 'nutanix', $cfg->{hostcfg_id}, $cfg->{hostalias});
            
            # target destinations check
            if (defined $decoded->{'destinations'}) {
                $xormon->setDestinations($decoded->{'destinations'});
            }

            process($xormon, $cfg);

        } elsif ($decoded->{task} eq "test") {

            apitest($client, $cfg);

        } else {
            $response = "undefined task";

            $client->send($response);
            $client->shutdown(SHUT_WR);
        }

        exit();

    } else {
        $client->autoflush(1);    
    }
}

sub process {
    my $xormon = shift;
    my $cfg = shift;
    
    $xormon->startJob('fetch', $cfg->{hostalias}, $$);

    # get data from technology
    $cfg->{interval} = 60;

    eval {
      api2json($xormon, $cfg);
    };

    if ($@) {
      Xormon::error("Error: ".$@);
      $xormon->endJob('fetch', $cfg->{hostalias}, "error", $$);
    } else {
      $xormon->endJob('fetch', $cfg->{hostalias}, "successful", $$);
    }

    Xormon::log("Task [fetch] for ".$cfg->{hostalias}." is done");
}

sub api2json {
  my $xormon = shift;
  my $data = shift;

  my $connected = Test::connTest($data->{host}, $data->{port});
  if ($connected == 1) {
    $xormon->isConnected(1);
  } else {
    $xormon->isConnected(0, "TCP connection to " . $data->{host} . ":" . $data->{port} . " has failed!");
    return ();
  }

  my $async_slots = defined $data->{async_slots} ? $data->{async_slots} : 6;

  my $async = HTTP::Async->new( ssl_options => { SSL_verify_mode => 0, verify_hostname => 0 }, slots => $async_slots, timeout => 30 );

  my $service_data = $xormon->getData();

  my $timestamp_start = (defined $service_data->{timestamp}) ? $service_data->{timestamp} : time()-1200;
  my $timestamp_end = time()-180;

  $timestamp_start = $timestamp_start - ($timestamp_start % 60) - 180;
  $timestamp_end = $timestamp_end - ($timestamp_end % 60);

  if ($timestamp_end - $timestamp_start > 3600) { # 1 hour
   $timestamp_start = $timestamp_end - 3600;
  }

  # cluster + node, vms, disk + container + pool, virtual disks
  my $timestamp_start_cluster_node = defined $service_data->{timestamp_cluster_node} && $service_data->{timestamp_cluster_node} >= $timestamp_start 
    ? $service_data->{timestamp_cluster_node} 
    : $timestamp_start;
  my $timestamp_start_vm = defined $service_data->{timestamp_vm} && $service_data->{timestamp_vm} >= $timestamp_start
    ? $service_data->{timestamp_vm} 
    : $timestamp_start;
  my $timestamp_start_disk_container_pool = defined $service_data->{timestamp_disk_container_pool} && $service_data->{timestamp_disk_container_pool} >= $timestamp_start
    ? $service_data->{timestamp_disk_container_pool} 
    : $timestamp_start;
  my $timestamp_start_vdisk = defined $service_data->{timestamp_vdisk} && $service_data->{timestamp_vdisk} >= $timestamp_start
    ? $service_data->{timestamp_vdisk} 
    : $timestamp_start;

  my @architecture;
  my %conf;
  my %perf;
  my %tmp;
  my $index = 0;
  my $nutanix = Nutanix->new($data->{protocol}, $data->{host}, $data->{port}, $data->{username}, $data->{password});
  my $firstClusterId;
  my %savedIds;

  # clusters
  my $clusters = $nutanix->getClusters($api{$data->{type}}{'cluster'});
  Xormon::keepLastResponse($clusters, 'nutanix', $data->{hostcfg_id}, 'clusters');
  if ($api{$data->{type}}{'cluster'} eq "v3") {
      # API v3
    Nutanix::log("$data->{hostalias}: fetching ".$clusters->{metadata}{total_matches}." clusters");
    for (@{$clusters->{entities}}) {
      my $cluster = $_;
      $firstClusterId = $cluster->{metadata}->{uuid};

      #architecture
      my %cluster_arch = (
        "item_id" => $cluster->{metadata}->{uuid},
        "label" => $cluster->{status}->{name},
        "hw_type" => "nutanix",
        "subsystem" => "cluster",
        "parents" => [],
        "hostcfg_id" => $data->{hostcfg_id}
      );
      push(@architecture, \%cluster_arch);
      $savedIds{$cluster->{metadata}->{uuid}} = 1;
    }
  } elsif ($api{$data->{type}}{'cluster'} eq "v2") {
    # API v2 
    Nutanix::log("$data->{hostalias}: fetching ".$clusters->{metadata}{total_entities}." clusters");
    for (@{$clusters->{entities}}) {
      my $cluster = $_;
      $firstClusterId = $cluster->{uuid};

      #architecture
      my %cluster_arch = (
        "item_id" => $cluster->{uuid},
        "label" => $cluster->{name},
        "hw_type" => "nutanix",
        "subsystem" => "cluster",
        "parents" => [],
        "hostcfg_id" => $data->{hostcfg_id}
      );
      push(@architecture, \%cluster_arch);
      $savedIds{$cluster->{uuid}} = 1;
    }
  } elsif ($api{$data->{type}}{'cluster'} eq "v1") {
    # API v1
    Nutanix::log("$data->{hostalias}: fetching ".$clusters->{metadata}{totalEntities}." clusters");
    for (@{$clusters->{entities}}) {
      my $cluster = $_;
      $firstClusterId = $cluster->{uuid};

      #architecture
      my %cluster_arch = (
        "item_id" => $cluster->{uuid},
        "label" => $cluster->{name},
        "hw_type" => "nutanix",
        "subsystem" => "cluster",
        "parents" => [],
        "hostcfg_id" => $data->{hostcfg_id}
      );
      push(@architecture, \%cluster_arch);
      $savedIds{$cluster->{uuid}} = 1;
    }
  }
  $clusters = ();

  # nodes
  my $actualDataCheck = 0;
  my $nodes = $nutanix->getNodes($api{$data->{type}}{'host'});
  Xormon::keepLastResponse($nodes, 'nutanix', $data->{hostcfg_id}, 'nodes');
  if ($api{$data->{type}}{'host'} eq "v3") {
    # API v3
    Nutanix::log("$data->{hostalias}: fetching ".$nodes->{metadata}{total_matches}." hosts");
    for (@{$nodes->{entities}}) {
      my $host = $_;
      
      if (!defined $host->{status}{cluster_reference}{uuid}) {
         next;
      }

      #conf
      $conf{$host->{metadata}->{uuid}}{name} = $host->{spec}{name};
      $conf{$host->{metadata}->{uuid}}{address} = $host->{spec}{resources}{controller_vm}{ip};
      $conf{$host->{metadata}->{uuid}}{memory} = $host->{status}{resources}{memory_capacity_mib}*(1024*1024);
      $conf{$host->{metadata}->{uuid}}{cpu_count} = $host->{status}{resources}{num_cpu_cores};
      $conf{$host->{metadata}->{uuid}}{socket_count} = $host->{status}{resources}{num_cpu_sockets};
      $conf{$host->{metadata}->{uuid}}{cpu_model} = $host->{status}{resources}{cpu_model};
      $conf{$host->{metadata}->{uuid}}{version} = $host->{status}{resources}{hypervisor}{hypervisor_full_name};
      $conf{$host->{metadata}->{uuid}}{parent_cluster} = $host->{status}{cluster_reference}{uuid};

      $tmp{'host_cluster'}{$host->{metadata}->{uuid}} = $host->{status}{cluster_reference}{uuid};

      my @parents;
      if (defined $savedIds{$host->{status}{cluster_reference}{uuid}}) {
        push(@parents, $host->{status}{cluster_reference}{uuid});
      }

      #architecture
      my %host_arch = (
        "item_id" => $host->{metadata}->{uuid},
        "label" => $host->{spec}{name},
        "hw_type" => "nutanix",
        "subsystem" => "host",
        "parents" => \@parents,
        "hostcfg_id" => $data->{hostcfg_id}
      );
      $savedIds{$host->{metadata}->{uuid}} = 1;
      push(@architecture, \%host_arch);

      my $tmpTime;
      my $metrics = "hypervisor_cpu_usage_ppm,hypervisor_memory_usage_ppm";
      my $nodePerf = $nutanix->getNodePerf($host->{metadata}->{uuid}, $metrics, $timestamp_start_cluster_node, $timestamp_end, $data->{interval}, $api{$data->{type}}{'host'});
      for (@{$nodePerf->{"stats_specific_responses"}}) {
        my $nodeMetrics = $_;
        $tmpTime = ceil($nodeMetrics->{start_time_in_usecs}/1000000);
        for (@{$nodeMetrics->{values}}){
          my $value = $_;
          if ($value < 0) { $value = 0; }
          if ($nodeMetrics->{metric} eq "hypervisor_memory_usage_ppm") {
             $perf{'nutanix'}{'host'}{$host->{metadata}->{uuid}}{$tmpTime}{'memory_used'} = ($host->{status}{resources}{memory_capacity_mib}*1024*1024*($value/1000000));
             $perf{'nutanix'}{'host'}{$host->{metadata}->{uuid}}{$tmpTime}{'memory_free'} = ($host->{status}{resources}{memory_capacity_mib}*1024*1024*(1-($value/1000000)));
          }
          if ($nodeMetrics->{metric} eq "hypervisor_cpu_usage_ppm") {
            $perf{'nutanix'}{'host'}{$host->{metadata}->{uuid}}{$tmpTime}{'cpu_utilization_cores'} = ($host->{status}{resources}{num_cpu_cores}*$value/1000000);
            #$perf{'nutanix'}{'host'}{$host->{metadata}->{uuid}}{$tmpTime}{'cpu_utilization_cores_max'} = $perf{'nutanix'}{'host'}{$host->{metadata}->{uuid}}{$tmpTime}{'cpu_utilization_cores'};

            if (defined $dictionary{$nodeMetrics->{metric}}{'division'}) {
              #$perf{'nutanix'}{'host'}{$host->{metadata}->{uuid}}{$tmpTime}{'cpu_utilization_percent_max'} = $value/$dictionary{$nodeMetrics->{metric}}{'division'};
            }
          } 
          if (defined $dictionary{$nodeMetrics->{metric}}) {
            if (defined $dictionary{$nodeMetrics->{metric}}{'division'}) {
              $perf{'nutanix'}{'host'}{$host->{metadata}->{uuid}}{$tmpTime}{$dictionary{$nodeMetrics->{metric}}{'label'}} = $value/$dictionary{$nodeMetrics->{metric}}{'division'};
            } elsif (defined $dictionary{$nodeMetrics->{metric}}{'multiplication'}) {
              $perf{'nutanix'}{'host'}{$host->{metadata}->{uuid}}{$tmpTime}{$dictionary{$nodeMetrics->{metric}}{'label'}} = $value*$dictionary{$nodeMetrics->{metric}}{'multiplication'};
            } else {
              $perf{'nutanix'}{'host'}{$host->{metadata}->{uuid}}{$tmpTime}{$dictionary{$nodeMetrics->{metric}}{'label'}} = $value;
            }
          } else {
            $perf{'nutanix'}{'host'}{$host->{metadata}->{uuid}}{$tmpTime}{$nodeMetrics->{metric}} = $value;
          }
          $perf{'nutanix'}{'host'}{$host->{metadata}->{uuid}}{$tmpTime}{'cpu_cores'} = $host->{status}{resources}{num_cpu_cores};
          $perf{'nutanix'}{'host'}{$host->{metadata}->{uuid}}{$tmpTime}{'memory_total'} = $host->{status}{resources}{memory_capacity_mib}*1024*1024;
          $tmpTime += $data->{interval};
        }
        if ($actualDataCheck == 0 && time() - $tmpTime < (20 * 60)) {
          $actualDataCheck = 1;
        }
      }
    }
  } elsif ($api{$data->{type}}{'host'} eq "v2") {
    # API v2 
    Nutanix::log("$data->{hostalias}: fetching ".$nodes->{metadata}{total_entities}." hosts");
    for (@{$nodes->{entities}}) {
      my $host = $_;

      if (!defined $host->{cluster_uuid}) {
      next;
      }

      # conf
      $conf{$host->{uuid}}{name} = $host->{name};
      $conf{$host->{uuid}}{address} = $host->{hypervisor_address};
      $conf{$host->{uuid}}{memory} = $host->{memory_capacity_in_bytes};
      $conf{$host->{uuid}}{cpu_count} = $host->{num_cpu_cores};
      $conf{$host->{uuid}}{socket_count} = $host->{num_cpu_sockets};
      $conf{$host->{uuid}}{cpu_model} = $host->{cpu_model};
      $conf{$host->{uuid}}{version} = $host->{hypervisor_full_name};
      $conf{$host->{uuid}}{parent_cluster} = $host->{cluster_uuid};

      $tmp{'host_cluster'}{$host->{uuid}} = $host->{cluster_uuid};

      my @parents;
      if (defined $savedIds{$host->{cluster_uuid}}) {
        push(@parents, $host->{cluster_uuid});
      }

      #architecture
      my %host_arch = (
        "item_id" => $host->{uuid},
        "label" => $host->{name},
        "hw_type" => "nutanix",
        "subsystem" => "host",
        "parents" => \@parents,
        "hostcfg_id" => $data->{hostcfg_id}
      );
      $savedIds{$host->{uuid}} = 1;
      push(@architecture, \%host_arch);

      my $tmpTime;
      my $metrics = "hypervisor_cpu_usage_ppm,hypervisor_memory_usage_ppm";
      my $nodePerf = $nutanix->getNodePerf($host->{uuid}, $metrics, $timestamp_start_cluster_node, $timestamp_end, $data->{interval}, $api{$data->{type}}{'host'});
      for (@{$nodePerf->{"stats_specific_responses"}}) {
        my $nodeMetrics = $_;
        $tmpTime = ceil($nodeMetrics->{start_time_in_usecs}/1000000);
        for (@{$nodeMetrics->{values}}){
          my $value = $_;
          if ($value < 0) { $value = 0; }
          if ($nodeMetrics->{metric} eq "hypervisor_memory_usage_ppm") {
             $perf{'nutanix'}{'host'}{$host->{uuid}}{$tmpTime}{'memory_used'} = $host->{memory_capacity_in_bytes}*($value/1000000);
             $perf{'nutanix'}{'host'}{$host->{uuid}}{$tmpTime}{'memory_free'} = $host->{memory_capacity_in_bytes}*(1-($value/1000000));
          }
          if ($nodeMetrics->{metric} eq "hypervisor_cpu_usage_ppm") {
            $perf{'nutanix'}{'host'}{$host->{uuid}}{$tmpTime}{'cpu_utilization_cores'} = $host->{num_cpu_cores}*($value/1000000);
            #$perf{'nutanix'}{'host'}{$host->{uuid}}{$tmpTime}{'cpu_utilization_cores_max'} = $perf{'nutanix'}{'host'}{$host->{uuid}}{$tmpTime}{'cpu_utilization_cores'}; 

            if (defined $dictionary{$nodeMetrics->{metric}}{'division'}) {
              #$perf{'nutanix'}{'host'}{$host->{uuid}}{$tmpTime}{'cpu_utilization_percent_max'} = $value/$dictionary{$nodeMetrics->{metric}}{'division'};
            }
          }
          if (defined $dictionary{$nodeMetrics->{metric}}) {
            if (defined $dictionary{$nodeMetrics->{metric}}{'division'}) {
              $perf{'nutanix'}{'host'}{$host->{uuid}}{$tmpTime}{$dictionary{$nodeMetrics->{metric}}{'label'}} = $value/$dictionary{$nodeMetrics->{metric}}{'division'};
            } elsif (defined $dictionary{$nodeMetrics->{metric}}{'multiplication'}) {
              $perf{'nutanix'}{'host'}{$host->{uuid}}{$tmpTime}{$dictionary{$nodeMetrics->{metric}}{'label'}} = $value*$dictionary{$nodeMetrics->{metric}}{'multiplication'};
            } else {
              $perf{'nutanix'}{'host'}{$host->{uuid}}{$tmpTime}{$dictionary{$nodeMetrics->{metric}}{'label'}} = $value;
            }
          } else {
            $perf{'nutanix'}{'host'}{$host->{uuid}}{$tmpTime}{$nodeMetrics->{metric}} = $value;
          }
          $perf{'nutanix'}{'host'}{$host->{uuid}}{$tmpTime}{'cpu_cores'} = $host->{num_cpu_cores};
          $perf{'nutanix'}{'host'}{$host->{uuid}}{$tmpTime}{'memory_total'} = $host->{memory_capacity_in_bytes};
          $tmpTime += $data->{interval};
        }
        if ($actualDataCheck == 0 && time() - $tmpTime < (20 * 60)) {
          $actualDataCheck = 1;
        }
      }
    }
  } elsif ($api{$data->{type}}{'host'} eq "v1") {
    # API v1
    Nutanix::log("$data->{hostalias}: fetching ".$nodes->{metadata}{totalEntities}." hosts");
    for (@{$nodes->{entities}}) {
      my $host = $_;

      if (!defined $host->{clusterUuid}) {
      next;
      }

      # conf
      $conf{$host->{uuid}}{name} = $host->{name};
      $conf{$host->{uuid}}{address} = $host->{hypervisorAddress};
      $conf{$host->{uuid}}{memory} = $host->{memoryCapacityInBytes};
      $conf{$host->{uuid}}{cpu_count} = $host->{numCpuCores};
      $conf{$host->{uuid}}{socket_count} = $host->{numCpuSockets};
      $conf{$host->{uuid}}{cpu_model} = $host->{cpuModel};
      $conf{$host->{uuid}}{version} = $host->{hypervisorFullName};
      $conf{$host->{uuid}}{parent_cluster} = $host->{clusterUuid};

      $tmp{'host_cluster'}{$host->{uuid}} = $host->{clusterUuid};

      my @parents;
      if (defined $savedIds{$host->{clusterUuid}}) {
        push(@parents, $host->{clusterUuid});
      }

      #architecture
      my %host_arch = (
        "item_id" => $host->{uuid},
        "label" => $host->{name},
        "hw_type" => "nutanix",
        "subsystem" => "host",
        "parents" => \@parents,
        "hostcfg_id" => $data->{hostcfg_id}
      );
      $savedIds{$host->{uuid}} = 1;
      push(@architecture, \%host_arch);

      my $tmpTime;
      my $metrics = "hypervisor_cpu_usage_ppm,hypervisor_memory_usage_ppm";
      my $nodePerf = $nutanix->getNodePerf($host->{uuid}, $metrics, $timestamp_start_cluster_node, $timestamp_end, $data->{interval}, $api{$data->{type}}{'host'});
      for (@{$nodePerf->{"statsSpecificResponses"}}) {
        my $nodeMetrics = $_;
        $tmpTime = ceil($nodeMetrics->{startTimeInUsecs}/1000000);
        for (@{$nodeMetrics->{values}}){
          my $value = $_;
          if ($value < 0) { $value = 0; }
          if ($nodeMetrics->{metric} eq "hypervisor_memory_usage_ppm") {
             $perf{'nutanix'}{'host'}{$host->{uuid}}{$tmpTime}{'memory_used'} = $host->{memoryCapacityInBytes}*($value/1000000);
             $perf{'nutanix'}{'host'}{$host->{uuid}}{$tmpTime}{'memory_free'} = $host->{memoryCapacityInBytes}*(1-($value/1000000));
          }
          if ($nodeMetrics->{metric} eq "hypervisor_cpu_usage_ppm") {
            $perf{'nutanix'}{'host'}{$host->{uuid}}{$tmpTime}{'cpu_utilization_cores'} = $host->{numCpuCores}*($value/1000000);
            #$perf{'nutanix'}{'host'}{$host->{uuid}}{$tmpTime}{'cpu_utilization_cores_max'} = $perf{'nutanix'}{'host'}{$host->{uuid}}{$tmpTime}{'cpu_utilization_cores'};

            if (defined $dictionary{$nodeMetrics->{metric}}{'division'}) {
              #$perf{'nutanix'}{'host'}{$host->{uuid}}{$tmpTime}{'cpu_utilization_percent_max'} = $value/$dictionary{$nodeMetrics->{metric}}{'division'};
            }
          }
          if (defined $dictionary{$nodeMetrics->{metric}}) {
            if (defined $dictionary{$nodeMetrics->{metric}}{'division'}) {
              $perf{'nutanix'}{'host'}{$host->{uuid}}{$tmpTime}{$dictionary{$nodeMetrics->{metric}}{'label'}} = $value/$dictionary{$nodeMetrics->{metric}}{'division'};
            } elsif (defined $dictionary{$nodeMetrics->{metric}}{'multiplication'}) {
              $perf{'nutanix'}{'host'}{$host->{uuid}}{$tmpTime}{$dictionary{$nodeMetrics->{metric}}{'label'}} = $value*$dictionary{$nodeMetrics->{metric}}{'multiplication'};
            } else {
              $perf{'nutanix'}{'host'}{$host->{uuid}}{$tmpTime}{$dictionary{$nodeMetrics->{metric}}{'label'}} = $value;
            }
          } else {
            $perf{'nutanix'}{'host'}{$host->{uuid}}{$tmpTime}{$nodeMetrics->{metric}} = $value;
          }
          $perf{'nutanix'}{'host'}{$host->{uuid}}{$tmpTime}{'cpu_cores'} = $host->{numCpuCores};
          $perf{'nutanix'}{'host'}{$host->{uuid}}{$tmpTime}{'memory_total'} = $host->{memoryCapacityInBytes};
          $tmpTime += $data->{interval};
        }
        if ($actualDataCheck == 0 && time() - $tmpTime < (20 * 60)) {
          $actualDataCheck = 1;
        }
      }
    }
  }
  $nodes = ();

  if ($actualDataCheck == 1) {
    $xormon->hasData(1);
  } else {
    $xormon->hasData(0, "Unable to get actual node perf data");
  }

  # store cluster & nodes data
  $xormon->saveJSON2(\%perf);
  %perf = ();

  updateServiceData($xormon, 'timestamp_cluster_node', $timestamp_end-300);

  # vms 
  my $page = 1;
  my $loop = 1;
  while($loop eq "1") {

    my $vms = $nutanix->getVMs( $api{$data->{type}}{'vm'}, $page_max, $page );
    Xormon::keepLastResponse($vms, 'nutanix', $data->{hostcfg_id}, 'vms_'.$page);
    if ($api{$data->{type}}{'vm'} eq "v3") {
      # API v3

      if ($vms->{metadata}{total_matches} < ($page-1)*$page_max || $vms->{metadata}{length} eq "0") {
        $loop = 0;
        last;
      }

      Nutanix::log( "$data->{hostalias}: fetching " . $vms->{metadata}{length} . " vms" . " [step ".$page." / " . ceil($vms->{metadata}{total_matches}/$page_max) . "]" );
      for (@{$vms->{entities}}) {
        my $vm = $_;

        # if no cluster uuid
        if (!defined $vm->{spec}->{cluster_reference}->{uuid}) {
          next;
        }

        # if vm already saved
        if (defined $savedIds{$vm->{metadata}->{uuid}}) {
          next;
        }

        # conf
        $conf{$vm->{metadata}->{uuid}}{name} = $vm->{status}->{name};
        $conf{$vm->{metadata}->{uuid}}{memory} = $vm->{spec}->{resources}->{memory_size_mib}*(1024*1024);
        $conf{$vm->{metadata}->{uuid}}{cpu_count} = $vm->{spec}->{resources}->{num_vcpus_per_socket}*$vm->{spec}->{resources}->{num_sockets};
        $conf{$vm->{metadata}->{uuid}}{os} = " ";
        $conf{$vm->{metadata}->{uuid}}{parent_host} = $vm->{status}->{resources}->{host_reference}->{uuid};
        $conf{$vm->{metadata}->{uuid}}{parent_cluster} = $vm->{spec}->{cluster_reference}->{uuid};
        $conf{$vm->{metadata}->{uuid}}{hypervisor} = $vm->{status}->{resources}->{hypervisor_type};

        my @parents;
        if (defined $vm->{status}->{resources}->{host_reference}->{uuid} && defined $savedIds{$vm->{status}->{resources}->{host_reference}->{uuid}}) {
          push(@parents, $vm->{status}->{resources}->{host_reference}->{uuid});
        }
        if (defined $vm->{spec}->{cluster_reference}->{uuid} && defined $savedIds{$vm->{spec}->{cluster_reference}->{uuid}}) {
          push(@parents, $vm->{spec}->{cluster_reference}->{uuid});
        }

        #architecture
        my %vm_arch = (
          "item_id" => $vm->{metadata}->{uuid},
          "label" => $vm->{status}->{name},
          "hw_type" => "nutanix",
          "subsystem" => "vm",
          "parents" => \@parents,
          "hostcfg_id" => $data->{hostcfg_id}
        );
        $savedIds{$vm->{metadata}->{uuid}} = 1;
        push(@architecture, \%vm_arch);

        #if (total_size(\%perf) >= $max_size) { savePerf(\%perf, $hostcfg_uuid, $timestamp_start, $timestamp_end, $index); $index++; %perf = (); }

        # my $tmpTime;
        # my $metrics = "controller_read_io_bandwidth_kBps,hypervisor_cpu_usage_ppm,controller_write_io_bandwidth_kBps,controller_io_bandwidth_kBps";
        # my $vmPerf = $nutanix->getVMPerf($vm->{metadata}->{uuid}, $metrics, $timestamp_start, $timestamp_end, $data->{interval});
        # for (@{$vmPerf->{"statsSpecificResponses"}}) {
        #   my $vmMetrics = $_;
        #   $tmpTime = ceil($vmMetrics->{startTimeInUsecs}/1000000);
        #   for (@{$vmMetrics->{values}}){
        #     my $value = $_;
        #     if ($value < 0) { $value = 0; }
        #     if ($vmMetrics->{metric} eq "hypervisor_cpu_usage_ppm") {
        #       $perf{'nutanix'}{'vm'}{$vm->{metadata}->{uuid}}{$tmpTime}{'cpu_utilization_cores'} = ($vm->{spec}->{resources}->{num_vcpus_per_socket}*$vm->{spec}->{resources}->{num_sockets})*($value/1000000);
        #       #$perf{'nutanix'}{'vm'}{$vm->{metadata}->{uuid}}{$tmpTime}{'cpu_utilization_cores_max'} = $perf{'nutanix'}{'vm'}{$vm->{metadata}->{uuid}}{$tmpTime}{'cpu_utilization_cores'};

        #       if (defined $dictionary{$vmMetrics->{metric}}{'division'}) {
        #         #$perf{'nutanix'}{'vm'}{$vm->{metadata}->{uuid}}{$tmpTime}{'cpu_utilization_percent_max'} = $value/$dictionary{$vmMetrics->{metric}}{'division'};
        #       }
        #     }
        #     if (defined $dictionary{$vmMetrics->{metric}}) {
        #       if (defined $dictionary{$vmMetrics->{metric}}{'division'}) {
        #         $perf{'nutanix'}{'vm'}{$vm->{metadata}->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = $value/$dictionary{$vmMetrics->{metric}}{'division'};
        #       } elsif (defined $dictionary{$vmMetrics->{metric}}{'multiplication'}) {
        #         $perf{'nutanix'}{'vm'}{$vm->{metadata}->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = $value*$dictionary{$vmMetrics->{metric}}{'multiplication'};
        #       } else {
        #         $perf{'nutanix'}{'vm'}{$vm->{metadata}->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = $value;
        #       }
        #     } else {
        #       $perf{'nutanix'}{'vm'}{$vm->{metadata}->{uuid}}{$tmpTime}{$vmMetrics->{metric}} = $value;
        #     }
        #     $tmpTime += $data->{interval};
        #   }
        # }

        # $metrics = "controller_avg_write_io_latency_usecs,controller_avg_read_io_latency_usecs,controller_avg_io_latency_usecs,memory_usage_ppm,guest.memory_usage_bytes";
        # $vmPerf = $nutanix->getVMPerf($vm->{metadata}->{uuid}, $metrics, $timestamp_start, $timestamp_end, $data->{interval});
        # for (@{$vmPerf->{"statsSpecificResponses"}}) {
        #   my $vmMetrics = $_;
        #   $tmpTime = ceil($vmMetrics->{startTimeInUsecs}/1000000);
        #   for (@{$vmMetrics->{values}}){
        #     my $value = $_;
        #     if ($value < 0) { $value = 0; }
        #     if ($vmMetrics->{metric} eq "memory_usage_ppm") {
        #       $perf{'nutanix'}{'vm'}{$vm->{metadata}->{uuid}}{$tmpTime}{'memory_used'} = ($vm->{spec}->{resources}->{memory_size_mib}*1024*1024)*($value/1000000);
        #       $perf{'nutanix'}{'vm'}{$vm->{metadata}->{uuid}}{$tmpTime}{'memory_free'} = ($vm->{spec}->{resources}->{memory_size_mib}*1024*1024)*(1-($value/1000000));
        #     }
        #     if (defined $dictionary{$vmMetrics->{metric}}) {
        #       if (defined $dictionary{$vmMetrics->{metric}}{'division'}) {
        #         $perf{'nutanix'}{'vm'}{$vm->{metadata}->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = $value/$dictionary{$vmMetrics->{metric}}{'division'};
        #       } elsif (defined $dictionary{$vmMetrics->{metric}}{'multiplication'}) {
        #         $perf{'nutanix'}{'vm'}{$vm->{metadata}->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = $value*$dictionary{$vmMetrics->{metric}}{'multiplication'};
        #       } else {
        #         $perf{'nutanix'}{'vm'}{$vm->{metadata}->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = $value;
        #       }
        #     } else {
        #       $perf{'nutanix'}{'vm'}{$vm->{metadata}->{uuid}}{$tmpTime}{$vmMetrics->{metric}} = $value;
        #     }
        #     $perf{'nutanix'}{'vm'}{$vm->{metadata}->{uuid}}{$tmpTime}{'cpu_cores'} = $vm->{spec}->{resources}->{num_vcpus_per_socket}*$vm->{spec}->{resources}->{num_sockets};
        #     $perf{'nutanix'}{'vm'}{$vm->{metadata}->{uuid}}{$tmpTime}{'memory_total'} = $vm->{spec}->{resources}->{memory_size_mib}*1024*1024;
        #     $tmpTime += $data->{interval};
        #   }
        # }

        # $metrics = "controller_num_read_io,controller_num_write_io,controller_num_io,hypervisor_num_transmitted_bytes,hypervisor_num_received_bytes";
        # $vmPerf = $nutanix->getVMPerf($vm->{metadata}->{uuid}, $metrics, $timestamp_start, $timestamp_end, $data->{interval});
        # for (@{$vmPerf->{"statsSpecificResponses"}}) {
        #   my $vmMetrics = $_;
        #   $tmpTime = ceil($vmMetrics->{startTimeInUsecs}/1000000);
        #   for (@{$vmMetrics->{values}}){
        #     my $value = $_;
        #     if ($value < 0) { $value = 0; }
        #     if (defined $dictionary{$vmMetrics->{metric}}) {
        #       if (defined $dictionary{$vmMetrics->{metric}}{'division'}) {
        #         $perf{'nutanix'}{'vm'}{$vm->{metadata}->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = ($value/$dictionary{$vmMetrics->{metric}}{'division'})/$data->{interval};
        #       } elsif (defined $dictionary{$vmMetrics->{metric}}{'multiplication'}) {
        #         $perf{'nutanix'}{'vm'}{$vm->{metadata}->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = ($value*$dictionary{$vmMetrics->{metric}}{'multiplication'})/$data->{interval};
        #       } else {
        #         $perf{'nutanix'}{'vm'}{$vm->{metadata}->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = $value/$data->{interval};;
        #       }
        #     } else {
        #       $perf{'nutanix'}{'vm'}{$vm->{metadata}->{uuid}}{$tmpTime}{$vmMetrics->{metric}} = $value/$data->{interval};
        #     }
        #     $tmpTime += $data->{interval};
        #   }
        # }

        # ASYNC 
        my $metrics = "controller_read_io_bandwidth_kBps,hypervisor_cpu_usage_ppm,controller_write_io_bandwidth_kBps,controller_io_bandwidth_kBps";
        $async->add(
          $nutanix->getVMPerfRequest($vm->{metadata}->{uuid}, $metrics, $timestamp_start_vm, $timestamp_end, $data->{interval})
        );
        
        $metrics = "controller_avg_write_io_latency_usecs,controller_avg_read_io_latency_usecs,controller_avg_io_latency_usecs,memory_usage_ppm,guest.memory_usage_bytes";
        $async->add(
          $nutanix->getVMPerfRequest($vm->{metadata}->{uuid}, $metrics, $timestamp_start_vm, $timestamp_end, $data->{interval})
        );
        
        $metrics = "controller_num_read_io,controller_num_write_io,controller_num_io,hypervisor_num_transmitted_bytes,hypervisor_num_received_bytes";
        $async->add(
          $nutanix->getVMPerfRequest($vm->{metadata}->{uuid}, $metrics, $timestamp_start_vm, $timestamp_end, $data->{interval})
        );
        # END ASYNC

      }
    } elsif ($api{$data->{type}}{'vm'} eq "v2") {
      # API v2

      if ($vms->{metadata}{total_entities} < ($page-1)*$page_max || $vms->{metadata}{start_index} eq "-1") {
        $loop = 0;
        last;
      }

      Nutanix::log( "$data->{hostalias}: fetching ".(($vms->{metadata}{end_index}+1)-$vms->{metadata}{start_index})." vms [step ".$page." / ".ceil($vms->{metadata}{total_entities}/$page_max)."]");
      for (@{$vms->{entities}}) {
        my $vm = $_;

        # if no host uuid
        if (!defined $vm->{host_uuid}) {
          next;
        }

        # if vm already saved
        if (defined $savedIds{$vm->{uuid}}) {
          next;
        }

        # conf
        $conf{ $vm->{uuid} }{name}           = $vm->{name};
        $conf{ $vm->{uuid} }{memory}         = $vm->{memory_mb}*(1024*1024);
        $conf{ $vm->{uuid} }{cpu_count}      = $vm->{num_vcpus};
        $conf{ $vm->{uuid} }{os}             = " ";
        $conf{ $vm->{uuid} }{parent_host}    = $vm->{host_uuid};
        $conf{ $vm->{uuid} }{parent_cluster} = $tmp{host_cluster}{$vm->{host_uuid}};
        $conf{ $vm->{uuid} }{hypervisor}     = " ";

        my @parents;
        if (defined $vm->{host_uuid} && defined $savedIds{$vm->{host_uuid}}) {
          push(@parents, $vm->{host_uuid});
        }
        if (defined $tmp{host_cluster}{$vm->{host_uuid}} && defined $savedIds{$tmp{host_cluster}{$vm->{host_uuid}}}) {
          push(@parents, $tmp{host_cluster}{$vm->{host_uuid}});
        }

        #architecture
        my %vm_arch = (
          "item_id" => $vm->{uuid},
          "label" => $vm->{name},
          "hw_type" => "nutanix",
          "subsystem" => "vm",
          "parents" => \@parents,
          "hostcfg_id" => $data->{hostcfg_id}
        );
        $savedIds{$vm->{uuid}} = 1;
        push(@architecture, \%vm_arch);

        #if (total_size(\%perf) >= $max_size) { savePerf(\%perf, $hostcfg_uuid, $timestamp_start, $timestamp_end, $index); $index++; %perf = (); }

        # my $tmpTime;
        # my $metrics = "controller_read_io_bandwidth_kBps,hypervisor_cpu_usage_ppm,controller_write_io_bandwidth_kBps,controller_io_bandwidth_kBps";
        # my $vmPerf = $nutanix->getVMPerf($vm->{uuid}, $metrics, $timestamp_start, $timestamp_end, $data->{interval});
        # for (@{$vmPerf->{"statsSpecificResponses"}}) {
        #   my $vmMetrics = $_;
        #   $tmpTime = ceil($vmMetrics->{startTimeInUsecs}/1000000);
        #   for (@{$vmMetrics->{values}}){
        #     my $value = $_;
        #     if ($value < 0) { $value = 0; }
        #     if ($vmMetrics->{metric} eq "hypervisor_cpu_usage_ppm") {
        #       $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{'cpu_utilization_cores'} = ($vm->{num_vcpus})*($value/1000000); 
        #       #$perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{'cpu_utilization_cores_max'} = $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{'cpu_utilization_cores'};

        #       if (defined $dictionary{$vmMetrics->{metric}}{'division'}) {
        #         #$perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{'cpu_utilization_percent_max'} = $value/$dictionary{$vmMetrics->{metric}}{'division'};
        #       }
        #     }
        #     if (defined $dictionary{$vmMetrics->{metric}}) {
        #       if (defined $dictionary{$vmMetrics->{metric}}{'division'}) {
        #         $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = $value/$dictionary{$vmMetrics->{metric}}{'division'};
        #       } elsif (defined $dictionary{$vmMetrics->{metric}}{'multiplication'}) {
        #         $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = $value*$dictionary{$vmMetrics->{metric}}{'multiplication'};
        #       } else {
        #         $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = $value;
        #       }
        #     } else {
        #       $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{$vmMetrics->{metric}} = $value;
        #     }
        #     $tmpTime += $data->{interval};
        #   }
        # }

        # $metrics = "controller_avg_write_io_latency_usecs,controller_avg_read_io_latency_usecs,controller_avg_io_latency_usecs,memory_usage_ppm,guest.memory_usage_bytes";
        # $vmPerf = $nutanix->getVMPerf($vm->{uuid}, $metrics, $timestamp_start, $timestamp_end, $data->{interval});
        # for (@{$vmPerf->{"statsSpecificResponses"}}) {
        #   my $vmMetrics = $_;
        #   $tmpTime = ceil($vmMetrics->{startTimeInUsecs}/1000000);
        #   for (@{$vmMetrics->{values}}){
        #     my $value = $_;
        #     if ($value < 0) { $value = 0; }
        #     if ($vmMetrics->{metric} eq "memory_usage_ppm") {
        #       $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{'memory_used'} = ($vm->{memory_mb}*1024*1024)*($value/1000000);
        #       $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{'memory_free'} = ($vm->{memory_mb}*1024*1024)*(1-($value/1000000));
        #     }
        #     if (defined $dictionary{$vmMetrics->{metric}}) {
        #       if (defined $dictionary{$vmMetrics->{metric}}{'division'}) {
        #         $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = $value/$dictionary{$vmMetrics->{metric}}{'division'};
        #       } elsif (defined $dictionary{$vmMetrics->{metric}}{'multiplication'}) {
        #         $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = $value*$dictionary{$vmMetrics->{metric}}{'multiplication'};
        #       } else {
        #         $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = $value;
        #       }
        #     } else {
        #       $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{$vmMetrics->{metric}} = $value;
        #     }
        #     $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{'cpu_cores'} = $vm->{num_vcpus};
        #     $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{'memory_total'} = $vm->{memory_mb}*1024*1024;
        #     $tmpTime += $data->{interval};
        #   }
        # }

        # $metrics = "controller_num_read_io,controller_num_write_io,controller_num_io,hypervisor_num_transmitted_bytes,hypervisor_num_received_bytes";
        # $vmPerf = $nutanix->getVMPerf($vm->{uuid}, $metrics, $timestamp_start, $timestamp_end, $data->{interval});
        # for (@{$vmPerf->{"statsSpecificResponses"}}) {
        #   my $vmMetrics = $_;
        #   $tmpTime = ceil($vmMetrics->{startTimeInUsecs}/1000000);
        #   for (@{$vmMetrics->{values}}){
        #     my $value = $_;
        #     if ($value < 0) { $value = 0; }
        #     if (defined $dictionary{$vmMetrics->{metric}}) {
        #       if (defined $dictionary{$vmMetrics->{metric}}{'division'}) {
        #         $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = ($value/$dictionary{$vmMetrics->{metric}}{'division'})/$data->{interval};
        #       } elsif (defined $dictionary{$vmMetrics->{metric}}{'multiplication'}) {
        #         $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = ($value*$dictionary{$vmMetrics->{metric}}{'multiplication'})/$data->{interval};
        #       } else {
        #         $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = $value/$data->{interval};
        #       }
        #     } else {
        #       $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{$vmMetrics->{metric}} = $value/$data->{interval};
        #     }
        #     $tmpTime += $data->{interval};
        #   }
        # }

        # ASYNC 
        my $metrics = "controller_read_io_bandwidth_kBps,hypervisor_cpu_usage_ppm,controller_write_io_bandwidth_kBps,controller_io_bandwidth_kBps";
        $async->add(
          $nutanix->getVMPerfRequest($vm->{uuid}, $metrics, $timestamp_start_vm, $timestamp_end, $data->{interval})
        );
        
        $metrics = "controller_avg_write_io_latency_usecs,controller_avg_read_io_latency_usecs,controller_avg_io_latency_usecs,memory_usage_ppm,guest.memory_usage_bytes";
        $async->add(
          $nutanix->getVMPerfRequest($vm->{uuid}, $metrics, $timestamp_start_vm, $timestamp_end, $data->{interval})
        );
        
        $metrics = "controller_num_read_io,controller_num_write_io,controller_num_io,hypervisor_num_transmitted_bytes,hypervisor_num_received_bytes";
        $async->add(
          $nutanix->getVMPerfRequest($vm->{uuid}, $metrics, $timestamp_start_vm, $timestamp_end, $data->{interval})
        );
        # END ASYNC

      }
    } elsif ($api{$data->{type}}{'vm'} eq "v1") {
      # API v1

      if ($vms->{metadata}{totalEntities} < ($page-1)*$page_max || $vms->{metadata}{startIndex} eq "-1") {
        $loop = 0;
        last;
      }

      Nutanix::log( "$data->{hostalias}: fetching ".(($vms->{metadata}{endIndex}+1)-$vms->{metadata}{startIndex})." vms [step ".$page." / ".ceil($vms->{metadata}{totalEntities}/$page_max)."]");
      for (@{$vms->{entities}}) {
        my $vm = $_;

        # if no cluster uuid
        if (!defined $vm->{clusterUuid}) {
          next;
        }

        # if vm already saved
        if (defined $savedIds{$vm->{uuid}}) {
          next;
        }

        # conf
        $conf{$vm->{uuid}}{name} = $vm->{vmName};
        $conf{$vm->{uuid}}{memory} = $vm->{memoryCapacityInBytes};
        $conf{$vm->{uuid}}{cpu_count} = $vm->{numVCpus};
        $conf{$vm->{uuid}}{os} = " ";
        $conf{$vm->{uuid}}{parent_host} = $vm->{hostUuid};
        $conf{$vm->{uuid}}{parent_cluster} = $vm->{clusterUuid};
        $conf{$vm->{uuid}}{hypervisor} = $vm->{hypervisorType};

        my @parents;
        if (defined $vm->{hostUuid} && defined $savedIds{$vm->{hostUuid}}) {
          push(@parents, $vm->{hostUuid});
        }
        if (defined $vm->{clusterUuid} && defined $savedIds{$vm->{clusterUuid}}) {
          push(@parents, $vm->{clusterUuid});
        }

        #architecture
        my %vm_arch = (
          "item_id" => $vm->{uuid},
          "label" => $vm->{vmName},
          "hw_type" => "nutanix",
          "subsystem" => "vm",
          "parents" => \@parents,
          "hostcfg_id" => $data->{hostcfg_id}
        );
        $savedIds{$vm->{uuid}} = 1;
        push(@architecture, \%vm_arch);

        #if (total_size(\%perf) >= $max_size) { savePerf(\%perf, $hostcfg_uuid, $timestamp_start, $timestamp_end, $index); $index++; %perf = (); }

        # my $tmpTime;
        # my $metrics = "controller_read_io_bandwidth_kBps,hypervisor_cpu_usage_ppm,controller_write_io_bandwidth_kBps,controller_io_bandwidth_kBps";
        # my $vmPerf = $nutanix->getVMPerf($vm->{uuid}, $metrics, $timestamp_start, $timestamp_end, $data->{interval});
        # for (@{$vmPerf->{"statsSpecificResponses"}}) {
        #   my $vmMetrics = $_;
        #   $tmpTime = ceil($vmMetrics->{startTimeInUsecs}/1000000);
        #   for (@{$vmMetrics->{values}}){
        #     my $value = $_;
        #     if ($value < 0) { $value = 0; }
        #     if ($vmMetrics->{metric} eq "hypervisor_cpu_usage_ppm") {
        #       $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{'cpu_utilization_cores'} = ($vm->{numVCpus})*($value/1000000);
        #       #$perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{'cpu_utilization_cores_max'} = $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{'cpu_utilization_cores'};

        #       if (defined $dictionary{$vmMetrics->{metric}}{'division'}) {
        #         #$perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{'cpu_utilization_percent_max'} = $value/$dictionary{$vmMetrics->{metric}}{'division'};
        #       }
        #     }
        #     if (defined $dictionary{$vmMetrics->{metric}}) {
        #       if (defined $dictionary{$vmMetrics->{metric}}{'division'}) {
        #         $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = $value/$dictionary{$vmMetrics->{metric}}{'division'};
        #       } elsif (defined $dictionary{$vmMetrics->{metric}}{'multiplication'}) {
        #         $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = $value*$dictionary{$vmMetrics->{metric}}{'multiplication'};
        #       } else {
        #         $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = $value;
        #       }
        #     } else {
        #       $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{$vmMetrics->{metric}} = $value;
        #     }
        #     $tmpTime += $data->{interval};
        #   }
        # }

        # $metrics = "controller_avg_write_io_latency_usecs,controller_avg_read_io_latency_usecs,controller_avg_io_latency_usecs,memory_usage_ppm,guest.memory_usage_bytes";
        # $vmPerf = $nutanix->getVMPerf($vm->{uuid}, $metrics, $timestamp_start, $timestamp_end, $data->{interval});
        # for (@{$vmPerf->{"statsSpecificResponses"}}) {
        #   my $vmMetrics = $_;
        #   $tmpTime = ceil($vmMetrics->{startTimeInUsecs}/1000000);
        #   for (@{$vmMetrics->{values}}){
        #     my $value = $_;
        #     if ($value < 0) { $value = 0; }
        #     if ($vmMetrics->{metric} eq "memory_usage_ppm") {
        #       $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{'memory_used'} = ($vm->{memoryCapacityInBytes})*($value/1000000);
        #       $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{'memory_free'} = ($vm->{memoryCapacityInBytes})*(1-($value/1000000));
        #     }
        #     if (defined $dictionary{$vmMetrics->{metric}}) {
        #       if (defined $dictionary{$vmMetrics->{metric}}{'division'}) {
        #         $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = $value/$dictionary{$vmMetrics->{metric}}{'division'};
        #       } elsif (defined $dictionary{$vmMetrics->{metric}}{'multiplication'}) {
        #         $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = $value*$dictionary{$vmMetrics->{metric}}{'multiplication'};
        #       } else {
        #         $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = $value;
        #       }
        #     } else {
        #       $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{$vmMetrics->{metric}} = $value;
        #     }
        #     $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{'cpu_cores'} = $vm->{numVCpus};
        #     $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{'memory_total'} = $vm->{memoryCapacityInBytes};
        #     $tmpTime += $data->{interval};
        #   }
        # }

        # $metrics = "controller_num_read_io,controller_num_write_io,controller_num_io,hypervisor_num_transmitted_bytes,hypervisor_num_received_bytes";
        # $vmPerf = $nutanix->getVMPerf($vm->{uuid}, $metrics, $timestamp_start, $timestamp_end, $data->{interval});
        # for (@{$vmPerf->{"statsSpecificResponses"}}) {
        #   my $vmMetrics = $_;
        #   $tmpTime = ceil($vmMetrics->{startTimeInUsecs}/1000000);
        #   for (@{$vmMetrics->{values}}){
        #     my $value = $_;
        #     if ($value < 0) { $value = 0; }
        #     if ($value < 0) { $value = 0; }
        #     if (defined $dictionary{$vmMetrics->{metric}}) {
        #       if (defined $dictionary{$vmMetrics->{metric}}{'division'}) {
        #         $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = ($value/$dictionary{$vmMetrics->{metric}}{'division'})/$data->{interval};
        #       } elsif (defined $dictionary{$vmMetrics->{metric}}{'multiplication'}) {
        #         $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = ($value*$dictionary{$vmMetrics->{metric}}{'multiplication'})/$data->{interval};
        #       } else {
        #         $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = $value/$data->{interval};
        #       }
        #     } else {
        #       $perf{'nutanix'}{'vm'}{$vm->{uuid}}{$tmpTime}{$vmMetrics->{metric}} = $value/$data->{interval};
        #     }
        #     $tmpTime += $data->{interval};
        #   }
        # }

        # ASYNC 
        my $metrics = "controller_read_io_bandwidth_kBps,hypervisor_cpu_usage_ppm,controller_write_io_bandwidth_kBps,controller_io_bandwidth_kBps";
        $async->add(
          $nutanix->getVMPerfRequest($vm->{uuid}, $metrics, $timestamp_start_vm, $timestamp_end, $data->{interval})
        );
        
        $metrics = "controller_avg_write_io_latency_usecs,controller_avg_read_io_latency_usecs,controller_avg_io_latency_usecs,memory_usage_ppm,guest.memory_usage_bytes";
        $async->add(
          $nutanix->getVMPerfRequest($vm->{uuid}, $metrics, $timestamp_start_vm, $timestamp_end, $data->{interval})
        );
        
        $metrics = "controller_num_read_io,controller_num_write_io,controller_num_io,hypervisor_num_transmitted_bytes,hypervisor_num_received_bytes";
        $async->add(
          $nutanix->getVMPerfRequest($vm->{uuid}, $metrics, $timestamp_start_vm, $timestamp_end, $data->{interval})
        );
        # END ASYNC

      }
    }
    $page++;
    $vms = ();
  }

  # VM ASYNC 
  my $asyncTime = time();
  Nutanix::log("$data->{hostalias}: ".$async->info);
  while ( my $response = $async->wait_for_next_response ) {

    if ($response->{_rc} != 200) {
      Nutanix::error("$data->{hostalias}: [Error] [VM] Request: " . $response->{_request}{_uri} . " Message: " . $response->{_msg});
    }

    my $uuid = $response->{_request}{_headers}{uuid};
    my $vm = $conf{$uuid};

    my $vmPerf = Nutanix::decode_json_eval($response->content, $data->{hostalias}. ": [VM Perf]");
    my $tmpTime;

    my @divisionMetrics = ('controller_num_read_io','controller_num_write_io','controller_num_io','hypervisor_num_transmitted_bytes','hypervisor_num_received_bytes');
    for (@{$vmPerf->{"statsSpecificResponses"}}) {
      my $vmMetrics = $_;
      $tmpTime = ceil($vmMetrics->{startTimeInUsecs}/1000000);
      for (@{$vmMetrics->{values}}){
        my $value = $_;
        if ($value < 0) { $value = 0; }

        if ($vmMetrics->{metric} eq "hypervisor_cpu_usage_ppm") { 
          $perf{'nutanix'}{'vm'}{$uuid}{$tmpTime}{'cpu_utilization_cores'} = ($vm->{cpu_count})*($value/1000000); 
        } elsif ($vmMetrics->{metric} eq "memory_usage_ppm") {
          $perf{'nutanix'}{'vm'}{$uuid}{$tmpTime}{'memory_used'} = ($vm->{memory})*($value/1000000);
          $perf{'nutanix'}{'vm'}{$uuid}{$tmpTime}{'memory_free'} = ($vm->{memory})*(1-($value/1000000));
        }

        if (defined $dictionary{$vmMetrics->{metric}}) {
          if (defined $dictionary{$vmMetrics->{metric}}{'division'}) {
            $perf{'nutanix'}{'vm'}{$uuid}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = $value/$dictionary{$vmMetrics->{metric}}{'division'};
          } elsif (defined $dictionary{$vmMetrics->{metric}}{'multiplication'}) {
            $perf{'nutanix'}{'vm'}{$uuid}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = $value*$dictionary{$vmMetrics->{metric}}{'multiplication'};
          } else {
            $perf{'nutanix'}{'vm'}{$uuid}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = $value;
          }
        } else {
          $perf{'nutanix'}{'vm'}{$uuid}{$tmpTime}{$vmMetrics->{metric}} = $value;
        }

        if ( grep( /^$vmMetrics->{metric}$/, @divisionMetrics ) ) { 
          $perf{'nutanix'}{'vm'}{$uuid}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} = $perf{'nutanix'}{'vm'}{$uuid}{$tmpTime}{$dictionary{$vmMetrics->{metric}}{'label'}} / $data->{interval};
        }

        $perf{'nutanix'}{'vm'}{$uuid}{$tmpTime}{'cpu_cores'} = $vm->{cpu_count};
        $perf{'nutanix'}{'vm'}{$uuid}{$tmpTime}{'memory_total'} = $vm->{memory};
        $tmpTime += $data->{interval};
      }
    }
  }
  my $asyncDuration = time() - $asyncTime;
  Nutanix::log("$data->{hostalias}: [VM] Async duration: $asyncDuration secs");
  # END VM ASYNC 

  # store vms data
  $xormon->saveJSON2(\%perf);
  %perf = ();

  updateServiceData($xormon, 'timestamp_vm', $timestamp_end-300);
  
  # disk
  $page = 1;
  $loop = 1;
  while($loop eq "1") {
    if (!defined $api{$data->{type}}{'disk'}) { last; }
    #if (total_size(\%perf) >= $max_size) { savePerf(\%perf, $hostcfg_uuid, $timestamp_start, $timestamp_end, $index); $index++; %perf = (); }
    my $disks = $nutanix->getDisks($api{$data->{type}}{'disk'}, $page_max, $page);
    Xormon::keepLastResponse($disks, 'nutanix', $data->{hostcfg_id}, 'disks_'.$page);
    if ($api{$data->{type}}{'disk'} eq "v3" || $api{$data->{type}}{'disk'} eq "v2") {
      if ($disks->{metadata}{total_entities} < ($page-1)*$page_max || $disks->{metadata}{start_index} eq "-1") {
        $loop = 0;
        last;
      }
      # API v3 & v2
      Nutanix::log("$data->{hostalias}: fetching ".(($disks->{metadata}{end_index}+1)-$disks->{metadata}{start_index})." disks [step ".$page." / ".ceil($disks->{metadata}{total_entities}/$page_max)."]");
      for (@{$disks->{entities}}) {
        my $disk = $_;

        $conf{$disk->{disk_uuid}}{label} = $disk->{node_name}." - ".$disk->{location};
        $conf{$disk->{disk_uuid}}{type} = $disk->{storage_tier_name};
        $conf{$disk->{disk_uuid}}{node_uuid} = $disk->{node_uuid};
        $conf{$disk->{disk_uuid}}{physical_size} = ceil($disk->{usage_stats}->{'storage.capacity_bytes'});

        my @parents;
        if (defined $savedIds{$disk->{node_uuid}}) {
          push(@parents, $disk->{node_uuid});
        }
        if (defined $savedIds{$disk->{cluster_uuid}}) {
          push(@parents, $disk->{cluster_uuid});
        }

        #architecture
        my %disk_arch = (
          "item_id" => $disk->{disk_uuid},
          "label" => $disk->{node_name}." - ".$disk->{location},
          "hw_type" => "nutanix",
          "subsystem" => "disk",
          "parents" => \@parents,
          "hostcfg_id" => $data->{hostcfg_id}
        );
        $savedIds{$disk->{disk_uuid}} = 1;
        push(@architecture, \%disk_arch);


        my $tmpTime;
        my $metrics = "read_io_bandwidth_kBps,write_io_bandwidth_kBps,avg_io_latency_usecs";
        my $diskPerf = $nutanix->getDiskPerf($disk->{disk_uuid}, $metrics, $timestamp_start_disk_container_pool, $timestamp_end, $data->{interval}, $api{$data->{type}}{'disk'});
        for (@{$diskPerf->{"stats_specific_responses"}}) {
          my $diskMetrics = $_;
          $tmpTime = ceil($diskMetrics->{start_time_in_usecs}/1000000);
          for (@{$diskMetrics->{values}}){
            my $value = $_;
            if ($value < 0) { $value = 0; }
            if (defined $dictionary{$diskMetrics->{metric}}) {
              if (defined $dictionary{$diskMetrics->{metric}}{'division'}) {
                $perf{'nutanix'}{'disk'}{$disk->{disk_uuid}}{$tmpTime}{$dictionary{$diskMetrics->{metric}}{'label'}} = $value/$dictionary{$diskMetrics->{metric}}{'division'};
              } elsif (defined $dictionary{$diskMetrics->{metric}}{'multiplication'}) {
                $perf{'nutanix'}{'disk'}{$disk->{disk_uuid}}{$tmpTime}{$dictionary{$diskMetrics->{metric}}{'label'}} = $value*$dictionary{$diskMetrics->{metric}}{'multiplication'};
              } else {
                $perf{'nutanix'}{'disk'}{$disk->{disk_uuid}}{$tmpTime}{$dictionary{$diskMetrics->{metric}}{'label'}} = $value;
              }
            } else {
              $perf{'nutanix'}{'disk'}{$disk->{disk_uuid}}{$tmpTime}{$diskMetrics->{metric}} = $value;
            }
            
            $tmpTime += $data->{interval};
          }
        }

        $metrics = "num_io,num_write_io,num_read_io";
        $diskPerf = $nutanix->getDiskPerf($disk->{disk_uuid}, $metrics, $timestamp_start_disk_container_pool, $timestamp_end, $data->{interval}, $api{$data->{type}}{'disk'});
        for (@{$diskPerf->{"stats_specific_responses"}}) {
          my $diskMetrics = $_;
          $tmpTime = ceil($diskMetrics->{start_time_in_usecs}/1000000);
          for (@{$diskMetrics->{values}}){
            my $value = $_;
            if ($value < 0) { $value = 0; }
            if (defined $dictionary{$diskMetrics->{metric}}) {
              if (defined $dictionary{$diskMetrics->{metric}}{'division'}) {
                $perf{'nutanix'}{'disk'}{$disk->{disk_uuid}}{$tmpTime}{$dictionary{$diskMetrics->{metric}}{'label'}} = ($value/$dictionary{$diskMetrics->{metric}}{'division'})/$data->{interval};
              } elsif (defined $dictionary{$diskMetrics->{metric}}{'multiplication'}) {
                $perf{'nutanix'}{'disk'}{$disk->{disk_uuid}}{$tmpTime}{$dictionary{$diskMetrics->{metric}}{'label'}} = ($value*$dictionary{$diskMetrics->{metric}}{'multiplication'})/$data->{interval};
              } else {
                $perf{'nutanix'}{'disk'}{$disk->{disk_uuid}}{$tmpTime}{$dictionary{$diskMetrics->{metric}}{'label'}} = $value/$data->{interval};
              }
            } else {
              $perf{'nutanix'}{'disk'}{$disk->{disk_uuid}}{$tmpTime}{$diskMetrics->{metric}} = $value/$data->{interval};
            }
            
            $tmpTime += $data->{interval};
          }
        }
      }
    } elsif ($api{$data->{type}}{'disk'} eq "v1") {
      if ($disks->{metadata}{totalEntities} < ($page-1)*$page_max || $disks->{metadata}{startIndex} eq "-1") {
        $loop = 0;
        last;
      }
      # API v1
      Nutanix::log("$data->{hostalias}: fetching ".(($disks->{metadata}{endIndex}+1)-$disks->{metadata}{startIndex})." disks [step ".$page." / ".ceil($disks->{metadata}{totalEntities}/$page_max)."]");
      for (@{$disks->{entities}}) {
        my $disk = $_;

        $conf{$disk->{diskUuid}}{label} = $disk->{nodeName}." - ".$disk->{location};
        $conf{$disk->{diskUuid}}{type} = $disk->{storageTierName};
        $conf{$disk->{diskUuid}}{node_uuid} = $disk->{nodeUuid};
        $conf{$disk->{diskUuid}}{physical_size} = ceil($disk->{usageStats}->{'storage.capacity_bytes'});

        my @parents;
        if (defined $savedIds{$disk->{nodeUuid}}) {
          push(@parents, $disk->{nodeUuid});
        }
        if (defined $savedIds{$disk->{clusterUuid}}) {
          push(@parents, $disk->{clusterUuid});
        }

        #architecture
        my %disk_arch = (
          "item_id" => $disk->{diskUuid},
          "label" => $disk->{nodeName}." - ".$disk->{location},
          "hw_type" => "nutanix",
          "subsystem" => "disk",
          "parents" => \@parents,
          "hostcfg_id" => $data->{hostcfg_id}
        );
        $savedIds{$disk->{diskUuid}} = 1;
        push(@architecture, \%disk_arch);


        my $tmpTime;
        my $metrics = "read_io_bandwidth_kBps,write_io_bandwidth_kBps,avg_io_latency_usecs";
        my $diskPerf = $nutanix->getDiskPerf($disk->{diskUuid}, $metrics, $timestamp_start_disk_container_pool, $timestamp_end, $data->{interval}, $api{$data->{type}}{'disk'});
        for (@{$diskPerf->{"statsSpecificResponses"}}) {
          my $diskMetrics = $_;
          $tmpTime = ceil($diskMetrics->{startTimeInUsecs}/1000000);
          for (@{$diskMetrics->{values}}){
            my $value = $_;
            if ($value < 0) { $value = 0; }
            if (defined $dictionary{$diskMetrics->{metric}}) {
              if (defined $dictionary{$diskMetrics->{metric}}{'division'}) {
                $perf{'nutanix'}{'disk'}{$disk->{diskUuid}}{$tmpTime}{$dictionary{$diskMetrics->{metric}}{'label'}} = $value/$dictionary{$diskMetrics->{metric}}{'division'};
              } elsif (defined $dictionary{$diskMetrics->{metric}}{'multiplication'}) {
                $perf{'nutanix'}{'disk'}{$disk->{diskUuid}}{$tmpTime}{$dictionary{$diskMetrics->{metric}}{'label'}} = $value*$dictionary{$diskMetrics->{metric}}{'multiplication'};
              } else {
                $perf{'nutanix'}{'disk'}{$disk->{diskUuid}}{$tmpTime}{$dictionary{$diskMetrics->{metric}}{'label'}} = $value;
              }
            } else {
              $perf{'nutanix'}{'disk'}{$disk->{diskUuid}}{$tmpTime}{$diskMetrics->{metric}} = $value;
            }
            
            $tmpTime += $data->{interval};
          }
        }

        $metrics = "num_io,num_write_io,num_read_io";
        $diskPerf = $nutanix->getDiskPerf($disk->{diskUuid}, $metrics, $timestamp_start_disk_container_pool, $timestamp_end, $data->{interval}, $api{$data->{type}}{'disk'});
        for (@{$diskPerf->{"statsSpecificResponses"}}) {
          my $diskMetrics = $_;
          $tmpTime = ceil($diskMetrics->{startTimeInUsecs}/1000000);
          for (@{$diskMetrics->{values}}){
            my $value = $_;
            if ($value < 0) { $value = 0; }
            if (defined $dictionary{$diskMetrics->{metric}}) {
              if (defined $dictionary{$diskMetrics->{metric}}{'division'}) {
                $perf{'nutanix'}{'disk'}{$disk->{diskUuid}}{$tmpTime}{$dictionary{$diskMetrics->{metric}}{'label'}} = ($value/$dictionary{$diskMetrics->{metric}}{'division'})/$data->{interval};
              } elsif (defined $dictionary{$diskMetrics->{metric}}{'multiplication'}) {
                $perf{'nutanix'}{'disk'}{$disk->{diskUuid}}{$tmpTime}{$dictionary{$diskMetrics->{metric}}{'label'}} = ($value*$dictionary{$diskMetrics->{metric}}{'multiplication'})/$data->{interval};
              } else {
                $perf{'nutanix'}{'disk'}{$disk->{diskUuid}}{$tmpTime}{$dictionary{$diskMetrics->{metric}}{'label'}} = $value/$data->{interval};
              }
            } else {
              $perf{'nutanix'}{'disk'}{$disk->{diskUuid}}{$tmpTime}{$diskMetrics->{metric}} = $value/$data->{interval};
            }
            $tmpTime += $data->{interval};
          }
        }

      }
    }
    $page++;
  }

  # storage containers
  my $containers = $nutanix->getStorageContainers($api{$data->{type}}{'container'});
  Xormon::keepLastResponse($containers, 'nutanix', $data->{hostcfg_id}, 'containers');
  if ($api{$data->{type}}{'container'} eq "v3" || $api{$data->{type}}{'container'} eq "v2") {
    # API v3 & v2
    Nutanix::log("$data->{hostalias}: fetching ".$containers->{metadata}{total_entities}." storage containers");
    for (@{$containers->{entities}}) {
      my $container = $_;

      $conf{$container->{storage_container_uuid}}{label} = $container->{name};
      $conf{$container->{storage_container_uuid}}{uuid} = $container->{storage_container_uuid};
      $conf{$container->{storage_container_uuid}}{compression_enabled} = $container->{compression_enabled};
      $conf{$container->{storage_container_uuid}}{is_nutanix_managed} = defined $container->{is_nutanix_managed} ? $container->{is_nutanix_managed} : 'false';
      $conf{$container->{storage_container_uuid}}{capacity_size} = ceil($container->{usage_stats}->{'storage.capacity_bytes'});
      $conf{$container->{storage_container_uuid}}{capacity_used} = ceil($container->{usage_stats}->{'storage.usage_bytes'});
      $conf{$container->{storage_container_uuid}}{capacity_free} = ceil(($container->{usage_stats}->{'storage.capacity_bytes'}-$container->{usage_stats}->{'storage.usage_bytes'}));

      my @parents;
      if (defined $savedIds{$container->{cluster_uuid}}) {
        push(@parents, $container->{cluster_uuid});
      }

      #architecture
      my %sc_arch = (
        "item_id" => $container->{storage_container_uuid},
        "label" => $container->{name},
        "hw_type" => "nutanix",
        "subsystem" => "storage_container",
        "parents" => \@parents,
        "hostcfg_id" => $data->{hostcfg_id}
      );
      $savedIds{$container->{storage_container_uuid}} = 1;
      push(@architecture, \%sc_arch);

      my $tmpTime;
      my $metrics = "controller_read_io_bandwidth_kBps,controller_write_io_bandwidth_kBps,controller_avg_io_latency_usecs";
      my $containerPerf = $nutanix->getStorageContainerPerf($container->{storage_container_uuid}, $metrics, $timestamp_start_disk_container_pool, $timestamp_end, $data->{interval}, $api{$data->{type}}{'container'});
      for (@{$containerPerf->{"stats_specific_responses"}}) {
        my $containerMetrics = $_;
        $tmpTime = ceil($containerMetrics->{start_time_in_usecs}/1000000);
        for (@{$containerMetrics->{values}}){
          my $value = $_;
          if ($value < 0) { $value = 0; }
          if (defined $dictionary{$containerMetrics->{metric}}) {
            if (defined $dictionary{$containerMetrics->{metric}}{'division'}) {
              $perf{'nutanix'}{'storage_container'}{$container->{storage_container_uuid}}{$tmpTime}{$dictionary{$containerMetrics->{metric}}{'label'}} = $value/$dictionary{$containerMetrics->{metric}}{'division'};
            } elsif (defined $dictionary{$containerMetrics->{metric}}{'multiplication'}) {
              $perf{'nutanix'}{'storage_container'}{$container->{storage_container_uuid}}{$tmpTime}{$dictionary{$containerMetrics->{metric}}{'label'}} = $value*$dictionary{$containerMetrics->{metric}}{'multiplication'};
            } else {
              $perf{'nutanix'}{'storage_container'}{$container->{storage_container_uuid}}{$tmpTime}{$dictionary{$containerMetrics->{metric}}{'label'}} = $value;
            }
          } else {
            $perf{'nutanix'}{'storage_container'}{$container->{storage_container_uuid}}{$tmpTime}{$containerMetrics->{metric}} = $value;
          }
          $tmpTime += $data->{interval};
        }
      }

      $metrics = "controller_num_io,controller_num_write_io,controller_num_read_io";
      $containerPerf = $nutanix->getStorageContainerPerf($container->{storage_container_uuid}, $metrics, $timestamp_start_disk_container_pool, $timestamp_end, $data->{interval}, $api{$data->{type}}{'container'});
      for (@{$containerPerf->{"stats_specific_responses"}}) {
        my $containerMetrics = $_;
        $tmpTime = ceil($containerMetrics->{start_time_in_usecs}/1000000);
        for (@{$containerMetrics->{values}}){
          my $value = $_;
          if ($value < 0) { $value = 0; }
          if (defined $dictionary{$containerMetrics->{metric}}) {
            if (defined $dictionary{$containerMetrics->{metric}}{'division'}) {
              $perf{'nutanix'}{'storage_container'}{$container->{storage_container_uuid}}{$tmpTime}{$dictionary{$containerMetrics->{metric}}{'label'}} = ($value/$dictionary{$containerMetrics->{metric}}{'division'})/$data->{interval};
            } elsif (defined $dictionary{$containerMetrics->{metric}}{'multiplication'}) {
              $perf{'nutanix'}{'storage_container'}{$container->{storage_container_uuid}}{$tmpTime}{$dictionary{$containerMetrics->{metric}}{'label'}} = ($value*$dictionary{$containerMetrics->{metric}}{'multiplication'})/$data->{interval};
            } else {
              $perf{'nutanix'}{'storage_container'}{$container->{storage_container_uuid}}{$tmpTime}{$dictionary{$containerMetrics->{metric}}{'label'}} = $value/$data->{interval};
            }
          } else {
            $perf{'nutanix'}{'storage_container'}{$container->{storage_container_uuid}}{$tmpTime}{$containerMetrics->{metric}} = $value;
          }
          $tmpTime += $data->{interval};
        }
      }

    }
  } elsif ($api{$data->{type}}{'container'} eq "v1") {
    # API v1
    Nutanix::log("$data->{hostalias}: fetching ".$containers->{metadata}{totalEntities}." storage containers");
    for (@{$containers->{entities}}) {
      my $container = $_;

      $conf{$container->{containerUuid}}{uuid} = $container->{containerUuid};
      $conf{$container->{containerUuid}}{compression_enabled} = $container->{compressionEnabled};
      $conf{$container->{containerUuid}}{is_nutanix_managed} = defined $container->{isNutanixManaged} ? $container->{isNutanixManaged} : 'false';
      $conf{$container->{containerUuid}}{capacity_size} = ceil($container->{usageStats}->{'storage.capacity_bytes'});
      $conf{$container->{containerUuid}}{capacity_used} = ceil($container->{usageStats}->{'storage.usage_bytes'});
      $conf{$container->{containerUuid}}{capacity_free} = ceil(($container->{usageStats}->{'storage.capacity_bytes'}-$container->{usageStats}->{'storage.usage_bytes'}));

      my @parents;
      if (defined $savedIds{$container->{clusterUuid}}) {
        push(@parents, $container->{clusterUuid});
      }

      #architecture
      my %sc_arch = (
        "item_id" => $container->{containerUuid},
        "label" => $container->{name},
        "hw_type" => "nutanix",
        "subsystem" => "storage_container",
        "parents" => \@parents,
        "hostcfg_id" => $data->{hostcfg_id}
      );
      $savedIds{$container->{containerUuid}} = 1;
      push(@architecture, \%sc_arch);

      my $tmpTime;
      my $metrics = "controller_read_io_bandwidth_kBps,controller_write_io_bandwidth_kBps,controller_avg_io_latency_usecs";
      my $containerPerf = $nutanix->getStorageContainerPerf($container->{containerUuid}, $metrics, $timestamp_start_disk_container_pool, $timestamp_end, $data->{interval}, $api{$data->{type}}{'container'});
      for (@{$containerPerf->{"statsSpecificResponses"}}) {
        my $containerMetrics = $_;
        $tmpTime = ceil($containerMetrics->{startTimeInUsecs}/1000000);
        for (@{$containerMetrics->{values}}){
          my $value = $_;
          if ($value < 0) { $value = 0; }
          if (defined $dictionary{$containerMetrics->{metric}}) {
            if (defined $dictionary{$containerMetrics->{metric}}{'division'}) {
              $perf{'nutanix'}{'storage_container'}{$container->{containerUuid}}{$tmpTime}{$dictionary{$containerMetrics->{metric}}{'label'}} = $value/$dictionary{$containerMetrics->{metric}}{'division'};
            } elsif (defined $dictionary{$containerMetrics->{metric}}{'multiplication'}) {
              $perf{'nutanix'}{'storage_container'}{$container->{containerUuid}}{$tmpTime}{$dictionary{$containerMetrics->{metric}}{'label'}} = $value*$dictionary{$containerMetrics->{metric}}{'multiplication'};
            } else {
              $perf{'nutanix'}{'storage_container'}{$container->{containerUuid}}{$tmpTime}{$dictionary{$containerMetrics->{metric}}{'label'}} = $value;
            }
          } else {
            $perf{'nutanix'}{'storage_container'}{$container->{containerUuid}}{$tmpTime}{$containerMetrics->{metric}} = $value;
          }
          $tmpTime += $data->{interval};
        }
      }

      $metrics = "controller_num_io,controller_num_write_io,controller_num_read_io";
      $containerPerf = $nutanix->getStorageContainerPerf($container->{containerUuid}, $metrics, $timestamp_start_disk_container_pool, $timestamp_end, $data->{interval}, $api{$data->{type}}{'container'});
      for (@{$containerPerf->{"statsSpecificResponses"}}) {
        my $containerMetrics = $_;
        $tmpTime = ceil($containerMetrics->{startTimeInUsecs}/1000000);
        for (@{$containerMetrics->{values}}){
          my $value = $_;
          if ($value < 0) { $value = 0; }
          if (defined $dictionary{$containerMetrics->{metric}}) {
            if (defined $dictionary{$containerMetrics->{metric}}{'division'}) {
              $perf{'nutanix'}{'storage_container'}{$container->{containerUuid}}{$tmpTime}{$dictionary{$containerMetrics->{metric}}{'label'}} = ($value/$dictionary{$containerMetrics->{metric}}{'division'})/$data->{interval};
            } elsif (defined $dictionary{$containerMetrics->{metric}}{'multiplication'}) {
              $perf{'nutanix'}{'storage_container'}{$container->{containerUuid}}{$tmpTime}{$dictionary{$containerMetrics->{metric}}{'label'}} = ($value*$dictionary{$containerMetrics->{metric}}{'multiplication'})/$data->{interval};
            } else {
              $perf{'nutanix'}{'storage_container'}{$container->{containerUuid}}{$tmpTime}{$dictionary{$containerMetrics->{metric}}{'label'}} = $value/$data->{interval};
            }
          } else {
            $perf{'nutanix'}{'storage_container'}{$container->{containerUuid}}{$tmpTime}{$containerMetrics->{metric}} = $value/$data->{interval};
          }
          $tmpTime += $data->{interval};
        }
      }
    }
  }
  $containers = ();

  # storage pools
  my $pools = $nutanix->getStoragePools($api{$data->{type}}{'pool'});
  Xormon::keepLastResponse($pools, 'nutanix', $data->{hostcfg_id}, 'pools');
  if ($api{$data->{type}}{'pool'} eq "v3" || $api{$data->{type}}{'pool'} eq "v2" || $api{$data->{type}}{'pool'} eq "v1") {
    # API v3 & v2 & v1
    Nutanix::log("$data->{hostalias}: fetching ".$pools->{metadata}{totalEntities}." storage pools");
    for (@{$pools->{entities}}) {
      my $pool = $_;

      $conf{$pool->{storagePoolUuid}}{label} = $pool->{name};
      $conf{$pool->{storagePoolUuid}}{uuid} = $pool->{storagePoolUuid};
      $conf{$pool->{storagePoolUuid}}{cluster} = $pool->{clusterUuid};
      $conf{$pool->{storagePoolUuid}}{capacity_size} = ceil($pool->{capacity});
      $conf{$pool->{storagePoolUuid}}{capacity_used} = ceil($pool->{usageStats}->{'storage.usage_bytes'});
      $conf{$pool->{storagePoolUuid}}{capacity_free} = ceil($pool->{usageStats}->{'storage.free_bytes'});

      my @parents;
      if (defined $savedIds{$pool->{clusterUuid}}) {
        push(@parents, $pool->{clusterUuid});
      }

      #architecture
      my %sp_arch = (
        "item_id" => $pool->{storagePoolUuid},
        "label" => $pool->{name},
        "hw_type" => "nutanix",
        "subsystem" => "storage_pool",
        "parents" => \@parents,
        "hostcfg_id" => $data->{hostcfg_id}
      );
      $savedIds{$pool->{storagePoolUuid}} = 1;
      push(@architecture, \%sp_arch);

      my $tmpTime;
      my $metrics = "read_io_bandwidth_kBps,write_io_bandwidth_kBps,avg_io_latency_usecs";
      my $poolPerf = $nutanix->getStoragePoolPerf($pool->{storagePoolUuid}, $metrics, $timestamp_start_disk_container_pool, $timestamp_end, $data->{interval}, $api{$data->{type}}{'pool'});
      for (@{$poolPerf->{"statsSpecificResponses"}}) {
        my $poolMetrics = $_;
        $tmpTime = ceil($poolMetrics->{startTimeInUsecs}/1000000);
        for (@{$poolMetrics->{values}}){
          my $value = $_;
          if ($value < 0) { $value = 0; }
          if (defined $dictionary{$poolMetrics->{metric}}) {
            if (defined $dictionary{$poolMetrics->{metric}}{'division'}) {
              $perf{'nutanix'}{'storage_pool'}{$pool->{storagePoolUuid}}{$tmpTime}{$dictionary{$poolMetrics->{metric}}{'label'}} = $value/$dictionary{$poolMetrics->{metric}}{'division'};
            } elsif (defined $dictionary{$poolMetrics->{metric}}{'multiplication'}) {
              $perf{'nutanix'}{'storage_pool'}{$pool->{storagePoolUuid}}{$tmpTime}{$dictionary{$poolMetrics->{metric}}{'label'}} = $value*$dictionary{$poolMetrics->{metric}}{'multiplication'};
            } else {
              $perf{'nutanix'}{'storage_pool'}{$pool->{storagePoolUuid}}{$tmpTime}{$dictionary{$poolMetrics->{metric}}{'label'}} = $value;
            }
          } else {
            $perf{'nutanix'}{'storage_pool'}{$pool->{storagePoolUuid}}{$tmpTime}{$poolMetrics->{metric}} = $value;
          }
          $tmpTime += $data->{interval};
        }
      }

      $metrics = "num_io,num_write_io,num_read_io";
      $poolPerf = $nutanix->getStoragePoolPerf($pool->{storagePoolUuid}, $metrics, $timestamp_start_disk_container_pool, $timestamp_end, $data->{interval}, $api{$data->{type}}{'pool'});
      for (@{$poolPerf->{"statsSpecificResponses"}}) {
        my $poolMetrics = $_;
        $tmpTime = ceil($poolMetrics->{startTimeInUsecs}/1000000);
        for (@{$poolMetrics->{values}}){
          my $value = $_;
          if ($value < 0) { $value = 0; }
          if (defined $dictionary{$poolMetrics->{metric}}) {
            if (defined $dictionary{$poolMetrics->{metric}}{'division'}) {
              $perf{'nutanix'}{'storage_pool'}{$pool->{storagePoolUuid}}{$tmpTime}{$dictionary{$poolMetrics->{metric}}{'label'}} = ($value/$dictionary{$poolMetrics->{metric}}{'division'})/$data->{interval};
            } elsif (defined $dictionary{$poolMetrics->{metric}}{'multiplication'}) {
              $perf{'nutanix'}{'storage_pool'}{$pool->{storagePoolUuid}}{$tmpTime}{$dictionary{$poolMetrics->{metric}}{'label'}} = ($value*$dictionary{$poolMetrics->{metric}}{'multiplication'})/$data->{interval};
            } else {
              $perf{'nutanix'}{'storage_pool'}{$pool->{storagePoolUuid}}{$tmpTime}{$dictionary{$poolMetrics->{metric}}{'label'}} = $value/$data->{interval};
            }
          } else {
            $perf{'nutanix'}{'storage_pool'}{$pool->{storagePoolUuid}}{$tmpTime}{$poolMetrics->{metric}} = $value/$data->{interval};
          }
          $tmpTime += $data->{interval};
        }
      }
    }
  }
  $pools = ();

  # store disks, storage containers & storage pools data
  $xormon->saveJSON2(\%perf);
  %perf = ();

  updateServiceData($xormon, 'timestamp_disk_container_pool', $timestamp_end-300);

  # virtual disks
  $page = 1;
  $loop = 1;
  while($loop eq "1") {
    if (!defined $api{$data->{type}}{'vdisk'}) { last; }
    #if (total_size(\%perf) >= $max_size) { savePerf(\%perf, $hostcfg_uuid, $timestamp_start, $timestamp_end, $index); $index++; %perf = (); }
    my $vdisks = $nutanix->getVdisks($api{$data->{type}}{'vdisk'}, $page_max, $page);
    Xormon::keepLastResponse($vdisks, 'nutanix', $data->{hostcfg_id}, 'vdisks_'.$page);
    if ($api{$data->{type}}{'vdisk'} eq "v3" || $api{$data->{type}}{'vdisk'} eq "v2") {
      # API v3 & v2
      if ($vdisks->{metadata}{total_entities} < ($page-1)*$page_max || $vdisks->{metadata}{start_index} eq "-1") {
        $loop = 0;
        last;
      }
      Nutanix::log("$data->{hostalias}: fetching ".(($vdisks->{metadata}{end_index}+1)-$vdisks->{metadata}{start_index})." virtual disks [step ".$page." / ".ceil($vdisks->{metadata}{total_entities}/$page_max)."]");
      for (@{$vdisks->{entities}}) {
        my $vdisk = $_;

        if (!defined $vdisk->{attached_vm_uuid}) {
          next;
        }
        
        $conf{$vdisk->{uuid}}{label} = $vdisk->{attached_vmname}." - ".$vdisk->{disk_address};
        $conf{$vdisk->{uuid}}{uuid} = $vdisk->{uuid};
        $conf{$vdisk->{uuid}}{attached_vm_uuid} = $vdisk->{attached_vm_uuid};
        $conf{$vdisk->{uuid}}{attached_vmname} = $vdisk->{attached_vmname};
        $conf{$vdisk->{uuid}}{attached_volume_group_id} = $vdisk->{attached_volume_group_id};
        $conf{$vdisk->{uuid}}{capacity} = ceil($vdisk->{disk_capacity_in_bytes});
        $conf{$vdisk->{uuid}}{cluster_uuid} = $vdisk->{cluster_uuid};
        $conf{$vdisk->{uuid}}{storage_container_uuid} = $vdisk->{storage_container_uuid};

        my @parents;
        if (defined $savedIds{$vdisk->{cluster_uuid}}) {
          push(@parents, $vdisk->{cluster_uuid});
        }
        if (defined $savedIds{$vdisk->{storage_container_uuid}}) {
          push(@parents, $vdisk->{storage_container_uuid});
        }
        if (defined $savedIds{$vdisk->{attached_vm_uuid}}) {
          push(@parents, $vdisk->{attached_vm_uuid});
        }

        #architecture
        my %vdisk_arch = (
          "item_id" => $vdisk->{uuid},
          "label" => $vdisk->{attached_vmname}." - ".$vdisk->{disk_address},
          "hw_type" => "nutanix",
          "subsystem" => "vdisk",
          "parents" => \@parents,
          "hostcfg_id" => $data->{hostcfg_id}
        );
        $savedIds{$vdisk->{uuid}} = 1;
        push(@architecture, \%vdisk_arch);

        # my $tmpTime;
        # my $metrics = "controller_read_io_bandwidth_kBps,controller_write_io_bandwidth_kBps,controller_avg_io_latency_usecs";
        # my $vdiskPerf = $nutanix->getVdiskPerf($vdisk->{uuid}, $metrics, $timestamp_start, $timestamp_end, $data->{interval}, $api{$data->{type}}{'vdisk'});
        # for (@{$vdiskPerf->{"stats_specific_responses"}}) {
        #   my $vdiskMetrics = $_;
        #   $tmpTime = ceil($vdiskMetrics->{start_time_in_usecs}/1000000);
        #   for (@{$vdiskMetrics->{values}}){
        #     my $value = $_;
        #     if ($value < 0) { $value = 0; }
        #       if (defined $dictionary{$vdiskMetrics->{metric}}) {
        #       if (defined $dictionary{$vdiskMetrics->{metric}}{'division'}) {
        #         $perf{'nutanix'}{'vdisk'}{$vdisk->{uuid}}{$tmpTime}{$dictionary{$vdiskMetrics->{metric}}{'label'}} = $value/$dictionary{$vdiskMetrics->{metric}}{'division'};
        #       } elsif (defined $dictionary{$vdiskMetrics->{metric}}{'multiplication'}) {
        #         $perf{'nutanix'}{'vdisk'}{$vdisk->{uuid}}{$tmpTime}{$dictionary{$vdiskMetrics->{metric}}{'label'}} = $value*$dictionary{$vdiskMetrics->{metric}}{'multiplication'};
        #       } else {
        #         $perf{'nutanix'}{'vdisk'}{$vdisk->{uuid}}{$tmpTime}{$dictionary{$vdiskMetrics->{metric}}{'label'}} = $value;
        #       }
        #     } else {
        #       $perf{'nutanix'}{'vdisk'}{$vdisk->{uuid}}{$tmpTime}{$vdiskMetrics->{metric}} = $value;
        #     }
        #     $tmpTime += $data->{interval};
        #   }
        # }

        # $metrics = "controller_num_io,controller_num_write_io,controller_num_read_io";
        # $vdiskPerf = $nutanix->getVdiskPerf($vdisk->{uuid}, $metrics, $timestamp_start, $timestamp_end, $data->{interval}, $api{$data->{type}}{'vdisk'});
        # for (@{$vdiskPerf->{"stats_specific_responses"}}) {
        #   my $vdiskMetrics = $_;
        #   $tmpTime = ceil($vdiskMetrics->{start_time_in_usecs}/1000000);
        #   for (@{$vdiskMetrics->{values}}){
        #     my $value = $_;
        #     if ($value < 0) { $value = 0; }
        #       if (defined $dictionary{$vdiskMetrics->{metric}}) {
        #       if (defined $dictionary{$vdiskMetrics->{metric}}{'division'}) {
        #         $perf{'nutanix'}{'vdisk'}{$vdisk->{uuid}}{$tmpTime}{$dictionary{$vdiskMetrics->{metric}}{'label'}} = ($value/$dictionary{$vdiskMetrics->{metric}}{'division'})/$data->{interval};
        #       } elsif (defined $dictionary{$vdiskMetrics->{metric}}{'multiplication'}) {
        #         $perf{'nutanix'}{'vdisk'}{$vdisk->{uuid}}{$tmpTime}{$dictionary{$vdiskMetrics->{metric}}{'label'}} = ($value*$dictionary{$vdiskMetrics->{metric}}{'multiplication'})/$data->{interval};
        #       } else {
        #         $perf{'nutanix'}{'vdisk'}{$vdisk->{uuid}}{$tmpTime}{$dictionary{$vdiskMetrics->{metric}}{'label'}} = $value/$data->{interval};
        #       }
        #     } else {
        #       $perf{'nutanix'}{'vdisk'}{$vdisk->{uuid}}{$tmpTime}{$vdiskMetrics->{metric}} = $value/$data->{interval};
        #     }
        #     $tmpTime += $data->{interval};
        #   }
        # }

        # ASYNC 
        my $metrics = "controller_read_io_bandwidth_kBps,controller_write_io_bandwidth_kBps,controller_avg_io_latency_usecs";
        $async->add(
          $nutanix->getVdiskPerfRequest($vdisk->{uuid}, $metrics, $timestamp_start_vdisk, $timestamp_end, $data->{interval}, $api{$data->{type}}{'vdisk'})
        );
        
        $metrics = "controller_num_io,controller_num_write_io,controller_num_read_io";
        $async->add(
          $nutanix->getVdiskPerfRequest($vdisk->{uuid}, $metrics, $timestamp_start_vdisk, $timestamp_end, $data->{interval}, $api{$data->{type}}{'vdisk'})
        );
        # END ASYNC

      }

    } elsif ($api{$data->{type}}{'vdisk'} eq "v1") {
      # API v1
      if ($vdisks->{metadata}{totalEntities} < ($page-1)*$page_max || $vdisks->{metadata}{startIndex} eq "-1") {
        $loop = 0;
        last;
      }
      Nutanix::log("$data->{hostalias}: fetching ".(($vdisks->{metadata}{endIndex}+1)-$vdisks->{metadata}{startIndex})." virtual disks [step ".$page." / ".ceil($vdisks->{metadata}{totalEntities}/$page_max)."]");
      for (@{$vdisks->{entities}}) {
        my $vdisk = $_;

        if (!defined $vdisk->{attachedVmUuid}) {
          next;
        }
        
        $conf{$vdisk->{uuid}}{label} = $vdisk->{attachedVMName}." - ".$vdisk->{diskAddress};
        $conf{$vdisk->{uuid}}{uuid} = $vdisk->{uuid};
        $conf{$vdisk->{uuid}}{attached_vm_uuid} = $vdisk->{attachedVmUuid};
        $conf{$vdisk->{uuid}}{attached_vmname} = $vdisk->{attachedVMName};
        $conf{$vdisk->{uuid}}{attached_volume_group_id} = $vdisk->{attachedVolumeGroupId};
        $conf{$vdisk->{uuid}}{capacity} = ceil($vdisk->{diskCapacityInBytes});
        $conf{$vdisk->{uuid}}{cluster_uuid} = $vdisk->{clusterUuid};
        $conf{$vdisk->{uuid}}{storage_container_uuid} = $vdisk->{containerUuid};

        my @parents;
        if (defined $savedIds{$vdisk->{clusterUuid}}) {
          push(@parents, $vdisk->{clusterUuid});
        }
        if (defined $savedIds{$vdisk->{containerUuid}}) {
          push(@parents, $vdisk->{containerUuid});
        }
        if (defined $savedIds{$vdisk->{attachedVmUuid}}) {
          push(@parents, $vdisk->{attachedVmUuid});
        }

        #architecture
        my %vdisk_arch = (
          "item_id" => $vdisk->{uuid},
          "label" => $vdisk->{attachedVMName}." - ".$vdisk->{diskAddress},
          "hw_type" => "nutanix",
          "subsystem" => "vdisk",
          "parents" => \@parents,
          "hostcfg_id" => $data->{hostcfg_id}
        );
        $savedIds{$vdisk->{uuid}} = 1;
        push(@architecture, \%vdisk_arch);

        # my $tmpTime;
        # my $metrics = "controller_read_io_bandwidth_kBps,controller_write_io_bandwidth_kBps,controller_avg_io_latency_usecs";
        # my $vdiskPerf = $nutanix->getVdiskPerf($vdisk->{uuid}, $metrics, $timestamp_start, $timestamp_end, $data->{interval}, $api{$data->{type}}{'vdisk'});
        # for (@{$vdiskPerf->{"statsSpecificResponses"}}) {
        #   my $vdiskMetrics = $_;
        #   $tmpTime = ceil($vdiskMetrics->{startTimeInUsecs}/1000000);
        #   for (@{$vdiskMetrics->{values}}){
        #     my $value = $_;
        #     if ($value < 0) { $value = 0; }
        #     if (defined $dictionary{$vdiskMetrics->{metric}}) {
        #       if (defined $dictionary{$vdiskMetrics->{metric}}{'division'}) {
        #         $perf{'nutanix'}{'vdisk'}{$vdisk->{uuid}}{$tmpTime}{$dictionary{$vdiskMetrics->{metric}}{'label'}} = $value/$dictionary{$vdiskMetrics->{metric}}{'division'};
        #       } elsif (defined $dictionary{$vdiskMetrics->{metric}}{'multiplication'}) {
        #         $perf{'nutanix'}{'vdisk'}{$vdisk->{uuid}}{$tmpTime}{$dictionary{$vdiskMetrics->{metric}}{'label'}} = $value*$dictionary{$vdiskMetrics->{metric}}{'multiplication'};
        #       } else {
        #         $perf{'nutanix'}{'vdisk'}{$vdisk->{uuid}}{$tmpTime}{$dictionary{$vdiskMetrics->{metric}}{'label'}} = $value;
        #       }
        #     } else {
        #       $perf{'nutanix'}{'vdisk'}{$vdisk->{uuid}}{$tmpTime}{$vdiskMetrics->{metric}} = $value;
        #     }
        #     $tmpTime += $data->{interval};
        #   }
        # }

        # $metrics = "controller_num_io,controller_num_write_io,controller_num_read_io";
        # $vdiskPerf = $nutanix->getVdiskPerf($vdisk->{uuid}, $metrics, $timestamp_start, $timestamp_end, $data->{interval}, $api{$data->{type}}{'vdisk'});
        # for (@{$vdiskPerf->{"statsSpecificResponses"}}) {
        #   my $vdiskMetrics = $_;
        #   $tmpTime = ceil($vdiskMetrics->{startTimeInUsecs}/1000000);
        #   for (@{$vdiskMetrics->{values}}){
        #     my $value = $_;
        #     if ($value < 0) { $value = 0; }
        #     if (defined $dictionary{$vdiskMetrics->{metric}}) {
        #       if (defined $dictionary{$vdiskMetrics->{metric}}{'division'}) {
        #         $perf{'nutanix'}{'vdisk'}{$vdisk->{uuid}}{$tmpTime}{$dictionary{$vdiskMetrics->{metric}}{'label'}} = ($value/$dictionary{$vdiskMetrics->{metric}}{'division'})/$data->{interval};
        #       } elsif (defined $dictionary{$vdiskMetrics->{metric}}{'multiplication'}) {
        #         $perf{'nutanix'}{'vdisk'}{$vdisk->{uuid}}{$tmpTime}{$dictionary{$vdiskMetrics->{metric}}{'label'}} = ($value*$dictionary{$vdiskMetrics->{metric}}{'multiplication'})/$data->{interval};
        #       } else {
        #         $perf{'nutanix'}{'vdisk'}{$vdisk->{uuid}}{$tmpTime}{$dictionary{$vdiskMetrics->{metric}}{'label'}} = $value/$data->{interval};
        #       }
        #     } else {
        #       $perf{'nutanix'}{'vdisk'}{$vdisk->{uuid}}{$tmpTime}{$vdiskMetrics->{metric}} = $value/$data->{interval};
        #     }
        #     $tmpTime += $data->{interval};
        #   }
        # }

        # ASYNC 
        my $metrics = "controller_read_io_bandwidth_kBps,controller_write_io_bandwidth_kBps,controller_avg_io_latency_usecs";
        $async->add(
          $nutanix->getVdiskPerfRequest($vdisk->{uuid}, $metrics, $timestamp_start_vdisk, $timestamp_end, $data->{interval}, $api{$data->{type}}{'vdisk'})
        );
        
        $metrics = "controller_num_io,controller_num_write_io,controller_num_read_io";
        $async->add(
          $nutanix->getVdiskPerfRequest($vdisk->{uuid}, $metrics, $timestamp_start_vdisk, $timestamp_end, $data->{interval}, $api{$data->{type}}{'vdisk'})
        );
        # END ASYNC

      }
    }
    $page++;
  }

  # VDISK ASYNC 
  $asyncTime = time();
  Nutanix::log("$data->{hostalias}: ".$async->info);
  while ( my $response = $async->wait_for_next_response ) {

    if ($response->{_rc} != 200) {
      Nutanix::error("$data->{hostalias}: [Error] [VDISK] Request: " . $response->{_request}{_uri} . " Message: " . $response->{_msg});
    }

    my $uuid = $response->{_request}{_headers}{uuid};
    my $vm = $conf{$uuid};

    my $vdiskPerf = Nutanix::decode_json_eval($response->content);
    my $tmpTime;

    my @divisionMetrics = ('controller_num_io','controller_num_write_io','controller_num_read_io');

    if ($api{$data->{type}}{'vdisk'} eq "v3" || $api{$data->{type}}{'vdisk'} eq "v2") {
      for (@{$vdiskPerf->{"stats_specific_responses"}}) {
        my $vdiskMetrics = $_;
        $tmpTime = ceil($vdiskMetrics->{start_time_in_usecs}/1000000);
        for (@{$vdiskMetrics->{values}}){
          my $value = $_;
          if ($value < 0) { $value = 0; }

          if (defined $dictionary{$vdiskMetrics->{metric}}) {
            if (defined $dictionary{$vdiskMetrics->{metric}}{'division'}) {
              $perf{'nutanix'}{'vdisk'}{$uuid}{$tmpTime}{$dictionary{$vdiskMetrics->{metric}}{'label'}} = $value/$dictionary{$vdiskMetrics->{metric}}{'division'};
            } elsif (defined $dictionary{$vdiskMetrics->{metric}}{'multiplication'}) {
              $perf{'nutanix'}{'vdisk'}{$uuid}{$tmpTime}{$dictionary{$vdiskMetrics->{metric}}{'label'}} = $value*$dictionary{$vdiskMetrics->{metric}}{'multiplication'};
            } else {
              $perf{'nutanix'}{'vdisk'}{$uuid}{$tmpTime}{$dictionary{$vdiskMetrics->{metric}}{'label'}} = $value;
            }
          } else {
            $perf{'nutanix'}{'vdisk'}{$uuid}{$tmpTime}{$vdiskMetrics->{metric}} = $value;
          }

          if ( grep( /^$vdiskMetrics->{metric}$/, @divisionMetrics ) ) { 
            $perf{'nutanix'}{'vdisk'}{$uuid}{$tmpTime}{$dictionary{$vdiskMetrics->{metric}}{'label'}} = $perf{'nutanix'}{'vdisk'}{$uuid}{$tmpTime}{$dictionary{$vdiskMetrics->{metric}}{'label'}} / $data->{interval};
          }

          $tmpTime += $data->{interval};
        }
      }
    } elsif ($api{$data->{type}}{'vdisk'} eq "v1") {
      for (@{$vdiskPerf->{"statsSpecificResponses"}}) {
        my $vdiskMetrics = $_;
        $tmpTime = ceil($vdiskMetrics->{startTimeInUsecs}/1000000);
        for (@{$vdiskMetrics->{values}}){
          my $value = $_;
          if ($value < 0) { $value = 0; }

          if (defined $dictionary{$vdiskMetrics->{metric}}) {
            if (defined $dictionary{$vdiskMetrics->{metric}}{'division'}) {
              $perf{'nutanix'}{'vdisk'}{$uuid}{$tmpTime}{$dictionary{$vdiskMetrics->{metric}}{'label'}} = $value/$dictionary{$vdiskMetrics->{metric}}{'division'};
            } elsif (defined $dictionary{$vdiskMetrics->{metric}}{'multiplication'}) {
              $perf{'nutanix'}{'vdisk'}{$uuid}{$tmpTime}{$dictionary{$vdiskMetrics->{metric}}{'label'}} = $value*$dictionary{$vdiskMetrics->{metric}}{'multiplication'};
            } else {
              $perf{'nutanix'}{'vdisk'}{$uuid}{$tmpTime}{$dictionary{$vdiskMetrics->{metric}}{'label'}} = $value;
            }
          } else {
            $perf{'nutanix'}{'vdisk'}{$uuid}{$tmpTime}{$vdiskMetrics->{metric}} = $value;
          }

          if ( grep( /^$vdiskMetrics->{metric}$/, @divisionMetrics ) ) { 
            $perf{'nutanix'}{'vdisk'}{$uuid}{$tmpTime}{$dictionary{$vdiskMetrics->{metric}}{'label'}} = $perf{'nutanix'}{'vdisk'}{$uuid}{$tmpTime}{$dictionary{$vdiskMetrics->{metric}}{'label'}} / $data->{interval};
          }

          $tmpTime += $data->{interval};
        }
      }
    }
  }
  $asyncDuration = time() - $asyncTime;
  Nutanix::log("$data->{hostalias}: [VDISK] Async duration: $asyncDuration secs");
  # END VDISK ASYNC 

  # store virtual disks data
  $xormon->saveJSON2(\%perf);
  %perf = ();

  updateServiceData($xormon, 'timestamp_vdisk', $timestamp_end-300);

  #health status
  Nutanix::log("$data->{hostalias}: fetching health status");
      
  my $gs;
  my @hs;
  my %hsMapping;
  my $healths = $nutanix->getHealthSummary();
  Xormon::keepLastResponse($healths, 'nutanix', $data->{hostcfg_id}, 'health');
  if ( ref($healths) eq 'ARRAY' ) {
    for ( @{$healths} ) {
      my $health = $_;

      if ( defined $health->{entityType} ) {
        if ($health->{entityType} eq "CLUSTER" || $health->{entityType} eq "HOST" || $health->{entityType} eq "VM") {
          my $status = 'ok';

          foreach my $id ( keys %{ $health->{detailedCheckSummary} } ) {
            my $test = $health->{detailedCheckSummary}->{$id};
            if ( ( $test->{Critical} eq "1" ) || ( $test->{Error} eq "1" ) || ( $test->{Warning} eq "1" )) {
              
              if (( $test->{Critical} eq "1" ) || ( $test->{Error} eq "1" )) {
                $status = 'error'; 
              } else {
                $status = 'warning';
              }

              if (!defined $gs) {
                $gs = $status;
              } elsif ($status = 'error') {
                $gs = $status;
              } elsif ($status = 'warning' && $gs ne 'error') {
                $gs = $status;
              }

              my $test_response = $nutanix->getHealthCheck($id);
              if ( exists $test_response->{name} ) {

                my @split = split("::", $test_response->{id});
                my $id = $split[0];

                if (!defined $hsMapping{$id}) {
                  $hsMapping{$id}{status} = $status;
                  $hsMapping{$id}{data} = $test_response->{description};
                } elsif ($hsMapping{$id}{status} ne 'error') {
                  $hsMapping{$id}{status} = $status;
                  $hsMapping{$id}{data} = $test_response->{description};
                }

              }
            }
          }

          if (scalar keys %{$health->{detailedCheckSummary}} == 0 && $health->{entityType} eq "CLUSTER" && defined $health->{healthSummary}) {
            if ($health->{healthSummary}{Critical} != 0 || $health->{healthSummary}{Error} != 0 ) {
              $status = 'error';
            } elsif ($health->{healthSummary}{Warning} != 0) {
              $status = 'warning';
            }

            $hsMapping{$firstClusterId}{status} = $status;
            $hsMapping{$firstClusterId}{data} = "API did not return status details";
          }

          if (!defined $hsMapping{$firstClusterId}) {
            $hsMapping{$firstClusterId}{status} = $status;
          }
        }
      }
    }
  } 
  foreach my $id (keys %hsMapping) {
    my %status = (
      "item_id" => $id,
      "status" => $hsMapping{$id}{status},
      "data" => $hsMapping{$id}{data}
    );
    push(@hs, \%status)
  }

  # hs details 
  my %hsDetail;
  my @health_set = ( { "item" => "vms", "uuid_text" => "uuid" }, { "item" => "hosts", "uuid_text" => "uuid" }, { "item" => "clusters", "uuid_text" => "uuid" } );
  for (@health_set) {
    my $health_item = $_;
    my $healths     = $nutanix->getItemHealth( $health_item->{item} );
    for ( @{ $healths->{entities} } ) {
      my $health = $_;
      $hsDetail{ $health_item->{item} }{ $health->{ $health_item->{uuid_text} } }{status} = $health->{healthSummary}->{healthStatus};
      if ( $health->{healthSummary}->{healthStatus} ne "Good" ) {
        foreach my $key ( keys %{ $health->{healthSummary}->{healthCheckSummaries} } ) {
          if ( $health->{healthSummary}->{healthCheckSummaries}->{$key}->{healthStatus} ne "Good" ) {
            $hsDetail{ $health_item->{item} }{ $health->{ $health_item->{uuid_text} } }{errors}{$key}{id}     = $key;
            $hsDetail{ $health_item->{item} }{ $health->{ $health_item->{uuid_text} } }{errors}{$key}{status} = $health->{healthSummary}->{healthCheckSummaries}->{$key}->{healthStatus};
            my $test_response = $nutanix->getHealthCheck($key);
            if ( exists $test_response->{name} ) {
              $hsDetail{ $health_item->{item} }{ $health->{ $health_item->{uuid_text} } }{errors}{$key}{name}        = $test_response->{name};
              $hsDetail{ $health_item->{item} }{ $health->{ $health_item->{uuid_text} } }{errors}{$key}{description} = $test_response->{description};
            }
          }
        }
      }
    }
  }
  $conf{$firstClusterId}{hs} = encode_json(\%hsDetail);

  # Alerts
  my $alerts_start = time() - 172800; # last 2 days
  my $alerts       = $nutanix->getAlerts($alerts_start);
  my %alert;
  Nutanix::log("$data->{hostalias}: fetching alerts");
  for ( @{ $alerts->{"entities"} } ) {
    my $al = $_;

    $alert{ $al->{id} }{created_time_stamp_in_usecs}         = $al->{created_time_stamp_in_usecs};
    $alert{ $al->{id} }{last_occurrence_time_stamp_in_usecs} = $al->{last_occurrence_time_stamp_in_usecs};
    $alert{ $al->{id} }{severity}                            = $al->{severity};

    my $cc = 0;
    my %dict;
    my $dict;
    for ( @{ $al->{context_types} } ) {
      my $context = $_;
      $dict->{$context} = $al->{context_values}->[$cc];
      $cc += 1;
    }

    $alert{ $al->{id} }{message} = $al->{message};
    for my $key ( keys %{$dict} ) {
      $alert{ $al->{id} }{message} =~ s/{$key}/$dict->{$key}/g;
    }

    $alert{ $al->{id} }{cluster}     = $al->{cluster_uuid};
    $alert{ $al->{id} }{alert_title} = $al->{alert_title};

    for my $key ( keys %{$dict} ) {
      $alert{ $al->{id} }{alert_title} =~ s/{$key}/$dict->{$key}/g;
    }
    $alert{ $al->{id} }{resolved_time_stamp_in_usecs} = $al->{resolved_time_stamp_in_usecs};
  }
  $conf{$firstClusterId}{alerts} = encode_json(\%alert);

  # Events
  my $events_start = time() - 172800;  # last 2 days
  my $events       = $nutanix->getEvents($events_start);
  my %event;
  Nutanix::log("$data->{hostalias}: fetching events");
  for ( @{ $events->{"entities"} } ) {
    my $event = $_;
    $event{ $event->{id} }{created_time_stamp_in_usecs}         = $event->{created_time_stamp_in_usecs};
    $event{ $event->{id} }{last_occurrence_time_stamp_in_usecs} = $event->{last_occurrence_time_stamp_in_usecs};
    $event{ $event->{id} }{severity}                            = $event->{severity};

    my $cc = 0;
    my %dict;
    my $dict;
    for ( @{ $event->{context_types} } ) {
      my $context = $_;
      $dict->{$context} = $event->{context_values}->[$cc];
      $cc += 1;
    }

    $event{ $event->{id} }{message} = $event->{message};
    for my $key ( keys %{$dict} ) {
      $event{ $event->{id} }{message} =~ s/{$key}/$dict->{$key}/g;
    }

    $event{ $event->{id} }{cluster}                      = $event->{cluster_uuid};
    $event{ $event->{id} }{alert_title}                  = $event->{alert_title};
    $event{ $event->{id} }{resolved_time_stamp_in_usecs} = $event->{resolved_time_stamp_in_usecs};
  }
  $conf{$firstClusterId}{events} = encode_json(\%event);

  #save last read
  updateServiceData($xormon, 'timestamp', $timestamp_end-300);

  #save architecture
  $xormon->saveArchitecture(\@architecture);

  #save configuration
  $xormon->saveConfJSON(\%conf);

  #save hs 
  $xormon->saveStatus(\@hs);  

  return 1;
}

sub updateServiceData {
  my $xormon = shift;
  my $key = shift;
  my $value = shift;

  my $service_data = $xormon->getData();
  if (!defined $service_data->{$key} || $service_data->{$key} < $value) {
    $service_data->{$key} = $value;
    $xormon->saveData($service_data);
  }
}

sub apitest {
  my $client = shift;
  my $data = shift;

  my $test = Test->new($client);

  my $connected = Test::connTest($data->{host}, $data->{port});
  if ($connected == 1) {
    $test->addSection("TCP connection test", 1);
  } else {
    $test->addSection("TCP connection test", 0, "TCP connection to " . $data->{host} . ":" . $data->{port} . " has failed! Open it on the firewall.");
    $test->stop();
    return;
  }


  my $nutanix = Nutanix->new($data->{protocol}, $data->{host}, $data->{port}, $data->{username}, $data->{password});

  my $authTest = $nutanix->authTest();
  if ($authTest == 1) {
    $test->addSection("Auth test", 1);
  } else {
    $test->addSection("Auth test", 0, "Wrong credentials");
    $test->stop();
    return;
  }


  my $clusters = $nutanix->getClusters('v1');

  if ($clusters->{metadata}{totalEntities} >= 1) {
    my $clusters = $clusters->{metadata}{totalEntities};
    $test->addSection("Data test", 1, "Number of clusters: $clusters");
  } else {
    $test->addSection("Data test", 0, "Can not list clusters");
  }

  $test->stop();
}


$server->close();