package OVirt;

use strict;
use warnings;

use JSON;
use Data::Dumper;
use Digest::MD5 qw(md5 md5_hex md5_base64);

################################################################################
# Conf queries
################################################################################

our $dc_current_stmt = <<END;
SELECT datacenter_configuration.datacenter_id as uuid,
  datacenter_configuration.datacenter_name
FROM datacenter_configuration
  WHERE ( datacenter_configuration.history_id
          IN ( SELECT max(a.history_id) AS max
               FROM datacenter_configuration a
               GROUP BY a.datacenter_id ))
  AND datacenter_configuration.delete_date IS NULL;
END

our $cl_current_stmt = <<END;
SELECT cluster_configuration.cluster_id as uuid,
  cluster_configuration.cluster_name,
  cluster_configuration.datacenter_id,
  cluster_configuration.compatibility_version
FROM cluster_configuration
  WHERE ( cluster_configuration.history_id
          IN ( SELECT max(a.history_id) AS max
               FROM cluster_configuration a
               GROUP BY a.cluster_id ))
  AND cluster_configuration.delete_date IS NULL;
END

our $host_current_stmt = <<END;
SELECT host_configuration.host_id as uuid,
  host_configuration.host_unique_id,
  host_configuration.host_name,
  host_configuration.cluster_id,
  host_configuration.fqdn_or_ip,
  host_configuration.memory_size_mb,
  host_configuration.swap_size_mb,
  host_configuration.cpu_model,
  host_configuration.cpu_speed_mh,
  host_configuration.number_of_cores,
  host_configuration.number_of_sockets,
  host_configuration.host_os,
  host_configuration.kernel_version,
  host_configuration.kvm_version,
  host_configuration.threads_per_core,
  host_configuration.hardware_product_name,
  host_configuration.hardware_serial_number
FROM host_configuration
  WHERE ( host_configuration.history_id
          IN ( SELECT max(a.history_id) AS max
               FROM host_configuration a
               GROUP BY a.host_id ))
  AND host_configuration.delete_date IS NULL;
END

our $host_nic_current_stmt = <<END;
SELECT host_interface_configuration.host_interface_id as uuid,
  host_interface_configuration.host_interface_name,
  host_interface_configuration.host_id
FROM host_interface_configuration
  WHERE ( host_interface_configuration.history_id
          IN ( SELECT max(a.history_id) AS max
               FROM host_interface_configuration a
               GROUP BY a.host_interface_id ))
  AND host_interface_configuration.delete_date IS NULL;
END

our $storage_domain_current_stmt = <<END;
SELECT sdc.storage_domain_id as uuid,
  sdc.storage_domain_name,
  sd_type.storage_domain_type,
  s_type.storage_type,
  sd_dc.datacenter_id
FROM ( SELECT storage_domain_id,
         datacenter_id
       FROM datacenter_storage_domain_map
       WHERE ( datacenter_storage_domain_map.history_id
               IN ( SELECT max(a.history_id) AS max
                    FROM datacenter_storage_domain_map a
                    GROUP BY a.storage_domain_id ))
         AND datacenter_storage_domain_map.detach_date IS NULL ) AS sd_dc
JOIN ( SELECT storage_domain_id,
         storage_domain_name,
         storage_domain_type,
         storage_type
       FROM storage_domain_configuration
       WHERE ( storage_domain_configuration.history_id
               IN ( SELECT max(a.history_id) AS max
                    FROM storage_domain_configuration a
                    GROUP BY a.storage_domain_id ))
         AND storage_domain_configuration.delete_date IS NULL ) as sdc
ON sdc.storage_domain_id = sd_dc.storage_domain_id
JOIN ( SELECT enum_key,
         value AS storage_domain_type
       FROM enum_translator
       WHERE enum_type = 'STORAGE_DOMAIN_TYPE'
         AND language_code = 'en_US' ) AS sd_type
ON sdc.storage_domain_type = sd_type.enum_key
JOIN ( SELECT enum_key, value AS storage_type
       FROM enum_translator
       WHERE enum_type = 'STORAGE_TYPE'
         AND language_code = 'en_US' ) AS s_type
ON sdc.storage_type = s_type.enum_key;
END

our $storage_domain_size_stmt = <<END;
SELECT storage_domain_id as uuid,
  available_disk_size_gb,
  used_disk_size_gb
FROM storage_domain_samples_history
  WHERE ( storage_domain_samples_history.history_id
          IN ( SELECT max(a.history_id) AS max
               FROM storage_domain_samples_history a
               GROUP BY a.storage_domain_id ));
END

# note: SELECT 2 AS enum_key, 'High Performance' AS vm_type is a workaround for missing entry in data warehouse (as of oVirt up to 4.5)
our $vm_current_stmt = <<END;
SELECT vm_conf.*,
   os_type.operating_system,
   vm_type.vm_type
FROM
  ( SELECT vm_id AS uuid,
      vm_name,
      cluster_id,
      cpu_per_socket,
      number_of_sockets,
      memory_size_mb,
      operating_system,
      vm_type
    FROM vm_configuration
    WHERE (vm_configuration.history_id
           IN ( SELECT max(a.history_id) AS max
                FROM vm_configuration a
                GROUP BY a.vm_id ))
    AND vm_configuration.delete_date IS NULL ) AS vm_conf
  JOIN
  ( SELECT enum_key, value AS operating_system FROM enum_translator
    WHERE enum_type = 'OS_TYPE'
      AND language_code = 'en_US' ) AS os_type
  ON vm_conf.operating_system = os_type.enum_key
  JOIN
  ( SELECT enum_key, value AS vm_type FROM enum_translator
    WHERE enum_type = 'VM_TYPE'
      AND language_code = 'en_US' 
    UNION ALL
    SELECT 2 AS enum_key, 'High Performance' AS vm_type ) AS vm_type
  ON vm_conf.vm_type = vm_type.enum_key;
END

our $vm_running_on_host_stmt = <<END;
SELECT vm_id AS uuid,
  currently_running_on_host,
  vm_status
FROM vm_samples_history
  WHERE ( vm_samples_history.history_id
          IN ( SELECT max(a.history_id) AS max
               FROM vm_samples_history a
               GROUP BY a.vm_id ));
END

our $vm_devices_current_stmt = <<END;
SELECT vm_device_history.vm_id,
  vm_device_history.device_id as uuid,
  vm_device_history.type
FROM vm_device_history
  WHERE ( vm_device_history.history_id
          IN ( SELECT max(a.history_id) AS max
               FROM vm_device_history a
               GROUP BY a.vm_id, a.device_id ))
  AND vm_device_history.delete_date IS NULL;
END

our $vm_disk_current_stmt = <<END;
SELECT vm_disk_configuration.vm_disk_id as uuid,
  CASE
    WHEN vm_disk_configuration.vm_disk_name IS NOT NULL
      THEN vm_disk_configuration.vm_disk_name::text
    ELSE 'disk '::text || vm_disk_configuration.vm_internal_drive_mapping::character varying::text
  END AS disk_name,
  vm_disk_configuration.storage_domain_id,
  vm_disk_size_mb
  FROM vm_disk_configuration
  WHERE ( vm_disk_configuration.history_id
          IN ( SELECT max(a.history_id) AS max
               FROM vm_disk_configuration a
               GROUP BY a.vm_disk_id ))
  AND vm_disk_configuration.delete_date IS NULL;
END

our $vm_nic_current_stmt = <<END;
SELECT vm_interface_configuration.vm_interface_id as uuid,
  vm_interface_configuration.vm_interface_name
FROM vm_interface_configuration
  WHERE ( vm_interface_configuration.history_id
          IN ( SELECT max(a.history_id) AS max
               FROM vm_interface_configuration a
               GROUP BY a.vm_interface_id ))
  AND vm_interface_configuration.delete_date IS NULL;
END

################################################################################
# Perf queries
################################################################################

sub get_host_perf_stmt {
  my $last_update = shift;
  my $act_db_time = shift;

  my $host_samples_stmt = <<END;
SELECT perf.host_id,
  CAST(EXTRACT(EPOCH FROM perf.history_datetime) AS BIGINT) AS timestamp,
  perf.cpu_usage_percent,
  perf.user_cpu_usage_percent,
  perf.system_cpu_usage_percent,
  perf.ksm_cpu_percent,
  perf.cpu_load,
  perf.total_vms_vcpus,
  perf.memory_usage_percent,
  perf.ksm_shared_memory_mb,
  perf.swap_used_mb,
  perf.host_status
FROM host_samples_history AS perf
WHERE history_datetime > to_timestamp($last_update)
  AND to_timestamp($act_db_time) > history_datetime
ORDER BY host_id, timestamp;
END

  return \$host_samples_stmt;
}

sub get_storage_domain_perf_stmt {
  my $last_update = shift;
  my $act_db_time = shift;

  my $storage_domain_samples_stmt = <<END;
SELECT perf.storage_domain_id,
  CAST(EXTRACT(EPOCH FROM perf.history_datetime) AS BIGINT) AS timestamp,
  perf.available_disk_size_gb,
  perf.used_disk_size_gb
FROM storage_domain_samples_history AS perf
WHERE history_datetime > to_timestamp($last_update)
  AND to_timestamp($act_db_time) > history_datetime
ORDER BY storage_domain_id, timestamp;
END

  return \$storage_domain_samples_stmt;
}

sub get_vm_perf_stmt {
  my $last_update = shift;
  my $act_db_time = shift;

  my $vm_samples_stmt = <<END;
SELECT perf.vm_id,
  CAST(EXTRACT(EPOCH FROM perf.history_datetime) AS BIGINT) AS timestamp,
  perf.cpu_usage_percent,
  perf.user_cpu_usage_percent,
  perf.system_cpu_usage_percent,
  perf.memory_usage_percent,
  perf.memory_buffered_kb,
  perf.memory_cached_kb,
  perf.vm_status
FROM vm_samples_history AS perf
WHERE perf.history_datetime > to_timestamp($last_update)
  AND to_timestamp($act_db_time) > perf.history_datetime
ORDER BY vm_id, timestamp;
END

  return \$vm_samples_stmt;
}

sub get_vm_disk_perf_stmt {
  my $last_update    = shift;
  my $act_db_time    = shift;
  my $schema_version = shift || 0;

  my $vm_disk_samples_stmt = <<END;
SELECT disk.vm_disk_id,
  CAST(EXTRACT(EPOCH FROM disk.history_datetime) AS BIGINT) AS timestamp,
  disk.write_rate_bytes_per_second,
  disk.read_rate_bytes_per_second,
  disk.write_latency_seconds,
  disk.read_latency_seconds,
  disk.vm_disk_actual_size_mb,
  disk.vm_disk_status
FROM vm_disk_samples_history AS disk
WHERE disk.history_datetime > to_timestamp($last_update)
  AND to_timestamp($act_db_time) > disk.history_datetime
ORDER BY disk.vm_disk_id, timestamp;
END

  if ( $schema_version eq '404' ) {
    $vm_disk_samples_stmt = <<END;
SELECT disk.vm_disk_id,
  CAST(EXTRACT(EPOCH FROM disk.history_datetime) AS BIGINT) AS timestamp,
  disk.write_rate_bytes_per_second,
  disk.read_rate_bytes_per_second,
  disk.write_latency_seconds,
  disk.read_latency_seconds,
  disk.vm_disk_actual_size_mb,
  disk.vm_disk_status,
  disk.write_ops_per_second,
  disk.read_ops_per_second
FROM vm_disk_samples_history AS disk
WHERE disk.history_datetime > to_timestamp($last_update)
  AND to_timestamp($act_db_time) > disk.history_datetime
ORDER BY disk.vm_disk_id, timestamp;
END

  }
  elsif ( $schema_version eq '405' ) {
    $vm_disk_samples_stmt = <<END;
SELECT disk.vm_disk_id,
  CAST(EXTRACT(EPOCH FROM disk.history_datetime) AS BIGINT) AS timestamp,
  disk.write_rate_bytes_per_second,
  disk.read_rate_bytes_per_second,
  disk.write_latency_seconds,
  disk.read_latency_seconds,
  disk.vm_disk_actual_size_mb,
  disk.vm_disk_status,
  disk.write_ops_total_count,
  disk.read_ops_total_count
FROM vm_disk_samples_history AS disk
WHERE disk.history_datetime > to_timestamp($last_update)
  AND to_timestamp($act_db_time) > disk.history_datetime
ORDER BY disk.vm_disk_id, timestamp;
END
  }

  return \$vm_disk_samples_stmt;
}

sub get_vm_nic_perf_stmt {
  my $last_update = shift;
  my $act_db_time = shift;

  my $vm_nic_samples_stmt = <<END;
SELECT nic.vm_interface_id,
  CAST(EXTRACT(EPOCH FROM nic.history_datetime) AS BIGINT) AS timestamp,
  nic.received_total_byte,
  nic.transmitted_total_byte
FROM vm_interface_samples_history AS nic
WHERE history_datetime > to_timestamp($last_update) - INTERVAL '1 MINUTES'
  AND to_timestamp($act_db_time) > history_datetime
ORDER BY vm_interface_id, timestamp;
END

  return \$vm_nic_samples_stmt;
}

sub get_host_nic_perf_stmt {
  my $last_update = shift;
  my $act_db_time = shift;

  my $host_nic_samples_stmt = <<END;
SELECT host_interface_id,
  CAST(EXTRACT(EPOCH FROM history_datetime) AS BIGINT) AS timestamp,
  received_total_byte,
  transmitted_total_byte
FROM host_interface_samples_history
WHERE history_datetime > to_timestamp($last_update) - INTERVAL '1 MINUTES'
  AND to_timestamp($act_db_time) > history_datetime
ORDER BY host_interface_id, timestamp;
END

  return \$host_nic_samples_stmt;
}

################################################################################


1;
