package SanCisco;

use strict;
use warnings;

use FindBin;
use lib "$FindBin::Bin/"; # use lib in this folder
use lib "../../../perl-modules";
use Data::Dumper;
use Xormon;
use JSON qw(encode_json decode_json);
use SNMP_lib;
use Math::BigInt;
use XormonLogger;

my $xormon;
my $cfg;
my $wrkdir;
my $logger;

my %data;
my $sess;

my $cfg_time           = 1;
my $act_time           = "";
my $sample_rate        = 300;
my $snmp_request_sleep = 1; # use one second as a default sleep between each snmp request

my $switch_name;
my $switch_uuid;

sub collect {
  $xormon   = shift;
  $cfg      = shift;
  $wrkdir   = shift;
  $act_time = time();

  #
  # test main variables
  #
  $switch_uuid = ( exists $cfg->{'hostcfg_id'} )  ? $cfg->{'hostcfg_id'}  : Xormon::error("Cannot continue, 'hostcfg_id' is not configured!")  && return "error";
  $switch_name = ( exists $cfg->{'hostalias'} )   ? $cfg->{'hostalias'}   : Xormon::error("Cannot continue, 'hostalias' is not configured!")   && return "error";
  $logger = XormonLogger->new( 'sancisco', $cfg->{hostcfg_id}, $cfg->{hostalias} );

  #
  # additional variables
  #
  if ( exists $cfg->{'snmp_request_sleep'} && Xormon::isdigit($cfg->{'snmp_request_sleep'}) ) {
    $snmp_request_sleep = $cfg->{'snmp_request_sleep'};
    $logger->info("snmp_request_sleep=$snmp_request_sleep");
  }

  #
  # get microservice data for this device
  #
  my $device_data_in  = $xormon->getData();
  my $device_data_out = {};

  #
  # check if it is time to collect configuration (only once per hour)
  #
  if ( exists $device_data_in->{last_conf_timestamp} && time() - $device_data_in->{last_conf_timestamp} < 3600 ) {
    $cfg_time = 0;
    $device_data_out->{last_conf_timestamp} = $device_data_in->{last_conf_timestamp};
  }

  #
  # initialize SNMP
  #
  #print Dumper $cfg;
  SNMP_lib::initialize_snmp($cfg);
  SNMP_lib::initialize_mibs('san');
  $sess = SNMP_lib::new_session();

  #
  # connection status
  #
  my ( $conn_status, $reason ) = SNMP_lib::collect_san( \%data, $sess, 'configuration', 'sysName', '', 1 ); # 1 -> conn test
  if ( $conn_status ) {
    $xormon->isConnected(1);
  }
  else {
    $xormon->isConnected(0, $reason);
  }

  #
  # perf data
  #
  collect_perf();
  my $perf_out = prepare_perf();
  my $rc       = $xormon->saveJSON2($perf_out);
  my $status   = "error";
  if ( defined $rc ) {
    $logger->info("save perf: $rc");

    if ( $rc eq "201" ) {
      $status = "successful";
    }
    elsif ( $rc eq "401" ) {
      $status = "unauthorized";
      last;
    }
    else {
      $status = "error";
    }
  }
  else {
    $logger->info("save perf: UN");
  }

  #
  # perf data status
  #
  if ( $status eq "successful" ) {
    $xormon->hasData(1);
  }
  else {
    $xormon->hasData(0, "no available stats: $status");
  }

  #
  # health status
  #
  health_status();

  #
  # conf data
  #
  if ( $cfg_time ) { # only once per hour
    collect_conf();
    send_conf();

    $device_data_out->{last_conf_timestamp} = time();
  }

  #
  # save microservice data for this device
  #
  $xormon->saveData(\%{ $device_data_out });


  return $status;
}

sub health_status {
  my @hs             = ();
  my @status_reason  = ();
  my $status_summary = "unknown";

  #
  # switch status
  #
  if ( exists $data{'configuration'}{'connUnitStatus'} && $data{'configuration'}{'connUnitStatus'} ne '' ) {
    my $switch_status = $data{'configuration'}{'connUnitStatus'};

    if ( $switch_status eq "ok" ) {
      #push(@hs, { "item_id" => $switch_uuid, "status" => "ok", "data" => "The operational status of the switch $switch_name is Healthy/ok." });
      $status_summary = "ok";
      push(@status_reason, "The operational status of the switch $switch_name is Healthy/ok.")
    }
    elsif ( $switch_status eq "unused" ) {
      #push(@hs, { "item_id" => $switch_uuid, "status" => "ok", "data" => "The operational status of the switch $switch_name is Unmonitored." });
      $status_summary = "ok";
      push(@status_reason, "The operational status of the switch $switch_name is Unmonitored.")
    }
    elsif ( $switch_status eq "unknown" ) {
      #push(@hs, { "item_id" => $switch_uuid, "status" => "unknown", "data" => "The operational status of the switch $switch_name is Unknown." });
      $status_summary = "unknown";
      push(@status_reason, "The operational status of the switch $switch_name is Unknown.")
    }
    elsif ( $switch_status eq "warning" ) {
      #push(@hs, { "item_id" => $switch_uuid, "status" => "warning", "data" => "The operational status of the switch $switch_name is Marginal/Warning." });
      $status_summary = "warning";
      push(@status_reason, "The operational status of the switch $switch_name is Marginal/Warning.")
    }
    elsif ( $switch_status eq "failed" ) {
      #push(@hs, { "item_id" => $switch_uuid, "status" => "error", "data" => "The operational status of the switch $switch_name is Down/Failed." });
      $status_summary = "error";
      push(@status_reason, "The operational status of the switch $switch_name is Down/Failed.")
    }
  }
  #else {
  #  push(@hs, { "item_id" => $switch_uuid, "status" => "unknown" });
  #}

  # load operational status reason file
  my $reason_list;
  my $reason_file = "$wrkdir/../../operational_status_reason.json";
  #if ( -f "$wrkdir/../../custom-operational_status_reason.json" ) {
  #  # possibility how to use custom statuses
  #  $reason_file = "$wrkdir/../../custom-operational_status_reason.json";
  #}
  if ( -f $reason_file ) {
    $reason_list = Xormon::read_json($reason_file);
  }
  else {
    $logger->error("$reason_file doesn't exist!" . __FILE__ . ":" . __LINE__ ) && return 0;
  }

  #
  # sanport statuses
  #
  if ( exists $data{ports} ) {
    foreach ( sort { $a <=> $b } keys( %{ $data{ports} } ) ) {
      my $var = $data{ports}{$_};

      if ( exists $var->{ifName} && $var->{ifName} ne '' ) {
        unless ( supported_port($var->{ifName}) ) { next; } # skip unsupported ports
        my $label = set_port_name($var->{ifName});

        if ( exists $var->{'ifAlias'} )  {
          my $alias = $var->{'ifAlias'};
          $alias    =~ s/\\//g;
          $alias    =~ s/\r//g;
          $alias    =~ s/'//g;

          # add alias to the label
          $label .= " [$alias]";
        }

        my $id = lc( Xormon::create_hash_key($switch_uuid."sanport".$var->{ifName}) );

        my $phys_status        = ( exists $var->{'ifAdminStatus'} )       ? $var->{'ifAdminStatus'}       : "";
        my $oper_status        = ( exists $var->{'ifOperStatus'} )        ? $var->{'ifOperStatus'}        : "";
        my $oper_status_reason = ( exists $var->{'fcIfOperStatusCause'} ) ? $var->{'fcIfOperStatusCause'} : "";

        # only ok or error
        if ( $oper_status eq "dormant" || $oper_status eq "lowerLayerDown" ) {
          push(@hs, { "item_id" => $id, "status" => "error", "data" => "Port $label: The operational status is $oper_status, the physical status is $phys_status and operational status reason is $oper_status_reason." });
          $status_summary = "error";
          push(@status_reason, "Port $label: The operational status is $oper_status, the physical status is $phys_status and operational status reason is $oper_status_reason.")
        }
        elsif ( $oper_status_reason ne '' && exists $cfg->{"OSR_$oper_status_reason"} ) {
          if ( $cfg->{"OSR_$oper_status_reason"} eq "red" ) {
            # at first check if there is a red color for this status in the ms advance cfg
            push(@hs, { "item_id" => $id, "status" => "error", "data" => "Port $label: The operational status is $oper_status, the physical status is $phys_status and operational status reason is $oper_status_reason." });
            $status_summary = "error";
            push(@status_reason, "Port $label: The operational status is $oper_status, the physical status is $phys_status and operational status reason is $oper_status_reason.")
          }
          else {
            push(@hs, { "item_id" => $id, "status" => "ok", "data" => "Port $label: The operational status is $oper_status, the physical status is $phys_status and operational status reason is $oper_status_reason." });
          }
        }
        elsif ( $oper_status_reason ne '' && exists $reason_list->{$oper_status_reason} && $reason_list->{$oper_status_reason} eq "red" ) {
          push(@hs, { "item_id" => $id, "status" => "error", "data" => "Port $label: The operational status is $oper_status, the physical status is $phys_status and operational status reason is $oper_status_reason." });
          $status_summary = "error";
          push(@status_reason, "Port $label: The operational status is $oper_status, the physical status is $phys_status and operational status reason is $oper_status_reason.")
        }
        else {
          push(@hs, { "item_id" => $id, "status" => "ok", "data" => "Port $label: The operational status is $oper_status, the physical status is $phys_status and operational status reason is $oper_status_reason." });
        }
      }
    }
  }

  #
  # add switch status summary (including red states of the ports)
  #
  push(@hs, { "item_id" => $switch_uuid, "status" => $status_summary, "data" => join("\n", @status_reason) });

  # send statuses
  my $rc = $xormon->saveStatus(\@hs);
  $logger->info("save health status: $rc");
  #print Dumper @hs;
  #print "HS : send statuses : $rc\n";

  return 1;
}

sub prepare_perf {
  my %stats_last;
  my %stats_now;
  my %stats_out;

  #
  # switch stats
  #
  if ( exists $data{'configuration'}{'cseSysCPUUtilization'} ) {
    $data{'configuration'}{'cseSysCPUUtilization'} =~ s/\s+%//;
    $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{cpu_system} = Xormon::formatNumberByType("float4",$data{'configuration'}{'cseSysCPUUtilization'});
  }
  if ( exists $data{'configuration'}{'cseSysMemoryUtilization'} ) {
    $data{'configuration'}{'cseSysMemoryUtilization'} =~ s/\s+%//;
    $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{memory_used_percent} = Xormon::formatNumberByType("float4",$data{'configuration'}{'cseSysMemoryUtilization'});
  }

  #
  # load last stats (it's necessary for converting counters to gauge)
  #
  my $last_stats_file  = "$wrkdir/last_stats.json";
  my $convert_counters = 0;
  my $timediff         = "";
  my @add_timestamps   = ();
  if ( -f $last_stats_file ) {
    eval { %stats_last = %{ Xormon::read_json($last_stats_file) }; };
    if ( exists $stats_last{timestamp} ) {
      $timediff = $act_time - $stats_last{timestamp};
      #if ( $timediff < ( 2 * $sample_rate ) && $timediff >= ( $sample_rate - 60 ) ) { # convert counter only when timediff is between 4 and 10 minutes
      if ( $timediff < ( 12 * $sample_rate ) && $timediff >= ( $sample_rate - 60 ) ) { # convert counter only when timediff is between 4 and 10 minutes
        $convert_counters = 1;

        # if fetch task is longer than 2x300, then is necessary to save stats also for the previous 5 minutes samples
        if ( $timediff > ( 2 * $sample_rate ) ) {
          my $max = int( $timediff / $sample_rate );
          for ( my $i = 1; $i < $max; $i++ ) {
            my $prev_timestamp = $act_time - ( $i * $sample_rate);

            if ( $prev_timestamp > $stats_last{timestamp} && ( $prev_timestamp - $stats_last{timestamp}) > ( $sample_rate - 60 ) ) {
              push(@add_timestamps, $prev_timestamp);
            }
          }
          @add_timestamps = sort @add_timestamps;
        }
      }
      else {
        $logger->error( "Cannot convert counters to gauge. \$act_time=$act_time (".localtime($act_time)."), \$stats_last{timestamp}=$stats_last{timestamp} (".localtime($stats_last{timestamp})."), \$timediff=$timediff " . __FILE__ . ":" . __LINE__ );
      }
    }
    else {
      $logger->error( "Something wrong \$stats_last{timestamp} doesn't exist!" . __FILE__ . ":" . __LINE__ );
    }
  }

  #
  # ports stats
  #
  my $ports_count = 0;
  if ( exists $data{ports} ) {
    foreach ( sort { $a <=> $b } keys( %{ $data{ports} } ) ) {
      my $var = $data{ports}{$_};

      if ( exists $var->{ifName} && $var->{ifName} ne '' ) {
        unless ( supported_port($var->{ifName}) ) { next; } # skip unsupported ports
        $var->{ifName} = set_port_name($var->{ifName});

        my $id = lc( Xormon::create_hash_key($switch_uuid."sanport".$var->{ifName}) );

        #
        # counters
        #
        # set current counters
        $stats_now{timestamp} = $act_time;

        if ( exists $var->{ifHCOutOctets}    && Xormon::isdigit($var->{ifHCOutOctets}) )    { $stats_now{$id}{data_out}   = $var->{ifHCOutOctets}; }
        if ( exists $var->{ifHCInOctets}     && Xormon::isdigit($var->{ifHCInOctets}) )     { $stats_now{$id}{data_in}    = $var->{ifHCInOctets}; }
        if ( exists $var->{fcIfC3OutFrames}  && Xormon::isdigit($var->{fcIfC3OutFrames}) )  { $stats_now{$id}{frames_out} = $var->{fcIfC3OutFrames}; }
        if ( exists $var->{fcIfC3InFrames}   && Xormon::isdigit($var->{fcIfC3InFrames}) )   { $stats_now{$id}{frames_in}  = $var->{fcIfC3InFrames}; }
        if ( exists $var->{fcIfInvalidCrcs}  && Xormon::isdigit($var->{fcIfInvalidCrcs}) )  { $stats_now{$id}{crc_errors} = $var->{fcIfInvalidCrcs}; }
        if ( exists $var->{fcIfLinkFailures} && Xormon::isdigit($var->{fcIfLinkFailures}) ) { $stats_now{$id}{link_fail}  = $var->{fcIfLinkFailures}; }
        if ( exists $var->{fcIfSyncLosses}   && Xormon::isdigit($var->{fcIfSyncLosses}) )   { $stats_now{$id}{loss_sync}  = $var->{fcIfSyncLosses}; }
        if ( $var->{ifName} =~ m/^fc/ || $var->{ifName} =~ m/^portChannel/ ) { # only for these ports
          if ( exists $var->{fcIfBBCreditTransistionFromZero} && Xormon::isdigit($var->{fcIfBBCreditTransistionFromZero}) ) { $stats_now{$id}{bb_credits} = $var->{fcIfBBCreditTransistionFromZero}; }
        }

        # prepare stats to DB (convert counters to gauge)
        if ( $convert_counters && exists $stats_last{$id} ) {
          if ( exists $stats_now{$id}{data_out}   && exists $stats_last{$id}{data_out} )   { $stats_out{sancisco}{sanport}{$id}{$act_time}{data_out}   = make_gauge($timediff,$stats_now{$id}{data_out},$stats_last{$id}{data_out}); }
          if ( exists $stats_now{$id}{data_in}    && exists $stats_last{$id}{data_in} )    { $stats_out{sancisco}{sanport}{$id}{$act_time}{data_in}    = make_gauge($timediff,$stats_now{$id}{data_in},$stats_last{$id}{data_in}); }
          if ( exists $stats_now{$id}{frames_out} && exists $stats_last{$id}{frames_out} ) { $stats_out{sancisco}{sanport}{$id}{$act_time}{frames_out} = make_gauge_int($timediff,$stats_now{$id}{frames_out},$stats_last{$id}{frames_out}); }
          if ( exists $stats_now{$id}{frames_in}  && exists $stats_last{$id}{frames_in} )  { $stats_out{sancisco}{sanport}{$id}{$act_time}{frames_in}  = make_gauge_int($timediff,$stats_now{$id}{frames_in},$stats_last{$id}{frames_in}); }
          if ( exists $stats_now{$id}{bb_credits} && exists $stats_last{$id}{bb_credits} ) { $stats_out{sancisco}{sanport}{$id}{$act_time}{bb_credits} = make_gauge($timediff,$stats_now{$id}{bb_credits},$stats_last{$id}{bb_credits}); }
          if ( exists $stats_now{$id}{crc_errors} && exists $stats_last{$id}{crc_errors} ) { $stats_out{sancisco}{sanport}{$id}{$act_time}{crc_errors} = make_gauge($timediff,$stats_now{$id}{crc_errors},$stats_last{$id}{crc_errors}); }
          if ( exists $stats_now{$id}{link_fail}  && exists $stats_last{$id}{link_fail} )  { $stats_out{sancisco}{sanport}{$id}{$act_time}{link_fail}  = make_gauge($timediff,$stats_now{$id}{link_fail},$stats_last{$id}{link_fail}); }
          if ( exists $stats_now{$id}{loss_sync}  && exists $stats_last{$id}{loss_sync} )  { $stats_out{sancisco}{sanport}{$id}{$act_time}{loss_sync}  = make_gauge($timediff,$stats_now{$id}{loss_sync},$stats_last{$id}{loss_sync}); }

          # port speed + utilization + check if max speed was ixceeded
          if ( exists $var->{ifHighSpeed} && Xormon::isdigit($var->{ifHighSpeed}) ) { # IF-MIB::ifHighSpeed.16777216 = Gauge32: 4000 (Mb / 1000 = 4 Gbps)
            #my $port_speed = Xormon::returnBytes($var->{ifHighSpeed}, "KB"); # 1024
            my $port_speed = ( $var->{ifHighSpeed} * 1000 * 1000 ) / 8; # from Mb/s to B/s
                                                                        # must be (*1000*1000)/8, ( 4000 Mb / 1000 ) = 4 Gbps )

            if ( exists $stats_out{sancisco}{sanport}{$id}{$act_time}{data_in} && $stats_out{sancisco}{sanport}{$id}{$act_time}{data_in} ne '' ) {

              if ( $stats_out{sancisco}{sanport}{$id}{$act_time}{data_in} > $port_speed ) {
                $logger->error("The port $var->{ifName}:$id max speed has been exceeded! data_in max=$port_speed, current=$stats_out{sancisco}{sanport}{$id}{$act_time}{data_in}");
                delete($stats_out{sancisco}{sanport}{$id}{$act_time}{data_in});
              }
              elsif ( $port_speed > 0 ) {
                $stats_out{sancisco}{sanport}{$id}{$act_time}{utilization_in} = Xormon::formatNumberByType("float4", ( $stats_out{sancisco}{sanport}{$id}{$act_time}{data_in} / $port_speed ) * 100 );
              }
            }
            if ( exists $stats_out{sancisco}{sanport}{$id}{$act_time}{data_out} && $stats_out{sancisco}{sanport}{$id}{$act_time}{data_out} ne '' ) {

              if ( $stats_out{sancisco}{sanport}{$id}{$act_time}{data_out} > $port_speed ) {
                $logger->error("The port $var->{ifName}:$id max speed has been exceeded! data_out max=$port_speed, current=$stats_out{sancisco}{sanport}{$id}{$act_time}{data_out}");
                delete($stats_out{sancisco}{sanport}{$id}{$act_time}{data_out});
              }
              elsif ( $port_speed > 0 ) {
                $stats_out{sancisco}{sanport}{$id}{$act_time}{utilization_out} = Xormon::formatNumberByType("float4", ( $stats_out{sancisco}{sanport}{$id}{$act_time}{data_out} / $port_speed ) * 100 );
              }
            }
          }

          # missing credits in % of time
          if ( exists $stats_out{sancisco}{sanport}{$id}{$act_time}{bb_credits} && Xormon::isdigit($stats_out{sancisco}{sanport}{$id}{$act_time}{bb_credits}) ) {
            #  $val = 0.0000025;
            #  $cmd .= " DEF:value${i}=\\\"$rrd\\\":$item_rrd:AVERAGE";
            #  $cmd .= " CDEF:valmb1_${i}=value${i},$val,*";
            #  $cmd .= " CDEF:valmb2_${i}=valmb1_${i},60,/";
            #  $cmd .= " CDEF:valmb3_${i}=valmb2_${i},100,*";
            #  $cmd .= " CDEF:valmb${i}=valmb3_${i},1000,GT,UNKN,valmb3_${i},IF";
            #  $cmd .= " $gtype:valmb${i}$color[$col_indx]:\\\"$port_space\\\"";
            #my $value = ( $stats_out{sancisco}{sanport}{$id}{$act_time}{bb_credits} * 0.00025 ) / 300; # replace 60 by 300, bcs step is 5 minutes now instead of 1 minute in stor
            my $value = ( $stats_out{sancisco}{sanport}{$id}{$act_time}{bb_credits} * 0.00025 ) / $timediff;
            if ( $value > 1000 ) {
              $stats_out{sancisco}{sanport}{$id}{$act_time}{bb_credits} = "";
            }
            else {
              $stats_out{sancisco}{sanport}{$id}{$act_time}{bb_credits} = sprintf( '%.2f', $value );
            }
          }

          # aggregate stats
          if ( exists $stats_out{sancisco}{sanport}{$id}{$act_time}{data_out} && Xormon::isdigit($stats_out{sancisco}{sanport}{$id}{$act_time}{data_out}) ) {
            if ( exists $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{data_out} ) {
              $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{data_out} += $stats_out{sancisco}{sanport}{$id}{$act_time}{data_out};
            }
            else {
              $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{data_out} = $stats_out{sancisco}{sanport}{$id}{$act_time}{data_out};
            }
          }
          if ( exists $stats_out{sancisco}{sanport}{$id}{$act_time}{data_in} && Xormon::isdigit($stats_out{sancisco}{sanport}{$id}{$act_time}{data_in}) ) {
            if ( exists $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{data_in} ) {
              $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{data_in} += $stats_out{sancisco}{sanport}{$id}{$act_time}{data_in};
            }
            else {
              $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{data_in} = $stats_out{sancisco}{sanport}{$id}{$act_time}{data_in};
            }
          }
          if ( exists $stats_out{sancisco}{sanport}{$id}{$act_time}{frames_out} && Xormon::isdigit($stats_out{sancisco}{sanport}{$id}{$act_time}{frames_out}) ) {
            if ( exists $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{frames_out} ) {
              $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{frames_out} += $stats_out{sancisco}{sanport}{$id}{$act_time}{frames_out};
            }
            else {
              $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{frames_out} = $stats_out{sancisco}{sanport}{$id}{$act_time}{frames_out};
            }
          }
          if ( exists $stats_out{sancisco}{sanport}{$id}{$act_time}{frames_in} && Xormon::isdigit($stats_out{sancisco}{sanport}{$id}{$act_time}{frames_in}) ) {
            if ( exists $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{frames_in} ) {
              $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{frames_in} += $stats_out{sancisco}{sanport}{$id}{$act_time}{frames_in};
            }
            else {
              $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{frames_in} = $stats_out{sancisco}{sanport}{$id}{$act_time}{frames_in};
            }
          }
          if ( exists $stats_out{sancisco}{sanport}{$id}{$act_time}{bb_credits} && Xormon::isdigit($stats_out{sancisco}{sanport}{$id}{$act_time}{bb_credits}) ) {
            if ( exists $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{bb_credits} ) {
              $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{bb_credits} += $stats_out{sancisco}{sanport}{$id}{$act_time}{bb_credits};
            }
            else {
              $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{bb_credits} = $stats_out{sancisco}{sanport}{$id}{$act_time}{bb_credits};
            }
          }
          if ( exists $stats_out{sancisco}{sanport}{$id}{$act_time}{crc_errors} && Xormon::isdigit($stats_out{sancisco}{sanport}{$id}{$act_time}{crc_errors}) ) {
            if ( exists $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{crc_errors} ) {
              $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{crc_errors} += $stats_out{sancisco}{sanport}{$id}{$act_time}{crc_errors};
            }
            else {
              $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{crc_errors} = $stats_out{sancisco}{sanport}{$id}{$act_time}{crc_errors};
            }
          }
          if ( exists $stats_out{sancisco}{sanport}{$id}{$act_time}{link_fail} && Xormon::isdigit($stats_out{sancisco}{sanport}{$id}{$act_time}{link_fail}) ) {
            if ( exists $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{link_fail} ) {
              $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{link_fail} += $stats_out{sancisco}{sanport}{$id}{$act_time}{link_fail};
            }
            else {
              $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{link_fail} = $stats_out{sancisco}{sanport}{$id}{$act_time}{link_fail};
            }
          }
          if ( exists $stats_out{sancisco}{sanport}{$id}{$act_time}{loss_sync} && Xormon::isdigit($stats_out{sancisco}{sanport}{$id}{$act_time}{loss_sync}) ) {
            if ( exists $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{loss_sync} ) {
              $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{loss_sync} += $stats_out{sancisco}{sanport}{$id}{$act_time}{loss_sync};
            }
            else {
              $stats_out{sancisco}{device}{$switch_uuid}{$act_time}{loss_sync} = $stats_out{sancisco}{sanport}{$id}{$act_time}{loss_sync};
            }
          }

          # if fetch task is longer than 2x300, then is necessary to save stats also for the previous 5 minutes samples
          if ( scalar(@add_timestamps) > 0 && exists $stats_out{sancisco}{sanport}{$id}{$act_time} ) {
            foreach my $new_timestamp (@add_timestamps) {
              $stats_out{sancisco}{sanport}{$id}{$new_timestamp} = $stats_out{sancisco}{sanport}{$id}{$act_time};
            }
          }
        }

      }
      $ports_count++;
    }
  }

  if ( $ports_count == 0 ) {
    $logger->error( "No port found" . __FILE__ . ":" . __LINE__ );
  }

  if ( scalar(@add_timestamps) > 0 && exists $stats_out{sancisco}{device}{$switch_uuid}{$act_time} ) {
    foreach my $new_timestamp (@add_timestamps) {
      $logger->info("prepare perf for additional timestamp: $new_timestamp ".localtime($new_timestamp));
      $stats_out{sancisco}{device}{$switch_uuid}{$new_timestamp} = $stats_out{sancisco}{device}{$switch_uuid}{$act_time};
    }
    $logger->info("prepare perf for current timestamp: $act_time ".localtime($act_time));
  }

  # save current counters for the further usage
  eval { Xormon::save_as_json($last_stats_file,\%stats_now); };

  return \%stats_out;
}

sub send_conf {
  my @architecture;
  my %conf;

  #
  # device (switch)
  #
  my %device_arch = (
    "item_id"    => $switch_uuid,
    "hostcfg_id" => $switch_uuid,
    "label"      => $switch_name,
    "hw_type"    => "sancisco",
    "subsystem"  => "device",
    "parents"    => []
  );
  push(@architecture, \%device_arch);
  # conf
  if ( exists $data{configuration}{connUnitSn} )      { $conf{$switch_uuid}{'serial'} = $data{configuration}{connUnitSn}; }
  if ( exists $data{configuration}{fcFeElementName} ) { $conf{$switch_uuid}{'wwn'}    = Xormon::formatWWN($data{configuration}{fcFeElementName}); }
  if ( exists $data{configuration}{connUnitUrl} ) {
    $conf{$switch_uuid}{'ip'} = $data{configuration}{connUnitUrl};
    $conf{$switch_uuid}{'ip'} =~ s/http://;
    $conf{$switch_uuid}{'ip'} =~ s/\///g;
  }
  if ( exists $data{configuration}{sysDescr} ) {
    my ( $switchType, $sysDescr, $os_version, undef ) = split( ", ", $data{configuration}{sysDescr} );
    $conf{$switch_uuid}{'model'}      = "$switchType";
    $conf{$switch_uuid}{'version'} = "$sysDescr $os_version";
  }
  #if ( exists $data{configuration}{sysUpTime} ) {
  #  ( $conf{$switch_uuid}{'days_up'}, undef ) = split( ":", $data{configuration}{sysUpTime} );
  #}
  my $total_ports  = 0;
  my $free_ports   = 0;
  my $avail_ports  = 0;
  my $unused_ports = 0;
  my $isl_ports    = "-"; #TODO

  # ports cfg
  #SNMP_lib::collect_san( \%data, $sess, 'ISL',        'connUnitLinkPortWwnY' );
  #SNMP_lib::collect_san( \%data, $sess, 'ports_host', 'fcIfNxPortName' );
  # vsan
  #SNMP_lib::collect_san( \%data, $sess, 'ports_vsan', 'vsanIfVsan' );
  #SNMP_lib::collect_san( \%data, $sess, 'vsan',       'vsanName' );
  #SNMP_lib::collect_san( \%data, $sess, 'vsan',       'dmDomainId' );
  #SNMP_lib::collect_san( \%data, $sess, 'vsan',       'dmPrincipalSwitchWwn' );
  #SNMP_lib::collect_san( \%data, $sess, 'vsan',       'dmLocalSwitchWwn' );

  # find host WWNs
  if ( exists $data{'ports_host'} ) {
    foreach ( keys %{ $data{'ports_host'} } ) {
      if ( exists $data{'ports_host'}{$_}{fcIfNxPortName} ) {
        my ( $port_id, $vsan_id, $val_id ) = split( /\./, $_ );
        if ( exists $data{'ports'}{$port_id} ) {
          $data{'ports'}{$port_id}{fcIfNxPortName}{$val_id}{'host_wwn'} = $data{'ports_host'}{$_}{fcIfNxPortName};
          $data{'ports'}{$port_id}{fcIfNxPortName}{$val_id}{'vsan_id'}  = $vsan_id;
        }
      }
    }
  }

  #
  # sanport
  #
  my %vsan;
  if ( exists $data{ports} ) {
    foreach ( sort { $a <=> $b } keys( %{ $data{ports} } ) ) {
      my $var = $data{ports}{$_};

      if ( exists $var->{ifName} && $var->{ifName} ne '' ) {
        unless ( supported_port($var->{ifName}) ) { next; } # skip unsupported ports
        $var->{ifName} = set_port_name($var->{ifName});

        my $id = lc( Xormon::create_hash_key($switch_uuid."sanport".$var->{ifName}) );

        # Architecture
        my %arch = (
          "item_id"    => $id,
          "label"      => $var->{ifName},
          "hw_type"    => "sancisco",
          "subsystem"  => "sanport",
          "hostcfg_id" => $switch_uuid,
          "parents"    => [$switch_uuid]
        );
        # port WWN
        if ( exists $var->{'connUnitPortWwn'} ) {
          my $port_wwn = Xormon::formatWWN($var->{'connUnitPortWwn'});
          $conf{$id}{'wwn'} = $port_wwn;
          $arch{wwn}        = $port_wwn;
        }
        # conf
        if ( exists $var->{'ifAdminStatus'} )       { $conf{$id}{'physical_status'}           = $var->{'ifAdminStatus'}; }
        if ( exists $var->{'ifOperStatus'} )        { $conf{$id}{'operational_status'}        = $var->{'ifOperStatus'}; }
        if ( exists $var->{'fcIfOperStatusCause'} ) { $conf{$id}{'operational_status_reason'} = $var->{'fcIfOperStatusCause'}; }
        if ( exists $var->{'ifAlias'} )  {
          $conf{$id}{'alias'} = $var->{'ifAlias'};
          $conf{$id}{'alias'} =~ s/\\//g;
          $conf{$id}{'alias'} =~ s/\r//g;
          $conf{$id}{'alias'} =~ s/'//g;

          # add alias to the label
          $arch{label} = "$var->{ifName} [$conf{$id}{'alias'}]";
        }
        my $host_wwns    = "";
        my $host_wwn_idx = 0;
        if ( exists $data{ISL}{$_}{connUnitLinkPortWwnY} ) {
          $data{ISL}{$_}{connUnitLinkPortWwnY} =~ s/\s//g;
          $data{ISL}{$_}{connUnitLinkPortWwnY} =~ s/0000000000000000$//;
          $data{ISL}{$_}{connUnitLinkPortWwnY} = Xormon::formatWWN($data{ISL}{$_}{connUnitLinkPortWwnY});
          $conf{$id}{isl} = $data{ISL}{$_}{connUnitLinkPortWwnY};
          push(@{ $arch{target_wwns} }, $data{ISL}{$_}{connUnitLinkPortWwnY});
          $host_wwns = $data{ISL}{$_}{connUnitLinkPortWwnY};
          $host_wwn_idx++;
        }
        # add host WWN to architecture for SAN topology purposes
        if ( exists $var->{'fcIfNxPortName'} && $var->{'fcIfNxPortName'} ne '' ) {
          foreach my $val_id ( sort { $a <=> $b } keys( %{ $var->{'fcIfNxPortName'} } ) ) {
            if ( exists $var->{'fcIfNxPortName'}{$val_id}{'host_wwn'} && $var->{'fcIfNxPortName'}{$val_id}{'host_wwn'} ne '' ) {
              my $host_wwn_act = Xormon::formatWWN($var->{'fcIfNxPortName'}{$val_id}{'host_wwn'});
              if ( exists $conf{$id}{isl} && $conf{$id}{isl} eq $host_wwn_act ) { next; } # already discovered WWN
              $host_wwn_idx++;
              push(@{ $arch{target_wwns} }, $host_wwn_act);
              if ( $host_wwn_idx == 1 ) {
                $host_wwns = $host_wwn_act;
              }
              else {
                $host_wwns .= ", $host_wwn_act";
              }
            }
          }
        }
        if ( $host_wwn_idx > 0 ) { $conf{$id}{'host_wwn'} = $host_wwns; }
        if ( exists $var->{ifHighSpeed} && $var->{ifHighSpeed} ne '' && Xormon::isdigit( $var->{ifHighSpeed} ) && $var->{ifHighSpeed} ne '0' ) {
          $conf{$id}{'speed'}    = $var->{ifHighSpeed} / 1000 . " Gbps";      # from Mbits to Gbits
          $conf{$id}{'speedBps'} = ( $var->{ifHighSpeed} * 1000 * 1000 ) / 8; # from Mbits to Bytes per second (Bps or B/s)
        }
        # switch ports summary
        #if ( $var->{ifName} =~ m/^portChannel/ || $var->{ifName} =~ m/^fcip/ ) { next; } # exclude these ports from the following stats
        # KZ: cannot be used like above, port was missing in arch... use different way below
        if ( $var->{ifName} !~ m/^portChannel/ && $var->{ifName} !~ m/^fcip/ ) { # exclude these ports from the following stats
          $total_ports++;
          if ( exists $conf{$id}{'operational_status_reason'} && exists $conf{$id}{'operational_status'} && exists $conf{$id}{'physical_status'} ) {
            if ( $conf{$id}{'operational_status_reason'} =~ m/fcotNotPresent/ ) {
              $unused_ports++;
              $conf{$id}{'working_state'} = "unused";
            }
            if ( $conf{$id}{'operational_status'} eq "up" && $conf{$id}{'physical_status'} eq "up" && $conf{$id}{'operational_status_reason'} eq "none" ) {
              $avail_ports++;
            }
            if ( $conf{$id}{'operational_status'} eq "down" && $conf{$id}{'operational_status_reason'} =~ m/^none$|^offline$|^adminDown$/ ) {
              $avail_ports++;
              $free_ports++;
            }
          }
        }
        # vsan
        if ( exists $data{ports_vsan}{$_} ) {
          if ( exists $data{ports_vsan}{$_}{vsanIfVsan} && $data{ports_vsan}{$_}{vsanIfVsan} ne '' ) {
            #8|sancisco  |           '1' => {
            #8|sancisco  |                    'dmLocalSwitchWwn' => '20 01 00 0D EC 4C 8E 41',
            #8|sancisco  |                    'vsanName' => 'fabric-blue',
            #8|sancisco  |                    'dmPrincipalSwitchWwn' => '20 01 00 0D EC 4C 8E 41',
            #8|sancisco  |                    'dmDomainId' => '51'
            #8|sancisco  |                  }
            my $vsan_id = $data{ports_vsan}{$_}{vsanIfVsan};
            if ( exists $data{vsan}{$vsan_id}{vsanName} && $data{vsan}{$vsan_id}{vsanName} ne '' ) {
              $vsan{$vsan_id}{label}  = $data{vsan}{$vsan_id}{vsanName};
              $vsan{$vsan_id}{id}     = lc( Xormon::create_hash_key("sanciscofabric".$data{vsan}{$vsan_id}{vsanName}.$vsan_id) );
              $conf{$id}{'vsan_name'} = $data{vsan}{$vsan_id}{vsanName};
              $conf{$id}{'vsan_id'}   = $vsan_id;
            }
            push(@{ $arch{parents} }, $vsan{$vsan_id}{id});
          }
        }
        push(@architecture, \%arch);
      }
    }
  }
  # licensed ports
  if ( exists $conf{$switch_uuid}{'model'} ) {
    $conf{$switch_uuid}{'licensed_ports'} = find_licensed_ports($total_ports,$conf{$switch_uuid}{'model'});
  }

  # switch ports summary
  $conf{$switch_uuid}{'total_ports'}     = $total_ports;
  $conf{$switch_uuid}{'free_ports'}      = $free_ports;
  $conf{$switch_uuid}{'available_ports'} = $avail_ports;
  $conf{$switch_uuid}{'unused_ports'}    = $unused_ports;
  $conf{$switch_uuid}{'isl_ports'}       = $isl_ports;

  # add vsans to architecture
  foreach (keys %vsan) {
    if ( exists $vsan{$_}{label} && exists $vsan{$_}{id} ) {
      # Architecture
      my %arch = (
        "item_id"   => $vsan{$_}{id},
        "label"     => $vsan{$_}{label},
        "hw_type"   => "sancisco",
        "subsystem" => "fabric",
        "parents"   => [],
      );
      push(@architecture, \%arch);
    }
  }

  #print "#################### Architecture\n";
  #print Dumper @architecture;
  #print "#################### CONF\n";
  #print Dumper \%conf;

  #save architecture
  my $rc_arch = $xormon->saveArchitecture(\@architecture);
  $logger->info("save architecture: $rc_arch");

  #save configuration
  my $rc_conf = $xormon->saveConfJSON(\%conf);
  $logger->info("save conf: $rc_conf");

  return 1;
}

sub collect_perf {
  # switch
  SNMP_lib::collect_san( \%data, $sess, 'configuration', 'connUnitStatus' );           # status of this switch
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'configuration', 'cseSysCPUUtilization' );     # CPU
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'configuration', 'cseSysMemoryUtilization' );  # Memory
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  # ports conf
  SNMP_lib::collect_san( \%data, $sess, 'ports', 'ifName' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports', 'ifAlias' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports', 'ifOperStatus' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports', 'ifAdminStatus' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports', 'fcIfOperStatusCause' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports', 'connUnitPortWwn' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports', 'ifHighSpeed' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  # ports perf
  SNMP_lib::collect_san( \%data, $sess, 'ports', 'ifHCOutOctets' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports', 'ifHCInOctets' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports', 'fcIfC3OutFrames' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports', 'fcIfC3InFrames' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports', 'fcIfInvalidCrcs' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports', 'fcIfBBCreditTransistionFromZero' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports', 'fcIfLinkFailures' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports', 'fcIfSyncLosses' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }

  return 1;
}

sub collect_conf {
  # switch cfg
  SNMP_lib::collect_san( \%data, $sess, 'configuration', 'sysDescr' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  #SNMP_lib::collect_san( \%data, $sess, 'configuration', 'sysUpTime' );
  #if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'configuration', 'connUnitSn' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'configuration', 'connUnitUrl' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'configuration', '1.3.6.1.2.1.75.1.1.2', 'fcFeElementName' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  # ports cfg
  SNMP_lib::collect_san( \%data, $sess, 'ISL',        'connUnitLinkPortWwnY' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'ports_host', 'fcIfNxPortName' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  # vsan
  SNMP_lib::collect_san( \%data, $sess, 'ports_vsan', 'vsanIfVsan' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'vsan',       'vsanName' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'vsan',       'dmDomainId' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'vsan',       'dmPrincipalSwitchWwn' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  SNMP_lib::collect_san( \%data, $sess, 'vsan',       'dmLocalSwitchWwn' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }
  # cisco count of the licensed ports
  SNMP_lib::collect_san( \%data, $sess, 'license', 'clmNoOfLicenseCurrentUsages' );
  if ( $snmp_request_sleep ) { sleep($snmp_request_sleep); }

  #print Dumper \%{ $data{vsan} };
  #print Dumper \%{ $data{ports_vsan} };

  return 1;
}

sub snmp_connection_test {
  $cfg = shift;

  #
  # initialize SNMP
  #
  SNMP_lib::initialize_snmp($cfg);
  SNMP_lib::initialize_mibs('san');
  #$sess = SNMP_lib::new_session( { 'Timeout' => 3000000 } ); # use 3 secs, default '1000000', micro-seconds before retry
  $sess = SNMP_lib::new_session( { 'Timeout' => 1000000 } ); # use only 1 sec, don't know why, but it always takes 6 times longer, so timeout is 6 seconds in this case

  my ( $status, $reason ) = SNMP_lib::collect_san( \%data, $sess, 'configuration', 'sysName', '', 1 ); # 1 -> conn test

  if ( $status ) {
    return 1, $data{configuration}{sysName}, SNMP_lib::build_snmpwalk_cmd($sess);
  }
  else {
    return 0, $reason, SNMP_lib::build_snmpwalk_cmd($sess);
  }
}

sub set_port_name {
  my $name = shift;

  $name =~ s/\//-/g;
  $name =~ s/^port-channel/portChannel/;
  $name =~ s/\s+/_/;

  return $name;
}

sub supported_port {
  my $name = shift;
  my $supp = 0;

  # exclude unsupported ports
  if ( $name =~ m/^fc|^port-channel|^Ten|^Ethernet|^GigabitEthernet|^Vlan|^bay|^ext/ ) { $supp = 1; }

  # exclude lan ports for Cisco Nexus switches, let there only fibre channel ports
  if ( defined $ENV{'EXCLUDE_LAN_FROM_CISCO_SAN_SWITCHES'} && isdigit( $ENV{'EXCLUDE_LAN_FROM_CISCO_SAN_SWITCHES'} ) && $ENV{'EXCLUDE_LAN_FROM_CISCO_SAN_SWITCHES'} == 1 ) {
    if ( $name !~ m/^fc/ ) { $supp = 0; }
  }

  return $supp;
}

sub find_licensed_ports {
  my $switch_ports = shift;
  my $switch_model = shift;

  if ( is_director($switch_model) ) {
    return $switch_ports; # Director switch has all ports licensed
  }

  my @supported_ids = ("20.80.79.82.84.95.65.67.84.73.86.95.57.49.52.56.83.95.80.75.71",
                       "20.80.79.82.84.95.65.67.84.73.86.95.57.49.52.56.84.95.80.75.71",
                       "14.80.79.82.84.95.65.67.84.73.86.95.50.48.80",
                       "19.80.79.82.84.95.65.67.84.73.86.65.84.73.79.78.95.80.75.71");

  if ( exists $data{license} ) {
    foreach my $lic_id ( keys( %{ $data{license} } ) ) {
      if ( ( grep {/^$lic_id$/} @supported_ids ) > 0 ) {
        if ( exists $data{license}{$lic_id}{clmNoOfLicenseCurrentUsages} && Xormon::isdigit($data{license}{$lic_id}{clmNoOfLicenseCurrentUsages}) && $data{license}{$lic_id}{clmNoOfLicenseCurrentUsages} != 0 ) {
          #print localtime()." : clmNoOfLicenseCurrentUsages : $lic_id : $data{license}{$lic_id}{clmNoOfLicenseCurrentUsages}\n";

          return $data{license}{$lic_id}{clmNoOfLicenseCurrentUsages};
        }
      }
    }
  }

  return "";
}

sub is_director {
  my $switch_model = shift;

  # Directors: Cisco MDS 9506, 9509, 9513, 9706, 9710 and 9718

  #Cisco NX-OS(tm) m9500
  #Cisco NX-OS(tm) m9700

  if ( ( grep {/m95[0-9][0-9]/} $switch_model) > 0 ) { return 1; }
  if ( ( grep {/m97[0-9][0-9]/} $switch_model) > 0 ) { return 1; }

  return 0;
}

sub make_gauge_int {
  my $timediff     = shift;
  my $counter_now  = shift;
  my $counter_last = shift;

  if ( Xormon::isdigit($timediff) && Xormon::isdigit($counter_now) && Xormon::isdigit($counter_last) && $timediff > 0 ) {
    if ( $counter_now > $counter_last ) {
      return int( ( $counter_now - $counter_last ) / $timediff );
    }
    elsif ( $counter_now < $counter_last ) {
      # probably counter overflow
      return "";
    }
    else {
      return 0;
    }
  }

  return "";
}

sub make_gauge {
  my $timediff     = shift;
  my $counter_now  = shift;
  my $counter_last = shift;

  if ( Xormon::isdigit($timediff) && Xormon::isdigit($counter_now) && Xormon::isdigit($counter_last) && $timediff > 0 ) {
    if ( $counter_now > $counter_last ) {
      #return Xormon::formatNumberByType("float4", ( $counter_now - $counter_last ) / $timediff );
      return ( $counter_now - $counter_last ) / $timediff;
    }
    elsif ( $counter_now < $counter_last ) {
      # probably counter overflow
      return "";
    }
    else {
      return 0;
    }
  }

  return "";
}

1;
