package SYNOLOGY;

use strict;
use warnings;
use SNMPHandler;
use JSON qw(encode_json decode_json);
use SNMP_lib;

my $logger;
my $xormon;
my $cfg;
my %performance_data;
my $st_type;
my $st_serial;
my @architecture = ();
my $timestamp    = "";
my $end_time     = "";
my $HTTP_BAD_REQUEST    = 400;
my $HTTP_INTERNAL_ERROR = 500;

# The value of sample_rate (which is set to 300) represents the sampling interval, i.e.,
# the time period over which the counter values were collected. 5 * 60 sec see more in (collect_perf)
my $sample_rate   = 300;
my $act_time      = "";
my $act_log_level = 0;     # 0-none,1-info,2-performance stats,3-debug

my %DATA_TO_COLLECT = (
  statistics => {
    "cpu" => {
      section => "CPU-CORE Level Statistics",
      metrics => [
        { mib => "ssCpuUser",   metric => "cpu_user" },
        { mib => "ssCpuSystem", metric => "cpu_system" },
        { mib => "ssCpuIdle",   metric => "cpu_idle" },
        {
          mib       => "enterprises.6574.108.2",
          metric    => "cpu_gpu_util_percent",
          use_index => 0
        }
      ],
      samples => [ 1, 2, 3, 4, 5 ],
    },
    "memory" => {
      section => "Memory Level Statistics",
      metrics => [
        {
          mib       => "hrStorageDescr",
          metric    => "description",
          use_index => 0
        },
        { mib => "hrStorageSize", metric => "memory_total" },
        { mib => "hrStorageUsed", metric => "memory_used" },
        {
          mib       => "memAvailReal",
          metric    => "memory_free",
          use_index => 0
        },
        {
          mib       => "enterprises.6574.108.3",
          metric    => "memory_gpu_util_percent",
          use_index => 0
        },
        {
          mib       => "enterprises.6574.108.4",
          metric    => "memory_gpu_free",
          use_index => 0
        },
        {
          mib       => "enterprises.6574.108.5",
          metric    => "memory_gpu_used",
          use_index => 0
        },
        {
          mib       => "enterprises.6574.108.6",
          metric    => "memory_gpu_total",
          use_index => 0
        },
      ],
      samples => [ 1, 2, 3, 4, 5 ],
    },
    "lun" => {
      section => "Lun Level Statistics",
      metrics => [
        { mib => "iSCSILUNUUID", metric => "uid", use_index => 0 },
        { mib => "iSCSILUNThroughputReadLow",  metric => "read_low" },
        { mib => "iSCSILUNThroughputWriteLow", metric => "write_low" },
        { mib => "iSCSILUNThroughputReadHigh", metric => "read_high" },
        {
          mib    => "iSCSILUNThroughputWriteHigh",
          metric => "write_high"
        },
        { mib => "iSCSILUNIopsRead",        metric => "read_io" },
        { mib => "iSCSILUNIopsWrite",       metric => "write_io" },
        { mib => "iSCSILUNDiskLatencyRead", metric => "disk_resp_t_r" },
        {
          mib    => "iSCSILUNDiskLatencyWrite",
          metric => "disk_resp_t_w"
        },
        {
          mib    => "iSCSILUNNetworkLatencyTx",
          metric => "port_resp_t_r"
        },
        {
          mib    => "iSCSILUNNetworkLatencyRx",
          metric => "port_resp_t_w"
        },
        { mib => "iSCSILUNQueueDepth", metric => "depth" },
      ],
      samples => [ 1, 2, 3, 4, 5 ],
    },
    "port" => {
      section => "Port Level Statistics",
      metrics => [
        { mib => "ifName",        metric => "name", use_index => 0 },
        { mib => "ifHCInOctets",  metric => "read" },
        { mib => "ifHCOutOctets", metric => "write" },
      ],
      samples => [ 1, 5 ],
    },
    "drive" => {
      section => "Drive Level Statistics",
      metrics => [
        { mib => "storageIOIndex",    metric => "id",   use_index => 0 },
        { mib => "storageIOLA5",      metric => "util", use_index => 0 },
        { mib => "storageIONRead",    metric => "read" },
        { mib => "storageIONWritten", metric => "write" },
        { mib => "storageIOReads",    metric => "read_io" },
        { mib => "storageIOWrites",   metric => "write_io" },
      ],
      samples => [ 1, 5 ],
    },
    "volume" => {
      section => "Volume Level Statistics",
      metrics => [
        { mib => "spaceIOIndex", metric => "id",   use_index => 0 },
        { mib => "spaceIOLA5",   metric => "util", use_index => 0 },
        { mib => "raidName",     metric => "name", use_index => 0 },
        {
          mib       => "raidFreeSize",
          metric    => "capacity_free",
          use_index => 0
        },
        {
          mib       => "raidTotalSize",
          metric    => "capacity_total",
          use_index => 0
        },
        { mib => "spaceIONRead",    metric => "read" },
        { mib => "spaceIONWritten", metric => "write" },
        { mib => "spaceIOReads",    metric => "read_io" },
        { mib => "spaceIOWrites",   metric => "write_io" },
      ],
      samples => [ 1, 5 ],
    },
  },
  configuration => {
    "storage" => {
      section => "Storage Level Configuration",
      metrics => [
        { mib => "modelName",       metric => "ModelName" },
        { mib => "version",         metric => "Version" },
        { mib => "serialNumber",    metric => "serial" },
        { mib => "systemStatus",    metric => "systemstatus" },
        { mib => "temperature",     metric => "temperature" },
        { mib => "powerStatus",     metric => "powerstatus" },
        { mib => "systemFanStatus", metric => "systemfanstatus" },
        { mib => "cpuFanStatus",    metric => "cpufanstatus" },
        { mib => "sysName",         metric => "name" },
      ],
    },
    "cpu" => {
      section => "CPU-CORE Level Configuration",
      metrics => [ { mib => "ssCpuUser", metric => "cpu_user" }, ],
    },
    "memory" => {
      section => "Memory Level Configuration",
      metrics => [
        { mib => "hrStorageDescr", metric => "description" },
        { mib => "hrStorageSize",  metric => "memory_total" },
        { mib => "hrStorageUsed",  metric => "memory_used" },
        { mib => "memAvailReal",   metric => "memory_free" },
      ],
    },
    "drive" => {
      section => "Drive Level Configuration",
      metrics => [
        { mib => "diskID",          metric => "name" },
        { mib => "diskStatus",      metric => "health_status" },
        { mib => "diskModel",       metric => "model" },
        { mib => "diskType",        metric => "type" },
        { mib => "diskTemperature", metric => "temperature" },
      ],
    },
    "port" => {
      section => "Port Level Configuration",
      metrics => [
        { mib => "ifName",  metric => "name" },
        { mib => "ifAlias", metric => "alias" },
      ],
    },
    "volume" => {
      section => "Volume Level Configuration",
      metrics => [
        { mib => "spaceIOIndex",  metric => "id" },
        { mib => "raidName",      metric => "name" },
        { mib => "raidFreeSize",  metric => "capacity_free" },
        { mib => "raidTotalSize", metric => "capacity_total" },
        { mib => "raidStatus",    metric => "health_status" },
      ],
    },
    "pool" => {
      section => "Pool Level Configuration",
      metrics => [
        { mib => "raidName",      metric => "name" },
        { mib => "raidFreeSize",  metric => "capacity_free" },
        { mib => "raidTotalSize", metric => "capacity_total" },
        { mib => "raidStatus",    metric => "health_status" },
      ],
    },
    "lun" => {
      section => "Lun Level Configuration",
      metrics => [
        { mib => "iSCSILUNUUID", metric => "uid" },
        { mib => "iSCSILUNName", metric => "name" },
      ],
    },
  },
);

# Method: collect_data
#
# This method collects performance and configuration data from a Synology device using SNMP.
# It initializes necessary objects, tests connection, gathers performance data, and logs the results.
# Depending on the success of the operations, it will return a status indicating the outcome.
#
# Parameters:
# - $xormon: A reference to the Xormon object responsible for handling data storage and communication.
# - $cfg: A reference to the configuration hash that contains necessary settings for the SNMP session.
#
# Returns:
# - $status: A string indicating whether the operation was "successful", "unauthorized", or "error".
sub collect_data {
  $xormon   = shift;
  $cfg      = shift;
  $act_time = time();

  # Test and assign main variables
  $cfg->{'uuid'} =
    ( exists $cfg->{'hostcfg_id'} )
    ? $cfg->{'hostcfg_id'}
    : Xormon::error("Cannot continue, 'hostcfg_id' is not configured!")
    && return "error";

  $st_type       = "synology";
  $st_serial     = $cfg->{'uuid'};
  $act_log_level = $cfg->{'log_level'} // 0;

  $logger =
    XormonLogger->new( $st_type, $cfg->{hostcfg_id}, $cfg->{hostalias} );

  message_log( 3, "Act log level", $act_log_level );

  # Initialize SNMPHandler
  my $snmp = SNMPHandler->new( $cfg, $logger );

  # Check connection
  my ( $conn_status, $reason ) = $snmp->check_connection();

  # Save to logs
  if ($conn_status) {
    $xormon->isConnected(1);
  }
  else {
    $xormon->isConnected( 0, $reason );
  }

  # Collect performance data
  my $start_time = time();
  message_log( 2, "Starting to collect performance data." );

  # Collect perf data
  collect_perf($snmp);

  my $end_time = time();                    # End timing this section
  my $duration = $end_time - $start_time;

  message_log( 2, "End collect performance data in $duration seconds." );
  message_log( 1, "Performance data were collected" );
  message_log( 1, "Trying to save performance data" );
  my $rc     = $xormon->saveJSON2( \%performance_data );
  my $status = "error";

  if ( defined $rc ) {
    message_log( 1, "Save perf", $rc );

    $logger->info("save perf: $rc");

    if ( $rc eq "201" ) {
      $status = "successful";
    }
    elsif ( $rc eq "401" ) {
      $status = "unauthorized";
    }
    else {
      $status = "error";
    }
  }
  else {
    message_log( 1, "Save perf failed: Undefined status" );
    $logger->info("save perf: UN");
  }

  # Performance data status
  if ( $status eq "successful" ) {
    $xormon->hasData(1);
    message_log( 1, "Performance data was successfully saved." );
  }
  else {
    $xormon->hasData( 0, "no available stats: $status" );
    message_log( 1, "No available stats.", $status );
  }

  message_log( 3, "Obj SNMP Data", $snmp->{data} );

  # Get configuration data
  $start_time = time();
  message_log( 2, "Starting to collect configuration data." );

  ###
  # Collect conf
  ###
  collect_conf($snmp);

  $end_time = time();
  $duration = $end_time - $start_time;
  message_log( 2, "End collect configuration data in $duration seconds." );


  message_log( 3, "Performance Data",   \%performance_data );
  message_log( 3, "Configuration Data", $snmp->{data}->{configuration} );

  ###
  # Save arch.
  ###
  message_log( 1, "Start saving architecture." );

  if ( set_architecture($snmp) == 1 ) {
    message_log( 1, "Architecture was successfully saved." );
  }
  else {
    message_log( 1, "Failure while saving architecture data." );
  }

  return $status;
}

# This subroutine 'collect_conf' collects configuration data by iterating over predefined
# metrics within the 'DATA_TO_COLLECT' hash. For each metric, it retrieves SNMP data
# using the 'collect_synology' method. After collecting the necessary data, it calls
# various 'set_data_*_conf' methods to store the collected information related to
# different system components such as storage, memory, CPU, drives, ports, volumes,
# pools, and LUNs.
#
# Parameters:
# - $snmp: An SNMP object used to collect data from the Synology device.
#
# Returns:
# - None;
sub collect_conf {
  my $snmp = shift;

  # Iterate over the configuration section of DATA_TO_COLLECT
  for my $key ( keys %{ $DATA_TO_COLLECT{configuration} } ) {
    my $section_data = $DATA_TO_COLLECT{configuration}{$key};

    # Collect data for each metric in the configuration section
    for my $metric ( @{ $section_data->{metrics} } ) {
      my $metric_name = $metric->{metric};

      $snmp->collect_synology( $section_data->{section},
        $metric->{mib}, $metric_name );
    }
  }

  # Call the set_data_*_conf
  set_data_storage_conf($snmp);
  set_data_memory_conf($snmp);
  set_data_cpu_conf($snmp);
  set_data_drive_conf($snmp);
  set_data_port_conf($snmp);
  set_data_volume_conf($snmp);
  set_data_pool_conf($snmp);
  set_data_lun_conf($snmp);
}

# Method: health_status
#
# This method is responsible for collecting and assessing the health status of various system components,
# including system status, power status, and fan statuses. It gathers data from different sections such
# as "Storage Level Configuration", "Drive Level Configuration", "Volume Level Configuration", and
# "Pools Level Configuration". The method checks if any components are in a failed state, and if so,
# records these failures. It then compiles the overall health status of the system, encoding it into JSON
# format, and saves the results into the database.
#
# Parameters:
# - $snmp: A reference to an SNMP session object that contains methods and data for SNMP communication.
#
sub health_status {
  my $snmp = shift;

  my %hs_data;
  my @hs                = ();
  my @failures          = ();
  my $snmp_section      = "Storage Level Configuration";
  my $system_status     = "";
  my $power_status      = "";
  my $system_fan_status = "";
  my $cpu_fan_status    = "";

  # Check and collect system status metrics
  my $metric = "systemstatus";
  if ( defined $snmp->{data}{$snmp_section}{0}{$metric} ) {
    $system_status = $snmp->{data}{$snmp_section}{0}{$metric};
    if ( $system_status =~ m/failed/ ) {
      push( @failures, "System Status : $system_status" );
    }
  }

  $metric = "powerstatus";
  if ( defined $snmp->{data}{$snmp_section}{0}{$metric} ) {
    $power_status = $snmp->{data}{$snmp_section}{0}{$metric};
    if ( $power_status =~ m/failed/ ) {
      push( @failures, "Power Status : $power_status" );
    }
  }

  $metric = "cpufanstatus";
  if ( defined $snmp->{data}{$snmp_section}{0}{$metric} ) {
    $cpu_fan_status = $snmp->{data}{$snmp_section}{0}{$metric};
    if ( $cpu_fan_status =~ m/failed/ ) {
      push( @failures, "CPU FAN Status : $cpu_fan_status" );
    }
  }

  $metric = "systemfanstatus";
  if ( defined $snmp->{data}{$snmp_section}{0}{$metric} ) {
    $system_fan_status = $snmp->{data}{$snmp_section}{0}{$metric};
    if ( $system_fan_status =~ m/failed/ ) {
      push( @failures, "System FAN Status : $system_fan_status" );
    }
  }

  # Check and collect health status for drives, volumes, and pools
  my @sections = (
    "Drive Level Configuration",
    "Volume Level Configuration",
    "Pools Level Configuration"
  );
  foreach my $section (@sections) {
    foreach my $id ( keys %{ $snmp->{data}{$section} } ) {
      my $name   = $snmp->{data}{$section}{$id}{name};
      my $status = $snmp->{data}{$section}{$id}{health_status};
      my $sec    = $section;
      $sec =~ s/ Level Configuration//g;
      $sec =~ s/Pools/Pool/g;
      if ( !defined $name || !defined $status ) { next; }
      if ( $sec eq "Pool" && $name !~ m/Storage Pool/ ) { next; }
      if ( $status =~ m/Crashed|SystemPartitionFailed/ ) {
        push( @failures, "$sec : $name : $status" );
      }

      # Store component health status
      my %component;
      $component{section} = $sec;
      $component{name}    = $name;
      $component{status}  = $status;
      push(
        @{ $hs_data{ $cfg->{'uuid'} }{health_status_data}{components} },
        \%component
      );
    }
  }

  # Store system health status
  my %system;
  $system{systemstatus}    = $system_status;
  $system{powerstatus}     = $power_status;
  $system{cpufanstatus}    = $cpu_fan_status;
  $system{systemfanstatus} = $system_fan_status;
  push( @{ $hs_data{ $cfg->{'uuid'} }{health_status_data}{system} }, \%system );

  # Determine overall status
  my $status = "ok";
  my $reason = "Optimal";

  if (@failures) {
    $status = "error";
    $reason = join( ", ", @failures );
  }

  # Log the status and reason for debugging
  message_log( 3, "Overall status: $status" );
  message_log( 3, "Reason: $reason" );

  push(
    @hs,
    {
      "item_id" => $cfg->{'uuid'},
      "status"  => $status,
      "data"    => $reason
    }
  );

  # Encode health status data to JSON
  if ( exists $hs_data{ $cfg->{'uuid'} }{health_status_data} ) {
    $hs_data{ $cfg->{'hostcfg_id'} }{health_status_data} =
      encode_json( $hs_data{ $cfg->{'hostcfg_id'} }{health_status_data} );
  }

  # Save to database and log
  my $rc = $xormon->saveStatus( \@hs );
  $logger->info("save health status: $rc");

  my $rc_hs = $xormon->saveConfJSON( \%hs_data, 1 );
  $logger->info("save health status table: $rc_hs");

}

# Method: set_architecture
#
# This method configures the architecture of the device by creating and storing
# the hierarchy and relationships of device components. It builds a list of components
# and their attributes, such as item ID, label, hardware type, and subsystem.
# The method also handles the removal of unnecessary data and saves the architecture
# and configuration data to the database.
#
# Parameters:
# - $snmp: A reference to an SNMP session object containing methods and data for SNMP communication.
#
# Returns:
# - 1 if both the configuration and architecture data were successfully saved,
#   or 0 if any save operation failed.

sub set_architecture {
  my $snmp = shift;

  # Remove unwanted test connection data if it exists
  if ( exists $snmp->{data}->{configuration}->{sysName} ) {
    delete $snmp->{data}->{configuration}->{sysName};
  }

  # DEVICE: Define the architecture for the main device
  my %device_arch = (
    "item_id"    => $cfg->{'uuid'},         # Unique ID for the device
    "label"      => $cfg->{'hostalias'},    # Device alias (e.g., "Synology")
    "hw_type"    => $st_type,               # Hardware type (e.g., "Synology")
    "subsystem"  => "device",               # Subsystem type
    "parents"    => [],                     # No parent devices
    "hostcfg_id" => $cfg->{'uuid'}          # Configuration ID
  );

  # Add device architecture to the architecture array
  push( @architecture, \%device_arch );

  # Process and add configuration data for each component
  foreach my $uid ( keys %{ $snmp->{data}{configuration} } ) {
    if ( ref( $snmp->{data}{configuration}{$uid} ) eq 'HASH' ) {
      my $name = $snmp->{data}{configuration}{$uid}{name};
      my $type = $snmp->{data}{configuration}{$uid}{TYPE};
      if ( !defined $name || !defined $type ) { next; }

      # Define the architecture for each component
      my %arch = (
        "item_id"   => $uid,
        "label"     => $name,
        "hw_type"   => $st_type,
        "subsystem" => $type,
        "parents"   => [ $cfg->{'uuid'} ]
      );

      # If the component is a volume, add WWN for mapping
      if ( $type eq "volume" ) {
        $arch{wwn} = Xormon::formatWWN($uid);
      }

      # Add the component architecture to the architecture array
      push( @architecture, \%arch );

      # Remove the TYPE key after processing
      delete $snmp->{data}{configuration}{$uid}{TYPE};
    }
  }

  # Save the architecture data to the database
  my $rc_arch = $xormon->saveArchitecture( \@architecture );
  # Check if the architecture was successfully saved (response code not 500 or 400)
  if ( $rc_arch != $HTTP_INTERNAL_ERROR && $rc_arch != $HTTP_BAD_REQUEST ) {

      # Start collecting health status 
      message_log( 2, "Starting collect data for health status." );
      health_status($snmp);
      message_log( 2, "End collect data for health status." );

      # Save the configuration data to the database
      my $rc_conf = $xormon->saveConfJSON( $snmp->{data}{configuration} );

      message_log( 3, "Status of save Conf", $rc_conf );
  }

  message_log( 3, "Status of save Arch", $rc_arch );

  # Return the status: 1 if saveArchitecture was successful, otherwise 0
  return $rc_arch != $HTTP_INTERNAL_ERROR && $rc_arch != $HTTP_BAD_REQUEST ? 1 : 0;
}

# Function: collect_perf
#
# This function gathers performance metrics using SNMP by repeatedly collecting data
# over a series of five intervals. It collects CPU, memory, LUN, port, drive, and volume
# data depending on the iteration and stores the gathered data for later use.
#
# Parameters:
# - $snmp: Reference to the SNMP session object.
#
# Returns:
# - None (performs data collection and processing).
sub collect_perf {
  my $snmp = shift;

  # Capture the current timestamp
  $timestamp = time();

  # Main loop for data collection, repeats 5 times
  for ( my $i = 1 ; $i <= 5 ; $i++ ) {
    my $t1 = time();
    message_log( 3, "Step $i" );

    # Iterate over the statistics section of DATA_TO_COLLECT
    for my $key ( keys %{ $DATA_TO_COLLECT{statistics} } ) {
      my $section_data = $DATA_TO_COLLECT{statistics}{$key};

      # Check if $i is in array
      if ( grep { $_ == $i } @{ $section_data->{samples} } ) {
        my $start_time = time();

        for my $metric ( @{ $section_data->{metrics} } ) {
          my $use_index =
            exists $metric->{use_index} ? $metric->{use_index} : 1;
          my $metric_name = $metric->{metric};

          if ($use_index) {
            $metric_name .= $i;
          }

          # Collect data using the collect_synology method
          $snmp->collect_synology( $section_data->{section},
            $metric->{mib}, $metric_name );
          message_log(
            3,
            sprintf(
              "For Section %s are vars: %s\n",
              $section_data->{section} // 'undefined section',
              ref( $snmp->{data} ) eq 'HASH'
                && exists $snmp->{data}->{ $section_data->{section} }
              ? format_data( $snmp->{data}->{ $section_data->{section} } )
              : 'undefined value'
            )
          );
        }

        my $end_time = time();                    # End timing this section
        my $duration = $end_time - $start_time;
        message_log( 2, "Section $key collected in $duration seconds." );
      }
    }

    # Last collect step
    if ( $i == 5 ) {

      # TODO: Call the refactored set_data_*
      set_data_cpu( $i, $snmp );
      set_data_memory( $i, $snmp );
      set_data_port( $i, $snmp );
      set_data_drive( $i, $snmp );
      set_data_volume( $i, $snmp );
      set_data_lun( $i, $snmp );
    }

    my $d = 60 - ( time() - $t1 );

    message_log( 2, "Sleep $d" );
    sleep($d) if $d > 0;
  }

  # Capture the end time
  $end_time = time();

  # Collect pool data and device
  set_data_pool($snmp);
  set_data_device($snmp);
}

# Method: set_data_storage_conf
#
# This method collects configuration information from a device using SNMP.
# It retrieves various system metrics such as model name, version, serial number,
# temperature, system status, power status, and fan statuses. The collected data is
# then stored in the `$snmp->{data}{configuration}` hash for further processing or storage.
#
# Parameters:
# - $snmp: A reference to an SNMP session object that contains methods and data for SNMP communication.
#
# Returns:
# - Always returns 1 indicating successful execution.
sub set_data_storage_conf {
  my $snmp = shift;

  my $snmp_section = "Storage Level Configuration";
  my $data_ref     = $snmp->{data}{$snmp_section}{0} || {};

  # Define keys to collect data for
  my @keys =
    qw(ModelName Version serial systemstatus temperature powerstatus systemfanstatus cpufanstatus name);

  # Define corresponding storage keys
  my @storage_keys =
    qw(model version serial system_status temperature power_status system_fan_status cpu_fan_status machine_name);

  # Iterate over all keys and store data
  for my $i ( 0 .. $#keys ) {
    my $key         = $keys[$i];
    my $storage_key = $storage_keys[$i];
    $snmp->{data}{configuration}{$st_serial}{$storage_key} = $data_ref->{$key}
      if defined $data_ref->{$key};
  }

  return 1;
}

# Method: set_data_pool_conf
#
# This method collects and configures pool-related data from the Synology device using SNMP.
# It aggregates capacity metrics from volume-level configuration and stores them under the pool section.
#
# Parameters:
# - $snmp: A reference to the SNMP session object.
#
# Returns:
# - None.
sub set_data_pool_conf {
  my $snmp             = shift;
  my $section          = "pool";
  my $snmp_section     = "Pool Level Configuration";
  my $snmp_section_vol = "Volume Level Configuration";
  my $pool_id          = 0;

  my $free  = 0;
  my $total = 0;
  my $used  = 0;

  foreach my $volume ( keys %{ $snmp->{data}{$snmp_section_vol} } ) {
    my $free_l = $snmp->{data}{$snmp_section_vol}{$volume}{"capacity_free"};
    my $total_l =
      $snmp->{data}{$snmp_section_vol}{$volume}{"capacity_total"};
    my $used_l = $snmp->{data}{$snmp_section_vol}{$volume}{"capacity_used"};

    if ( Xormon::isdigit($total_l)
      && Xormon::isdigit($free_l)
      && Xormon::isdigit($used_l) )
    {
      $total += $total_l;
      $used  += $used_l;
      $free  += $free_l;
    }
  }

  if ( Xormon::isdigit($total)
    && Xormon::isdigit($free)
    && Xormon::isdigit($used)
    && $total != 0
    && $free != 0
    && $used != 0 )
  {
    $snmp->{data}{$snmp_section}{$pool_id}{"capacity_used"}  = $used;
    $snmp->{data}{$snmp_section}{$pool_id}{"capacity_total"} = $total;
    $snmp->{data}{$snmp_section}{$pool_id}{"capacity_free"}  = $free;
    $snmp->{data}{$snmp_section}{$pool_id}{"name"}           = "array_total";
    $snmp->{data}{$snmp_section}{$pool_id}{"id"}             = 0;

    # Create a unique identifier for the pool configuration
    my $uid = create_uid( $st_serial, $section, $pool_id, $snmp->{data},
      $snmp_section );

    $snmp->{data}{configuration}{$uid} =
      $snmp->{data}{$snmp_section}{$pool_id};
    $snmp->{data}{configuration}{$uid}{TYPE} = $section;
  }
}

# Method: set_data_pool
#
# This method collects and aggregates pool-related statistics from the Synology device using SNMP.
# It sums the relevant metrics across all volumes and stores the results under the pool section.
#
# Parameters:
# - $snmp: A reference to the SNMP session object.
#
# Returns:
# - None.
sub set_data_pool {
  my $snmp         = shift;
  my $section      = "pool";
  my $snmp_section = "Pool Level Statistics";
  my @metrics      = (
    "read_io", "write_io",  "io_rate",        "read",
    "write",   "data_rate", "capacity_total", "capacity_used",
    "capacity_free"
  );
  my $snmp_section_vol = "Volume Level Statistics";

  my $pool_name = "array_total";
  my $pool_id   = 0;

  foreach my $volume ( keys %{ $snmp->{data}{$snmp_section_vol} } ) {
    foreach my $metric (@metrics) {
      if (
        Xormon::isdigit( $snmp->{data}{$snmp_section_vol}{$volume}{$metric} ) )
      {
        if (
          Xormon::isdigit( $snmp->{data}{$snmp_section}{$pool_id}{$metric} ) )
        {
          $snmp->{data}{$snmp_section}{$pool_id}{$metric} +=
            $snmp->{data}{$snmp_section_vol}{$volume}{$metric};
        }
        else {
          $snmp->{data}{$snmp_section}{$pool_id}{$metric} =
            $snmp->{data}{$snmp_section_vol}{$volume}{$metric};
        }
      }
    }
    $snmp->{data}{$snmp_section}{$pool_id}{id}   = $pool_id;
    $snmp->{data}{$snmp_section}{$pool_id}{name} = $pool_name;
  }

  # Capacity
  if (
    Xormon::isdigit( $snmp->{data}{$snmp_section}{$pool_id}{"capacity_total"} )
    )
  {
    $snmp->{data}{$snmp_section}{$pool_id}{"capacity_total"} =
      $snmp->{data}{$snmp_section}{$pool_id}{"capacity_total"};
  }
  if (
    Xormon::isdigit( $snmp->{data}{$snmp_section}{$pool_id}{"capacity_used"} ) )
  {
    $snmp->{data}{$snmp_section}{$pool_id}{"capacity_used"} =
      $snmp->{data}{$snmp_section}{$pool_id}{"capacity_used"};
  }
  if (
    Xormon::isdigit( $snmp->{data}{$snmp_section}{$pool_id}{"capacity_free"} ) )
  {
    $snmp->{data}{$snmp_section}{$pool_id}{"capacity_free"} =
      $snmp->{data}{$snmp_section}{$pool_id}{"capacity_free"};
  }

  # Capacity in percent
  if (
    Xormon::isdigit( $snmp->{data}{$snmp_section}{$pool_id}{"capacity_total"} )
    && Xormon::isdigit(
      $snmp->{data}{$snmp_section}{$pool_id}{"capacity_used"} ) )
  {
    if ( $snmp->{data}{$snmp_section}{$pool_id}{"capacity_total"} == 0 ) {
      $snmp->{data}{$snmp_section}{$pool_id}{"capacity_used_percent"} = 0;
    }
    else {
      my $pom = ( $snmp->{data}{$snmp_section}{$pool_id}{"capacity_used"} /
          $snmp->{data}{$snmp_section}{$pool_id}{"capacity_total"} ) * 100;
      $snmp->{data}{$snmp_section}{$pool_id}{"capacity_used_percent"} =
        Xormon::round( $pom, 2 );
    }
  }

  # BlockSize (read, write)
  my $data_act = $snmp->{data}{$snmp_section}{$pool_id};
  if ( exists $data_act->{io_rate}
    && defined $data_act->{io_rate} 
    && $data_act->{io_rate} ne ''
    && Xormon::isdigit( $data_act->{io_rate} )
    && exists $data_act->{data_rate}
    && defined $data_act->{data_rate}
    && $data_act->{data_rate} ne ''
    && Xormon::isdigit( $data_act->{data_rate} ))
  {
    if ( $data_act->{io_rate} == 0 ) {
      $snmp->{data}{$snmp_section}{$pool_id}{block_size} = 0;
    }
    else {
      $snmp->{data}{$snmp_section}{$pool_id}{block_size} =
      Xormon::formatNumberByType( "float4",        
      $data_act->{data_rate} / $data_act->{io_rate} );
    }
  }
  if ( exists $data_act->{read_io}
    && Xormon::isdigit( $data_act->{read_io} )
    && exists $data_act->{read} )
  {
    if ( $data_act->{read_io} == 0 ) {
      $snmp->{data}{$snmp_section}{$pool_id}{block_size_r} = 0;
    }
    else {
      $snmp->{data}{$snmp_section}{$pool_id}{block_size_r} =
        Xormon::formatNumberByType( "float4",
        $data_act->{read} / $data_act->{read_io} );
    }
  }
  if ( exists $data_act->{write_io}
    && Xormon::isdigit( $data_act->{write_io} )
    && exists $data_act->{write} )
  {
    if ( $data_act->{write_io} == 0 ) {
      $snmp->{data}{$snmp_section}{$pool_id}{block_size_w} = 0;
    }
    else {
      $snmp->{data}{$snmp_section}{$pool_id}{block_size_w} =
        Xormon::formatNumberByType( "float4",
        $data_act->{write} / $data_act->{write_io} );
    }
  }

  # Create a unique identifier for the pool
  my $uid =
    create_uid( $st_serial, $section, $pool_id, $snmp->{data}, $snmp_section );

  foreach my $metric ( keys %{ $snmp->{data}{$snmp_section}{$pool_id} } ) {
    $performance_data{$st_type}{$section}{$uid}{$timestamp}{$metric} =
      $snmp->{data}{$snmp_section}{$pool_id}{$metric};
  }
}

# Just copy device perf from pool perf
sub set_data_device {
  my $snmp         = shift;
  my $section      = "device";
  my $snmp_section = "Pool Level Statistics";
  my $pool_id      = 0;

  foreach my $metric ( keys %{ $snmp->{data}{$snmp_section}{$pool_id} } ) {
    $performance_data{$st_type}{$section}{$st_serial}{$timestamp}{$metric} =
      $snmp->{data}{$snmp_section}{$pool_id}{$metric};
  }
}

# Method: set_data_volume_conf
#
# This method collects and configures volume-related data from the Synology device using SNMP.
# It processes each volume, computes used capacities, and stores the results.
#
# Parameters:
# - $snmp: A reference to the SNMP session object.
#
# Returns:
# - None.
sub set_data_volume_conf {
  my $snmp         = shift;
  my $snmp_section = "Volume Level Configuration";
  my $section      = "volume";

  foreach my $volume ( keys %{ $snmp->{data}{$snmp_section} } ) {
    if ( $volume !~ m/-/
      || !defined $snmp->{data}{$snmp_section}{$volume}{name} )
    {
      next;
    }
    my $name  = $snmp->{data}{$snmp_section}{$volume}{name};
    my $total = $snmp->{data}{$snmp_section}{$volume}{"capacity_total"};
    my $free  = $snmp->{data}{$snmp_section}{$volume}{"capacity_free"};

    if ( $name =~ m/Volume/ ) {
      $name =~ s/Volume//g;
      $name = remove_whitespace($name);
      if ( defined $snmp->{data}{$snmp_section}{$name} ) {
        if ( Xormon::isdigit($total) && Xormon::isdigit($free) ) {
          $snmp->{data}{$snmp_section}{$name}{"capacity_used"} =
            Xormon::round( $total - $free, 0 );
          $snmp->{data}{$snmp_section}{$name}{"capacity_total"} =
            $total;
          $snmp->{data}{$snmp_section}{$name}{"capacity_free"} =
            $free;
          $snmp->{data}{$snmp_section}{$name}{name} =
            $snmp->{data}{$snmp_section}{$volume}{name};
          $snmp->{data}{$snmp_section}{$name}{volume_id} = $name;

          if ( defined $snmp->{data}{$snmp_section}{$volume}{"health_status"} )
          {
            $snmp->{data}{$snmp_section}{$name}{"health_status"} =
              $snmp->{data}{$snmp_section}{$volume}{"health_status"};
          }

          $snmp->{data}{$snmp_section}{$name}{"pool_id"} = 0;
          $snmp->{data}{$snmp_section}{$name}{"pool_name"} =
            "array_total";
        }
        delete $snmp->{data}{$snmp_section}{$volume};
      }
    }
    else {
      delete $snmp->{data}{$snmp_section}{$volume};
      next;
    }
  }

  foreach my $volume ( keys %{ $snmp->{data}{$snmp_section} } ) {

    # Create a unique identifier for the volume configuration
    my $uid =
      create_uid( $st_serial, $section, $volume, $snmp->{data}, $snmp_section );

    $snmp->{data}{configuration}{$uid} =
      $snmp->{data}{$snmp_section}{$volume};
    if ( !defined $snmp->{data}{configuration}{$uid}{TYPE} ) {
      $snmp->{data}{configuration}{$uid}{TYPE} = $section;
    }
  }
}

# Method: set_data_volume
#
# This method collects and processes volume-level statistics using SNMP. It retrieves various
# performance metrics for volumes and calculates additional metrics like data rate, IO rate,
# and block sizes. The method stores the processed data in the performance data structure.
#
# Parameters:
# - $sample: Sampling rate indicator, used to differentiate between different data collections.
# - $snmp: A reference to the SNMP session object.
#
# Returns:
# - None.
sub set_data_volume {
  my $sample = shift;
  my $snmp   = shift;

  my $snmp_section = "Volume Level Statistics";
  my $section      = "volume";

  if ( $sample eq "5" ) {
    my @metrics = ( "read_io", "write_io", "read", "write" );

    foreach my $port ( keys %{ $snmp->{data}{$snmp_section} } ) {
      foreach my $metric (@metrics) {
        $snmp->{data}{$snmp_section}{$port}{$metric} = get_value_from_counters(
          $snmp->{data}{$snmp_section}{$port}{ $metric . "1" },
          $snmp->{data}{$snmp_section}{$port}{ $metric . "5" },
          $metric, $snmp_section
        );
      }

      if ( Xormon::isdigit( $snmp->{data}{$snmp_section}{$port}{read} )
        && Xormon::isdigit( $snmp->{data}{$snmp_section}{$port}{write} )
        && !Xormon::isdigit( $snmp->{data}{$snmp_section}{$port}{data_rate} ) )
      {
        $snmp->{data}{$snmp_section}{$port}{data_rate} =
          $snmp->{data}{$snmp_section}{$port}{read} +
          $snmp->{data}{$snmp_section}{$port}{write};
      }

      if ( Xormon::isdigit( $snmp->{data}{$snmp_section}{$port}{read_io} )
        && Xormon::isdigit( $snmp->{data}{$snmp_section}{$port}{write_io} )
        && !Xormon::isdigit( $snmp->{data}{$snmp_section}{$port}{io_rate} ) )
      {
        $snmp->{data}{$snmp_section}{$port}{io_rate} =
          $snmp->{data}{$snmp_section}{$port}{read_io} +
          $snmp->{data}{$snmp_section}{$port}{write_io};
      }

      my $data_act = $snmp->{data}{$snmp_section}{$port};
      if ( exists $data_act->{io_rate}
        && Xormon::isdigit( $data_act->{io_rate} )
        && $data_act->{io_rate} ne ''
        && exists $data_act->{data_rate} 
        && $data_act->{data_rate} ne '')
      {
        if ( $data_act->{io_rate} == 0 ) {
          $snmp->{data}{$snmp_section}{$port}{block_size} = 0;
        }
        else {
          $snmp->{data}{$snmp_section}{$port}{block_size} =
            Xormon::formatNumberByType( "float4",
            $data_act->{data_rate} / $data_act->{io_rate} );
        }
      }
      if ( exists $data_act->{read_io}
        && Xormon::isdigit( $data_act->{read_io} )
        && $data_act->{read_io} ne ''
        && exists $data_act->{read} 
        && $data_act->{read} ne '' )
      {
        if ( $data_act->{read_io} == 0 ) {
          $snmp->{data}{$snmp_section}{$port}{block_size_r} = 0;
        }
        else {
          $snmp->{data}{$snmp_section}{$port}{block_size_r} =
            Xormon::formatNumberByType( "float4",
            $data_act->{read} / $data_act->{read_io} );
        }
      }
      if ( exists $data_act->{write_io}
        && Xormon::isdigit( $data_act->{write_io} )
        && $data_act->{write_io} ne ''
        && exists $data_act->{write} 
        && $data_act->{write} ne '')
      {

        if ( defined $data_act->{write}
          && defined $data_act->{write_io}
          && $data_act->{write_io} != 0 )
        {
          $snmp->{data}{$snmp_section}{$port}{block_size_w} =
            Xormon::formatNumberByType( "float4",
            $data_act->{write} / $data_act->{write_io} );
        }
        else {
          $snmp->{data}{$snmp_section}{$port}{block_size_w} = 0;
        }
      }
    }

    foreach my $volume ( keys %{ $snmp->{data}{$snmp_section} } ) {
      if ( $volume !~ m/-/
        || !defined $snmp->{data}{$snmp_section}{$volume}{name} )
      {
        next;
      }

      my $name      = $snmp->{data}{$snmp_section}{$volume}{name};
      my $total     = $snmp->{data}{$snmp_section}{$volume}{"capacity_total"};
      my $free      = $snmp->{data}{$snmp_section}{$volume}{"capacity_free"};
      my $pool_name = $snmp->{data}{$snmp_section}{$volume}{"pool_name"};
      my $pool_id   = $snmp->{data}{$snmp_section}{$volume}{"pool_id"};

      if ( $name =~ m/Volume/ ) {
        $name =~ s/Volume//g;
        $name = remove_whitespace($name);

        # Based on the number in the name, determine if it is a volume or pool
        if ( defined $snmp->{data}{$snmp_section}{$name} ) {
          if ( Xormon::isdigit($total) && Xormon::isdigit($free) ) {
            $snmp->{data}{$snmp_section}{$name}{"capacity_used"} =
              $total - $free;
            $snmp->{data}{$snmp_section}{$name}{"capacity_total"} = $total;
            $snmp->{data}{$snmp_section}{$name}{"capacity_free"}  = $free;
            $snmp->{data}{$snmp_section}{$name}{"capacity_used_percent"} = 0;

            # Capacity in percent
            if (
              Xormon::isdigit(
                $snmp->{data}{$snmp_section}{$name}{"capacity_total"}
              )
              && Xormon::isdigit(
                $snmp->{data}{$snmp_section}{$name}{"capacity_used"}
              )
              )
            {
              if ( $snmp->{data}{$snmp_section}{$name}{"capacity_total"} == 0 )
              {
                $snmp->{data}{$snmp_section}{$name}{"capacity_used_percent"} =
                  0;
              }
              else {
                my $tmp =
                  ( $snmp->{data}{$snmp_section}{$name}{"capacity_used"} /
                    $snmp->{data}{$snmp_section}{$name}{"capacity_total"} ) *
                  100;
                $snmp->{data}{$snmp_section}{$name}{"capacity_used_percent"} =
                  Xormon::round( $tmp, 2 );
              }
            }

            if ( defined $pool_name ) {
              $snmp->{data}{$snmp_section}{$name}{"pool_name"} =
                $pool_name;
            }
            if ( defined $pool_id ) {
              $snmp->{data}{$snmp_section}{$name}{"pool_id"} =
                $pool_id;
            }

          }
          delete $snmp->{data}{$snmp_section}{$volume};
        }
      }
      else {
        delete $snmp->{data}{$snmp_section}{$volume};
        next;
      }
    }

    foreach my $volume ( keys %{ $snmp->{data}{$snmp_section} } ) {

      # Create a unique identifier for the volume
      my $uid = create_uid( $st_serial, $section, $volume, $snmp->{data},
        $snmp_section );
      $performance_data{$st_type}{$section}{$uid}{$timestamp} =
        $snmp->{data}{$snmp_section}{$volume};
    }
  }
}

# Function: set_data_drive_conf
#
# This function collects and configures information about drives
# on a Synology device using SNMP. The data is stored in the configuration
# structure for further processing.
#
# Parameters:
# - $snmp: Reference to the SNMP session object.
#
# Returns:
# - None (performs data collection and storage).

sub set_data_drive_conf {
  my $snmp               = shift;
  my $snmp_section       = "Drive Level Configuration";
  my $snmp_section_stats = "Drive Level Statistics";
  my $section            = "drive";

  # Iterate over all drives and store them in the configuration
  foreach my $serial ( keys %{ $snmp->{data}{$snmp_section} } ) {

    # Skip drives without a defined ID
    if ( !defined $snmp->{data}{$snmp_section}{$serial}{id} ) { next; }

    # Create a unique identifier for the drive
    my $uid =
      create_uid( $st_serial, $section, $serial, $snmp->{data}, $snmp_section );

    # Store drive information in the configuration
    foreach my $item ( keys %{ $snmp->{data}{$snmp_section}{$serial} } ) {
      $snmp->{data}{configuration}{$uid}{$item} =
        $snmp->{data}{$snmp_section}{$serial}{$item};

      # Set the TYPE field if not already defined
      if ( !defined $snmp->{data}{configuration}{$uid}{TYPE} ) {
        $snmp->{data}{configuration}{$uid}{TYPE} = $section;
      }
    }
  }
}

# Function: set_data_drive
#
# This function collects and processes drive-level statistics from a Synology device using SNMP.
# It retrieves performance metrics for the drives, calculates additional metrics like data rate and IO rate,
# and stores the processed data in the performance data structure.
#
# Parameters:
# - $sample: Sampling rate indicator, used to differentiate between different data collections.
# - $snmp: Reference to the SNMP session object.
#
# Returns:
# - None (performs data collection and processing).

sub set_data_drive {
  my $sample = shift;
  my $snmp   = shift;

  my $snmp_section = "Drive Level Statistics";
  my $section      = "drive";

  # If the sample rate is 5, collect additional data and compute metrics
  if ( $sample eq "5" ) {

    my @metrics = ( "read_io", "write_io", "read", "write" );

    foreach my $port ( keys %{ $snmp->{data}{$snmp_section} } ) {

      # Create a unique identifier for the drive
      my $uid =
        create_uid( $st_serial, $section, $port, $snmp->{data}, $snmp_section );

      foreach my $metric (@metrics) {
        $performance_data{$st_type}{$section}{$uid}{$timestamp}{$metric} =
          get_value_from_counters(
          $snmp->{data}{$snmp_section}{$port}{ $metric . "1" },
          $snmp->{data}{$snmp_section}{$port}{ $metric . "5" },
          $metric, $section
          );
      }

      # Calculate data rate if both read and write values are present
      if (
        Xormon::isdigit(
          $performance_data{$st_type}{$section}{$uid}{$timestamp}{read}
        )
        && Xormon::isdigit(
          $performance_data{$st_type}{$section}{$uid}{$timestamp}{write}
        )
        && !Xormon::isdigit(
          $performance_data{$st_type}{$section}{$uid}{$timestamp}{data_rate}
        )
        )
      {
        $performance_data{$st_type}{$section}{$uid}{$timestamp}{data_rate} =
          $performance_data{$st_type}{$section}{$uid}{$timestamp}{read} +
          $performance_data{$st_type}{$section}{$uid}{$timestamp}{write};
      }

      # Calculate IO rate if both read_io and write_io values are present
      if (
        Xormon::isdigit(
          $performance_data{$st_type}{$section}{$uid}{$timestamp}{read_io}
        )
        && Xormon::isdigit(
          $performance_data{$st_type}{$section}{$uid}{$timestamp}{write_io}
        )
        && !Xormon::isdigit(
          $performance_data{$st_type}{$section}{$uid}{$timestamp}{io_rate}
        )
        )
      {
        $performance_data{$st_type}{$section}{$uid}{$timestamp}{io_rate} =
          $performance_data{$st_type}{$section}{$uid}{$timestamp}{read_io} +
          $performance_data{$st_type}{$section}{$uid}{$timestamp}{write_io};
      }
    }
  }
}

# Function: set_data_port_conf
#
# This function collects and configures port-level data from a Synology device using SNMP.
# It retrieves the name and alias of each port and stores the data in the configuration structure.
#
# Parameters:
# - $snmp: Reference to the SNMP session object.
#
# Returns:
# - None (performs data collection and storage).

sub set_data_port_conf {
  my $snmp         = shift;
  my $snmp_section = "Port Level Configuration";
  my $section      = "port";

  # Iterate over all ports and store them in the configuration
  foreach my $serial ( keys %{ $snmp->{data}{$snmp_section} } ) {

    # Create unique identifier for drive configuration
    my $uid =
      create_uid( $st_serial, $section, $serial, $snmp->{data}, $snmp_section );

    # Store port information in the configuration
    foreach my $item ( keys %{ $snmp->{data}{$snmp_section}{$serial} } ) {
      $snmp->{data}{configuration}{$uid}{$item} =
        $snmp->{data}{$snmp_section}{$serial}{$item};

      # Set the TYPE field if not already defined
      if ( !defined $snmp->{data}{configuration}{$uid}{TYPE} ) {
        $snmp->{data}{configuration}{$uid}{TYPE} = $section;
      }
    }
  }
}

# Function: set_data_port
#
# This function collects and processes port-level statistics from a Synology device using SNMP.
# It retrieves performance metrics such as read and write operations for the ports,
# calculates additional metrics like data rate, and stores the processed data in the performance data structure.
#
# Parameters:
# - $sample: Sampling rate indicator, used to differentiate between different data collections.
# - $snmp: Reference to the SNMP session object.
#
# Returns:
# - None (performs data collection and processing).

sub set_data_port {
  my $sample = shift;
  my $snmp   = shift;

  my $snmp_section = "Port Level Statistics";
  my $section      = "port";

  ### If the sample rate is 5, calculate additional metrics
  if ( $sample eq "5" ) {
    my @metrics = ( "read", "write" );

    foreach my $port ( keys %{ $snmp->{data}{$snmp_section} } ) {

      # Create unique identifier for port
      my $uid =
        create_uid( $st_serial, $section, $port, $snmp->{data}, $snmp_section );

      foreach my $metric (@metrics) {
        $performance_data{$st_type}{$section}{$uid}{$timestamp}{$metric} =
          get_value_from_counters(
          $snmp->{data}{$snmp_section}{$port}{ $metric . "1" },
          $snmp->{data}{$snmp_section}{$port}{ $metric . "5" },
          $metric, $section
          );
      }

      # Calculate data rate if both read and write values are present
      if (
        Xormon::isdigit(
          $performance_data{$st_type}{$section}{$uid}{$timestamp}{read}
        )
        && Xormon::isdigit(
          $performance_data{$st_type}{$section}{$uid}{$timestamp}{write}
        )
        && !Xormon::isdigit(
          $performance_data{$st_type}{$section}{$uid}{$timestamp}{data_rate}
        )
        )
      {
        $performance_data{$st_type}{$section}{$uid}{$timestamp}{data_rate} =
          $performance_data{$st_type}{$section}{$uid}{$timestamp}{read} +
          $performance_data{$st_type}{$section}{$uid}{$timestamp}{write};
      }
    }
  }
}

# Function: set_data_memory_conf
#
# This function collects and configures memory-related information from a Synology device using SNMP.
# It stores the memory configuration data in the configuration structure under the 'memory' section.
#
# Parameters:
# - $snmp: Reference to the SNMP session object.
#
# Returns:
# - None (performs data collection and configuration).

sub set_data_memory_conf {
  my $snmp = shift;

  my $section      = "system";
  my $snmp_section = "Memory Level Configuration";

  # Set name and ID for total memory
  $snmp->{data}{$snmp_section}{0}{name} = "total";
  $snmp->{data}{$snmp_section}{0}{id}   = "0";

  # Create unique identifier for memory configuration
  my $uid = create_uid( $st_serial, $section, 0, $snmp->{data}, $snmp_section );

  # Store memory configuration data
  $snmp->{data}{configuration}{$uid} = $snmp->{data}{$snmp_section}{0};
  $snmp->{data}{configuration}{$uid}{TYPE} = $section;
}

# Function: set_data_memory
#
# This function collects and processes memory-level statistics from a Synology device using SNMP.
# It retrieves various memory metrics, such as total and used memory, calculates additional values,
# and stores the processed data in the performance data structure.
#
# Parameters:
# - $sample: Sampling rate indicator, used to differentiate between different data collections.
# - $snmp: Reference to the SNMP session object.
#
# Returns:
# - None (performs data collection and processing).

sub set_data_memory {
  my $sample = shift;
  my $snmp   = shift;

  my $section      = "system";
  my $snmp_section = "Memory Level Statistics";
  my $mib          = "";
  my $metric       = "";
  my @metrics      = ( "memory_total", "memory_used" );

  # If the sample rate is 5, perform additional calculations
  if ( $sample == 5 ) {
    for ( my $i = 1 ; $i <= 5 ; $i++ ) {
      foreach my $node ( keys %{ $snmp->{data}{$snmp_section} } ) {
        foreach my $metric (@metrics) {
          my $val =
            $snmp->{data}{$snmp_section}{$node}{ $metric . $i };
          if ( Xormon::isdigit($val) ) {
            if (
              Xormon::isdigit( $snmp->{data}{$snmp_section}{$node}{$metric} ) )
            {
              $snmp->{data}{$snmp_section}{$node}{$metric} +=
                $val;
            }
            else {
              $snmp->{data}{$snmp_section}{$node}{$metric} = $val;
            }
          }
        }
      }
    }

    foreach my $node ( keys %{ $snmp->{data}{$snmp_section} } ) {
      foreach my $metric (@metrics) {
        if ( Xormon::isdigit( $snmp->{data}{$snmp_section}{$node}{$metric} ) ) {
          $snmp->{data}{$snmp_section}{$node}{$metric} =
            Xormon::round( $snmp->{data}{$snmp_section}{$node}{$metric} / 5,
            0 );
        }
      }
    }
  }

  # Return early if sample rate is not 5
  return if $sample != 5;

  my $id                          = 0;
  my $util                        = "";
  my $total                       = "";
  my $used_total                  = "";
  my $free                        = "";
  my $buffer                      = "";
  my $cache                       = "";
  my $shared                      = "";
  my $tmp_memory_gpu_util_percent = "";
  my $tmp_memory_gpu_free         = "";
  my $tmp_memory_gpu_used         = "";
  my $tmp_memory_gpu_total        = "";

  # Calculate physical used memory, buffers, cached memory, and shared memory
  foreach my $node ( keys %{ $snmp->{data}{$snmp_section} } ) {
    my $name = $snmp->{data}{$snmp_section}{$node}{description};
    next unless defined $name;

    my $size = $snmp->{data}{$snmp_section}{$node}{"memory_total"};
    my $used = $snmp->{data}{$snmp_section}{$node}{"memory_used"};

    if ( $name eq "Physical memory" && Xormon::isdigit($size) ) {
      $total      = $size;
      $used_total = $used;
    }
    if ( $name eq "Memory buffers" && Xormon::isdigit($used) ) {
      $buffer = $used;
    }
    if ( $name eq "Cached memory" && Xormon::isdigit($used) ) {
      $cache = $used;
    }
    if ( $name eq "Shared memory" && Xormon::isdigit($used) ) {
      $shared = $used;
    }

    # before delete - store gpu_* statistics
    $tmp_memory_gpu_util_percent =
      $snmp->{data}{$snmp_section}{$node}{"memory_gpu_util_percent"};
    $tmp_memory_gpu_free =
      $snmp->{data}{$snmp_section}{$node}{"memory_gpu_free"};
    $tmp_memory_gpu_used =
      $snmp->{data}{$snmp_section}{$node}{"memory_gpu_used"};
    $tmp_memory_gpu_total =
      $snmp->{data}{$snmp_section}{$node}{"memory_gpu_total"};
  }

  delete $snmp->{data}{$snmp_section};

  if ( Xormon::isdigit($used_total)
    && Xormon::isdigit($buffer)
    && Xormon::isdigit($shared)
    && Xormon::isdigit($cache)
    && Xormon::isdigit($total) )
  {
    $snmp->{data}{$snmp_section}{$id}{id} = $id;
    my $used_real = $used_total - $buffer - $shared - $cache;
    my $free      = $total - ( $used_real + $buffer + $cache );
    $snmp->{data}{$snmp_section}{$id}{"memory_used"} =
      Xormon::round( $used_real, 0 );
    $snmp->{data}{$snmp_section}{$id}{"memory_buffer"} = $buffer;
    $snmp->{data}{$snmp_section}{$id}{"memory_cached"} = $cache;
    $snmp->{data}{$snmp_section}{$id}{"memory_total"}  = $total;
    $snmp->{data}{$snmp_section}{$id}{"memory_free"} =
      Xormon::round( $free, 0 );
    $snmp->{data}{$snmp_section}{$id}{"memory_util_percent"} =
      Xormon::round( ( $used_real / $total ) * 100, 1 );

    $snmp->{data}{$snmp_section}{$id}{"memory_gpu_util_percent"} =
      (    defined $tmp_memory_gpu_util_percent
        && $tmp_memory_gpu_util_percent ne ''
        && Xormon::isdigit($tmp_memory_gpu_util_percent) )
      ? $tmp_memory_gpu_util_percent
      : 0;
    $snmp->{data}{$snmp_section}{$id}{"memory_gpu_free"} =
      (    defined $tmp_memory_gpu_free
        && $tmp_memory_gpu_free ne ''
        && Xormon::isdigit($tmp_memory_gpu_free) )
      ? $tmp_memory_gpu_free
      : 0;
    $snmp->{data}{$snmp_section}{$id}{"memory_gpu_used"} =
      (    defined $tmp_memory_gpu_used
        && $tmp_memory_gpu_used ne ''
        && Xormon::isdigit($tmp_memory_gpu_used) )
      ? $tmp_memory_gpu_used
      : 0;
    $snmp->{data}{$snmp_section}{$id}{"memory_gpu_total"} =
      (    defined $tmp_memory_gpu_total
        && $tmp_memory_gpu_total ne ''
        && Xormon::isdigit($tmp_memory_gpu_total) )
      ? $tmp_memory_gpu_total
      : 0;

    # Calculate reserved memory
    $free = $snmp->{data}{$snmp_section}{$id}{"memory_free"};
    if ( $total - ( $used_real + $free + $buffer + $cache ) >= 0 ) {
      my $reserve = $total - ( $used_real + $free + $buffer + $cache );
      $snmp->{data}{$snmp_section}{$id}{"memory_reserved"} =
        Xormon::round( $reserve + $shared, 0 );
    }

    # Create unique identifier for memory
    my $uid =
      create_uid( $st_serial, $section, $id, $snmp->{data}, $snmp_section );

    foreach my $metric ( keys %{ $snmp->{data}{$snmp_section}{$id} } ) {
      $performance_data{$st_type}{$section}{$uid}{$timestamp}{$metric} =
        $snmp->{data}{$snmp_section}{$id}{$metric};
    }
  }
}

# Function: set_data_cpu_conf
#
# This function collects and configures CPU-related data from a Synology device using SNMP.
# It stores the CPU configuration data in the configuration structure under the 'cpu' section.
#
# Parameters:
# - $snmp: Reference to the SNMP session object.
#
# Returns:
# - None (performs data collection and configuration).

sub set_data_cpu_conf {
  my $snmp         = shift;
  my $section      = "system";
  my $snmp_section = "CPU-CORE Level Configuration";

  # Set name for total CPU usage
  $snmp->{data}{$snmp_section}{0}{name} = "total";

  # Create unique identifier for CPU configuration
  my $uid = create_uid( $st_serial, $section, 0, $snmp->{data}, $snmp_section );

  # Store CPU configuration data
  $snmp->{data}{configuration}{$uid} = $snmp->{data}{$snmp_section}{0};
  $snmp->{data}{configuration}{$uid}{TYPE} = $section;
}

# Function: set_data_cpu
#
# This function collects and processes CPU statistics from a Synology device using SNMP.
# It retrieves various CPU metrics such as user, system, and idle times, and stores
# the processed data in the performance data structure.
#
# Parameters:
# - $sample: Sampling rate indicator, used to differentiate between different data collections.
# - $snmp: Reference to the SNMP session object.
#
# Returns:
# - None (performs data collection and processing).

sub set_data_cpu {
  my $sample       = shift;
  my $snmp         = shift;
  my $section      = "system";
  my $snmp_section = "CPU-CORE Level Statistics";

  # Define metrics for further calculations
  my @metrics          = ( "cpu_user", "cpu_system", "cpu_idle" );
  my @metrics_to_store = ( @metrics, "cpu_gpu_util_percent" );

  # Perform additional calculations if the sample rate is 5
  if ( $sample == 5 ) {

#$snmp->collect_synology($snmp_section, "enterprises.6574.108.2", "gpu_util(%)");

    for ( my $i = 1 ; $i <= 5 ; $i++ ) {
      foreach my $node ( keys %{ $snmp->{data}{$snmp_section} } ) {
        foreach my $metric (@metrics) {
          my $val =
            $snmp->{data}{$snmp_section}{$node}{ $metric . $i };

          if ( Xormon::isdigit($val) ) {
            if (
              Xormon::isdigit( $snmp->{data}{$snmp_section}{$node}{$metric} ) )
            {
              $snmp->{data}{$snmp_section}{$node}{$metric} +=
                $val;
            }
            else {
              $snmp->{data}{$snmp_section}{$node}{$metric} = $val;
            }

          }
        }
      }

      foreach my $node ( keys %{ $snmp->{data}{$snmp_section} } ) {
        foreach my $metric (@metrics) {
          if ( Xormon::isdigit( $snmp->{data}{$snmp_section}{$node}{$metric} ) )
          {
            $snmp->{data}{$snmp_section}{$node}{$metric} =
              Xormon::round( $snmp->{data}{$snmp_section}{$node}{$metric} / 5,
              2 );
          }
        }
      }

    }

    # collect perf data
    foreach my $node ( keys %{ $snmp->{data}{$snmp_section} } ) {
      my $uid =
        create_uid( $st_serial, $section, $node, $snmp->{data}, $snmp_section );

      # Store processed data in the performance data structure
      foreach my $metric (@metrics_to_store) {
        $performance_data{$st_type}{$section}{$uid}{$timestamp}{$metric} =
          $snmp->{data}{$snmp_section}{$node}{$metric};
      }

# foreach my $metric (@metrics) {
#    $performance_data{$st_type}{$section}{$uid}{$timestamp}{$metric} = $snmp->{data}{$snmp_section}{$node}{$metric};
# }
    }
  }
}

# Function: set_data_lun_conf
#
# This function collects and configures LUN (Logical Unit Number) data from a Synology device using SNMP.
# It retrieves the LUN configuration details and stores the data in the configuration structure.
#
# Parameters:
# - $snmp: Reference to the SNMP session object.
#
# Returns:
# - None (performs data collection and configuration).

sub set_data_lun_conf {
  my $snmp         = shift;
  my $section      = "lun";
  my $snmp_section = "Lun Level Configuration";

  foreach my $node ( keys %{ $snmp->{data}{$snmp_section} } ) {

    # Create a unique identifier for LUN configuration
    my $uid =
      create_uid( $st_serial, $section, $node, $snmp->{data}, $snmp_section );

    $snmp->{data}{configuration}{$uid} =
      $snmp->{data}{$snmp_section}{$node};
    if ( !defined $snmp->{data}{configuration}{$uid}{TYPE} ) {
      $snmp->{data}{configuration}{$uid}{TYPE} = $section;
    }
  }

}

# Function: set_data_lun
#
# This function collects and processes LUN (Logical Unit Number) statistics from a Synology device using SNMP.
# It retrieves various LUN metrics such as throughput, IOPS, and latency, and stores the processed data
# in the performance data structure.
#
# Parameters:
# - $sample: Sampling rate indicator, used to differentiate between different data collections.
# - $snmp: Reference to the SNMP session object.
#
# Returns:
# - None (performs data collection and processing).

sub set_data_lun {
  my $sample       = shift;
  my $snmp         = shift;
  my $section      = "lun";
  my $snmp_section = "Lun Level Statistics";

  my @metrics = (
    "read_low",      "write_low",     "read_high",     "write_high",
    "read_io",       "write_io",      "disk_resp_t_r", "disk_resp_t_w",
    "port_resp_t_r", "port_resp_t_w", "depth"
  );
  my @metrics_to_store = (
    "resp_t_r",      "resp_t_w",      "resp_t",        "port_resp_t",
    "disk_resp_t",   "data_rate",     "io_rate",       "read_io",
    "write_io",      "disk_resp_t_r", "disk_resp_t_w", "port_resp_t_r",
    "port_resp_t_w", "read",          "write"
  );

  # Perform additional calculations if the sample rate is 5
  if ( $sample == 5 ) {
    for ( my $i = 1 ; $i <= 5 ; $i++ ) {
      foreach my $node ( keys %{ $snmp->{data}{$snmp_section} } ) {
        foreach my $metric (@metrics) {
          my $val =
            $snmp->{data}{$snmp_section}{$node}{ $metric . $i };
          if ( Xormon::isdigit($val) ) {
            if (
              Xormon::isdigit( $snmp->{data}{$snmp_section}{$node}{$metric} ) )
            {
              $snmp->{data}{$snmp_section}{$node}{$metric} +=
                $val;
            }
            else {
              $snmp->{data}{$snmp_section}{$node}{$metric} = $val;
            }
          }
        }
      }
    }

    # Average or scale the metrics after summing them up
    foreach my $node ( keys %{ $snmp->{data}{$snmp_section} } ) {
      foreach my $metric (@metrics) {
        if ( Xormon::isdigit( $snmp->{data}{$snmp_section}{$node}{$metric} ) ) {
          if ( $metric =~ m/resp/ ) {
            $snmp->{data}{$snmp_section}{$node}{$metric} =
              Xormon::round( $snmp->{data}{$snmp_section}{$node}{$metric} / 5,
              3 );
          }
          elsif ( $metric =~ /^read-|^write-/ ) {
            my $denominator = ( 1 * 5 );
            $snmp->{data}{$snmp_section}{$node}{$metric} = Xormon::round(
              $snmp->{data}{$snmp_section}{$node}{$metric} / $denominator, 0 );
          }
          else {
            $snmp->{data}{$snmp_section}{$node}{$metric} =
              Xormon::round( $snmp->{data}{$snmp_section}{$node}{$metric} / 5,
              0 );
          }
        }
      }
    }

    # Calculate combined metrics for read and write operations
    foreach my $node ( keys %{ $snmp->{data}{$snmp_section} } ) {
      my $read_io  = $snmp->{data}{$snmp_section}{$node}{read_io}  // undef;
      my $write_io = $snmp->{data}{$snmp_section}{$node}{write_io} // undef;

      my $read_low  = $snmp->{data}{$snmp_section}{$node}{"read_low"};
      my $write_low = $snmp->{data}{$snmp_section}{$node}{"write_low"};

      my $read_high  = $snmp->{data}{$snmp_section}{$node}{"read_high"};
      my $write_high = $snmp->{data}{$snmp_section}{$node}{"write_high"};

      my $read  = "";
      my $write = "";

      # if (Xormon::isdigit($read_low)) {
      #     $read = $read_low;
      # }
      # if (Xormon::isdigit($write_low)) {
      #     $write = $write_low;
      # }
      # if (Xormon::isdigit($read_high) && $read_high != 0) {
      #     $read = "$read_high$read";
      # }
      # if (Xormon::isdigit($write_high) && $write_high != 0) {
      #     $write = "$write_high$write";
      # }

      if ( Xormon::isdigit($read_low) ) {
        $read = Xormon::isdigit($read_high)
          && $read_high != 0 ? $read_high . $read_low : $read_low;
      }

      if ( Xormon::isdigit($write_low) ) {
        $write = Xormon::isdigit($write_high)
          && $write_high != 0 ? $write_high . $write_low : $write_low;
      }
      $snmp->{data}{$snmp_section}{$node}{read}  = $read;
      $snmp->{data}{$snmp_section}{$node}{write} = $write;

      my $data_rate = "";
      my $io_rate   = "";
      if ( Xormon::isdigit($read_io) ) {
        $io_rate = $read_io;
      }
      if ( Xormon::isdigit($write_io) ) {
        if ( Xormon::isdigit($io_rate) ) {
          $io_rate = $io_rate + $write_io;
        }
        else {
          $io_rate = $write_io;
        }
      }
      $snmp->{data}{$snmp_section}{$node}{io_rate} = $io_rate;

      if ( Xormon::isdigit($read) ) {
        $data_rate = $read;
      }
      if ( Xormon::isdigit($write) ) {
        if ( Xormon::isdigit($data_rate) ) {
          $data_rate = $data_rate + $write;
        }
        else {
          $data_rate = $write;
        }
      }
      $snmp->{data}{$snmp_section}{$node}{data_rate} = $data_rate;

      # Compute latency
      my $disk_resp_t_r = $snmp->{data}{$snmp_section}{$node}{"disk_resp_t_r"}
        // undef;
      my $disk_resp_t_w = $snmp->{data}{$snmp_section}{$node}{"disk_resp_t_w"}
        // undef;
      my $port_resp_t_r = $snmp->{data}{$snmp_section}{$node}{"port_resp_t_r"}
        // undef;
      my $port_resp_t_w = $snmp->{data}{$snmp_section}{$node}{"port_resp_t_w"}
        // undef;

      my $disk_resp_t = 0;
      if ( Xormon::isdigit($disk_resp_t_r)
        && Xormon::isdigit($disk_resp_t_w) )
      {
        $disk_resp_t = $disk_resp_t_r + $disk_resp_t_w;
      }

      my $port_resp_t = 0;
      if ( Xormon::isdigit($port_resp_t_r)
        && Xormon::isdigit($port_resp_t_w) )
      {
        $port_resp_t = $port_resp_t_r + $port_resp_t_w;
      }

      my $resp_t_r = 0;
      if ( Xormon::isdigit($disk_resp_t_r)
        && Xormon::isdigit($port_resp_t_r) )
      {
        $resp_t_r = $disk_resp_t_r + $port_resp_t_r;
      }

      my $resp_t_w = 0;
      if ( Xormon::isdigit($disk_resp_t_w)
        && Xormon::isdigit($port_resp_t_w) )
      {
        $resp_t_w = $disk_resp_t_w + $port_resp_t_w;
      }

      my $resp_t = 0;
      if ( Xormon::isdigit($resp_t_r)
        && Xormon::isdigit($resp_t_w)
        && Xormon::isdigit($read_io)
        && Xormon::isdigit($write_io)
        && Xormon::isdigit($io_rate)
        && $io_rate != 0 )
      {
        $resp_t =
          ( ( $resp_t_r * $read_io ) + ( $resp_t_w * $write_io ) ) / $io_rate;
      }

      $snmp->{data}{$snmp_section}{$node}{"disk_resp_t"} = $disk_resp_t;
      $snmp->{data}{$snmp_section}{$node}{"port_resp_t"} = $port_resp_t;
      $snmp->{data}{$snmp_section}{$node}{"resp_t_r"}    = $resp_t_r;
      $snmp->{data}{$snmp_section}{$node}{"resp_t_w"}    = $resp_t_w;
      $snmp->{data}{$snmp_section}{$node}{"resp_t"}      = $resp_t;

      # Create unique identifier for LUN
      my $uid =
        create_uid( $st_serial, $section, $node, $snmp->{data}, $snmp_section );

      # For each supported metric in node crate perf data
      foreach my $metric (@metrics_to_store) {
        $performance_data{$st_type}{$section}{$uid}{$timestamp}{$metric} =
          $snmp->{data}{$snmp_section}{$node}{$metric};
      }

# # Store processed data in the performance data structure
# foreach my $metric (keys %{ $snmp->{data}{$snmp_section}{$node} }) {
#     $performance_data{$st_type}{$section}{$uid}{$timestamp}{$metric} = $snmp->{data}{$snmp_section}{$node}{$metric};
# }
    }
  }
}

# Method: remove_whitespace
#
# This method removes leading and trailing whitespace from the input string.
#
# Parameters:
# - $value: The string from which to remove whitespace.
#
# Returns:
# - The trimmed string.

sub remove_whitespace {
  my $value = shift;
  $value =~ s/^\s+|\s+$//g;
  return $value;
}

# Method: get_value_from_counters
#
# This method calculates the difference between two counters and converts
# the result based on the metric and section provided. The conversion factor
# is adjusted depending on the type of metric and section.
#
# Parameters:
# - $val1: The first counter value.
# - $val2: The second counter value.
# - $metric: The metric type.
# - $section: The section type.
#
# Returns:
# - The calculated difference, possibly converted, or an empty string if the
#   inputs are not valid numbers.

sub get_value_from_counters {
  my $val1    = shift;
  my $val2    = shift;
  my $metric  = shift;
  my $section = shift;

  if ( !Xormon::isdigit($val1) || !Xormon::isdigit($val2) ) {
    return "";
  }

  # if (
  #     ($section eq "port" ||
  #      $section eq "drive" ||
  #      $section eq "volume") &&
  #     $metric =~ m/^read$|^write$|^data_rate$/
  # ) {
  #     $convert = 1024;
  # }

  my $diff = ( $val2 - $val1 ) / $sample_rate;

  #if (Xormon::isdigit($diff) && $diff !~ /^-/) {
  if ( Xormon::isdigit($diff) && $diff >= 0 ) {
    if ( $metric =~ m/io/ ) {
      my $rounded_diff = Xormon::round( $diff, 0 );
      return $rounded_diff;
    }

    my $rounded_diff = Xormon::round($diff);
    return $rounded_diff;
  }
  return "";
}

# Method: snmp_connection_test
#
# This method tests the SNMP connection using the provided configuration.
# It initializes the SNMP handler, tests the connection, and returns the
# connection status along with additional information or the reason for failure.
#
# Parameters:
# - $cfg: A reference to the configuration used to initiate the SNMP connection.
#
# Returns:
# - $conn_status: 1 if the connection is successful, 0 if it fails.
# - $reason: The reason for the connection failure if it fails.
# - SNMP walk command: The command used to perform an SNMP walk for additional diagnostics.
#

sub snmp_connection_test {
  my $cfg = shift;

  # Initialization SNMP Handler
  my $snmp = SNMPHandler->new( $cfg, undef, 1 );

  # Get result of test connection
  my ( $conn_status, $reason ) = $snmp->check_connection();

  #return status of connection
  return $conn_status, $reason, $snmp->build_snmpwalk_cmd();
}

# Method: create_uid
# This method creates a unique identifier (UID) for a given serial, section, and node.
# It also logs debug information about the UID creation process.
#
# Parameters:
# - $st_serial: The serial number of the device.
# - $section: The section of the device or data being processed.
# - $node: The node identifier within the SNMP data.
# - $snmp_data: Reference to the SNMP data structure.
# - $snmp_section: The SNMP section key within the SNMP data.
#
# Returns:
# - $uid: The generated unique identifier (UID).
sub create_uid {
  my ( $st_serial, $section, $node, $snmp_data, $snmp_section ) = @_;

  my $uid = Xormon::create_hash_key(
      ( defined $st_serial ? $st_serial : '' )
    . ( defined $section   ? $section   : '' )
      . (
      defined $snmp_data->{$snmp_section}{$node}{id}
      ? $snmp_data->{$snmp_section}{$node}{id}
      : ''
      )
  );

  message_log(
    3,
    "Generated UID: $uid for val "
      . ( defined $st_serial ? $st_serial : '' ) . "."
      . ( defined $section   ? $section   : '' ) . "."
      . (
      defined $snmp_data->{$snmp_section}{$node}{id}
      ? $snmp_data->{$snmp_section}{$node}{id}
      : ''
      )
  );

  return $uid;
}

# Method: message_log
#
# This method logs a message at a specified log level if the actual log level
# is lower than the provided message level. It logs the message and, if a reference
# to a data structure is provided, it formats and includes the data in the log.
#
# Parameters:
# - $message_level: The required log level for the message to be logged.
# - $message: The message to log.
# - $data_ref: (Optional) A reference to a data structure to be formatted and logged.
#
# Returns:
# - None.

sub message_log {
  my ( $message_level, $message, $data_ref ) = @_;

  if ( $act_log_level >= $message_level ) {
    my $log_message = $message;

    if ( defined $data_ref ) {
      $log_message .= " " . format_data($data_ref);
    }
    if ( defined $logger ) {
      $logger->info( $log_message . "\n" );
    }
    else {
      print $log_message . "\n";
    }
  }
}

# This is like Dumper
sub format_data {
  my ($data_ref) = @_;

  # Handle undefined input
  return 'undef' unless defined $data_ref;

  if ( ref($data_ref) eq 'HASH' ) {
    my @formatted;
    foreach my $key ( keys %$data_ref ) {
      my $value = defined $data_ref->{$key} ? $data_ref->{$key} : 'undef';
      push @formatted, "$key => " . format_data($value);
    }
    return "{ " . join( ", ", @formatted ) . " }";
  }

  elsif ( ref($data_ref) eq 'ARRAY' ) {
    return "[ " . join( ", ", map { format_data($_) } @$data_ref ) . " ]";
  }

  elsif ( ref($data_ref) ) {

    # Handle other reference types
    return "Unsupported reference type: " . ref($data_ref);
  }

  else {
    return $data_ref;
  }
}

1;
