package Cmd;
use strict;
use warnings 'FATAL' => 'all';
use Xormon qw(VrDump);

my $cmd;
my $debug;
my $log;
my $timeout = 300;

###
# new($cmd, $debug = 0)
# class constructor
# @param
#   @var $cmd - string - cesta k souboru příkazu
#   @var $debug - int - 1 - prints debugging information (different numbers for different levels of information output in the future)
# @return object
sub new {
  my $class          = shift;
  $cmd               = shift;
  $debug             = shift;
  $timeout           = shift;
  my $self           = {};
  $self->{'timeout'} = $timeout ? $timeout : 300;
  $self              = bless $self, $class;
  return $self;
}

###
#addCoreParam($paramName, $paramValue, $ifNotEmpty = 0);
# adds a parameter to the stack - it is a normal parameter when resetParam is executed it will be cleared
# @param
#  @var $paramName - string - parameter name
#  @var $paramValue - string - parameter value
#  @var $ifNotEmpty - bool - if 1 then the parameter is added only if its value is non-empty
# @return void
sub addCoreParam {
  my $self       = shift;
  my $paramName  = shift;
  my $paramValue = shift;
  my $ifNotEmpty = shift;
  if( !$ifNotEmpty || $paramValue ) {
    push(@{$self->{'coreParams'}}, {'paramName' => $paramName, 'paramValue' => $paramValue});
  }
}

###
#addParam($paramName, $paramValue, $ifNotEmpty = 0);
# adds a parameter to the stack - it is a normal parameter when resetParam is executed it will be cleared
# @param
#  @var $paramName - string - parameter name
#  @var $paramValue - string - parameter value
#  @var $ifNotEmpty - bool - if 1 then the parameter is added only if its value is non-empty
# @return void
sub addParam {
  my $self       = shift;
  my $paramName  = shift;
  my $paramValue = shift;
  my $ifNotEmpty = shift;
  if( !$ifNotEmpty || $paramValue ) {
    push(@{$self->{'params'}}, {'paramName' => $paramName, 'paramValue' => $paramValue});
  }
}

###
# getCmd()
# builds and returns a complete command
# @return string - cmd + coreParams + params
sub getCmd {
  my $self = shift;
  my $parameterString;
  for my $param_ref ( @{$self->{'coreParams'}}, @{$self->{'params'}} ) {
    my %param        = %{$param_ref};
    $parameterString .= ' ' . $param{'paramName'};
    if( $param{'paramValue'} ) {
      $parameterString .= ' ' . $param{'paramValue'};
    }
  }
  return $cmd . $parameterString;
}

###
# resetParams()
# Empty parameter stack. Core parameters will not be deleted
# @return void
sub resetParams {
  my $self = shift;
  undef( $self->{'params'} );
}

sub setTimeout {
  my $self           = shift;
  my $timeout        = shift;
  $self->{'timeout'} = $timeout;
}

sub getTimeout {
  my $self = shift;
  return $self->{'timeout'};
}

sub set_password_remove_regex {
  my $self  = shift;
  my $regex = shift;
  my $subst = shift;

  $self->{'regex'} = $regex;
  $self->{'subst'} = $subst;
}

sub password_remove {
  my $self         = shift;
  my $input_string = shift;
  my $result       = $input_string;
  if( defined($self->{'regex'}) and defined($self->{'subst'}) ) {
    $result = $input_string =~ s/$self->{'regex'}/$self->{'subst'}/eerg;
  }
  return $result;
}

sub set_logger {
  my $self          = shift;
  my $logger        = shift;
  $self->{'logger'} = $logger;
}

sub get_logger {
  my $self = shift;
  return $self->{'logger'};
}


###
# run($resetAfterRun = 0)
# executes the command, and returns the result
#  @param
#    @var $resetAfterRun - bool - if it is 1 then the parameter array is reset after the end of the run
#  @return string - standard command output
sub run {
  my $self          = shift;
  my $resetAfterRun = shift;
  my $command       = $self->getCmd();
  my $out           = '';
  my $exit_code     = 0;
  my $error;
  my $command_safe = $self->password_remove($command);
  if( $debug ) {
    $log          = $self->get_logger();
    $log->info("running: # " . $command_safe);
  }

  eval {
    my $act_time       = localtime();
    local $SIG{'ALRM'} = sub { die "$act_time: I'm terminating the '$command_safe' command, it's been running longer than ".$self->getTimeout()."s."; };
    alarm($self->getTimeout());
    $out       = `$command`;
    if( $debug ) {
      $log->info("timemout: ".$self->getTimeout());
      $log->info(&VrDump($out, $command_safe, 0, 1, 1));
    }
    $exit_code = $?; # Gets a return code
    alarm(0);
  };

  if( $@ ) {
    $out   = '';
    $error = $@;
  } elsif( $out =~ /Error Code\:/ ) {
    $error = $out;
    $out = '';
    if( $error =~ /Unable to validate the specified certificate./ ) {
      $error .= "uemcli command needs to accept the certificate! It must work without that...\n";
      $error .= "Follow this:\n https://xormon.com/EMC-PowerMax-VMAX-VNX-Unity-VPLEX-DataDomain-Isilon-XtremIO-CLARiiON-Celerra-monitoring.php#UNITY\nSetting security level under root user\n";
    }
  }

  if( $resetAfterRun ) {
    $self->resetParams();
  }

  return ( $out, $error );
}

1;

=pod

=head2 run($resetAfterRun = 0)

executes the command, and returns the result

=head3 param

=over 4

=item

C<resetAfterRun - bollean> - if it is 1 then the parameter array is reset after the end of the run

=back

=head3 return

C<string>  - standard command output

=cut
