##!/usr/bin/perl
#
# svcconfig.pl
#
# $Revision: 2.3.10 $  2023-06-01
my ($revision) = ( '$Revision: 2.3.10 $' =~ /([\d\.-]+)/ );

# Changes:
# 1.3.1 Add FlashSystem 840/900 support
# 1.3.6 Correct FC / Global Mirror with Change Volume
# 1.4.x Reduce ssh/scp login sessions
# 1.5.x Add physical capacity informations
# 2.1.x Implement Net::OpenSSH
# 2.2.x Implement tier_scm
# 2.3.x Implement Capacity Monitoring

# XormonNG uses different libs
my $xormonNG = ( $ENV{XormonNG} ) ? $ENV{XormonNG} : 0;
if ($xormonNG) {

  #require FindBin;
  use FindBin;
  use lib "$FindBin::Bin/";    # use lib in this folder
  use lib "../../../perl-modules";
  require Xormon;
  require XML::Simple;
}
else {
  #use Xorux_lib;
  unless ( exists $ENV{BINDIR} ) { error( "BINDIR is not defined! $!" . __FILE__ . ":" . __LINE__ ) && exit 1; }
  require Xorux_lib;
  require "$ENV{BINDIR}/xml.pl";    # it replaces above line and fixes an issue on Perl 5.10.1 on AIX 7100-00-06-1216
}

# Modules
use strict;
use Data::Dumper;
use Storable qw(retrieve nstore);
use Date::Parse;
use SVC 2.2;

my %code_level;

# Storage Hardware Types
my %storage_info = ( "is_svc" => 0, "is_storwize" => 0, "is_fs840" => 0, "is_fs9100" => 0, "is_v9000" => 0, "is_inspur" => 0 );

# SVC Hardware Types
my %svc_g1 = ( "CG8" => 1, "CF8" => 1, "8A4" => 1, "8G4" => 1, "8F4" => 1, "8F2" => 1, "4F2" => 1 );
my %svc_g2 = ( "DH8" => 1, "SV1" => 1 );

# Time Zone
my %tzoffset = (
  "A"     => "-0100", "ACDT"    => "+1030", "ACST"    => "+0930", "ACT"    => "-0500", "ACWDT"   => "+0945", "ACWST"   => "+0845", "ADDT"   => "-0200", "ADT"     => "-0300", "AEDT"   => "+1100",
  "AEST"  => "+1000", "AFT"     => "+0430", "AHDT"    => "-0900", "AHST"   => "-1000", "AKDT"    => "-0800", "AKST"    => "-0900", "AKTST"  => "+0600", "AKTT"    => "+0500", "ALMST"  => "+0700",
  "ALMT"  => "+0600", "AMST"    => "-0300", "AMT"     => "-0400", "ANAST"  => "+1200", "ANAT"    => "+1200", "ANT"     => "-0430", "APT"    => "-0300", "AQTST"   => "+0500", "AQTT"   => "+0500",
  "ARST"  => "-0200", "ART"     => "-0300", "ASHST"   => "+0500", "ASHT"   => "+0400", "AST"     => "-0400", "AT-0200" => "",      "AWDT"   => "+0900", "AWST"    => "+0800", "AWT"    => "-0300",
  "AZOMT" => "+0000", "AZOST"   => "+0000", "AZOT"    => "-0100", "AZST"   => "+0500", "AZT"     => "+0400", "B"       => "-0200", "BAKST"  => "+0400", "BAKT"    => "+0400", "BDST"   => "+0700",
  "BDT"   => "+0600", "BEAT"    => "+0230", "BEAUT"   => "+0245", "BNT"    => "+0800", "BORT"    => "+0800", "BORTST"  => "+0820", "BOT"    => "-0400", "BRST"    => "-0200", "BRT"    => "-0300",
  "BST"   => "+0100", "BT+0300" => "",      "BTT"     => "+0600", "BURT"   => "+0630", "C"       => "-0300", "CADT"    => "+1030", "CANT"   => "-0100", "CAPT"    => "-0900", "CAST"   => "+1100", "CAT"   => "+0200",
  "CAWT"  => "-0900", "CCT"     => "+0630", "CDDT"    => "-0400", "CDT"    => "-0500", "CEMT"    => "+0300", "CEST"    => "+0200", "CET"    => "+0100", "CGST"    => "-0100", "CGT"    => "-0200", "CHADT" => "+1345",
  "CHAST" => "+1245", "CHDT"    => "-0530", "CHOST"   => "+1000", "CHOT"   => "+0800", "CHST"    => "+1000", "CHUT"    => "+1000", "CKHST"  => "-0930", "CKT"     => "-1000", "CLDT"   => "-0300", "CLST"  => "-0300",
  "CLT"   => "-0300", "CMT"     => "+0155", "COST"    => "-0400", "COT"    => "-0500", "CPT"     => "-0500", "CST"     => "-0600", "CUT"    => "+0220", "CVST"    => "-0100", "CVT"    => "-0100", "CWT"   => "-0500",
  "CXT"   => "+0700", "D"       => "-0400", "DACT"    => "+0600", "DAVT"   => "+0700", "DDUT"    => "+1000", "DUSST"   => "+0600", "DUST"   => "+0600", "E"       => "-0500", "EADT"   => "+1100", "EASST" => "-0500",
  "EAST"  => "-0500", "EAT"     => "+0300", "ECT"     => "-0500", "EDDT"   => "-0300", "EDT"     => "-0400", "EEST"    => "+0300", "EET"    => "+0200", "EETDST"  => "+0300", "EETEDT" => "+0300",
  "EGST"  => "+0000", "EGT"     => "-0100", "EHDT"    => "-0430", "EPT"    => "-0400", "EST"     => "-0500", "EWT"     => "-0400", "F"      => "-0600", "FET"     => "+0300", "FJST"   => "+1300", "FJT" => "+1200",
  "FKST"  => "-0300", "FKT"     => "-0400", "FNST"    => "-0100", "FNT"    => "-0200", "FORT"    => "+0500", "FRUST"   => "+0600", "FRUT"   => "+0600", "FST"     => "+0200", "FWT"    => "+0100", "G"   => "-0700",
  "GALT"  => "-0600", "GAMT"    => "-0900", "GB+0100" => "",      "GBGT"   => "-0345", "GEST"    => "+0400", "GET"     => "+0400", "GFT"    => "-0300", "GHST"    => "+0020", "GILT"   => "+1200", "GMT" => "+0000",
  "GMT+1" => "+0100", "GMT+10"  => "+1000", "GMT+11"  => "+1100", "GMT+12" => "+1200", "GMT+2"   => "+0200", "GMT+3"   => "+0300", "GMT+4"  => "+0400", "GMT+5"   => "+0500", "GMT+6"  => "+0600",
  "GMT+7" => "+0700", "GMT+8"   => "+0800", "GMT+9"   => "+0900", "GMT-1"  => "-0100", "GMT-10"  => "-1000", "GMT-11"  => "-1100", "GMT-12" => "-1200", "GMT-13"  => "-1300", "GMT-14" => "-1400",
  "GMT-2" => "-0200", "GMT-3"   => "-0300", "GMT-4"   => "-0400", "GMT-5"  => "-0500", "GMT-6"   => "-0600", "GMT-7"   => "-0700", "GMT-8"  => "-0800", "GMT-9"   => "-0900", "GST"    => "-0200",
  "GYT"   => "-0400", "H"       => "-0800", "HADT"    => "-0900", "HAST"   => "-1000", "HDT"     => "-0930", "HKST"    => "+0900", "HKT"    => "+0800", "HOVST"   => "+0800", "HOVT"   => "+0700", "HST"   => "-1000",
  "I"     => "-0900", "ICT"     => "+0700", "IDDT"    => "+0400", "IDLE"   => "+1200", "IDLW"    => "-1200", "IDT"     => "+0300", "IHST"   => "+0600", "IOT"     => "+0600", "IRDT"   => "+0430", "IRKST" => "+0900",
  "IRKT"  => "+0800", "IRST"    => "+0330", "ISST"    => "+0000", "IST"    => "+0200", "IT+0330" => "",      "JAVT"    => "+0720", "JCST"   => "+0900", "JDT"     => "+1000", "JST"    => "+0900", "JWST"  => "+0800",
  "K"     => "-1000", "KART"    => "+0500", "KDT"     => "+1000", "KGST"   => "+0600", "KGT"     => "+0600", "KIZST"   => "+0600", "KIZT"   => "+0500", "KOST"    => "+1100", "KRAST"  => "+0800", "KRAT"  => "+0700",
  "KST"   => "+0900", "KUYST"   => "+0500", "KUYT"    => "+0300", "KWAT"   => "-1200", "L"       => "-1100", "LHDT"    => "+1100", "LHST"   => "+1030", "LINT"    => "+1400", "LKT"    => "+0600", "M"     => "-1200",
  "MADMT" => "+0100", "MADST"   => "+0000", "MADT"    => "-0100", "MAGST"  => "+1200", "MAGT"    => "+1000", "MALST"   => "+0720", "MALT"   => "+0730", "MART"    => "-0930", "MAWT"   => "+0500", "MDDT"  => "-0500",
  "MDT"   => "-0600", "MEST"    => "+0200", "MESZ"    => "+0200", "MET"    => "+0100", "METDST"  => "+0200", "MEWT"    => "+0100", "MEZ"    => "+0100", "MHT"     => "+1200", "MIST"   => "+1100", "MMT"   => "+0630",
  "MOST"  => "+0900", "MOT"     => "+0800", "MPT"     => "+1000", "MSD"    => "+0400", "MSK"     => "+0300", "MSM"     => "+0500", "MST"    => "-0700", "MUST"    => "+0500", "MUT"    => "+0400", "MVT"   => "+0500",
  "MWT"   => "-0600", "MYT"     => "+0800", "N"       => "+0100", "NCST"   => "+1200", "NCT"     => "+1100", "NDDT"    => "-0130", "NDT"    => "-0230", "NEGT"    => "-0330", "NEST"   => "+0120", "NET"   => "+0020",
  "NFT"   => "+1130", "NMT"     => "+1112", "NOVST"   => "+0700", "NOVT"   => "+0600", "NPT"     => "+0545", "NRT"     => "+1200", "NST"    => "-0330", "NT-1100" => "",      "NUT"    => "-1100", "NWT"   => "-1000",
  "NZDT"  => "+1300", "NZMT"    => "+1130", "NZST"    => "+1200", "NZT"    => "+1200", "O"       => "+0200", "OMSST"   => "+0700", "OMST"   => "+0600", "ORAST"   => "+0500", "ORAT"   => "+0500", "P"     => "+0300",
  "PDDT"  => "-0600", "PDT"     => "-0700", "PEST"    => "-0400", "PET"    => "-0500", "PETST"   => "+1200", "PETT"    => "+1200", "PGT"    => "+1000", "PHOT"    => "+1300", "PHST"   => "+0900", "PHT"   => "+0800",
  "PKST"  => "+0600", "PKT"     => "+0500", "PMDT"    => "-0200", "PMST"   => "-0300", "PMT"     => "+1000", "PNT"     => "-0830", "PONT"   => "+1100", "PPMT"    => "-0449", "PPT"    => "-0700", "PST"   => "-0800",
  "PWT"   => "+0900", "PYST"    => "-0300", "PYT"     => "-0400", "Q"      => "+0400", "QMT"     => "-0514", "QYZST"   => "+0700", "QYZT"   => "+0600", "R"       => "+0500", "RET"    => "+0400", "ROK"   => "+0900",
  "ROTT"  => "-0300", "S"       => "+0600", "SAKST"   => "+1100", "SAKT"   => "+1000", "SAMST"   => "+0400", "SAMT"    => "+0400", "SAST"   => "+0200", "SAT"     => "-0400", "SBT"    => "+1100", "SCT"   => "+0400",
  "SDMT"  => "-0440", "SDT"     => "-1000", "SGT"     => "+0800", "SHEST"  => "+0600", "SHET"    => "+0500", "SMT"     => "+0216", "SRET"   => "+1100", "SRT"     => "-0300", "SST"    => "-1100", "STAT"  => "+0400",
  "SVEST" => "+0500", "SVET"    => "+0400", "SWAT"    => "+0130", "SWT"    => "+0100", "SYOT"    => "+0300", "T"       => "+0700", "TAHT"   => "-1000", "TASST"   => "+0600", "TAST"   => "+0600", "TBIST" => "+0400",
  "TBIT"  => "+0400", "TFT"     => "+0500", "TJT"     => "+0500", "TKT"    => "+1300", "TLT"     => "+0900", "TMT"     => "+0500", "TOST"   => "+1400", "TOT"     => "+1300", "TRST"   => "+0400", "TRT"   => "+0300",
  "TSAT"  => "+0300", "TVT"     => "+1200", "U"       => "+0800", "ULAST"  => "+0900", "ULAT"    => "+0800", "URAST"   => "+0500", "URAT"   => "+0400", "UT+0000" => "",      "UTC"    => "+0000", "UYHST" => "-0230",
  "UYST"  => "-0200", "UYT"     => "-0300", "UZST"    => "+0600", "UZT"    => "+0500", "V"       => "+0900", "VET"     => "-0430", "VLAST"  => "+1100", "VLAT"    => "+1000", "VOLST"  => "+0400", "VOLT"  => "+0400",
  "VOST"  => "+0600", "VUST"    => "+1200", "VUT"     => "+1100", "W"      => "+1000", "WAKT"    => "+1200", "WARST"   => "-0300", "WART"   => "-0400", "WAST"    => "+0200", "WAT"    => "+0100", "WEMT"  => "+0200",
  "WEST"  => "+0100", "WET"     => "+0000", "WFT"     => "+1200", "WGST"   => "-0200", "WGT"     => "-0300", "WIB"     => "+0700", "WIT"    => "+0900", "WITA"    => "+0800", "WMT"    => "+0124", "WSDT"  => "+1400",
  "WSST"  => "+1300", "X"       => "+1100", "XJT"     => "+0600", "Y"      => "+1200", "YAKST"   => "+1000", "YAKT"    => "+0900", "YDDT"   => "-0700", "YDT"     => "-0800", "YEKST"  => "+0600", "YEKT"  => "+0500",
  "YERST" => "+0400", "YERT"    => "+0400", "YPT"     => "-0800", "YST"    => "-0900", "YWT"     => "-0800", "Z"       => "+0000", "ZP4"    => "+0400", "ZP5"     => "+0500", "ZP6"    => "+0600",
);

# Oprions and their default values
my $storage;           # SVC/Storwize V7000 alias
my $svc;               # SVC/Storwize V7000 cluster host name (IP address)
my $user = "admin";    # SVC/Storwize V7000 user name
my $key  = "";         # SSH key filename
my $dir  = "..";       # Data directory

my $debug           = 0;        # Debug mode 0:off, 1:on
my $debug_full      = 0;        # Debug Full mode 0:off, 1:on
my $fullcfgint      = 86400;    # Full Config Interval (sec) ... 24H = 86400
my $capupd_interval = 3600;     # Capacity Config Update Interval (1 hour = 3600)
my $ssh_timeout     = 900;      # SSH Command timeout - 15 min
my $net_openssh     = 1;        # Use Net::OpenSSH Perl Modul

my $password        = "";
my $password_enable = "";

if ( defined $ENV{STORAGE_NAME} ) {
  $storage = $ENV{STORAGE_NAME};
}
else {
  &message("SVC/Storwize V7000 storage name alias is required.");
  exit(1);
}
if ( defined $ENV{SVC_IP} ) {
  $svc = $ENV{SVC_IP};
}
else {
  &message("SVC/Storwize V7000 host name or IP address is required.");
  exit(1);
}
if ( defined $ENV{SVC_USER} )            { $user            = $ENV{SVC_USER} }
if ( defined $ENV{SVC_KEY} )             { $key             = $ENV{SVC_KEY} }
if ( defined $ENV{SVC_DIR} )             { $dir             = $ENV{SVC_DIR} }
if ( defined $ENV{DEBUG} )               { $debug           = $ENV{DEBUG} }
if ( defined $ENV{SVC_PASSWORD_ENABLE} ) { $password_enable = $ENV{SVC_PASSWORD_ENABLE} }
if ( defined $ENV{SVC_PASSWORD} )        { $password        = $ENV{SVC_PASSWORD} }

if ( defined $ENV{SVC_DEBUG_FULL} ) {
  if    ( $ENV{SVC_DEBUG_FULL} =~ /^\s*(\d{1})\s*$/ ) { $debug      = $1; $debug_full = $1; }
  elsif ( $ENV{SVC_DEBUG_FULL} eq $storage )          { $debug      = 1; $debug_full = 1; }
  else                                                { $debug_full = 0; }
}
if ( defined $ENV{NET_OPENSSH} )    { $net_openssh     = $ENV{NET_OPENSSH} }
if ( defined $ENV{SVC_CAPUPD_INT} ) { $capupd_interval = $ENV{SVC_CAPUPD_INT} }

if ( defined $password && $password ne "" ) {
  unless ($xormonNG) {
    $password = Xorux_lib::unobscure_password($password);
  }
}

# Main process

# Display parameters
if ($debug_full) {
  print( "
svcconfig.pl starts with the following parameters ...
\tRevision:                             $revision
\tSVC/Storwize V7000 cluster name:      $storage
\tSVC cluster host name:                $svc
\tSVC admin user name:                  $user
\tSVC admin user SSH key filename:      $key
\tBase Directory for output files:      $dir
\tDebug:                                $debug
\n" );
}

#
if ($net_openssh) {

  # use require here to do not affect all storage whe it does not exist and "use" is on the top
  require Xorux::Net::OpenSSH;
  if ( $Xorux::Net::OpenSSH::VERSION < 0.78 ) {
    &error( "Current version of Xorux::Net::OpenSSH is " . $Xorux::Net::OpenSSH::VERSION . ". Use version 0.78 and higher." );
  }
  Xorux::Net::OpenSSH->import(qw(new disconnect capture scp_get pipe_out));
}
#
my $fullcfgflg = 0;
my $cfgdata;
my %cfg;
my $cfg_ref = \%cfg;
my %capdata;
my $capdata_ref = \%capdata;
my $net_ssh;
my $cmd_ssh;
my $cmd_scp;
my $cmd_ls   = "/bin/ls";
my $cmd_mv   = "/bin/mv";
my $cmd_rm   = "/bin/rm";
my $cmd_grep = "/bin/grep";

$cmd_ssh = $ENV{SSH};

my $html_chars = qr/[\x09\x0A\x0D\x20-\x{D7FF}\x{E000}-\x{FFFD}\x{10000}-\x{10FFFF}]/;
my @html_exclude_string;
if ( defined $ENV{SVC_HTML_EXCLUDE_STRING} ) {
  @html_exclude_string = split( / /, $ENV{SVC_HTML_EXCLUDE_STRING} );
}
else {
  @html_exclude_string = qw(&up; &dquot;);
}

# Date and Time
my $local_time  = localtime(time);
my $local_epoch = str2time($local_time);
my ( $sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst ) = localtime(time);
$year += 1900;
$mon++;
my $date = sprintf( "%4d%2.2d%2.2d",   $year, $mon, $mday );
my $time = sprintf( "%2.2d%2.2d%2.2d", $hour, $min, $sec );

# Get Latest Config Data
my $tmp_conf_file = &GetConfigFilename();

if ( defined $tmp_conf_file ) {
  if ($debug_full) { &message( "Retrive data from file: " . $dir . "/" . $tmp_conf_file ) }
  &RetriveData;
}
else {
  $fullcfgflg = 1;
  if ($debug_full) { &message("Config File is not present.") }
  $tmp_conf_file = $dir . "/" . $storage . "_svcconf_" . $date . "_" . $time . ".data";
}

# Open remote session to the storage (Net::OpenSSH) or setup SSH/SCP commands
if ($net_openssh) {
  if ($debug_full) { &message( "Open SSH session - host: " . $svc . ", user: " . $user . " key: " . $key ); }
  $net_ssh = SVC::openssh_new( $svc, "22", $user, $key, $ssh_timeout, $password, $password_enable );
  if ( not defined $net_ssh ) {
    $cmd_ssh = SVC::cmdssh_setup( $svc, "22", $user, $key, $cmd_ssh );
    $cmd_ssh =~ s/-q //g;
    $cmd_ssh =~ s/-o ConnectTimeout=\w+ //g;
    $cmd_ssh =~ s/-o PreferredAuthentications=publickey //g;
    $cmd_ssh =~ s/-o SendEnv=no //g;
    $cmd_ssh =~ s/-o StrictHostKeyChecking=no //g;
    error( "ERROR: Unable to connect to " . $svc . "\n", 1 );
    error( "       " . $cmd_ssh . " \"lssystem\"\n",     1 );
    exit 2;
  }
}
else {
  if ($debug_full) { &message( "Setup SSH command - host: " . $svc . ", user: " . $user . " key: " . $key ); }
  $cmd_ssh = SVC::cmdssh_setup( $svc, "22", $user, $key, $cmd_ssh );
  $cmd_scp = SVC::cmdscp_setup( $svc, "22", $user, $key, $cmd_scp );
}

# Get SVC/Storwize Code Level
my $sys_info = &GetSystemInfo;

if ( !defined $sys_info || !defined $code_level{'version'} ) {
  if ($net_openssh) {
    $cmd_ssh = SVC::cmdssh_setup( $svc, "22", $user, $key, $cmd_ssh );
  }
  $cmd_ssh =~ s/-q //g;
  $cmd_ssh =~ s/-o ConnectTimeout=\w+ //g;
  $cmd_ssh =~ s/-o PreferredAuthentications=publickey //g;
  $cmd_ssh =~ s/-o SendEnv=no //g;
  $cmd_ssh =~ s/-o StrictHostKeyChecking=no //g;
  error( "ERROR: storage access is not working without providing password, this must work and role must be Administrator: \n", 1 );
  error( "       " . $cmd_ssh . " \"lscurrentuser\"\n",                                                                        1 );
  exit 1;
}

# Define configurations file name
my $tmp_conf_full_file = $dir . "/" . $storage . "_svcconf" . ".datafull";
my $out_conf_file      = $dir . "/" . $storage . "_svcconf_" . $date . "_" . $time . ".out";
my $out_cap_file       = $dir . "/" . $storage . "_svccapacity_" . $date . "_" . $time . ".out";
my $tmp_cap_file       = $dir . "/" . $storage . "_svccapacity" . ".data";

if ($debug_full) {
  my $msg;
  my @a;
  foreach my $k ( sort( keys(%storage_info) ) ) {
    push( @a, $k . "=" . $storage_info{$k} );
  }
  message( "HW Type Info: " . join( ",", @a ) . "\n" );
}
if ($debug_full) {
  &message( "SVC/Storwize code level - Version: " . $code_level{'version'} . " Release: " . $code_level{'release'} . " Major: " . $code_level{'major'} . " Minor: " . $code_level{'minor'} );
}
my $codevrmn = $code_level{'version'} * 1000000 + $code_level{'release'} * 10000 + $code_level{'major'} * 100 + $code_level{'minor'};

# Update date-time in config data
if ( not defined $cfg_ref->{date}->{cfg_date} ) {
  $cfg_ref->{date} = { 'cfg_date' => undef, 'cfg_epoch' => undef, 'fullcfg_epoch' => undef };
}
$cfg_ref->{date}->{'cfg_date'}  = sprintf( "%2.2d.%2.2d.%4d", $mday, $mon, $year ) . " " . sprintf( "%2.2d:%2.2d:%2.2d", $hour, $min, $sec );
$cfg_ref->{date}->{'cfg_epoch'} = $local_epoch;
if ($fullcfgflg) {

  # Get SVC configuration file (XML file) from SVC cluster
  my $file_name = &GetXMLCronFile();
  &message( "(" . $revision . ") Config XML file: " . $file_name );
  if ( not defined $file_name ) {
    if ( $codevrmn >= 6040000 && $codevrmn <= 7010001 ) {

      # Command svcconfig backup failed with error message
      # CMMVC6202E This command can only be run by the superuser.
      # Suggested Fixes
      # 7.1.0.2  IC90824  Configuration backup could only be run by superuser
      &InsertDataCli();
    }
    else {
      $file_name = &GetXMLConfigFile();
      if ($debug_full) { &message( "Config Backup XML file: " . $file_name ); }
    }
  }
  if ( not defined $file_name ) {
    &error( "Upload XML Config File failed", 1 ) && exit 1;
  }
  my $xml_epoch = &GetXMLConfigTime($file_name);
  if ( not defined $xml_epoch ) {
    error( "ERROR: Cannot get time from XML file " . $file_name . ". Exiting.\n", 1 );
    exit 1;
  }
  my $pre_epoch = 0;
  if ( not defined $cfg_ref->{date}->{'fullcfg_epoch'} ) {

    # First time run
    # Insert new data from XML file
    $cfg_ref->{date}->{'fullcfg_epoch'} = $xml_epoch + 3600;
    if ($debug_full) {
      &message("No previous XML data in config data");
      &message( "Insert new data from XML file. Save epoch: " . $cfg_ref->{date}->{'fullcfg_epoch'} );
    }
    &CheckConfigFile($file_name);
    $cfgdata = &ReadConfigFile($file_name);
    &InsertDataXml();
    if ($debug_full) { &message("Store data."); }
    &StoreDataXml();
  }
  else {
    $pre_epoch = $cfg_ref->{date}->{'fullcfg_epoch'};
    if ($debug_full) { &message( "XML Config File: " . $file_name . " Time: " . localtime($xml_epoch) . " Epoch: " . $xml_epoch . " Last epoch: " . $pre_epoch ); }
    if ( $pre_epoch < $xml_epoch ) {

      # new data in XML file
      # Remove old data
      if ($debug_full) { &message("New data in XML file. Remove config data."); }
      &RemoveAllCfgData;

      # Insert new data from XML file
      $cfg_ref->{date}->{'fullcfg_epoch'} = $xml_epoch + 3600;
      if ($debug_full) { &message( "Insert new data from XML file. Save epoch: " . $cfg_ref->{date}->{'fullcfg_epoch'} ); }
      &CheckConfigFile($file_name);
      $cfgdata = &ReadConfigFile($file_name);
      &InsertDataXml();
      if ($debug_full) { &message("Store data."); }
      &StoreDataXml();
    }
    else {
      # old data in XML file - load new config file after 4 hours
      $cfg_ref->{date}->{'fullcfg_epoch'} = $local_epoch - $fullcfgint + 14400;
      if ($debug_full) { &message("Old data in XML file. Insert data from CLI."); }
      &RemoveOldEventsCfgData();
      &InsertDataCli();
    }
  }
}
else {
  if ($debug_full) { &message("Insert data from CLI."); }
  &RemoveOldEventsCfgData;
  &InsertDataCli();
}

# Config Data
&CalculateData();
&PrintData();

# Capacity Data
&CapCalculateData();
&CapStoreData();
&CapPrintData();

# Store data to file
# Define new filename with current time
$tmp_conf_file = $dir . "/" . $storage . "_svcconf_" . $date . "_" . $time . ".data";
if ($debug_full) { &message( "Store data to file: " . $tmp_conf_file ) }
&StoreData();

### health status
my $hs_html       = "";
my $error_status  = 0;
my $section_check = 0;
my $event_check   = 0;
my $summary_dir   = "";
my @error_message = ();
my @event_header = ("time", "event_id", "object_type", "object_name", "description", "status", "fixed", "sequence_number", "object_id", "copy_id", "error_code");
my @subsystem_header = ("section","id", "name", "status", "description");
my %health_status;
my %health_status_component;

unless ($xormonNG) {
  my $webdir   = $ENV{WEBDIR};
  my $inputdir = $ENV{INPUTDIR};
  $hs_html     = "$webdir/$storage/health_status.html";
  $summary_dir = "$inputdir/tmp/health_status_summary";

  if ( !-d "$webdir/$storage" ) {
    mkdir "$webdir/$storage" || error( "Cannot read $webdir/$storage: $!" . __FILE__ . ":" . __LINE__ );
  }
}

%health_status = get_event_log($out_conf_file);

#print Dumper \%health_status;
if ( $event_check > 0 || ( $section_check > 0 && ( ( defined $health_status_component{Pool} && defined $health_status_component{Port} ) || defined $health_status_component{Volume} ) ) ) {
  my $table = get_health_status_html();
  if ($xormonNG) {
    hs_xormon();
  }
  else {
    write_health_status_html( $hs_html, $table );
    summary_health_status();
  }
}

#else {
#  my $time = localtime();
#  $time = $time . "<br>\n";
#  my $report = "$time Storage $storage health status = Optimal\n";
#  if ($xormonNG) {
#    hs_xormon();
#  }
#  else {
#    write_health_status_html( $hs_html, $report );
#    summary_health_status();
#  }
#}

###
if ($net_openssh) {
  if ($debug_full) { &message("Close SSH session"); }
  my $out = SVC::openssh_disconnect( $net_ssh, $ssh_timeout );
  if ( not defined $out ) {
    &error("Unable to disconnect from remote machine");
  }
}

exit();

sub message {
  my $msg = shift;
  my $tm  = localtime();
  print( $tm. " - INFO    - svcconfig.pl: " . $msg . "\n" );
}

sub warning {
  my ( $msg, $rc ) = @_;
  my $tm = localtime();
  print STDERR ( $tm . " - WARNING - svcconfig.pl: " . $msg . "\n" );
}

sub error {
  my ( $msg, $rc ) = @_;
  my $tm = localtime();
  if ( defined $rc ) {
    print STDERR ( $tm . " - ERROR   - svcconfig.pl: " . $msg . " - Return code: " . $rc . "\n" );
  }
  else {
    print STDERR ( $tm . " - ERROR   - svcconfig.pl: " . $msg . "\n" );
  }
}

sub GetXMLConfigFile {

  # Create XML config backup file
  my $cmd = "svcconfig backup 2>/dev/null >/dev/null";
  if ($debug_full) { message( "Remote Command " . $cmd ); }
  my $out = ();
  if ($net_openssh) {
    $out = SVC::openssh_runcmd( $net_ssh, $cmd, $ssh_timeout );
  }
  else {
    $out = SVC::cmdssh_runcmd( $cmd_ssh, $cmd, $ssh_timeout );
  }
  if ( not defined $out ) {
    &error( "Remote Command " . $cmd . " failed", 1 ) && return undef;
  }

  # Upload XML config backup file
  my $upload_files = ["/tmp/svc.config.backup.xml"];
  if ($debug_full) { message( "Download files: " . join( " ", @{$upload_files} ) . ", in to directory " . $dir ); }
  $out = ();
  if ($net_openssh) {
    $out = SVC::openssh_getfile( $net_ssh, $upload_files, $dir, $ssh_timeout );
  }
  else {
    $out = SVC::cmdssh_getfile( $cmd_scp, $upload_files, $dir, $ssh_timeout );
  }
  if ( not defined $out ) {
    &error( "Download files: " . join( " ", @{$upload_files} ) . ", failed", 1 ) && return undef;
  }
  return ( $dir . "/svc.config.backup.xml" );
}

sub ID_front_panel {
  my ( $node, $dumps_name ) = @_;

  # Create $front_panel_id
  my $cmd = "svcinfo lsnodevpd -delim : " . $node;
  if ($debug_full) { message( "Remote Command: " . $cmd ); }
  my $fh;
  if ($net_openssh) {
    $fh = SVC::openssh_pipeout( $net_ssh, [$cmd], $ssh_timeout );
  }
  else {
    $fh = SVC::cmdssh_pipeout( $cmd_ssh, [$cmd], $ssh_timeout );
  }
  if ( not defined $fh ) {
    &error( "Remote Command: " . $cmd . " failed.", 1 );
    return 1;
  }

  while ( my $line_LSNODEVPD = <$fh> ) {

    #if($debug_full) {print("The svcinfo lsnodevpd command output: $line_LSNODEVPD");}
    my @field_LSNODEVPD = split( ':', $line_LSNODEVPD );
    if ( $field_LSNODEVPD[0] =~ $dumps_name ) {

      #if($debug_full) {message("Check: $field_LSNODEVPD[1]");}
      my $front_panel_id = $field_LSNODEVPD[1];
      chomp($front_panel_id);
      if ($debug_full) { message("Front panel ID/Dump name: $front_panel_id"); }
      return $front_panel_id;
    }
  }
  close($fh);
}

sub GetXMLCronFile {

  #$debug_full = 1;
  my $iostatsdumps_name;
  my $filename;

  # find XML Cron config file on config node
  my $cmd = "svcinfo lsnode -nohdr -delim :";
  if ($debug_full) { message( "Remote Command: " . $cmd ); }
  my $fh;
  if ($net_openssh) {
    $fh = SVC::openssh_pipeout( $net_ssh, [$cmd], $ssh_timeout );
  }
  else {
    $fh = SVC::cmdssh_pipeout( $cmd_ssh, [$cmd], $ssh_timeout );
  }
  if ( not defined $fh ) {
    &error( "Remote Command: " . $cmd . " failed.", 1 );
    return 1;
  }

  while ( my $line_LSNODE = <$fh> ) {

    # Get XML Cron Config Backup Filename
    #if($debug_full) {message("The svcinfo lsnode command output: $line_LSNODE");}
    my @field_LSNODE = split( ':', $line_LSNODE );

    # Check none configuration nodes
    if ( length( $field_LSNODE[2] ) == 0 && length( $field_LSNODE[15] ) != 0 ) {

      # Storwize V7000 = UPS_serial_number is blank AND enclosure_serial_number is not blank
      $iostatsdumps_name = "dump_name";    # Storwize V7000 iostats dumps file name
      if ($debug_full) { message("H/W type, iostats dumps file name (Storwize V7000): $iostatsdumps_name."); }
    }
    elsif ( length( $field_LSNODE[8] ) == 0 && length( $field_LSNODE[15] ) != 0 ) {

      # Storwize V7000 = UPS_unique_id is blank AND enclosure_serial_number is not blank
      $iostatsdumps_name = "dump_name";    # Storwize V7000 iostats dumps file name
      if ($debug_full) { message("H/W type, iostats dumps file name (Storwize V7000): $iostatsdumps_name."); }
    }
    elsif ( $svc_g1{ $field_LSNODE[9] } ) {
      $iostatsdumps_name = "front_panel_id";    # SVC - Gen1 - with UPS - iostats dumps file name
      if ($debug_full) { message("H/W type, iostats dumps file name (SVC): $iostatsdumps_name."); }
    }
    elsif ( $svc_g2{ $field_LSNODE[9] } ) {
      $iostatsdumps_name = "front_panel_id";    # SVC - Gen2 - with battery - iostats dumps file name
      if ($debug_full) { message("H/W type, iostats dumps file name (SVC): $iostatsdumps_name."); }
    }
    else {
      error( "H/W type is not recognized.\n" . $line_LSNODE );
      return undef;
    }
    if ( $field_LSNODE[4] eq "online" and $field_LSNODE[7] eq "yes" ) {
      my $node           = $field_LSNODE[0];
      my $front_panel_id = &ID_front_panel( $node, $iostatsdumps_name );
      if ( $storage_info{is_inspur} ) {
        $filename = "mcs.config.cron.xml_" . $front_panel_id;
      }
      else {
        $filename = "svc.config.cron.xml_" . $front_panel_id;
      }
      if ($debug_full) { message( "node: " . $node . " ID: " . $front_panel_id . " XML filename name: " . $filename ); }
      last;
    }
  }
  close($fh);

  # Upload XML Cron Config Backup File
  my $upload_files = [ "/dumps/" . $filename ];
  if ($debug_full) { message( "Download files: " . join( " ", @{$upload_files} ) . ", in to directory " . $dir ); }
  my $out;
  if ($net_openssh) {
    $out = SVC::openssh_getfile( $net_ssh, $upload_files, $dir, $ssh_timeout );
  }
  else {
    $out = SVC::cmdssh_getfile( $cmd_scp, $upload_files, $dir, $ssh_timeout );
  }
  if ( not defined $out ) {
    &error( "Download files: " . join( " ", @{$upload_files} ) . ", failed", 1 ) && return undef;
  }
  return ( $dir . "/" . $filename );
}

sub GetXMLConfigTime {
  my $filename = shift;
  if ( not -r $filename ) { &message( "Config file " . $filename . " is not readable." ); }
  open( FILE, "<$filename" ) || &message( "Cannot open file " . $filename . "." ) && return undef;
  my $epoch;
  while ( my $line = <FILE> ) {
    chomp($line);

    # timestamp="2016/10/08 01:00:10" >
    if ( $line =~ /^\s*timestamp="(.*)"\s+>\s*$/ ) {
      my @t = split( /\s+/, $1 );
      if ( defined $t[2] and exists( $tzoffset{ $t[2] } ) ) {
        $t[2] = $tzoffset{ $t[2] };
      }
      $epoch = str2time( join( " ", @t ) );
      close(FILE);
      if ( not defined $epoch ) { &error( "Get time \"" . join( " ", @t ) . "\" failed. Unknown time format.\n" ) }
      return ($epoch);
    }
  }
  close(FILE);
  return (undef);
}

sub RemoveOldEventsCfgData {
  foreach my $key ( keys( %{ $cfg_ref->{svc_event} } ) ) {
    if ($debug_full) { &message( "Remove Event id - " . $key ) }
    delete( $cfg_ref->{svc_event}->{$key} );
  }
}

sub RemoveAllCfgData {
  foreach my $key ( keys( %{$cfg_ref} ) ) {
    if ( $key eq "date" ) { next; }
    if ($debug_full)      { &message( "Remove objects - " . $key ) }
    delete( $cfg_ref->{$key} );
  }
}

sub GetSystemInfo {
  my $line;
  my $delim = ":";
  my $codeent;
  my $prodent;
  if ( defined $cfg_ref->{info} ) {
    foreach my $k ( sort( keys(%storage_info) ) ) {
      $storage_info{$k} = $cfg_ref->{info}->{$k};
    }
  }
  if ( defined $cfg_ref->{svc_system}->{code_level} ) {
    $codeent = $cfg_ref->{svc_system}->{code_level};
  }
  else {
    my $cmd = "svcinfo lssystem -delim " . $delim;
    if ($debug_full) { message( "Remote Command: " . $cmd ); }
    my $fh;
    if ($net_openssh) {
      $fh = SVC::openssh_pipeout( $net_ssh, [$cmd], $ssh_timeout );
    }
    else {
      $fh = SVC::cmdssh_pipeout( $cmd_ssh, [$cmd], $ssh_timeout );
    }
    if ( not defined $fh ) {
      error( "Remote Command: " . $cmd . " failed.", 1 );
      return undef;
    }

    # code_level:7.3.0.7 (build 97.5.1410080000)
    # product_name:IBM Storwize V7000
    while ( my $line = <$fh> ) {
      chomp($line);
      if ( $line =~ /^\s*code_level$delim(.*$)/ ) {    # Code Level
        $codeent = $1;
        if ($debug_full) { &message( "  - \"" . $line . "\"" ); }
      }
      if ( $line =~ /^\s*product_name$delim(.*$)/ ) {
        $prodent = $1;
        if ($debug_full) { &message( "  - \"" . $line . "\"" ); }
      }
    }
    close($fh);
  }    ## if ( defined $cfg_ref->{code_level} )

  # Set product
  if ( defined $prodent ) {
    my $ent = &ProductName($prodent);
    if ( defined $ent ) {
      $storage_info{$ent} = 1;
    }
    else {
      return (undef);
    }
  }

  # Set Code Level
  if ( defined $codeent && $codeent =~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)\s+\(build\s+(.*)\).*$/ ) {
    $code_level{version} = $1;
    $code_level{release} = $2;
    $code_level{major}   = $3;
    $code_level{minor}   = $4;
    $code_level{build}   = $5;
  }
  else {
    return (undef);
  }
  return (0);
}

sub GetConfigFilename {
  my @a;
  foreach my $line_LS (`$cmd_ls $dir`) {
    chomp($line_LS);

    # Select Config Data file
    if ( $line_LS =~ /^${storage}_svcconf_\d+_\d+.data$/ ) {
      push( @a, $line_LS );
    }
  }
  my @b = sort @a;
  return ( pop @b );
}

sub GetCronXMLConfigFile {

  # Change STOUT buffer attribute
  $| = 1;    # disable STDOUT buffer

  # Clear SVC/Storwize V7000 iostats files  (NOTE: channge to script)
  my $cmd = "svcinfo lsnode -nohdr -delim :";
  if ($debug_full) { message( "Remote Command: " . $cmd ); }
  my $fh;
  if ($net_openssh) {
    $fh = SVC::openssh_pipeout( $net_ssh, [$cmd], $ssh_timeout );
  }
  else {
    $fh = SVC::cmdssh_pipeout( $cmd_ssh, [$cmd], $ssh_timeout );
  }
  if ( not defined $fh ) {
    &error( "Remote Command: " . $cmd . " failed.", 1 );
    return 1;
  }

  while ( my $line_LSNODE = <$fh> ) {

    # if($debug_full) {message("The svcinfo lsnode command output: $line_LSNODE");}
    my @field_LSNODE = split( ':', $line_LSNODE );
    my $cmd2         = "svctask cleardumps -prefix /dumps/iostats " . $field_LSNODE[0];
    if ($debug_full) { message( "Remote Command: " . $cmd2 ); }
    my $out;
    if ($net_openssh) {
      $out = SVC::openssh_runcmd( $net_ssh, $cmd2, $ssh_timeout );
    }
    else {
      $out = SVC::cmdssh_runcmd( $cmd_ssh, $cmd2, $ssh_timeout );
    }
    if ( not defined $out ) {
      &error( "Remote Command: " . $cmd2 . " failed", 1 ) && return 1;
    }
    if ($debug_full) { message( "Clear iostas files: node " . $field_LSNODE[1] ); }
  }
  close($fh);

}

sub CheckConfigFile {
  my $xmlfile = shift;
  my $tmpfile = $dir . "/" . "svc.config.backup.xml.check";
  if ($debug_full) { &message(".") }
  open( my $data, '<', $xmlfile ) || die "svcconfig.pl: Cannot open file " . $xmlfile . ". Exiting.";
  open( my $out,  '>', $tmpfile ) || die "svcconfig.pl: Cannot open file " . $tmpfile . ". Exiting.";
  if ($debug_full) { &message( "Check XML file: " . $xmlfile ) }
  while ( my $line = <$data> ) {
    chomp($line);

    #$line =~ s/[:^print:]//g;
    $line =~ s{ ( &\# ( x[0-9A-Z]+ | [0-9]+ ) ; ) } {
            my ($entity, $code) = ($1, $2);
            $code = hex $code if $code =~ s/x//i;
            chr($code) =~ $html_chars ? $entity : '';
        }eixg;

    #if ($debug_full) {print("*")};
    foreach my $exclude (@html_exclude_string) {
      $line =~ s/$exclude/_/g;
    }
    print $out $line;
  }
  if ($debug_full) { print("\n") }
  close($data);
  close($out);
}

# remove all non ASCII characters at first, there could appear some UTF16 chars which cause failing of XML parsing include xmllint
sub ASCIIFilter {
  my $xml_file     = shift;
  my $xml_file_new = $xml_file . "-tmp";

  open( FD, "< $xml_file" ) || error( "Can't open $xml_file : $! " . __FILE__ . ":" . __LINE__ ) && return 0;
  my @filter = <FD>;
  close(FD);
  open( FW, "> $xml_file_new" ) || error( "Can't open $xml_file_new : $! " . __FILE__ . ":" . __LINE__ ) && return 0;

  foreach my $line (@filter) {
    $line =~ s/[^[:ascii:]]+//g;
    print FW "$line";
  }
  close(FW);

  if ( !-z $xml_file_new ) {
    unlink($xml_file);
    rename( $xml_file_new, $xml_file );
  }
  else {
    error( "ASCIIFilter produces en empty file " . __FILE__ . ":" . __LINE__ );
    unlink($xml_file_new);
  }
  return 0;
}

sub ReadConfigFile {
  my $xmldata;
  my $xmlfile = $dir . "/" . "svc.config.backup.xml.check";

  ASCIIFilter($xmlfile);
  #
  $XML::Simple::PREFERRED_PARSER = "XML::Parser";
  if ($debug) { &message( "XML Parser: " . $XML::Simple::PREFERRED_PARSER ) }
  #
  my $simple = XML::Simple->new( keyattr => [], ForceArray => 1 );
  eval { $xmldata = $simple->XMLin($xmlfile); };
  if ($@) {
    warning("XML parsing error: $@");
    message("XML parsing error. Trying to recover XML with xmllint");
    eval {
      my $linted = `xmllint --recover $xmlfile`;
      $xmldata = $simple->XMLin($linted);
    };
    if ($@) {
      &error( "XML parsing error: " . $@ . __FILE__ . ":" . __LINE__ . " file:" . $xmlfile );
      &message( "XML parsing error: File: " . $xmlfile );
      exit(1);
    }
  }
  my $data = $xmldata->{'object'};
  return $data;
}

# -------------------------------------------------------------------------------------------------
# Insert CLI

sub InsertDataCli {

  # Save time of full config.
  if ($fullcfgflg) {
    if ( not defined $cfg_ref->{date}->{'fullcfg_epoch'} ) { $cfg_ref->{date}->{'fullcfg_epoch'} = $local_epoch; }
    if ($debug_full)                                       { &message( "Insert Full Config Epoch " . $cfg_ref->{date}->{'fullcfg_epoch'} . " to config data." ) }
  }
  my $capupd_flag = 0;
  if ( not defined $cfg_ref->{date}->{'capupd_epoch'} ) {
    $cfg_ref->{date}->{'capupd_epoch'} = $cfg_ref->{date}->{'cfg_epoch'};
    $capupd_flag = 1;
  }
  else {
    my $capupd_epoch_diff = $cfg_ref->{date}->{'cfg_epoch'} - $cfg_ref->{date}->{'capupd_epoch'};
    if ( $capupd_epoch_diff >= $capupd_interval ) {
      $cfg_ref->{date}->{'capupd_epoch'} = $cfg_ref->{date}->{'cfg_epoch'};
      $capupd_flag = 1;
      if ($debug_full) { &message( "Run Capacity update [" . $capupd_epoch_diff . "]." ) }
    }
  }

  # List drive,mdisk and pool with hw compression
  my @drive_compr;
  my %mdisk_compr;
  my %pool_compr;
  my @reclist = keys( %{ $cfg_ref->{svc_drive} } );
  foreach my $rec ( sort(@reclist) ) {
    if ( defined $cfg_ref->{svc_drive}->{$rec}->{compressed} && $cfg_ref->{svc_drive}->{$rec}->{compressed} eq "yes" ) {
      push( @drive_compr, $cfg_ref->{svc_drive}->{$rec}->{id} );
      if ( defined $cfg_ref->{svc_drive}->{$rec}->{mdisk_id} && $cfg_ref->{svc_drive}->{$rec}->{mdisk_id} ne "" ) {
        $mdisk_compr{ $cfg_ref->{svc_drive}->{$rec}->{mdisk_id} } = 1;
      }
    }
  }
  if ( $cfg_ref->{info}->{is_fs840} ) {
    $pool_compr{'mdiskgrp0'} = 1;
  }
  else {
    foreach my $rec ( sort( keys(%mdisk_compr) ) ) {
      my $mdsk_id = sprintf( "%04d", $rec );
      if ( defined $cfg_ref->{svc_mdsk}->{$mdsk_id}->{mdisk_grp_name} ) { $pool_compr{ $cfg_ref->{svc_mdsk}->{$mdsk_id}->{mdisk_grp_id} } = 1; }
    }
  }
  if ($debug_full) { &message( "List Drive with Hardware Compression: #" . scalar(@drive_compr) . " - " . join( " ", @drive_compr ) ); }
  if ($debug_full) { &message( "List MDisks with Hardware Compression: #" . scalar( keys(%mdisk_compr) ) . " - " . join( " ", keys(%mdisk_compr) ) ); }
  if ($debug_full) { &message( "List Pools with Hardware Compression: #" . scalar( keys(%pool_compr) ) . " - " . join( " ", keys(%pool_compr) ) ); }

  # List volumes for capacity update
  # compressed_copy_count 0, se_copy_count 1, 'deduplicated_copy_count' => '0',
  my @volume_capupd;
  if ($capupd_flag) {
    my @reclist = keys( %{ $cfg_ref->{svc_vdsk} } );
    foreach my $rec ( sort(@reclist) ) {
      if ( defined $cfg_ref->{svc_vdsk}->{$rec}->{'se_copy_count'} && $cfg_ref->{svc_vdsk}->{$rec}->{'se_copy_count'} > 0 ) {
        push( @volume_capupd, $cfg_ref->{svc_vdsk}->{$rec}->{'id'} );
      }
      elsif ( defined $cfg_ref->{svc_vdsk}->{$rec}->{'compressed_copy_count'} && $cfg_ref->{svc_vdsk}->{$rec}->{'compressed_copy_count'} > 0 ) {
        push( @volume_capupd, $cfg_ref->{svc_vdsk}->{$rec}->{'id'} );
      }
      elsif ( defined $cfg_ref->{svc_vdsk}->{$rec}->{'deduplicated_copy_count'} && $cfg_ref->{svc_vdsk}->{$rec}->{'deduplicated_copy_count'} > 0 ) {
        push( @volume_capupd, $cfg_ref->{svc_vdsk}->{$rec}->{'id'} );
      }
    }
    if ($debug_full) { &message( "List Volumes for capacity update: #" . scalar(@volume_capupd) . " - " . join( " ", @volume_capupd ) ); }
  }

  my $line;
  my $delim = ":";
  my $cmd;
  if ($net_openssh) {
    $cmd = " ";
    $cmd .= " echo __SYSTEM__; svcinfo lssystem -delim " . $delim . ";";
    $cmd .= " echo __SYSTEM_END__;";
    $cmd .= " echo __NODE__; svcinfo lsnode -bytes -delim " . $delim . ";";
    $cmd .= " echo __ENCLOSURE__; svcinfo lsenclosure -delim " . $delim . ";";
    $cmd .= " echo __PORT__; svcinfo lsportfc -delim " . $delim . ";";
    $cmd .= " echo __DRIVE__; svcinfo lsdrive -bytes -delim " . $delim . ";";
    $cmd .= " echo __MDISK__; svcinfo lsmdisk -bytes -delim " . $delim . ";";
    $cmd .= " echo __POOL__; svcinfo lsmdiskgrp -nohdr |while read id line; do svcinfo lsmdiskgrp -bytes -delim " . $delim . " \$id;echo __POOL_END__;done;";
    $cmd .= " echo __VOLUME__; svcinfo lsvdisk -bytes -delim " . $delim . ";";
    $cmd .= " echo __RCREL__; svcinfo lsrcrelationship -delim " . $delim . ";";
    $cmd .= " echo __FCMAP__; svcinfo lsfcmap -delim " . $delim . ";";
    $cmd .= " echo __EVENT__; svcinfo lseventlog -alert yes -fixed no -message no -monitoring no -expired no -order date -delim " . $delim . ";";
    $cmd .= " echo __SVCSN__; svcinfo lsnode -nohdr | while read id name rest; do [[ \$rest = \*\\ spare\* ]] && continue;echo node_name:\$name;svcinfo lsnodevpd -delim " . $delim . " \$id|while read line;do [[ \$line = \*system_serial_number\* ]] && echo \$line;done;echo __LSNODEVPD_END__;done;";

    if ( scalar(@drive_compr) ) {
      $cmd .= " echo __DRIVEALL__; for id in " . join( " ", @drive_compr ) . "; do svcinfo lsdrive -bytes -delim " . $delim . " \${id}; done; echo __DRIVEALL_END__;";
    }
    if ( scalar( keys(%mdisk_compr) ) ) {
      $cmd .= " echo __MDISKALL__; for id in " . join( " ", keys(%mdisk_compr) ) . "; do svcinfo lsmdisk -bytes -delim " . $delim . " \${id}; done; echo __MDISKALL_END__;";
    }
    if ( scalar(@volume_capupd) ) {
      $cmd .= " echo __VOLUMEALL__; for id in " . join( " ", @volume_capupd ) . "; do svcinfo lsvdisk -bytes -delim " . $delim . " \${id}; done; echo __VOLUMEALL_END__;";
    }
    $cmd .= " echo __END__;";
    $cmd .= " ";
  }
  else {
    $cmd = " \"";
    $cmd .= " echo __SYSTEM__; svcinfo lssystem -delim " . $delim . ";";
    $cmd .= " echo __SYSTEM_END__;";
    $cmd .= " echo __NODE__; svcinfo lsnode -bytes -delim " . $delim . ";";
    $cmd .= " echo __ENCLOSURE__; svcinfo lsenclosure -delim " . $delim . ";";
    $cmd .= " echo __PORT__; svcinfo lsportfc -delim " . $delim . ";";
    $cmd .= " echo __DRIVE__; svcinfo lsdrive -bytes -delim " . $delim . ";";
    $cmd .= " echo __MDISK__; svcinfo lsmdisk -bytes -delim " . $delim . ";";
    $cmd .= " echo __POOL__; svcinfo lsmdiskgrp -nohdr |while read id line; do svcinfo lsmdiskgrp -bytes -delim " . $delim . " \\\$id;echo __POOL_END__;done;";
    $cmd .= " echo __VOLUME__; svcinfo lsvdisk -bytes -delim " . $delim . ";";
    $cmd .= " echo __RCREL__; svcinfo lsrcrelationship -delim " . $delim . ";";
    $cmd .= " echo __FCMAP__; svcinfo lsfcmap -delim " . $delim . ";";
    $cmd .= " echo __EVENT__; svcinfo lseventlog -alert yes -fixed no -message no -monitoring no -expired no -order date -delim " . $delim . ";";
    $cmd .= " echo __SVCSN__; svcinfo lsnode -nohdr | while read id name rest; do do [[ \\\$rest = \\\*\\\\ spare\\\* ]] && continue;echo node_name:\\\$name;svcinfo lsnodevpd -delim " . $delim . " \\\$id|while read line;do [[ \\\$line = \\\*system_serial_number\\\* ]] && echo \\\$line;done;echo __LSNODEVPD_END__;done;";

    if ( scalar(@drive_compr) ) {
      $cmd .= " echo __DRIVEALL__; for id in " . join( " ", @drive_compr ) . "; do svcinfo lsdrive -bytes -delim " . $delim . " \\\${id}; done; echo __DRIVEALL_END__;";
    }
    if ( scalar( keys(%mdisk_compr) ) ) {
      $cmd .= " echo __MDISKALL__; for id in " . join( " ", keys(%mdisk_compr) ) . "; do svcinfo lsmdisk -bytes -delim " . $delim . " \\\${id}; done; echo __MDISKALL_END__;";
    }
    if ( scalar(@volume_capupd) ) {
      $cmd .= " echo __VOLUMEALL__; for id in " . join( " ", @volume_capupd ) . "; do svcinfo lsvdisk -bytes -delim " . $delim . " \\\${id}; done; echo __VOLUMEALL_END__;";
    }
    $cmd .= " echo __END__;";
    $cmd .= "\"";
  }

  # List of current konfiguration objects
  my %currcfg;
  my $currcfg_ref = \%currcfg;
  my @header;
  my $flag = 0;    # 0 ... End; 1 ... System; 2 ... Node;
  my $mdisk_id;
  my $drive_id;
  my $pool_record;
  my $pool_tier     = 1;
  my $pool_tier_ref = \$pool_tier;
  my $volume_args   = { 'volume_id' => undef, 'copy_id' => undef, 'tier_name' => undef };
  my $svcsn_node_name;
  if ($debug_full) { message( "Remote Command: " . $cmd ); }
  my $out;
  my $fh;

  if ($net_openssh) {
    $fh = SVC::openssh_pipeout( $net_ssh, [$cmd], $ssh_timeout );
  }
  else {
    $fh = SVC::cmdssh_pipeout( $cmd_ssh, [$cmd], $ssh_timeout );
  }
  if ( not defined $fh ) {
    &error( "Remote Command: " . $cmd . " failed.", 1 );
    return (1);
  }

  while ( $line = <$fh> ) {
    chomp($line);
    if    ( $line =~ /^__END__$/ )        { $flag = 0; next; }     # End of commands output
    elsif ( $line =~ /^__SYSTEM__$/ )     { $flag = 1; next; }     # System
    elsif ( $line =~ /^__SYSTEM_END__$/ ) { $flag = 2; }           # System
    elsif ( $line =~ /^__NODE__$/ )       { $flag = 3;  next; }    # Node
    elsif ( $line =~ /^__ENCLOSURE__$/ )  { $flag = 4;  next; }    # Enclosure
    elsif ( $line =~ /^__PORT__$/ )       { $flag = 5;  next; }    # Port
    elsif ( $line =~ /^__DRIVE__$/ )      { $flag = 6;  next; }    # Drive
    elsif ( $line =~ /^__MDISK__$/ )      { $flag = 7;  next; }    # MDisk
    elsif ( $line =~ /^__POOL__$/ )       { $flag = 8;  next; }    # Pool
    elsif ( $line =~ /^__VOLUME__$/ )     { $flag = 9;  next; }    # Volume
    elsif ( $line =~ /^__RCREL__$/ )      { $flag = 10; next; }    # RC Relatinship
    elsif ( $line =~ /^__FCMAP__$/ )      { $flag = 11; next; }    # FC Map
    elsif ( $line =~ /^__EVENT__$/ )      { $flag = 12; next; }    # FC Map
    elsif ( $line =~ /^__SVCSN__$/ )      { $flag = 13; next; }    # SVC G1 - system serial number
    elsif ( $line =~ /^__DRIVEALL__$/ )   { $flag = 14; next; }    # Drive with Compression
    elsif ( $line =~ /^__MDISKALL__$/ )   { $flag = 15; next; }    # Mdisk with Compression
    elsif ( $line =~ /^__VOLUMEALL__$/ )  { $flag = 16; next; }    # Volume for capacity update
    elsif ( $line =~ /^\s*$/ )            { next; }                #
                                                                   #

    if ( $flag == 1 ) {                                            # System
      if ( $line =~ /^(\w+)$delim(.*)$/ ) {
        $cfg_ref->{svc_system}->{$1} = $2;
      }
    }
    elsif ( $flag == 2 ) {                                         # System end
      &GetSystemDataCli();
    }
    elsif ( $flag == 3 ) {                                         # Node
      if ( $line =~ /^id${delim}name${delim}UPS_serial_number.*$/ ) {    # Node header
        @header = split( /$delim/, $line );
      }
      else {
        my %record;
        @record{@header} = split( /$delim/, $line );
        if ( defined $cfg_ref->{svc_node}->{ $record{name} } ) {

          # Update values
          while ( my ( $key, $val ) = each(%record) ) {
            $cfg_ref->{svc_node}->{ $record{name} }->{$key} = $val;
          }
        }
        else {
          $cfg_ref->{svc_node}->{ $record{name} } = \%record;
        }
        $currcfg_ref->{svc_node}->{ $record{name} } = 1;
      }
    }
    elsif ( $flag == 4 ) {    # Enclosure
      if ( $line =~ /^id${delim}status${delim}type.*$/ ) {    # Enclosure header
        @header = split( /$delim/, $line );
      }
      else {
        my %record;
        @record{@header} = split( /$delim/, $line );
        if ( defined $cfg_ref->{svc_enclosure}->{ $record{id} } ) {

          # Update values
          while ( my ( $key, $val ) = each(%record) ) {
            $cfg_ref->{svc_enclosure}->{ $record{id} }->{$key} = $val;
          }
        }
        else {
          $cfg_ref->{svc_enclosure}->{ $record{id} } = \%record;
        }
        $currcfg_ref->{svc_enclosure}->{ $record{id} } = 1;
      }
    }
    elsif ( $flag == 5 ) {    # Port
      if ( $line =~ /^id${delim}fc_io_port_id${delim}.*$/ ) {    # Port header
        @header = split( /$delim/, $line );
      }
      elsif ( $line =~ /^id${delim}adapter_id${delim}.*$/ ) {    # Port header
        @header = split( /$delim/, $line );
      }
      else {
        my $id;
        my %record;
        @record{@header} = split( /$delim/, $line );
        if ( length( $record{node_id} ) == 0 ) { next; }
        if ( $cfg_ref->{info}->{is_fs840} and defined $record{adapter_id} ) {
          $id = $record{node_name} . "_p" . sprintf( "%d", $record{adapter_id} ) . sprintf( "%d", $record{port_id} );
          my $inerface_id = $record{node_name} . "_interface" . sprintf( "%d", $record{adapter_id} );
          if ( not defined $cfg_ref->{svc_interfacefc}->{$inerface_id} ) {
            my %h;
            $h{name}                                    = $inerface_id;
            $h{node_name}                               = $record{node_name};
            $h{canister_id}                             = $record{canister_id};
            $h{adapter_id}                              = $record{adapter_id};
            $cfg_ref->{svc_interfacefc}->{$inerface_id} = \%h;
          }
        }
        else {
          $id = $record{node_name} . "_p" . sprintf( "%02d", $record{port_id} );
        }
        if ( defined $cfg_ref->{svc_portfc}->{$id} ) {

          # Update values
          while ( my ( $key, $val ) = each(%record) ) {
            $cfg_ref->{svc_portfc}->{$id}->{$key} = $val;
          }
        }
        else {
          $cfg_ref->{svc_portfc}->{$id} = \%record;
        }
        $currcfg_ref->{svc_portfc}->{$id} = 1;
      }
    }
    elsif ( $flag == 6 ) {    # Drive
      if ( $line =~ /^id${delim}status${delim}.*$/ ) {    # Drive header
        @header = split( /$delim/, $line );
      }
      else {
        my %record;
        @record{@header} = split( /$delim/, $line );
        my $id = sprintf( "%04d", $record{id} );
        if ( defined $cfg_ref->{svc_drive}->{$id} ) {

          # Update values
          while ( my ( $key, $val ) = each(%record) ) {
            $cfg_ref->{svc_drive}->{$id}->{$key} = $val;
          }
        }
        else {
          $cfg_ref->{svc_drive}->{$id} = \%record;
        }
        $currcfg_ref->{svc_drive}->{$id} = 1;
      }
    }
    elsif ( $flag == 7 ) {    # MDisk
      if ( $line =~ /^id${delim}name${delim}.*$/ ) {    # MDisk header
        @header = split( /$delim/, $line );
      }
      elsif ( $line =~ /^id${delim}status${delim}.*$/ ) {    # MDisk header
        @header = split( /$delim/, $line );
      }
      else {
        my %record;
        @record{@header} = split( /$delim/, $line );
        if ( $cfg_ref->{info}->{is_fs840} ) {
          $record{name}           = "array0";
          $record{tier}           = "ssd";
          $record{mdisk_grp_id}   = 0;
          $record{mdisk_grp_name} = "mdiskgrp0";
        }
        my $id = sprintf( "%04d", $record{id} );
        if ( defined $cfg_ref->{svc_mdsk}->{$id} ) {

          # Update values
          while ( my ( $key, $val ) = each(%record) ) {
            $cfg_ref->{svc_mdsk}->{$id}->{$key} = $val;
          }
        }
        else {
          $cfg_ref->{svc_mdsk}->{$id} = \%record;
        }
        $currcfg_ref->{svc_mdsk}->{$id} = 1;
      }
    }
    elsif ( $flag == 8 ) {    # Pool
      if ( $line eq "__POOL_END__" ) {    # Record last line
        if ( not defined $pool_record->{name} and defined $pool_record->{parent_mdisk_grp_name} ) {    # FS840
          $pool_record->{name} = $pool_record->{parent_mdisk_grp_name};
        }
        if ( $cfg_ref->{info}->{is_fs840} ) {
          $pool_record->{tier_0}               = "ssd";
          $pool_record->{tier_0_mdisk_count}   = 1;
          $pool_record->{tier_0_capacity}      = $pool_record->{capacity};
          $pool_record->{tier_0_free_capacity} = $pool_record->{free_capacity};
        }
        $cfg_ref->{svc_pool}->{ $pool_record->{name} }     = $pool_record;
        $currcfg_ref->{svc_pool}->{ $pool_record->{name} } = 1;
        next;
      }
      else {
        my @a = split( ':', $line );
        if ( $a[0] eq "id" ) {    # Record first line ... init record
          my %rec;
          $pool_record = \%rec;
          &InitMdiskGrpRecord($pool_record);
          $pool_record->{id} = $a[1];
        }
        else {                    # Record other line
          MDiskGroupDataRecord( $pool_record, $pool_tier_ref, \@a );
        }
      }
    }
    elsif ( $flag == 9 ) {    # Volume
      if ( $line =~ /^id${delim}name${delim}.*$/ ) {    # Volume header
        @header = split( /$delim/, $line );
      }
      else {
        my %record;
        @record{@header} = split( /$delim/, $line );
        my $id = sprintf( "%04d", $record{id} );
        if ( defined $cfg_ref->{svc_vdsk}->{$id} ) {

          # Update values
          while ( my ( $key, $val ) = each(%record) ) {
            $cfg_ref->{svc_vdsk}->{$id}->{$key} = $val;
          }
        }
        else {
          $cfg_ref->{svc_vdsk}->{$id} = \%record;
        }
        $currcfg_ref->{svc_vdsk}->{$id} = 1;
      }
    }
    elsif ( $flag == 10 ) {    # RC Relationship
      if ( $line =~ /^id${delim}name${delim}master_cluster_id${delim}.*$/ ) {    # RC Relationship header - id,name,master_cluster_id,master_cluster_name
        @header = split( /$delim/, $line );
      }
      else {
        my %record;
        @record{@header} = split( /$delim/, $line );
        my $id = sprintf( "%04d", $record{id} );
        if ( defined $cfg_ref->{svc_rcrel}->{$id} ) {

          # Update values
          while ( my ( $key, $val ) = each(%record) ) {
            $cfg_ref->{svc_rcrel}->{$id}->{$key} = $val;
          }
        }
        else {
          $cfg_ref->{svc_rcrel}->{$id} = \%record;
        }
        $currcfg_ref->{svc_rcrel}->{$id} = 1;
      }
    }
    elsif ( $flag == 11 ) {    # FC Map
      if ( $line =~ /^id${delim}name${delim}source_vdisk_id${delim}.*$/ ) {    # FC Map header - id,name,source_vdisk_id,source_vdisk_name,target_vdisk_id,target_vdisk_name,
        @header = split( /$delim/, $line );
      }
      else {
        my %record;
        @record{@header} = split( /$delim/, $line );
        my $id = sprintf( "%04d", $record{id} );
        if ( defined $cfg_ref->{svc_fcmap}->{$id} ) {

          # Update values
          while ( my ( $key, $val ) = each(%record) ) {
            $cfg_ref->{svc_fcmap}->{$id}->{$key} = $val;
          }
        }
        else {
          $cfg_ref->{svc_fcmap}->{$id} = \%record;
        }
        $currcfg_ref->{svc_fcmap}->{$id} = 1;
      }
    }
    elsif ( $flag == 12 ) {    # Event
      if ( $line =~ /^sequence_number${delim}last_timestamp${delim}.*$/ ) {    # Event header
        @header = split( /$delim/, $line );
      }
      else {
        my %record;
        @record{@header} = split( /$delim/, $line );
        my $id = sprintf( "%06d", $record{sequence_number} );
        if ( defined $cfg_ref->{svc_event}->{$id} ) {

          # Update values
          while ( my ( $key, $val ) = each(%record) ) {
            $cfg_ref->{svc_event}->{$id}->{$key} = $val;
          }
        }
        else {
          $cfg_ref->{svc_event}->{$id} = \%record;
        }
        $currcfg_ref->{svc_event}->{$id} = 1;
      }
    }
    elsif ( $cfg_ref->{info}->{is_svc} && $flag == 13 ) {    # SVC G1 - system serial number
      if ( $line eq "__LSNODEVPD_END__" ) {                  # Record last line
        next;
      }
      else {
        my @a = split( $delim, $line );
        if ( $a[0] eq "node_name" ) {                        # Record ID
          $svcsn_node_name = $a[1];
        }
        if ( $a[0] eq "system_serial_number" ) {             # Save
          $cfg_ref->{svc_node}->{$svcsn_node_name}->{serial_number} = $a[1] if ( !defined $cfg_ref->{svc_node}->{$svcsn_node_name}->{serial_number} );
        }
        else {                                               # Record other line
          next;
        }
      }
    }
    elsif ( $flag == 14 ) {    # Drive with Compression
      if ( $line eq "__DRIVEALL_END__" ) {    # Record last line
        next;
      }
      else {
        my @a = split( ':', $line );
        if ( $a[0] eq "id" ) {                # Record first line ... init record
          $drive_id = sprintf( "%04d", $a[1] );
          $cfg_ref->{svc_drive}->{$drive_id}->{ $a[0] } = $a[1];
        }
        else {                                # Record other line
          $cfg_ref->{svc_drive}->{$drive_id}->{ $a[0] } = $a[1];
        }
      }
    }
    elsif ( $flag == 15 ) {    # MDisks with Compression
      if ( $line eq "__MDISKALL_END__" ) {    # Record last line
        if ( $cfg_ref->{info}->{is_fs840} ) {
          $cfg_ref->{svc_mdsk}->{$mdisk_id}->{name}           = "array0";
          $cfg_ref->{svc_mdsk}->{$mdisk_id}->{tier}           = "ssd";
          $cfg_ref->{svc_mdsk}->{$mdisk_id}->{mdisk_grp_id}   = 0;
          $cfg_ref->{svc_mdsk}->{$mdisk_id}->{mdisk_grp_name} = "mdiskgrp0";
        }
        next;
      }
      else {
        my @a = split( ':', $line );
        if ( $a[0] eq "id" ) {    # Record first line ... init record
          $mdisk_id = sprintf( "%04d", $a[1] );
          $cfg_ref->{svc_mdsk}->{$mdisk_id}->{ $a[0] } = $a[1];
        }
        else {                    # Record other line
          $cfg_ref->{svc_mdsk}->{$mdisk_id}->{ $a[0] } = $a[1];
        }
      }
    }
    elsif ( $flag == 16 ) {    # Volumes for capacity update
      if ( $line eq "__VOLUMEALL_END__" ) {    # Record last line
        if ( $cfg_ref->{info}->{is_fs840} ) {
          my $record = $cfg_ref->{svc_vdsk}->{ $volume_args->{'volume_id'} };
          $record->{tier_0_mdisk_count}   = 1;
          $record->{tier_0_capacity}      = $record->{capacity};
          $record->{tier_0_free_capacity} = $record->{free_capacity};
        }
        next;
      }
      else {
        &GetVdiskAllDataCli( $volume_args, $line );
      }
    }
  }
  close($fh);

  # Remove deleted objects in configuration data
  foreach $key ( keys( %{$cfg_ref} ) ) {
    if ( exists( $currcfg{$key} ) ) {
      my @reclist = keys( %{ $cfg_ref->{$key} } );
      foreach my $rec ( sort(@reclist) ) {
        if ( defined $currcfg_ref->{$key}->{$rec} && $currcfg_ref->{$key}->{$rec} == 1 ) {
          next;
        }
        else {
          if ($debug_full) { &message( "Remove deleted objects - " . $key . " - " . $rec ) }
          delete( $cfg_ref->{$key}->{$rec} );
        }
      }
    }
  }

}

sub MDiskGroupDataRecord {
  my $record_ref = shift;
  my $tier_level = shift;
  my $row        = shift;

  if ( $row->[0] eq "site_id" && !defined $row->[1] ) {
    $record_ref->{ $row->[0] } = "";
  }
  elsif ( $row->[0] eq "site_name" && !defined $row->[1] ) {
    $record_ref->{ $row->[0] } = "";
  }
  elsif ( defined $row->[0] && !defined $row->[1] ) {

    # Value is not set - suppress
    return;
  }
  elsif ( $row->[0] eq "tier" && $row->[1] eq "generic_ssd" || $row->[1] eq "ssd" || $row->[1] eq "tier0_flash" ) {    # set TIER Level 0
    ${$tier_level} = "0";
    $record_ref->{tier_0} = $row->[1];
  }
  elsif ( $row->[0] eq "tier" && $row->[1] eq "tier1_flash" ) {                                                        # set TIER Level 0
    ${$tier_level} = "0n";
    $record_ref->{tier_0n} = $row->[1];
  }
  elsif ( $row->[0] eq "tier" && $row->[1] eq "tier_scm" ) {                                                           # set TIER Level 0
    ${$tier_level} = "scm";
    $record_ref->{tier_scm} = $row->[1];
  }
  elsif ( $row->[0] eq "tier" && $row->[1] eq "generic_hdd" || $row->[1] eq "enterprise" || $row->[1] eq "tier_enterprise" ) {    # set TIER Level 1
    ${$tier_level} = "1";
    $record_ref->{tier_1} = $row->[1];
  }
  elsif ( $row->[0] eq "tier" || $row->[1] eq "nearline" || $row->[1] eq "tier_nearline" ) {                                      # set TIER Level 2
    ${$tier_level} = "2";
    $record_ref->{tier_2} = $row->[1];
  }
  elsif ( $row->[0] eq "tier_mdisk_count" ) {
    $record_ref->{ "tier_" . ${$tier_level} . "_mdisk_count" } = $row->[1];
  }
  elsif ( $row->[0] eq "tier_capacity" ) {
    $record_ref->{ "tier_" . ${$tier_level} . "_capacity" } = $row->[1];
  }
  elsif ( $row->[0] eq "tier_free_capacity" ) {
    $record_ref->{ "tier_" . ${$tier_level} . "_free_capacity" } = $row->[1];
  }
  else {
    $record_ref->{ $row->[0] } = $row->[1];
  }
}

sub GetVdiskAllDataCli {
  my $args = shift;    # volume_id, copy_id, tier_name
  my $line = shift;
  #
  my @a = split( ':', $line );
  if ( $a[0] eq "id" ) {    # Record first line ... init record
    $args->{'volume_id'} = sprintf( "%04d", $a[1] );
    $args->{'copy_id'}   = undef;
    $args->{'tier_name'} = undef;
  }
  else {
    my $record = $cfg_ref->{svc_vdsk}->{ $args->{'volume_id'} };
    if ( $a[0] eq "copy_id" ) {    # Record copy first line ... init record
      $args->{'copy_id'} = $a[1];
    }
    else {                         # Record other line
      if ( not defined $args->{'copy_id'} ) {
        $record->{ $a[0] } = $a[1];
      }
      else {
        # Volume Copy 0/1
        if ( $a[0] eq "tier" ) {

          # tier = tier0_flash	==>	copy_0_tier_0 =	tier0_flash
          if ( $a[1] =~ /ssd/ || $a[1] =~ /tier0_flash/ ) {

            # SSD - TIER0-FLASH
            $args->{'tier_name'} = "tier_0";
            $record->{ 'copy_' . $args->{'copy_id'} . '_' . $args->{'tier_name'} } = $a[1];
          }
          elsif ( $a[1] =~ /tier1_flash/ ) {

            # HCF - TIER1-FLASH
            $args->{'tier_name'} = "tier_0n";
            $record->{ 'copy_' . $args->{'copy_id'} . '_' . $args->{'tier_name'} } = $a[1];
          }
          elsif ( $a[1] =~ /tier_scm/ ) {

            # SCM
            $args->{'tier_name'} = "tier_scm";
            $record->{ 'copy_' . $args->{'copy_id'} . '_' . $args->{'tier_name'} } = $a[1];
          }
          elsif ( $a[1] =~ /enterprise/ || $a[1] =~ /hdd/ ) {

            # Enterprise
            $args->{'tier_name'} = "tier_1";
            $record->{ 'copy_' . $args->{'copy_id'} . '_' . $args->{'tier_name'} } = $a[1];
          }
          elsif ( $a[1] =~ /nearline/ || $a[1] =~ /Nearline/ ) {

            # Nearline
            $args->{'tier_name'} = "tier_2";
            $record->{ 'copy_' . $args->{'copy_id'} . '_' . $args->{'tier_name'} } = $a[1];
          }
          else {
            &warning( "Unknown tiername " . $a[1] . "." );
            $args->{'tier_name'} = "tier_unknown";
          }
        }
        elsif ( $a[0] eq "tier_capacity" ) {
          $record->{ "copy_" . $args->{'copy_id'} . "_" . $args->{'tier_name'} . "_capacity" } = $a[1];
        }
        elsif ( $a[0] eq "mdisk_grp_id" ) {
          $record->{ "copy_" . $args->{'copy_id'} . "_" . $a[0] } = $a[1];
        }
        elsif ( $a[0] eq "mdisk_grp_name" ) {
          $record->{ "copy_" . $args->{'copy_id'} . "_" . $a[0] } = $a[1];
        }
        elsif ( $a[0] eq "used_capacity" ) {
          $record->{ "copy_" . $args->{'copy_id'} . "_" . $a[0] } = $a[1];
        }
        elsif ( $a[0] eq "real_capacity" ) {
          $record->{ "copy_" . $args->{'copy_id'} . "_" . $a[0] } = $a[1];
        }
        elsif ( $a[0] eq "free_capacity" ) {
          $record->{ "copy_" . $args->{'copy_id'} . "_" . $a[0] } = $a[1];
        }
        elsif ( $a[0] eq "se_copy" ) {
          $record->{ "copy_" . $args->{'copy_id'} . "_" . $a[0] } = $a[1];
        }
        elsif ( $a[0] eq "overallocation" ) {
          $record->{ "copy_" . $args->{'copy_id'} . "_" . $a[0] } = $a[1];
        }
        elsif ( $a[0] eq "compressed_copy" ) {
          $record->{ "copy_" . $args->{'copy_id'} . "_" . $a[0] } = $a[1];
        }
        elsif ( $a[0] eq "uncompressed_used_capacity" ) {
          $record->{ "copy_" . $args->{'copy_id'} . "_" . $a[0] } = $a[1];
        }
        elsif ( $a[0] eq "easy_tier" ) {
          $record->{ "copy_" . $args->{'copy_id'} . "_" . $a[0] } = $a[1];
        }
        elsif ( $a[0] eq "used_capacity_before_reduction" ) {
          $record->{ "copy_" . $args->{'copy_id'} . "_" . $a[0] } = $a[1];
        }
        else {
          $record->{ $a[0] } = $a[1];
        }
      }
    }
  }    ## if ( $a[0] eq "id" )
}

# -------------------------------------------------------------------------------------------------
# Insert XML

sub InsertDataXml {

  # Insert data from XML file
  &GetSystemDataXml();
  &GetNodeDataXml();
  &GetEnclosureDataXml();
  &GetPortfcDataXml();
  &GetDriveDataXml();
  &GetMdiskDataXml();
  &GetMdiskGrpDataXml();
  &GetVdiskDataXml();
  &GetVdiskHostMapDataXml();
  &GetHostDataXml();
  &GetRcrelDataXml();
  &GetFcmapDataXml();

  # Get CLI Data
  my $line;
  my $delim = ":";
  my $cmd;
  if ($net_openssh) {
    $cmd = " ";
    $cmd .= " echo __EVENT__; svcinfo lseventlog -alert yes -fixed no -message no -monitoring no -expired no -order date -delim " . $delim . ";";
    if ( $cfg_ref->{info}->{is_fs840} ) {
      $cmd .= " echo __FABRIC__; svcinfo lsfabric840 -delim " . $delim . ";";
    }
    else {
      $cmd .= " echo __FABRIC__; svcinfo lsfabric -delim " . $delim . ";";
    }
    if ( $cfg_ref->{info}->{is_svc} ) {
      $cmd .= " echo __SVCSN__; svcinfo lsnode -nohdr | while read id name rest; do echo node_name:\$name;svcinfo lsnodevpd -delim " . $delim . " \$id|while read line;do [[ \$line = \*system_serial_number\* ]] && echo \$line;done;echo __LSNODEVPD_END__;done;";
    }
    $cmd .= " echo __END__;";
    $cmd .= " ";
  }
  else {
    $cmd = " \"";
    $cmd .= " echo __EVENT__; svcinfo lseventlog -alert yes -fixed no -message no -monitoring no -expired no -order date -delim " . $delim . ";";
    if ( $cfg_ref->{info}->{is_fs840} ) {
      $cmd .= " echo __FABRIC__; svcinfo lsfabric840 -delim " . $delim . ";";
    }
    else {
      $cmd .= " echo __FABRIC__; svcinfo lsfabric -delim " . $delim . ";";
    }
    if ( $cfg_ref->{info}->{is_svc} ) {
      $cmd .= " echo __SVCSN__; svcinfo lsnode -nohdr | while read id name rest; do echo node_name:\\\$name;svcinfo lsnodevpd -delim " . $delim . " \\\$id|while read line;do [[ \\\$line = \\\*system_serial_number\\\* ]] && echo \\\$line;done;echo __LSNODEVPD_END__;done;";
    }
    $cmd .= " echo __END__;";
    $cmd .= "\"";
  }

  my @header;
  my $svcsn_node_name;
  my $flag = 0;    # 0 ... End; 1 ... Event; 2 ... Fabric;
  if ($debug_full) { &message( "Remote Command: " . $cmd ); }
  my $fh;
  if ($net_openssh) {
    $fh = SVC::openssh_pipeout( $net_ssh, [$cmd], $ssh_timeout );
  }
  else {
    $fh = SVC::cmdssh_pipeout( $cmd_ssh, [$cmd], $ssh_timeout );
  }
  if ( not defined $fh ) {
    &error( "Remote Command: " . $cmd . " failed.", 1 );
    return (1);
  }

  while ( $line = <$fh> ) {
    chomp($line);
    if    ( $line =~ /^__END__$/ )    { $flag = 0; next; }    # End of commands output
    elsif ( $line =~ /^__EVENT__$/ )  { $flag = 1; next; }    # Event
    elsif ( $line =~ /^__FABRIC__$/ ) { $flag = 2; next; }    # Fabric
    elsif ( $line =~ /^__SVCSN__$/ )  { $flag = 3; next; }    # SVC G1 - system serial number
                                                              #
    if    ( $flag == 1 ) {                                    # Event
      if ( $line =~ /^sequence_number${delim}last_timestamp${delim}.*$/ ) {    # Event header
        @header = split( /$delim/, $line );
      }
      else {
        my %record;
        @record{@header} = split( /$delim/, $line );
        my $id = sprintf( "%06d", $record{sequence_number} );
        $cfg_ref->{svc_event}->{$id} = \%record;
      }
    }
    elsif ( $flag == 2 ) {    # Fabric
      if ( $line =~ /^remote_wwpn${delim}remote_nportid${delim}.*$/ ) {    # header
        @header = split( /$delim/, $line );
        my @data;
        $cfg_ref->{svc_fabric} = \@data;
      }
      elsif ( $line =~ /^remote_uid${delim}local_uid${delim}.*$/ ) {       # header
        @header = split( /$delim/, $line );
        my @data;
        $cfg_ref->{svc_fabric} = \@data;
      }
      else {
        my %record;
        @record{@header} = split( /$delim/, $line );
        push( @{ $cfg_ref->{svc_fabric} }, \%record );
      }
    }
    elsif ( $flag == 3 ) {    # SVC G1 - system serial number
      if ( $line eq "__LSNODEVPD_END__" ) {    # Record last line
        next;
      }
      else {
        my @a = split( $delim, $line );
        if ( $a[0] eq "node_name" ) {          # Record ID
          $svcsn_node_name = $a[1];
        }
        if ( $a[0] eq "system_serial_number" ) {    # Save
          $cfg_ref->{svc_node}->{$svcsn_node_name}->{serial_number} = $a[1] if ( !defined $cfg_ref->{svc_node}->{$svcsn_node_name}->{serial_number} );
        }
        else {                                      # Record other line
          next;
        }
      }
    }
  }
  close($fh);
}

sub StoreData {
  #
  nstore( $cfg_ref, $tmp_conf_file ) or die("svcconfig.pl: Store data to file $tmp_conf_file failed.");
}

sub StoreDataXml {
  #
  nstore( $cfg_ref, $tmp_conf_full_file ) or die("svcconfig.pl: Store data to file $tmp_conf_full_file failed.");
}

sub RetriveData {
  if ( -r $dir . "/" . $tmp_conf_file ) {
    eval { $cfg_ref = retrieve( $dir . "/" . $tmp_conf_file ); };
    if ($@) {
      &message( $@ . __FILE__ . ":" . __LINE__ . " file:" . $dir . "/" . $tmp_conf_file );
      &message( "Retrieve data from file " . $dir . "/" . $tmp_conf_file . " failed." );
      system( "$cmd_rm -f " . $dir . "/" . $tmp_conf_file );
      $fullcfgflg = 1;
    }
    if ( !defined $cfg_ref ) {
      &message("Retrieve data from file $tmp_conf_file failed.");
      system( "$cmd_rm -f " . $dir . "/" . $tmp_conf_file );
      $fullcfgflg = 1;
    }
    elsif ( !scalar( %{$cfg_ref} ) ) {
      if ($debug_full) { &message("Previous configuration data is not valid.") }
      system( "$cmd_rm -f " . $dir . "/" . $tmp_conf_file );
      $fullcfgflg = 1;
    }
    elsif ( !defined $cfg_ref->{date}->{'fullcfg_epoch'} ) {
      if ($debug_full) { &message("Full Config Epoch is not find in previous data.") }
      $fullcfgflg = 1;
    }
    elsif ( $local_epoch - $cfg_ref->{date}->{'fullcfg_epoch'} >= $fullcfgint ) {
      $fullcfgflg = 1;
      if ($debug_full) {
        &message( "Full Config Time = " . localtime( $cfg_ref->{date}->{'fullcfg_epoch'} ) . " - epoch = " . $cfg_ref->{date}->{'fullcfg_epoch'} );
        my $diff = $local_epoch - $cfg_ref->{date}->{'fullcfg_epoch'};
        &message( "Full Config Interval = " . $diff );

      }
    }
    else {
      if ($debug_full) {
        &message( "Full Config Time = " . localtime( $cfg_ref->{date}->{'fullcfg_epoch'} ) . " - epoch = " . $cfg_ref->{date}->{'fullcfg_epoch'} );
        my $diff = $local_epoch - $cfg_ref->{date}->{'fullcfg_epoch'};
        &message( "Full Config Interval = " . $diff );
      }
    }
  }
  else {
    $fullcfgflg = 1;
    system( "$cmd_rm -f " . $dir . "/" . $tmp_conf_file );
  }
}

sub PrintData {
  #
  if ($debug_full) { &message("Output file: ${out_conf_file}.tmp.") }
  open( CONFOUT, ">${out_conf_file}.tmp" ) or die("Print data to file ${out_conf_file}.tmp failed.");
  &PrintSystemData();
  &PrintNodeData();
  &PrintEnclosureData();
  &PrintPortfcData();
  &PrintDriveData();
  &PrintMdiskData();
  &PrintMdiskGrpData();
  &PrintVdiskData();
  &PrintRcrelData();
  &PrintFcmapData();

  if ($fullcfgflg) {
    &PrintHostData() if ( $codevrmn < 6400 || $codevrmn > 7101 );
    &PrintFabricData();
  }
  &PrintEventData();
  close(CONFOUT);
  if ($debug_full)                        { &message("Rename output file to: ${out_conf_file}.") }
  if ( not -f ${out_conf_file} . ".tmp" ) { return; }
  #my $ret = system( "$cmd_mv " . ${out_conf_file} . ".tmp " . ${out_conf_file} );
  my $ret = system( "$cmd_mv '${out_conf_file}.tmp' '${out_conf_file}'" );
  if ($ret) { error( "svcconfig.pl: Couldn't rename file " . ${tmp_conf_file} . ".tmp ." ); return; }
}

# -------------------------------------------------------------------------------------------------
# CLI

sub GetSystemDataCli {
  $cfg_ref->{info} = \%storage_info;
  if ( defined $cfg_ref->{svc_system}->{product_name} ) {
    my $isprod = &ProductName( $cfg_ref->{svc_system}->{product_name} );
    if ( defined $isprod ) {
      $cfg_ref->{info}->{$isprod} = 1;
      if ( $isprod eq "is_fs840" ) {

        # IBM FlashSystem 840/900
        $cfg_ref->{svc_system}->{ssd_tier_capacity}      = $cfg_ref->{svc_system}->{space_in_mdisk_grps};
        $cfg_ref->{svc_system}->{ssd_tier_free_capacity} = $cfg_ref->{svc_system}->{total_free_space};
        #
        $cfg_ref->{svc_system}->{enterprise_tier_capacity}      = "0.0GB";
        $cfg_ref->{svc_system}->{enterprise_tier_free_capacity} = "0.0GB";
        $cfg_ref->{svc_system}->{nearline_tier_capacity}        = "0.0GB";
        $cfg_ref->{svc_system}->{nearline_tier_free_capacity}   = "0.0GB";
        $cfg_ref->{svc_system}->{ssdnl_tier_capacity}           = "0.0GB";
        $cfg_ref->{svc_system}->{ssdnl_tier_free_capacity}      = "0.0GB";
      }
    }
  }
  if ($debug_full) {
    my $msg;
    my @a;
    foreach my $k ( sort( keys( %{ $cfg_ref->{'info'} } ) ) ) {
      push( @a, $k . "=" . $cfg_ref->{'info'}->{$k} );
    }
    message( "HW Type Info: " . join( ",", @a ) . "\n" );
  }
}

sub InitMdiskGrpRecord {
  my $c = shift;
  $c->{capacity}                          = 0;
  $c->{free_capacity}                     = 0;
  $c->{virtual_capacity}                  = 0;
  $c->{used_capacity}                     = 0;
  $c->{real_capacity}                     = 0;
  $c->{tier_0_mdisk_count}                = 0;
  $c->{tier_0_capacity}                   = 0;
  $c->{tier_0_free_capacity}              = 0;
  $c->{tier_0n_mdisk_count}               = 0;
  $c->{tier_0n_capacity}                  = 0;
  $c->{tier_0n_free_capacity}             = 0;
  $c->{tier_scm_mdisk_count}              = 0;
  $c->{tier_scm_capacity}                 = 0;
  $c->{tier_scm_free_capacity}            = 0;
  $c->{tier_1_mdisk_count}                = 0;
  $c->{tier_1_capacity}                   = 0;
  $c->{tier_1_free_capacity}              = 0;
  $c->{tier_2_mdisk_count}                = 0;
  $c->{tier_2_capacity}                   = 0;
  $c->{tier_2_free_capacity}              = 0;
  $c->{compression_active}                = "no";
  $c->{compression_virtual_capacity}      = 0;
  $c->{compression_compressed_capacity}   = 0;
  $c->{compression_uncompressed_capacity} = 0;
  $c->{extent_size}                       = 0;
  $c->{overallocation}                    = 0;
  $c->{warning}                           = 0;
  $c->{easy_tier}                         = "no";
  $c->{easy_tier_status}                  = "no";
  $c->{compression_active}                = "no";
  $c->{hw_compressed}                     = 'no';
  $c->{effective_used_capacity}           = 0;
}

# -------------------------------------------------------------------------------------------------
# XML
sub GetSystemDataXml {
  my $p;
  foreach my $n ( @{$cfgdata} ) {
    if ( $n->{'type'} eq "cluster" ) {
      $p = $n->{'property'};
    }
  }
  my $tiername;
  my $tierunknown;
  foreach my $m ( @{$p} ) {
    if ( $m->{'name'} eq "tier" ) {
      if ( $m->{'value'} =~ /ssd/ || $m->{'value'} =~ /tier0_flash/ ) {
        $tiername = "ssd";
      }
      elsif ( $m->{'value'} =~ /tier1_flash/ ) {
        $tiername = "ssdnl";
      }
      elsif ( $m->{'value'} =~ /tier_scm/ ) {
        $tiername = "scm";
      }
      elsif ( $m->{'value'} =~ /enterprise/ || $m->{'value'} =~ /hdd/ ) {
        $tiername = "enterprise";
      }
      elsif ( $m->{'value'} =~ /nearline/ || $m->{'value'} =~ /Nearline/ ) {
        $tiername = "nearline";
      }
      else {
        $tierunknown = "Unknown tiername " . $m->{'value'} . ".";
        $tiername    = "tier_unknown";
      }
    }
    elsif ( $m->{'name'} eq "tier_capacity" ) {
      $cfg_ref->{svc_system}->{ $tiername . "_" . $m->{'name'} } = $m->{'value'};
    }
    elsif ( $m->{'name'} eq "tier_free_capacity" ) {
      $cfg_ref->{svc_system}->{ $tiername . "_" . $m->{'name'} } = $m->{'value'};
    }
    else {
      $cfg_ref->{svc_system}->{ $m->{'name'} } = $m->{'value'};
    }
  }
  if ( !defined $cfg_ref->{svc_system}->{ssd_tier_capacity} )             { $cfg_ref->{svc_system}->{ssd_tier_capacity}             = 0; }
  if ( !defined $cfg_ref->{svc_system}->{ssdnl_tier_capacity} )           { $cfg_ref->{svc_system}->{ssdnl_tier_capacity}           = 0; }
  if ( !defined $cfg_ref->{svc_system}->{scm_tier_capacity} )             { $cfg_ref->{svc_system}->{scm_tier_capacity}             = 0; }
  if ( !defined $cfg_ref->{svc_system}->{enterprise_tier_capacity} )      { $cfg_ref->{svc_system}->{enterprise_tier_capacity}      = 0; }
  if ( !defined $cfg_ref->{svc_system}->{nearline_tier_capacity} )        { $cfg_ref->{svc_system}->{nearline_tier_capacity}        = 0; }
  if ( !defined $cfg_ref->{svc_system}->{ssd_tier_free_capacity} )        { $cfg_ref->{svc_system}->{ssd_tier_free_capacity}        = 0; }
  if ( !defined $cfg_ref->{svc_system}->{ssdnl_tier_free_capacity} )      { $cfg_ref->{svc_system}->{ssdnl_tier_free_capacity}      = 0; }
  if ( !defined $cfg_ref->{svc_system}->{scm_tier_free_capacity} )        { $cfg_ref->{svc_system}->{scm_tier_free_capacity}        = 0; }
  if ( !defined $cfg_ref->{svc_system}->{enterprise_tier_free_capacity} ) { $cfg_ref->{svc_system}->{enterprise_tier_free_capacity} = 0; }
  if ( !defined $cfg_ref->{svc_system}->{nearline_tier_free_capacity} )   { $cfg_ref->{svc_system}->{nearline_tier_free_capacity}   = 0; }
  #
  $cfg_ref->{info} = \%storage_info;
  if ( defined $cfg_ref->{svc_system}->{product_name} ) {
    my $isprod = &ProductName( $cfg_ref->{svc_system}->{product_name} );
    if ( defined $isprod ) {
      $cfg_ref->{info}->{$isprod} = 1;
      if ( $isprod eq "is_fs840" ) {

        # IBM FlashSystem 840/900
        $cfg_ref->{svc_system}->{ssd_tier_capacity}      = $cfg_ref->{svc_system}->{space_in_mdisk_grps};
        $cfg_ref->{svc_system}->{ssd_tier_free_capacity} = $cfg_ref->{svc_system}->{total_free_space};
        #
        $cfg_ref->{svc_system}->{enterprise_tier_capacity}      = "0.0GB";
        $cfg_ref->{svc_system}->{enterprise_tier_free_capacity} = "0.0GB";
        $cfg_ref->{svc_system}->{nearline_tier_capacity}        = "0.0GB";
        $cfg_ref->{svc_system}->{nearline_tier_free_capacity}   = "0.0GB";
        $cfg_ref->{svc_system}->{ssdnl_tier_capacity}           = "0.0GB";
        $cfg_ref->{svc_system}->{ssdnl_tier_free_capacity}      = "0.0GB";
      }
    }
  }
  else {
    if ($tierunknown) { &warning($tierunknown); }
  }
  if ($debug_full) {
    my $msg;
    my @a;
    foreach my $k ( sort( keys( %{ $cfg_ref->{'info'} } ) ) ) {
      push( @a, $k . "=" . $cfg_ref->{'info'}->{$k} );
    }
    message( "HW Type Info: " . join( ",", @a ) . "\n" );
  }
}

sub GetNodeDataXml {
  foreach my $n ( @{$cfgdata} ) {
    if ( $n->{'type'} eq "node" ) {
      my %record;
      foreach my $m ( @{ $n->{'property'} } ) {
        $record{ $m->{'name'} } = $m->{'value'};
      }
      $cfg_ref->{svc_node}->{ $record{name} } = \%record;
    }
  }
}

sub GetEnclosureDataXml {
  foreach my $n ( @{$cfgdata} ) {
    if ( $n->{'type'} eq "enclosure" ) {
      my %record;
      foreach my $m ( @{ $n->{'property'} } ) {
        $record{ $m->{'name'} } = $m->{'value'};
      }
      $cfg_ref->{svc_enclosure}->{ $record{id} } = \%record;
    }
  }
}

sub GetPortfcDataXml {
  foreach my $n ( @{$cfgdata} ) {
    if ( $n->{'type'} eq "portfc" ) {
      my %record;
      my $id;
      foreach my $m ( @{ $n->{'property'} } ) {
        $record{ $m->{'name'} } = $m->{'value'};
      }
      if ( length( $record{node_id} ) == 0 ) { next; }
      if ( $cfg_ref->{info}->{is_fs840} and defined $record{adapter_id} ) {
        $id = $record{node_name} . "_p" . sprintf( "%d", $record{adapter_id} ) . sprintf( "%d", $record{port_id} );
        my $interface_id = $record{node_name} . "_interface" . sprintf( "%d", $record{adapter_id} );
        if ( not defined $cfg_ref->{svc_interfacefc}->{$interface_id} ) {
          my %h;
          $h{name}                                     = $interface_id;
          $h{node_name}                                = $record{node_name};
          $h{canister_id}                              = $record{canister_id};
          $h{adapter_id}                               = $record{adapter_id};
          $cfg_ref->{svc_interfacefc}->{$interface_id} = \%h;
        }
      }
      else {
        $id = $record{node_name} . "_p" . sprintf( "%02d", $record{port_id} );
      }
      $cfg_ref->{svc_portfc}->{$id} = \%record;
    }
  }
}

sub GetDriveDataXml {
  foreach my $n ( @{$cfgdata} ) {
    if ( $n->{'type'} eq "drive" ) {
      my %record;
      foreach my $m ( @{ $n->{'property'} } ) {
        $record{ $m->{'name'} } = $m->{'value'};
      }
      my $id = sprintf( "%04d", $record{id} );
      $cfg_ref->{svc_drive}->{$id} = \%record;
    }
  }
}

sub GetMdiskDataXml {
  foreach my $n ( @{$cfgdata} ) {
    if ( $n->{'type'} eq "mdisk" ) {
      my %record;
      foreach my $m ( @{ $n->{'property'} } ) {
        $record{ $m->{'name'} } = $m->{'value'};
      }
      if ( $cfg_ref->{info}->{is_fs840} ) {
        $record{name}           = "array0";
        $record{tier}           = "ssd";
        $record{mdisk_grp_id}   = 0;
        $record{mdisk_grp_name} = "mdiskgrp0";
      }
      my $id = sprintf( "%04d", $record{id} );
      $cfg_ref->{svc_mdsk}->{$id} = \%record;
    }
  }
}

sub GetMdiskGrpDataXml {
  foreach my $n ( @{$cfgdata} ) {
    if ( $n->{'type'} eq "mdisk_grp" ) {
      my %record;
      my $tiername;
      foreach my $m ( @{ $n->{'property'} } ) {
        if ( $m->{'name'} eq "tier" ) {
          if ( $m->{'value'} =~ /ssd/ || $m->{'value'} =~ /tier0_flash/ ) {

            # SSD - TIER0-FLASH
            $tiername = "tier_0";
          }
          elsif ( $m->{'value'} =~ /tier1_flash/ ) {

            # HCF - TIER1-FLASH
            $tiername = "tier_0n";
          }
          elsif ( $m->{'value'} =~ /tier_scm/ ) {

            # SCM
            $tiername = "tier_scm";
          }
          elsif ( $m->{'value'} =~ /enterprise/ || $m->{'value'} =~ /hdd/ ) {

            # Enterprise
            $tiername = "tier_1";
          }
          elsif ( $m->{'value'} =~ /nearline/ || $m->{'value'} =~ /Nearline/ ) {

            # Nearline
            $tiername = "tier_2";
          }
          else {
            &warning( "Unknown tiername " . $m->{'value'} . "." );
            $tiername = "tier_unknown";
          }
          $record{$tiername} = $m->{'value'};
        }
        elsif ( $m->{'name'} eq "tier_mdisk_count" ) {
          $record{ $tiername . "_mdisk_count" } = $m->{'value'};
        }
        elsif ( $m->{'name'} eq "tier_capacity" ) {
          $record{ $tiername . "_capacity" } = $m->{'value'};
        }
        elsif ( $m->{'name'} eq "tier_free_capacity" ) {
          $record{ $tiername . "_free_capacity" } = $m->{'value'};
        }
        else {
          $record{ $m->{'name'} } = $m->{'value'};
        }
      }
      if ( not defined $record{easy_tier} )               { $record{easy_tier}               = "no"; }
      if ( not defined $record{easy_tier_status} )        { $record{easy_tier_status}        = "no"; }
      if ( not defined $record{compression_active} )      { $record{compression_active}      = "no"; }
      if ( not defined $record{extent_size} )             { $record{extent_size}             = 0; }
      if ( not defined $record{overallocation} )          { $record{overallocation}          = 0; }
      if ( not defined $record{warning} )                 { $record{warning}                 = 0; }
      if ( not defined $record{hw_compressed} )           { $record{hw_compressed}           = 'no'; }
      if ( not defined $record{effective_used_capacity} ) { $record{effective_used_capacity} = 0; }

      if ( $cfg_ref->{info}->{is_fs840} ) {
        $record{tier_0_mdisk_count}   = 1;
        $record{tier_0_capacity}      = $record{capacity};
        $record{tier_0_free_capacity} = $record{free_capacity};
      }
      my $id = $record{name};
      $cfg_ref->{svc_pool}->{$id} = \%record;
    }
  }
}

sub GetVdiskDataXml {
  foreach my $n ( @{$cfgdata} ) {
    if ( $n->{'type'} eq "vdisk" ) {
      my %record;
      $record{'mdisk_grp_id'}   = "";
      $record{'mdisk_grp_name'} = "";

      #         foreach my $m ( @{$n->{'property'}} ) {
      #             $record{$m->{'name'}} = $m->{'value'};
      #         }
      #         my $id = sprintf("%04d",$record{id});
      #         $cfg_ref->{svc_vdsk}->{$id} = \%record;
      my $tiername;
      my $copyid = 0;
      foreach my $m ( @{ $n->{'property'} } ) {

        #$record{$m->{'name'}} = $m->{'value'};
        if ( $m->{'name'} eq "copy_id" ) {
          $copyid = $m->{'value'};
        }
        elsif ( $m->{'name'} eq "tier" ) {
          if ( $m->{'value'} =~ /ssd/ || $m->{'value'} =~ /tier0_flash/ ) {

            # SSD - TIER0-FLASH
            $tiername = "tier_0";
            $record{ "copy_" . $copyid . "_" . $tiername } = $m->{'value'};
          }
          elsif ( $m->{'value'} =~ /tier1_flash/ ) {

            # HCF - TIER1-FLASH
            $tiername = "tier_0n";
            $record{ "copy_" . $copyid . "_" . $tiername } = $m->{'value'};
          }
          elsif ( $m->{'value'} =~ /tier_scm/ ) {

            # SCM
            $tiername = "tier_scm";
            $record{ "copy_" . $copyid . "_" . $tiername } = $m->{'value'};
          }
          elsif ( $m->{'value'} =~ /enterprise/ || $m->{'value'} =~ /hdd/ ) {

            # Enterprise
            $tiername = "tier_1";
            $record{ "copy_" . $copyid . "_" . $tiername } = $m->{'value'};
          }
          elsif ( $m->{'value'} =~ /nearline/ || $m->{'value'} =~ /Nearline/ ) {

            # Nearline
            $tiername = "tier_2";
            $record{ "copy_" . $copyid . "_" . $tiername } = $m->{'value'};
          }
          else {
            &warning( "Unknown tiername " . $m->{'value'} . "." );
            $tiername = "tier_unknown";
          }
        }
        elsif ( $m->{'name'} eq "tier_capacity" ) {
          $record{ "copy_" . $copyid . "_" . $tiername . "_capacity" } = $m->{'value'};
        }
        elsif ( $m->{'name'} eq "mdisk_grp_id" ) {
          $record{ "copy_" . $copyid . "_" . $m->{'name'} } = $m->{'value'};
        }
        elsif ( $m->{'name'} eq "mdisk_grp_name" ) {
          $record{ "copy_" . $copyid . "_" . $m->{'name'} } = $m->{'value'};
        }
        elsif ( $m->{'name'} eq "used_capacity" ) {
          $record{ "copy_" . $copyid . "_" . $m->{'name'} } = $m->{'value'};
        }
        elsif ( $m->{'name'} eq "real_capacity" ) {
          $record{ "copy_" . $copyid . "_" . $m->{'name'} } = $m->{'value'};
        }
        elsif ( $m->{'name'} eq "free_capacity" ) {
          $record{ "copy_" . $copyid . "_" . $m->{'name'} } = $m->{'value'};
        }
        elsif ( $m->{'name'} eq "se_copy" ) {
          $record{ "copy_" . $copyid . "_" . $m->{'name'} } = $m->{'value'};
        }
        elsif ( $m->{'name'} eq "overallocation" ) {
          $record{ "copy_" . $copyid . "_" . $m->{'name'} } = $m->{'value'};
        }
        elsif ( $m->{'name'} eq "compressed_copy" ) {
          $record{ "copy_" . $copyid . "_" . $m->{'name'} } = $m->{'value'};
        }
        elsif ( $m->{'name'} eq "uncompressed_used_capacity" ) {
          $record{ "copy_" . $copyid . "_" . $m->{'name'} } = $m->{'value'};
        }
        elsif ( $m->{'name'} eq "easy_tier" ) {
          $record{ "copy_" . $copyid . "_" . $m->{'name'} } = $m->{'value'};
        }
        else {
          $record{ $m->{'name'} } = $m->{'value'};
        }
      }
      if ( $cfg_ref->{info}->{is_fs840} ) {
        $record{copy_0_tier_0_capacity} = $record{copy_0_real_capacity};
      }
      my $id = sprintf( "%04d", $record{id} );
      $cfg_ref->{svc_vdsk}->{$id} = \%record;

      #         $cfg{$id} = \%record;

    }
  }
}

sub GetFcmapDataXml {
  foreach my $n ( @{$cfgdata} ) {
    if ( $n->{'type'} eq "fcmap" ) {
      my %record;
      foreach my $m ( @{ $n->{'property'} } ) {
        $record{ $m->{'name'} } = $m->{'value'};
      }
      my $id = sprintf( "%04d", $record{id} );
      $cfg_ref->{svc_fcmap}->{$id} = \%record;
    }
  }
}

sub GetRcrelDataXml {
  foreach my $n ( @{$cfgdata} ) {
    if ( $n->{'type'} eq "rcrelationship" ) {
      my %record;
      foreach my $m ( @{ $n->{'property'} } ) {
        $record{ $m->{'name'} } = $m->{'value'};
      }
      my $id = sprintf( "%04d", $record{id} );
      $cfg_ref->{svc_rcrel}->{$id} = \%record;
    }
  }
}

sub GetHostDataXml {
  foreach my $n ( @{$cfgdata} ) {
    if ( $n->{'type'} eq "host" ) {
      my %record;
      my $i = 0;
      foreach my $m ( @{ $n->{'property'} } ) {
        if ( $m->{'name'} eq "WWPN" ) {
          $record{ "FC" . $i . "_" . $m->{'name'} } = $m->{'value'};
          $i++;
        }
        elsif ( $m->{'name'} eq "node_logged_in_count" ) {
          $record{ "FC" . $i . "_" . $m->{'name'} } = $m->{'value'};
        }
        elsif ( $m->{'name'} eq "state" ) {
          $record{ "FC" . $i . "_" . $m->{'name'} } = $m->{'value'};
        }
        else {
          $record{ $m->{'name'} } = $m->{'value'};
        }
      }
      my $id = sprintf( "%04d", $record{id} );
      $cfg_ref->{svc_host}->{$id} = \%record;
    }
  }
}

sub GetVdiskHostMapDataXml {
  my @a;
  foreach my $n ( @{$cfgdata} ) {
    if ( $n->{'type'} eq "vdiskhostmap" ) {
      my %record;
      my $i = 0;
      foreach my $m ( @{ $n->{'property'} } ) {
        $record{ $m->{'name'} } = $m->{'value'};
      }
      my $id = sprintf( "%04d", $record{vdisk_id} );
      $record{vdisk_id} = $id;
      push( @a, \%record );
    }
  }
  $cfg_ref->{svc_vdiskhostmap} = \@a;
}

# -------------------------------------------------------------------------------------------------

sub PrintSystemData {
  print CONFOUT "\nConfiguration Data\n";
  print CONFOUT "------------------\n";
  print CONFOUT "\tMachine Name:                    ", $cfg_ref->{svc_system}->{name},        "\n";
  print CONFOUT "\tMachine Type-Model:              ", &MachineTypeModel,                     "\n";
  print CONFOUT "\tMachine Serial:                  ", &MachineSerial,                        "\n";
  print CONFOUT "\tCode Level:                      ", $cfg_ref->{svc_system}->{code_level},, "\n";
  if ( defined $cfg_ref->{svc_system}->{'product_name'} ) {
    print CONFOUT "\tProduct Name:                    ", $cfg_ref->{svc_system}->{'product_name'}, "\n";
  }
  print CONFOUT "\n";
  print CONFOUT "\tPerformance Statistic Status:    ", $cfg_ref->{svc_system}->{statistics_status},,    "\n";
  print CONFOUT "\tPerformance Statistic Frequency: ", $cfg_ref->{svc_system}->{statistics_frequency},, "\n";
  print CONFOUT "\n";
  print CONFOUT "\tDate:           ", $cfg_ref->{date}->{'cfg_date'}, "\n";
  print CONFOUT "\n";
  print CONFOUT "\tTotal Mdisk Capacity:            ", $cfg_ref->{svc_system}->{total_mdisk_capacity},      "\n";
  print CONFOUT "\tSpace in Pools:                  ", $cfg_ref->{svc_system}->{space_in_mdisk_grps},       "\n";
  print CONFOUT "\tSpace Allocated to Volumes:      ", $cfg_ref->{svc_system}->{space_allocated_to_vdisks}, "\n";
  print CONFOUT "\n";
  print CONFOUT "\tTotal Free_space:                ", $cfg_ref->{svc_system}->{total_free_space},                "\n";
  print CONFOUT "\tTotal Used_capacity:             ", $cfg_ref->{svc_system}->{total_used_capacity},             "\n";
  print CONFOUT "\tTotal Overallocation:            ", $cfg_ref->{svc_system}->{total_overallocation},            "\n";
  print CONFOUT "\tTotal Volume Capacity:           ", $cfg_ref->{svc_system}->{total_vdisk_capacity},            "\n";
  print CONFOUT "\tTotal Volumecopy Capacity:       ", $cfg_ref->{svc_system}->{total_vdiskcopy_capacity},        "\n";
  print CONFOUT "\tTotal Allocated Extent Capacity: ", $cfg_ref->{svc_system}->{total_allocated_extent_capacity}, "\n";
  print CONFOUT "\n";

  if ( defined $cfg_ref->{svc_system}->{scm_tier_capacity} ) {
    printf CONFOUT ( "\t%-30s\t%12s\t%12s\t%12s\t%12s\t%12s\n", "    ", "     ", "          ", "Nearline", "High Capacity", "Storage     " );
    printf CONFOUT ( "\t%-30s\t%12s\t%12s\t%12s\t%12s\t%12s\n", "    ", "Flash", "Enterprise", "HDD     ", "Flash        ", "Class Memory" );
    printf CONFOUT (
      "\t%-30s\t%12s\t%12s\t%12s\t%12s\t%12s\n",        "Tier Capacity:",
      $cfg_ref->{svc_system}->{ssd_tier_capacity},      $cfg_ref->{svc_system}->{enterprise_tier_capacity},
      $cfg_ref->{svc_system}->{nearline_tier_capacity}, $cfg_ref->{svc_system}->{ssdnl_tier_capacity},
      $cfg_ref->{svc_system}->{scm_tier_capacity}
    );
    printf CONFOUT (
      "\t%-30s\t%12s\t%12s\t%12s\t%12s\t%12s\n",             "Tier Free Capacity:",
      $cfg_ref->{svc_system}->{ssd_tier_free_capacity},      $cfg_ref->{svc_system}->{enterprise_tier_free_capacity},
      $cfg_ref->{svc_system}->{nearline_tier_free_capacity}, $cfg_ref->{svc_system}->{ssdnl_tier_free_capacity},
      $cfg_ref->{svc_system}->{scm_tier_free_capacity}
    );
  }
  elsif ( defined $cfg_ref->{svc_system}->{ssd_tier_capacity} ) {
    print CONFOUT "\t                                 Flash  \tEnterprise     \tNearline_HDD     \tHigh_Capacity_Flash\n";
    print CONFOUT "\tTier Capacity:                   ", $cfg_ref->{svc_system}->{ssd_tier_capacity},      "    \t", $cfg_ref->{svc_system}->{enterprise_tier_capacity},      "    \t", $cfg_ref->{svc_system}->{nearline_tier_capacity},      "    \t", $cfg_ref->{svc_system}->{ssdnl_tier_capacity},      "\n";
    print CONFOUT "\tTier Free Capacity:              ", $cfg_ref->{svc_system}->{ssd_tier_free_capacity}, "    \t", $cfg_ref->{svc_system}->{enterprise_tier_free_capacity}, "    \t", $cfg_ref->{svc_system}->{nearline_tier_free_capacity}, "    \t", $cfg_ref->{svc_system}->{ssdnl_tier_free_capacity}, "\n";
  }
  print CONFOUT "\n";
}

sub PrintNodeData {
  my $delim = ",";

  # Add product_mtm
  # my @keylist = split(/,/,"id,name,UPS_serial_number,WWNN,status,IO_group_id,IO_group_name,config_node,UPS_unique_id,hardware,iscsi_name,iscsi_alias,panel_name,enclosure_id,canister_id,enclosure_serial_number");
  my @keylist = split( /,/, "id,name,UPS_serial_number,WWNN,status,IO_group_id,IO_group_name,config_node,UPS_unique_id,hardware,iscsi_name,iscsi_alias,panel_name,enclosure_id,canister_id,product_mtm,enclosure_serial_number" );
  my $hdrline = join( $delim, @keylist );

  print CONFOUT "\nNode Level Configuration\n";
  print CONFOUT "------------------------\n";
  print CONFOUT $hdrline, "\n";

  my @reclist = keys( %{ $cfg_ref->{svc_node} } );
  foreach my $rec ( sort(@reclist) ) {
    my @a;
    foreach my $key (@keylist) {
      if ( defined $cfg_ref->{svc_node}->{$rec}->{$key} ) {
        push( @a, $cfg_ref->{svc_node}->{$rec}->{$key} );
      }
      else {
        push( @a, "" );
      }
    }
    print CONFOUT join( $delim, @a ), "\n";
  }
  print CONFOUT "\n";
}

sub MachineTypeModel {
  my @a;
  if ( defined $cfg_ref->{svc_enclosure} ) {
    my @reclist = keys( %{ $cfg_ref->{svc_enclosure} } );
    foreach my $rec ( sort(@reclist) ) {
      if ( $cfg_ref->{svc_enclosure}->{$rec}->{type} eq "control" ) {
        push( @a, $cfg_ref->{svc_enclosure}->{$rec}->{product_MTM} );
      }
    }
  }
  else {
    my @reclist = keys( %{ $cfg_ref->{svc_node} } );
    foreach my $rec ( sort(@reclist) ) {
      if ( defined $cfg_ref->{svc_node}->{$rec}->{product_mtm} ) {
        push( @a, $cfg_ref->{svc_node}->{$rec}->{product_mtm} );
      }
      else {
        push( @a, "2145-" . $cfg_ref->{svc_node}->{$rec}->{hardware} );
      }

    }
  }
  return ( join( ",", @a ) );
}

sub MachineSerial {
  my @a;
  if ( defined $cfg_ref->{svc_enclosure} ) {
    my @reclist = keys( %{ $cfg_ref->{svc_enclosure} } );
    foreach my $rec ( sort(@reclist) ) {
      if ( $cfg_ref->{svc_enclosure}->{$rec}->{type} eq "control" ) {
        push( @a, $cfg_ref->{svc_enclosure}->{$rec}->{serial_number} );
      }
    }
  }
  else {
    my @reclist = keys( %{ $cfg_ref->{svc_node} } );
    foreach my $rec ( sort(@reclist) ) {
      if ( defined $cfg_ref->{svc_node}->{$rec}->{serial_number} ) {
        push( @a, $cfg_ref->{svc_node}->{$rec}->{serial_number} );
      }
      elsif ( defined $cfg_ref->{svc_node}->{$rec}->{enclosure_serial_number} ) {
        push( @a, $cfg_ref->{svc_node}->{$rec}->{enclosure_serial_number} );
      }
    }
  }
  if (@a) {
    return ( join( ",", @a ) );
  }
  else {
    return;
  }
}

sub PrintEnclosureData {
  if ( not defined $cfg_ref->{svc_enclosure} ) { return; }
  my $delim   = ",";
  my @keylist = split( /,/, "id,status,type,managed,IO_group_id,IO_group_name,product_MTM,serial_number,total_canisters,online_canisters,total_PSUs,online_PSUs,drive_slots,machine_part_number" );
  my $hdrline = join( $delim, @keylist );

  print CONFOUT "\nEnclosure Level Configuration(V7000 Series)\n";
  print CONFOUT "-------------------------------------------\n";
  print CONFOUT $hdrline, "\n";

  my @reclist = keys( %{ $cfg_ref->{svc_enclosure} } );
  foreach my $rec ( sort(@reclist) ) {
    my @a;
    foreach my $key (@keylist) {
      if ( defined $cfg_ref->{svc_enclosure}->{$rec}->{$key} ) {
        push( @a, $cfg_ref->{svc_enclosure}->{$rec}->{$key} );
      }
      else {
        push( @a, "" );
      }
    }
    print CONFOUT join( $delim, @a ), "\n";
  }
  print CONFOUT "\n";
}

sub PrintPortfcData {
  my $delim = ",";
  my @keylist;
  if ( $cfg_ref->{'info'}->{'is_fs840'} ) {
    @keylist = split( /,/, "id,adapter_id,port_id,type,port_speed,node_id,node_name,WWPN,nportid,status,attachment" );
  }
  else {
    @keylist = split( /,/, "id,fc_io_port_id,port_id,type,port_speed,node_id,node_name,WWPN,nportid,status,attachment" );
  }
  my $hdrline = join( $delim, @keylist );

  print CONFOUT "\nPort Level Configuration - Fibre Channel Ports\n";
  print CONFOUT "-----------------------------------------------\n";
  print CONFOUT "port_name", $delim, $hdrline, "\n";

  my @reclist = keys( %{ $cfg_ref->{svc_portfc} } );
  foreach my $rec ( sort(@reclist) ) {
    my @a;
    push( @a, $rec );
    foreach my $key (@keylist) {
      if ( defined $cfg_ref->{svc_portfc}->{$rec}->{$key} ) {
        push( @a, $cfg_ref->{svc_portfc}->{$rec}->{$key} );
      }
      else {
        push( @a, "" );
      }
    }
    print CONFOUT join( $delim, @a ), "\n";
  }
  print CONFOUT "\n";
}

sub PrintDriveData {
  my $delim   = ",";
  my @keylist = split( /,/, "id,status,error_sequence_number,use,tech_type,capacity,mdisk_id,mdisk_name,member_id,enclosure_id,slot_id,node_id,node_name,transport_protocol,compressed,physical_capacity,physical_used_capacity,effective_used_capacity" );
  my $hdrline = join( $delim, @keylist );

  print CONFOUT "\nDrive Level Configuration\n";
  print CONFOUT "------------------------\n";
  print CONFOUT "drive_id", $delim, $hdrline, "\n";

  my @reclist = keys( %{ $cfg_ref->{svc_drive} } );
  foreach my $rec ( sort(@reclist) ) {
    my @a;
    push( @a, $rec );
    foreach my $key (@keylist) {
      if ( defined $cfg_ref->{svc_drive}->{$rec}->{$key} ) {
        push( @a, $cfg_ref->{svc_drive}->{$rec}->{$key} );
      }
      else {
        push( @a, "" );
      }
    }
    print CONFOUT join( $delim, @a ), "\n";
  }
  print CONFOUT "\n";
}

sub PrintMdiskData {
  my $delim   = ",";
  my @keylist = split( /,/, "id,name,status,mode,mdisk_grp_id,mdisk_grp_name,capacity,ctrl_LUN_#,controller_name,UID,tier,physical_capacity,physical_free_capacity,allocated_capacity,effective_used_capacity,dedupe,over_provisioned,hw_compressed," );
  my $hdrline = join( $delim, @keylist );

  print CONFOUT "\nManaged Disk Level Configuration\n";
  print CONFOUT "------------------------\n";
  print CONFOUT "mdisk_id", $delim, $hdrline, "\n";

  my @reclist = keys( %{ $cfg_ref->{svc_mdsk} } );
  foreach my $rec ( sort(@reclist) ) {
    my @a;
    push( @a, $rec );
    foreach my $key (@keylist) {
      if ( defined $cfg_ref->{svc_mdsk}->{$rec}->{$key} ) {
        push( @a, $cfg_ref->{svc_mdsk}->{$rec}->{$key} );
      }
      else {
        push( @a, "" );
      }
    }
    print CONFOUT join( $delim, @a ), "\n";
  }
  print CONFOUT "\n";
}

sub PrintMdiskGrpData {
  my $delim   = ",";
  my @keylist = split( /,/, "name,id,status,mdisk_count,vdisk_count,capacity,extent_size,free_capacity,virtual_capacity,used_capacity,real_capacity,overallocation,warning,easy_tier,easy_tier_status,compression_active,compression_virtual_capacity,compression_compressed_capacity,compression_uncompressed_capacity,parent_mdisk_grp_name,child_mdisk_grp_count,child_mdisk_grp_capacity,type,reclaimable_capacity,physical_capacity,physical_free_capacity,effective_used_capacity,hw_compressed" );
  my $hdrline = join( $delim, @keylist );

  print CONFOUT "\nPool Level Configuration\n";
  print CONFOUT "------------------------\n";
  print CONFOUT $hdrline, "\n";

  my @reclist = keys( %{ $cfg_ref->{svc_pool} } );
  foreach my $rec ( sort(@reclist) ) {
    my @a;
    foreach my $key (@keylist) {
      if ( defined $cfg_ref->{svc_pool}->{$rec}->{$key} ) {
        push( @a, $cfg_ref->{svc_pool}->{$rec}->{$key} );
      }
      else {
        push( @a, "" );
      }
    }
    print CONFOUT join( $delim, @a ), "\n";
  }
  print CONFOUT "\n";
}

sub PrintVdiskData {
  my $delim   = ",";
  my @keylist = split( /,/, "id,name,IO_group_id,IO_group_name,status,mdisk_grp_id,mdisk_grp_name,capacity,type,FC_id,FC_name,RC_id,RC_name,vdisk_UID,fc_map_count,copy_count,fast_write_state,se_copy_count,RC_change,compressed_copy_count" );
  my $hdrline = join( $delim, @keylist );

  print CONFOUT "\nVolume Level Configuration\n";
  print CONFOUT "--------------------------\n";
  print CONFOUT "volume_id", $delim, $hdrline, "\n";

  my @reclist = keys( %{ $cfg_ref->{svc_vdsk} } );
  foreach my $rec ( sort(@reclist) ) {
    my @a;
    push( @a, $rec );
    foreach my $key (@keylist) {
      if ( $key eq "mdisk_grp_id" or $key eq "mdisk_grp_name" ) {
        my @b;
        if ( not defined $cfg_ref->{svc_vdsk}->{$rec}->{$key} ) {
          if ( defined $cfg_ref->{svc_vdsk}->{$rec}->{ "copy_0_" . $key } ) {
            push( @b, $cfg_ref->{svc_vdsk}->{$rec}->{ "copy_0_" . $key } );
          }
          if ( defined $cfg_ref->{svc_vdsk}->{$rec}->{ "copy_1_" . $key } ) {
            push( @b, $cfg_ref->{svc_vdsk}->{$rec}->{ "copy_1_" . $key } );
          }
          if ( scalar(@b) ) {
            push( @a, join( " ", @b ) );
          }
          else {
            push( @a, "" );
          }
        }
        elsif ( $cfg_ref->{svc_vdsk}->{$rec}->{$key} eq "many" ) {
          if ( defined $cfg_ref->{svc_vdsk}->{$rec}->{ "copy_0_" . $key } ) {
            push( @b, $cfg_ref->{svc_vdsk}->{$rec}->{ "copy_0_" . $key } );
          }
          if ( defined $cfg_ref->{svc_vdsk}->{$rec}->{ "copy_1_" . $key } ) {
            push( @b, $cfg_ref->{svc_vdsk}->{$rec}->{ "copy_1_" . $key } );
          }
          if ( scalar(@b) ) {
            push( @a, join( " ", @b ) );
          }
          else {
            push( @a, "" );
          }
        }
        else {
          push( @a, $cfg_ref->{svc_vdsk}->{$rec}->{$key} );
        }
      }
      elsif ( defined $cfg_ref->{svc_vdsk}->{$rec}->{$key} ) {
        push( @a, $cfg_ref->{svc_vdsk}->{$rec}->{$key} );
      }
      else {
        push( @a, "" );
      }
    }
    print CONFOUT join( $delim, @a ), "\n";
  }
  print CONFOUT "\n";
}

sub PrintHostData {
  my $delim   = ",";
  my @keylist = split( /,/, "id,name,port_count,iogrp_count,status" );
  my $hdrline = join( $delim, @keylist ) . $delim . "IQN WWPN" . $delim . "Volume IDs" . $delim . "Volume Names" . $delim . "host_cluster_id" . $delim . "host_cluster_name";

  print CONFOUT "\nHost Level Configuration\n";
  print CONFOUT "--------------------------\n";
  print CONFOUT "host_id", $delim, $hdrline, "\n";

  my @reclist = keys( %{ $cfg_ref->{svc_host} } );
  foreach my $rec ( sort(@reclist) ) {
    my @a;
    my @w;
    my @d;
    my @n;
    push( @a, $rec );
    foreach my $key (@keylist) {
      if ( defined $cfg_ref->{svc_host}->{$rec}->{$key} ) {
        push( @a, $cfg_ref->{svc_host}->{$rec}->{$key} );
      }
      else {
        push( @a, "" );
      }
    }

    # Add IQN and WWPNs
    my $i = 0;
    while ( $i < $cfg_ref->{svc_host}->{$rec}->{port_count} ) {
      if ( $cfg_ref->{svc_host}->{$rec}->{'iscsi_name'} ) {
        push( @w, $cfg_ref->{svc_host}->{$rec}->{'iscsi_name'} );
      }
      elsif ( $cfg_ref->{svc_host}->{$rec}->{ "FC" . $i . "_WWPN" } ) {
        push( @w, $cfg_ref->{svc_host}->{$rec}->{ "FC" . $i . "_WWPN" } );
      }
      $i++;
    }
    push( @a, join( " ", @w ) );

    # Add Vdisk IDs
    foreach my $map ( @{ $cfg_ref->{svc_vdiskhostmap} } ) {
      if ( $cfg_ref->{svc_host}->{$rec}->{'id'} eq $map->{'host_id'} ) {
        push( @d, $map->{'vdisk_id'} );
      }
    }
    @d = sort(@d);
    push( @a, join( " ", @d ) );

    # Add Vdisk Names
    foreach my $id (@d) {
      $id = sprintf( "%04d", $id );
      push( @n, $cfg_ref->{svc_vdsk}->{$id}->{'name'} );
    }
    push( @a, join( " ", @n ) );

    # Add Host Cluster
    if ( defined $cfg_ref->{svc_host}->{$rec}->{'host_cluster_id'} ) {
      push( @a, $cfg_ref->{svc_host}->{$rec}->{'host_cluster_id'}, $cfg_ref->{svc_host}->{$rec}->{'host_cluster_name'} );
    }
    else {
      push( @a, "", "" );
    }

    print CONFOUT join( $delim, @a ), "\n";
  }
  print CONFOUT "\n";
}

sub PrintRcrelData {
  my $delim   = ",";
  my @keylist = split( /,/, "id,name,master_cluster_id,master_cluster_name,master_vdisk_id,master_vdisk_name,aux_cluster_id,aux_cluster_name,aux_vdisk_id,aux_vdisk_name,primary,consistency_group_id,consistency_group_name,state,bg_copy_priority,progress,copy_type,cycling_mode,freeze_time" );
  my $hdrline = join( $delim, @keylist );

  print CONFOUT "\nRC Relationship Level Configuration\n";
  print CONFOUT "------------------------\n";
  print CONFOUT $hdrline, "\n";

  my @reclist = keys( %{ $cfg_ref->{svc_rcrel} } );
  foreach my $rec ( sort(@reclist) ) {
    my @a;
    foreach my $key (@keylist) {
      if ( defined $cfg_ref->{svc_rcrel}->{$rec}->{$key} ) {
        push( @a, $cfg_ref->{svc_rcrel}->{$rec}->{$key} );
      }
      else {
        push( @a, "" );
      }
    }
    print CONFOUT join( $delim, @a ), "\n";
  }
  print CONFOUT "\n";
}

sub PrintFcmapData {
  my $delim   = ",";
  my @keylist = split( /,/, "id,name,source_vdisk_id,source_vdisk_name,target_vdisk_id,target_vdisk_name,group_id,group_name,status,progress,copy_rate,clean_progress,incremental,partner_FC_id,partner_FC_name,restoring,start_time,rc_controlled" );
  my $hdrline = join( $delim, @keylist );

  print CONFOUT "\nFC Map Level Configuration\n";
  print CONFOUT "------------------------\n";
  print CONFOUT $hdrline, "\n";

  my @reclist = keys( %{ $cfg_ref->{svc_fcmap} } );
  foreach my $rec ( sort(@reclist) ) {
    my @a;

    # Suppress list of the FC Maps for GlobalMirror w/Change Volumes
    if ( $cfg_ref->{svc_fcmap}->{$rec}->{'rc_controlled'} eq "yes" ) { next; }
    foreach my $key (@keylist) {
      if ( defined $cfg_ref->{svc_fcmap}->{$rec}->{$key} ) {
        push( @a, $cfg_ref->{svc_fcmap}->{$rec}->{$key} );
      }
      else {
        push( @a, "" );
      }
    }
    print CONFOUT join( $delim, @a ), "\n";
  }
  print CONFOUT "\n";
}

sub PrintEventData {
  my $delim   = ",";
  my @keylist = split( /,/, "sequence_number,last_timestamp,object_type,object_id,object_name,copy_id,status,fixed,event_id,error_code,description" );
  my $hdrline = join( $delim, @keylist );

  print CONFOUT "\nEvent Log Level\n";
  print CONFOUT "------------------------\n";
  print CONFOUT "event_id", $delim, $hdrline, "\n";

  my @reclist = keys( %{ $cfg_ref->{svc_event} } );
  foreach my $rec ( sort(@reclist) ) {
    my @a;
    push( @a, $rec );
    foreach my $key (@keylist) {
      if ( defined $cfg_ref->{svc_event}->{$rec}->{$key} ) {
        push( @a, $cfg_ref->{svc_event}->{$rec}->{$key} );
      }
      else {
        push( @a, "" );
      }
    }
    print CONFOUT join( $delim, @a ), "\n";
  }
  print CONFOUT "\n";
}

sub PrintFabricData {
  my $delim = ",";
  #
  my @keylist;
  if ( $cfg_ref->{info}->{is_fs840} ) {
    @keylist = split( /:/, "remote_uid:local_uid:canister_id:adapter_id:port_id:node_id:node_name:type" );
  }
  else {
    @keylist = split( /:/, "remote_wwpn:remote_nportid:id:node_name:local_wwpn:local_port:local_nportid:state:name:cluster_name:type" );
  }
  my $hdrline = join( $delim, @keylist );

  print CONFOUT "\nFabric Level Configuration\n";
  print CONFOUT "------------------------\n";
  print CONFOUT $hdrline, "\n";

  foreach my $rec ( @{ $cfg_ref->{svc_fabric} } ) {
    my @a;
    foreach my $key (@keylist) {
      if ( defined $rec->{$key} ) {
        push( @a, $rec->{$key} );
      }
      else {
        push( @a, "" );
      }
    }
    print CONFOUT join( $delim, @a ), "\n";
  }
  print CONFOUT "\n";
}

sub error_die {
  my $message  = shift;
  my $act_time = localtime();
  print STDERR "$act_time: $message : $!\n";
  exit(1);
}

sub remove_whitespace {
  my $value = shift;
  $value =~ s/^\s+|\s+$//g;
  return $value;

}

sub get_array_data {
  my $file = shift;
  open( FH, "< $file" ) || error_die( "Cannot read $file: $!" . __FILE__ . ":" . __LINE__ );
  my @file_all = <FH>;
  close(FH);
  return @file_all;
}

sub get_header_index {
  my $line   = shift;
  my @header = split( ",", $line );
  my $index  = -1;
  my %header_index;

  foreach my $element (@header) {
    chomp $element;
    $index++;
    if ( $element eq "" ) { next; }
    $element = remove_whitespace($element);
    if ( defined $header_index{$element} ) { next; }
    $header_index{$element} = $index;
  }
  return %header_index;

}

sub set_data_by_header_array {
  my ( $data_line_tmp, $header_index_tmp, $health_status_tmp, $section ) = @_;
  my @data_line     = @{$data_line_tmp};
  my %header_index  = %{$header_index_tmp};
  my %health_status = %{$health_status_tmp};
  my $id            = "";
  if ( defined $header_index{event_id} && $section eq "EVENT" ) {
    $id = $data_line[ $header_index{event_id} ];
  }
  else {
    if ( defined $data_line[0] && $data_line[0] ne "" ) {
      $id = $data_line[0];
    }
    else {
      return;
    }
  }
  foreach my $element ( keys %header_index ) {
    my $value = $data_line[ $header_index{$element} ];

    #$value =~ s/,/ /g;
    my $name_property = $element;
    $health_status{$section}{$id}{$name_property} = $value;
  }
  return %health_status;
}

sub get_event_log {
  my $conf_file = shift;
  my %health_status;
  my $active = 0;
  my %header_index;
  my $section = "";

  if ( !( -e $conf_file ) && -z $conf_file ) {
    return %health_status;
  }
  my @data = get_array_data($conf_file);

  foreach my $line (@data) {
    chomp $line;
    if ( $line eq "" ) {
      $active = 0;
    }
    if ( $line eq "" || $line =~ m/-----------|Fabric Level Configuration/ ) { next; }
    $line = remove_whitespace($line);
    if ( $line =~ m/^Event Log level/i ) {
      $active = 1;
      next;
    }
    if ( $line =~ m/Level Configuration/ && $line !~ m/RC Relationship|FC Map/ ) {
      $active  = 3;
      $section = $line;
      $section =~ s/Level Configuration.*//g;
      $section = remove_whitespace($section);
      next;
    }
    if ( $active == 3 ) {
      %header_index = get_header_index($line);
      $active       = 4;
      next;
    }
    if ( $active == 4 && $section ne "" ) {
      my @data_line = split( ",", $line );
      $section_check++;
      if ( !defined $health_status_component{$section} ) {
        $health_status_component{$section} = 0;
      }
      $health_status_component{$section} = $health_status_component{$section} + 1;
      %health_status = set_data_by_header_array( \@data_line, \%header_index, \%health_status, $section );
      next;
    }
    if ( $active == 1 ) {
      %header_index = get_header_index($line);
      $active       = 2;
      next;
    }
    if ( $active == 2 ) {
      my @data_line = split( ",", $line );
      $event_check++;
      $section       = "EVENT";
      %health_status = set_data_by_header_array( \@data_line, \%header_index, \%health_status, $section );
      $section       = "";
      next;
    }
  }

  return %health_status;

}

sub get_health_status_html {
  my $table = "<table class =\"tabcfgsum tablesorter tablesortercfgsum\">\n";
  my $thead = "<thead><tr><th class = \"sortable\" title=\"Date/Time\" nowrap=\"\">Date/Time</th>\n";
  $thead = $thead . "<th class = \"sortable\" title=\"event_id\" nowrap=\"\">event_id</th>\n";
  $thead = $thead . "<th class = \"sortable\" title=\"object_type\" nowrap=\"\">object_type</th>\n";
  $thead = $thead . "<th class = \"sortable\" title=\"object_name\" nowrap=\"\">object_name</th>\n";
  $thead = $thead . "<th class = \"sortable\" title=\"description\" nowrap=\"\">description</th>\n";
  $thead = $thead . "<th class = \"sortable\" title=\"status\" nowrap=\"\">status</th>\n";
  $thead = $thead . "<th class = \"sortable\" title=\"fixed\" nowrap=\"\">fixed</th>\n";
  $thead = $thead . "<th class = \"sortable\" title=\"seq_number\" nowrap=\"\">seq_number</th>\n";
  $thead = $thead . "<th class = \"sortable\" title=\"object_id\" nowrap=\"\">object_id</th>\n";
  $thead = $thead . "<th class = \"sortable\" title=\"copy_id\" nowrap=\"\">copy_id</th>\n";
  $thead = $thead . "<th class = \"sortable\" title=\"error_code\" nowrap=\"\">error_code</th>\n";
  $thead = $thead . "</tr>\n</thead>\n";
  my $tbody = "<tbody>\n";

  foreach my $id ( sort keys %{ $health_status{EVENT} } ) {
    my $tr = "<tr>";

    my $property = "last_timestamp";
    my $td       = "";
    $td = get_value_to_table( $id, $property, $td, "EVENT" );

    $property = "event_id";
    $td       = get_value_to_table( $id, $property, $td,, "EVENT" );

    $property = "object_type";
    $td       = get_value_to_table( $id, $property, $td, "EVENT" );

    $property = "object_name";
    $td       = get_value_to_table( $id, $property, $td, "EVENT" );

    $property = "description";
    $td       = get_value_to_table( $id, $property, $td, "EVENT" );

    $property = "status";
    $td       = get_value_to_table( $id, $property, $td, "EVENT" );

    $property = "fixed";
    $td       = get_value_to_table( $id, $property, $td, "EVENT" );

    $property = "sequence_number";
    $td       = get_value_to_table( $id, $property, $td, "EVENT" );

    $property = "object_id";
    $td       = get_value_to_table( $id, $property, $td, "EVENT" );

    $property = "copy_id";
    $td       = get_value_to_table( $id, $property, $td, "EVENT" );

    $property = "error_code";
    $td       = get_value_to_table( $id, $property, $td, "EVENT" );

    my $val   = $td;
    my $index = 0;
    my $line  = "";
    my @vals  = split( "\n", $val );
    my $size  = scalar @vals;
    foreach my $element (@vals) {
      $index++;
      $element = remove_whitespace($element);
      if ( $element eq "" && $index != $size ) {
        $line = $line . " : ";
        next;
      }
      $element =~ s/<.*">//g;
      $element =~ s/<\/td>//g;
      $element =~ s/<td>//g;
      $element = remove_whitespace($element);
      if ( $line eq "" ) {
        $line = $element;
      }
      else {
        $line = $line . " : " . $element;
      }
      if ( $index == $size && $element eq "" ) {
        $line = "";
        last;
      }
    }
    if ( $line ne "" ) {
      push( @error_message, "$line\n" );
    }

    $tr    = $tr . $td . "</tr>\n";
    $tbody = $tbody . "$tr\n";

  }
  $table = $table . $thead . "$tbody</tbody>\n" . "</table><br><br><br>\n";

  my $table_section = "<table class =\"tabcfgsum tablesorter tablesortercfgsum\">\n";
  $thead = "<thead><tr><th class = \"sortable\" title=\"section\" nowrap=\"\">section</th>\n";
  $thead = $thead . "<th class = \"sortable\" title=\"id\" nowrap=\"\">id</th>\n";
  $thead = $thead . "<th class = \"sortable\" title=\"name\" nowrap=\"\">name</th>\n";
  $thead = $thead . "<th class = \"sortable\" title=\"status\" nowrap=\"\">status</th>\n";
  $thead = $thead . "<th class = \"sortable\" title=\"description\" nowrap=\"\">description</th>\n";
  $thead = $thead . "</tr>\n</thead>\n";
  $tbody = "<tbody>\n";

  my $row_element = 0;
  #print Dumper \%health_status;
  foreach my $section ( keys %health_status ) {
    if ( $section eq "EVENT" || $section eq "Volume") { next; }
    foreach my $id ( keys %{ $health_status{$section} } ) {
      if ( defined $health_status{$section}{$id}{"status"} && $health_status{$section}{$id}{"status"} =~ m/online|degraded|active|inactive_unconfigured|degraded_paths|degraded_ports|spare|^$/i ) {
        next;    ### we show only rows with status error
      }
      $row_element++;
      my $td       = "";
      my $tr       = "<tr>";
      my $property = "section";
      $td = get_value_to_table( $id, $property, $td, $section );

      $property = "id";
      $td       = get_value_to_table( $id, $property, $td, $section );

      $property = "name";
      $td       = get_value_to_table( $id, $property, $td, $section );

      $property = "status";
      $td       = get_value_to_table( $id, $property, $td, $section );

      $property = "description";
      $td       = get_value_to_table( $id, $property, $td, $section );

      my $val   = $td;
      my $index = 0;
      my $line  = "";
      my @vals  = split( "\n", $val );
      my $size  = scalar @vals;

      my $type   = "";
      my $status = "";

      foreach my $element (@vals) {
        $index++;
        $element = remove_whitespace($element);
        if ( $element eq "" && $index != $size ) {
          $line = $line . " : ";
          next;
        }
        $element =~ s/<.*">//g;
        $element =~ s/<\/td>//g;
        $element =~ s/<td>//g;
        $element = remove_whitespace($element);
        if ( $line eq "" ) {
          $line = $element;
        }
        else {
          $line = $line . " : " . $element;
        }

        if ( $index == 1 ) { $type   = $element; }
        if ( $index == 4 ) { $status = $element; }
      }
      if ( $type eq "Host" && $status eq "offline" ) {
        warning("Host offline status skipped: $line");
      }
      elsif ( $line =~ m/Unable to send mail through the SMTP/i ) {
        warning("Alert skipped: $line");
      }
      else {
        # KZ: some hosts regularly report offline status around 2 am, do not report it as an error
        $error_status++;
        push( @error_message, "$line\n" );
      }

      $tr    = $tr . $td . "</tr>\n";
      $tbody = $tbody . "$tr\n";
    }
  }
  $table_section = $table_section . $thead . "$tbody</tbody>\n" . "</table>\n";
  if ( $row_element == 0 ) { $table_section = ""; }

  my $report = "";
  my $time   = localtime();
  $time = $time . "<br>\n";
  if ( $error_status == 0 && $row_element == 0 ) {
    $report = "Storage $storage health status = Optimal\n";
  }
  else {
    $report = "Storage $storage health status = Error\n";
  }
  $table = $time . $report . $table . "$table_section\n";

  return $table;

}

sub get_value_to_table {
  my $id       = shift;
  my $property = shift;
  my $td       = shift;
  my $section  = shift;

  if ( $property eq "description" ) {
    $td = $td . "<td>";
  }
  else {
    $td = $td . "<td nowrap=\"\">";
  }
  if ( defined $health_status{$section}{$id}{$property} ) {
    my $value = $health_status{$section}{$id}{$property};
    if ( $value =~ m/undef/i ) { $value = ""; }
    if ( $property eq "last_timestamp" && $value ne "" ) {
      my @pom = split( "", $value );
      $value = "$pom[2]$pom[3]/$pom[4]$pom[5]/20$pom[0]$pom[1] $pom[6]$pom[7]:$pom[8]$pom[9]:$pom[10]$pom[11]";
    }
    if ( $property eq "error_code" && $value ne "" ) {
      my $description = "";
      if ( defined $health_status{$section}{$id}{description} ) {
        $description = $health_status{$section}{$id}{description};
      }
      if ( $description !~ m/Invalid LDAP configuration/i && $description !~ m/Unable to send mail through the SMTP/i ) {
        $error_status++;
      }
    }
    if ( $property eq "status" ) {
      if ( $section eq "Node" && $value =~ m/online/i ) {
        $value = "OK";

      }
      if ( $section eq "Node" && $value =~ m/offline|service/i ) {
        $value = "Error";
      }
      if ( $section eq "Node" && $value =~ m/flushing|pending|adding/i ) {
        $value = "Warning";
      }
      if ( $section eq "Enclosure" && $value =~ m/online/i ) {
        $value = "OK";
        $health_status{$section}{$id}{description} = "the canister is present and working normally";

      }
      if ( $section eq "Enclosure" && $value =~ m/degraded/i ) {
        $value = "Warning";
        $health_status{$section}{$id}{description} = "the canister is present but not working normally";
      }
      if ( $section eq "Enclosure" && $value =~ m/offline/i ) {
        $value = "Error";
        $health_status{$section}{$id}{description} = "the canister could not be detected";
      }
      if ( $section eq "Port" && $value =~ m/active|inactive_unconfigured/i ) {
        $value = "OK";

      }
      if ( $section eq "Port" && $value =~ m/inactive_configured/i ) {
        $value = "Error";
      }
      if ( $section eq "Drive" && $value =~ m/online/i ) {
        $value = "OK";

      }
      if ( $section eq "Drive" && $value =~ m/degraded/i ) {
        $value = "Warning";
      }
      if ( $section eq "Drive" && $value =~ m/offline/i ) {
        $value = "Error";
      }
      if ( $section eq "Managed Disk" && $value =~ m/online/i ) {
        $value = "OK";
        $health_status{$section}{$id}{description} = "The MDisk is online and available";

      }
      if ( $section eq "Managed Disk" && $value =~ m/degraded_paths|degraded_ports|degraded/i ) {
        $value = "Warning";
        if ( $value =~ m/degraded_paths/i ) {
          $health_status{$section}{$id}{description} = "One or more paths to the MDisk have been lost; the MDisk is not online to every node in the system";
        }
      }
      if ( $value =~ m/degraded_ports/i ) {
        $health_status{$section}{$id}{description} = "MDisk port errors";
      }
      if ( $value =~ m/degraded/i ) {
        $health_status{$section}{$id}{description} = "The array has members that are degraded, or the raid_status is degraded. (applies only to internal MDisks)";
      }
    }
    if ( $section eq "Managed Disk" && $value =~ m/offline|excluded/i ) {
      $value = "Error";
      if ( $value =~ m/offline/i ) {
        $health_status{$section}{$id}{description} = "All paths to the MDisk are lost";
      }
      if ( $value =~ m/excluded/i ) {
        $health_status{$section}{$id}{description} = "The MDisk is excluded from use by the system; the MDisk port error count exceeded the threshold";
      }
    }
    if ( $section eq "Pool" && $value =~ m/online/i ) {
      $value = "OK";
      $health_status{$section}{$id}{description} = "Indicates that the storage pool is online and available";
    }
    if ( $section eq "Pool" && $value =~ m/offline/i ) {
      $value = "Error";
      $health_status{$section}{$id}{description} = "Indicates that all paths to the storage pool are lost";
    }
    if ( $section eq "Volume" && $value =~ m/online/i ) {
      $value = "OK";

    }
    if ( $section eq "Volume" && $value =~ m/degraded/i ) {
      $value = "Warning";
    }
    if ( $section eq "Volume" && $value =~ m/offline/i ) {
      $value = "Error";
    }
    if ( $section eq "Host" && $value =~ m/online/i ) {
      $value = "OK";

    }
    if ( $section eq "Host" && $value =~ m/degraded/i ) {
      $value = "Warning";
      if ( $section eq "Host" && $value =~ m/offline/i ) {
        $value = "Error";
      }

    }
    $td = $td . "$value</td>\n";
  }
  else {
    if ( $property eq "section" ) {
      $td = $td . "$section</td>\n";
    }
    else {
      if ( $property eq "name" && $section eq "Port" && $health_status{$section}{$id}{port_name} ) {
        my $value = $health_status{$section}{$id}{port_name};
        $td = $td . "$value</td>\n";
      }
      else {
        $td = $td . "</td>\n";
      }
    }
  }
  return $td;

}

sub summary_health_status {
  my $timestamp    = time();
  my $time         = localtime();
  my $alert_status = 0;

  ### test
  #Xorux_lib::health_status_alerting($storage);

  if ( !-d "$summary_dir" ) {
    mkdir( "$summary_dir", 0755 ) || error( "$time: Cannot mkdir $summary_dir: $!" . __FILE__ . ":" . __LINE__ );
  }

  if ( !-f "$summary_dir/$storage.ok" && !-f "$summary_dir/$storage.nok" ) {
    $alert_status = 1;
  }

  if ( -f "$summary_dir/$storage.ok" ) { unlink("$summary_dir/$storage.ok"); $alert_status = 1; }
  if ( -f "$summary_dir/$storage.nok" ) {
    unlink("$summary_dir/$storage.nok");
    if ( $error_status == 0 ) {
      if ( exists &Xorux_lib::run_cmd_OV ) {
        Xorux_lib::run_cmd_OV( "normal", $storage );
      }
    }
  }
  if ( $error_status == 0 ) {
    my $report = "STORAGE : $storage : OK : $timestamp";
    write_health_status_html( "$summary_dir/$storage.ok", $report );
    return;
  }
  else {
    my $report = "STORAGE : $storage : NOT_OK : $timestamp";
    write_health_status_html( "$summary_dir/$storage.nok", $report );
    if ( $alert_status == 1 ) {
      ### alert health status
      if ( exists &Xorux_lib::health_status_alerting ) {
        Xorux_lib::health_status_alerting( $storage, \@error_message );
      }
      if ( exists &Xorux_lib::run_cmd_OV ) {
        Xorux_lib::run_cmd_OV( "error", $storage );
      }
    }
    return;
  }

}

sub hs_xormon {
  my %hs_data;
  my %hs_data_subsystem;

  if ( $error_status == 0 ) {
    $hs_data{status} = "ok";
    $hs_data{reason} = "Storage $storage health status = Optimal";
  }
  else {
    $hs_data{status} = "error";
    $hs_data{reason} = "Storage $storage health status = Error";

    if ( scalar(@error_message) > 0 ) {
      foreach (@error_message) {
        chomp $_;

        my @values = split(" : ", $_, -1);

        if ( scalar(@values) == scalar(@event_header) ) {
          my %event_hash;
          my $idx = 0;
          foreach my $value (@values) {
            $event_hash{ $event_header[$idx] } = $value;
            $idx++;
          }
          push( @{ $hs_data{events} }, \%event_hash);
        }
        elsif (scalar(@values) == scalar(@subsystem_header)){
          my %event_hash;
          my $idx = 0;
          foreach my $value (@values) {
            $event_hash{ $subsystem_header[$idx] } = $value;
            $idx++;
          }
          push( @{ $hs_data_subsystem{hs_data} }, \%event_hash);
        }
        else {
          error( "There is not the same count of the items as in the header! $!" . __FILE__ . ":" . __LINE__ );
          error( "header: ".join(" : ", @event_header) );
          error( "items: $_" );
        }
      }
    }
  }
  $hs_data{timestamp} = time();
  $hs_data{time}      = localtime();

  Xormon::save_as_json( "$dir/health_status.json", \%hs_data );
  Xormon::save_as_json( "$dir/health_status_subsystem.json", \%hs_data_subsystem);

  return 1;
}

sub write_health_status_html {
  my $file  = shift;
  my $table = shift;
  open( DATA, ">$file" ) or error("Cannot open file: $file : $!") && return 0;
  print DATA $table;
  close(DATA);

}

# -------------------------------------------------------------------------------------------------
# Configuration Data
sub CalculateData {
  &CalculateMdiskData();
  &CalculateMdiskGrpData();
}

sub CalculateMdiskData {

  # Add hw_compressed = <yes/no> to svc_mdsk entry
  my @reclist = keys( %{ $cfg_ref->{svc_drive} } );
  foreach my $rec ( sort(@reclist) ) {
    if ( defined $cfg_ref->{svc_drive}->{$rec}->{mdisk_id} && $cfg_ref->{svc_drive}->{$rec}->{mdisk_id} ne '' ) {
      my $mid = sprintf( "%04d", $cfg_ref->{svc_drive}->{$rec}->{mdisk_id} );
      if ( defined $cfg_ref->{svc_drive}->{$rec}->{compressed} && $cfg_ref->{svc_drive}->{$rec}->{compressed} eq 'yes' ) {
        $cfg_ref->{svc_mdsk}->{$mid}->{hw_compressed} = 'yes';
      }
      else {
        $cfg_ref->{svc_mdsk}->{$mid}->{hw_compressed} = 'no';
      }
    }
  }

}

sub CalculateMdiskGrpData {

  # $cfg_ref->{svc_pool}
  # Calculate effective_used_capacity for svc_pool
  my @reclist = keys( %{ $cfg_ref->{svc_mdsk} } );
  foreach my $rec ( sort(@reclist) ) {
    if ( $cfg_ref->{svc_mdsk}->{$rec}->{mdisk_grp_name} ne '' ) {
      my $pid = $cfg_ref->{svc_mdsk}->{$rec}->{mdisk_grp_name};
      if ( defined $cfg_ref->{svc_mdsk}->{$rec}->{hw_compressed} && $cfg_ref->{svc_mdsk}->{$rec}->{hw_compressed} eq 'yes' ) {
        $cfg_ref->{svc_pool}->{$pid}->{hw_compressed} = 'yes';
        if ( defined $cfg_ref->{svc_pool}->{$pid}->{effective_used_capacity} ) {
          $cfg_ref->{svc_pool}->{$pid}->{effective_used_capacity} += $cfg_ref->{svc_mdsk}->{$rec}->{effective_used_capacity};
        }
        else {
          $cfg_ref->{svc_pool}->{$pid}->{effective_used_capacity} = $cfg_ref->{svc_mdsk}->{$rec}->{effective_used_capacity};
        }
      }
    }
  }
}

sub ProductName {
  my $prod = shift;
  if ( defined $prod ) {

    # IBM Storwize ('product_name' => 'IBM Storwize V7000')
    return ("is_storwize") if ( $prod =~ m/^IBM\s+Storwize\s+V/ );

    # IBM Storwize ('product_name' => 'Lenovo Storage V3700 V2 XP')
    return ("is_storwize") if ( $prod =~ m/^Lenovo\s+Storage\s+V3700/ );

    # IBM FlashSystem ('product_name' => 'IBM SAN Volume Controller')
    return ("is_svc") if ( $prod =~ m/^IBM\s+SAN\s+Volume\s+Controller/ );

    # IBM FlashSystem V9000 ('product_name' => 'IBM FlashSystem V9000')
    return ("is_v9000") if ( $prod =~ m/^IBM\s+FlashSystem\s+V9000$/ );

    # IBM FlashSystem 9100 ('product_name' => 'IBM FlashSystem 9100')
    return ("is_fs9100") if ( $prod =~ m/^IBM\s+FlashSystem\s+9100$/ );

    # Inspur AS5600/5500/2600/2200 ('product_name' => 'Inspur AS2600G2') --> it uses this xml file name: mcs.config.cron.xml_9793365-2
    return ("is_inspur") if ( $prod =~ m/^Inspur\s+AS/ );

    # IBM FlashSystem 840/900
    return ("is_fs840") if ( $prod =~ m/^IBM\s+FlashSystem\s+840$/
      || $prod =~ m/^IBM\s+FlashSystem\s+900$/ );
    return ("is_unknown");
  }
  return (undef);
}

# -------------------------------------------------------------------------------------------------
# Capacity Data
sub CapCalculateData {
  &CapInitializePoolData;
  &CapCalculatePoolData1;
  &CapCalculateVolumeData;
  &CapCalculatePoolData2;
}

sub CapInitializePoolData {
  my %h;
  my $pool    = $capdata_ref->{'svc_pool'} = \%h;
  my @reclist = keys( %{ $cfg_ref->{'svc_pool'} } );
  foreach my $rec ( sort(@reclist) ) {
    my %i;
    $pool->{$rec} = \%i;
    my $tier_list = &PoolTierList( $cfg_ref->{'svc_pool'}->{$rec} );
    for my $tier_level ( @{$tier_list} ) {
      &InitializePoolDataTier( $pool->{$rec}, $cfg_ref->{'svc_pool'}->{$rec}, $tier_level );
    }
  }
}

sub CapCalculatePoolData1 {
  my $pool    = $capdata_ref->{'svc_pool'};
  my @reclist = keys( %{ $cfg_ref->{'svc_pool'} } );
  foreach my $rec ( sort(@reclist) ) {
    my $tier_list = &PoolTierList( $cfg_ref->{'svc_pool'}->{$rec} );
    for my $tier_level ( @{$tier_list} ) {
      if ( $tier_level ne "-" ) {
        &PoolHwCompressionTier( $pool->{$rec}, $cfg_ref->{'svc_pool'}->{$rec}, $tier_level );
        &CalculatePoolDataTier1( $pool->{$rec}, $cfg_ref->{'svc_pool'}->{$rec}, $tier_level );
      }
      $pool->{$rec}->{$tier_level}->{'type'} = $cfg_ref->{'svc_pool'}->{$rec}->{'type'};
    }

    # TierSum
    &CalculatePoolDataTier1( $pool->{$rec}, $cfg_ref->{'svc_pool'}->{$rec}, "-" );
  }
}

sub InitializePoolDataTier {
  my $pool = shift;
  my $cfg  = shift;
  my $tier = shift;
  my @keys = qw(
    raid_type
    raw_capacity
    raw_used_capacity
    spare_capacity
    raw_raid_capacity
    raw_raid_free_capacity
    raw_snapshot_used_capacity
    raw_replicated_used_capacity
    fmt_raid_capacity
    fmt_raid_free_capacity
    fmt_data_used_capacity
    fmt_snapshot_used_capacity
    fmt_replicated_used_capacity
    fmt_system_used_capacity
    raw_fmt_raid_capacity_koef
    raw_data_used_capacity
    raw_system_used_capacity
    thinprov_savings
    compress_savings
  );
  $pool->{$tier}->{'id'}   = $cfg->{'id'};
  $pool->{$tier}->{'name'} = $cfg->{'name'};
  #
  for my $key (@keys) {
    $pool->{$tier}->{$key} = 0;
  }

}

sub CalculatePoolDataTier1 {
  my $pool = shift;
  my $cfg  = shift;
  my $tier = shift;

  $pool->{$tier}->{'raid_type'} = &PoolRaidType( $cfg, $tier );
  #
  $pool->{$tier}->{'raw_capacity'}      = &PoolRawCapacity( $cfg, $tier );
  $pool->{$tier}->{'raw_used_capacity'} = &PoolRawUsedCapacity( $cfg, $tier );
  $pool->{$tier}->{'spare_capacity'}    = &PoolSpareCapacity( $cfg, $tier );
  #
  if ( $tier ne "-" && $pool->{$tier}->{'hw_compression'} eq "yes" ) {
    $pool->{$tier}->{'raw_raid_capacity'} = &PoolRawRaidCapacityHwc( $cfg, $tier );
  }
  else {
    $pool->{$tier}->{'raw_raid_capacity'} = &PoolRawRaidCapacity( $pool, $cfg, $tier );
  }
  $pool->{$tier}->{'raw_raid_free_capacity'} = &PoolRawRaidFreeCapacity( $cfg, $tier );
  #
  $pool->{$tier}->{'fmt_raid_capacity'}      = &PoolFmtRaidCapacity( $cfg, $tier );
  $pool->{$tier}->{'fmt_raid_free_capacity'} = &PoolFmtRaidFreeCapacity( $cfg, $tier );
  #
  if ( $pool->{$tier}->{'fmt_raid_capacity'} > 0 ) {
    $pool->{$tier}->{'raw_fmt_raid_capacity_koef'} = $pool->{$tier}->{'raw_raid_capacity'} / $pool->{$tier}->{'fmt_raid_capacity'};
  }
  else {
    $pool->{$tier}->{'raw_fmt_raid_capacity_koef'} = 0;
  }
  #
}

sub PoolHwCompressionTier {
  my $pool = shift;
  my $data = shift;
  my $tier = shift;

  $pool->{$tier}->{'hw_compression'} = 'no';
  my $pool_id = $data->{id};
  my @reclist = keys( %{ $cfg_ref->{'svc_mdsk'} } );
  foreach my $rec ( sort(@reclist) ) {
    if ( $cfg_ref->{'svc_mdsk'}->{$rec}->{'mdisk_grp_id'} eq $pool_id && $cfg_ref->{'svc_mdsk'}->{$rec}->{'mode'} eq "array" ) {
      if ( &MdiskHwCompression( $cfg_ref->{'svc_mdsk'}->{$rec}->{'id'}, $tier ) ) {
        $pool->{$tier}->{'hw_compression'} = 'yes';
      }
    }
  }
}

sub PoolThinprovSavings {
  my $pool = shift;
  my $data = shift;
  my $tier = shift;

  if ( $tier eq "-" ) {
    if ( $data->{'compression_active'} eq "yes" ) {
      return ( $data->{'virtual_capacity'} - $data->{'compression_uncompressed_capacity'} );
    }
    else {
      return ( $data->{'virtual_capacity'} - $data->{'used_capacity'} );
    }
  }
  return (0);
}

sub PoolDedupSavings {
  my $pool = shift;
  my $data = shift;
  my $tier = shift;

  if ( $tier eq "-" ) {
    if ( defined $data->{'data_reduction'} && $data->{'data_reduction'} eq "yes" ) {
      return ( $data->{'deduplication_capacity_saving'} );
    }
    else {
      return (0);
    }
  }
  return (0);
}

sub PoolCompressSavings {
  my $pool = shift;
  my $data = shift;
  my $tier = shift;

  if ( $tier ne "-" ) {

    # HW Compression - in tier data only
    if ( defined $pool->{$tier}->{'hw_compression'} && $pool->{$tier}->{'hw_compression'} eq "yes" ) {
      return ( $pool->{$tier}->{'fmt_data_used_capacity'} - $pool->{$tier}->{'raw_data_used_capacity'} );
    }
    else {
      return (0);
    }
  }
  else {
    # SW Compression - in pool data only
    if ( $data->{'compression_active'} eq "yes" ) {
      return ( $data->{'compression_uncompressed_capacity'} - $data->{'compression_compressed_capacity'} );
    }
  }
  return (0);
}

sub CapCalculatePoolData2 {
  my $pool    = $capdata_ref->{'svc_pool'};
  my @reclist = keys( %{ $cfg_ref->{'svc_pool'} } );
  foreach my $rec ( sort(@reclist) ) {
    my $tier_list = &PoolTierList( $cfg_ref->{'svc_pool'}->{$rec} );

    # Initialize pool data
    for my $tier_level ( @{$tier_list} ) {
      if ( $tier_level ne "-" ) {
        &CalculatePoolDataTier2( $pool->{$rec}, $cfg_ref->{'svc_pool'}->{$rec}, $tier_level );
      }
    }

    # TierSum
    &CalculatePoolDataTier2( $pool->{$rec}, $cfg_ref->{'svc_pool'}->{$rec}, "-" );
  }
}

sub CalculatePoolDataTier2 {
  my $pool = shift;
  my $cfg  = shift;
  my $tier = shift;
  #
  $pool->{$tier}->{'fmt_system_used_capacity'} = $pool->{$tier}->{'fmt_data_used_capacity'} - $pool->{$tier}->{'fmt_snapshot_used_capacity'} - $pool->{$tier}->{'fmt_replicated_used_capacity'};
  #
  if ( $tier ne "-" ) {
    if ( $pool->{$tier}->{'hw_compression'} eq "yes" ) {
      $pool->{$tier}->{'raw_data_used_capacity'} = &PoolRawDataUsedCapacityHwc( $cfg, $tier );
    }
    else {
      #$pool->{$tier}->{'raw_data_used_capacity'} = &PoolRawDataUsedCapacity($pool,$cfg,$tier);
      $pool->{$tier}->{'raw_data_used_capacity'} = $pool->{$tier}->{'fmt_data_used_capacity'} * $pool->{$tier}->{'raw_fmt_raid_capacity_koef'};
    }
    $pool->{'-'}->{'raw_data_used_capacity'} += $pool->{$tier}->{'raw_data_used_capacity'};
    #
    $pool->{$tier}->{'compress_savings'} = &PoolCompressSavings( $pool, $cfg, $tier );
    $pool->{'-'}->{'compress_savings'} += $pool->{$tier}->{'compress_savings'};
    $pool->{$tier}->{'raw_snapshot_used_capacity'} = $pool->{$tier}->{'fmt_snapshot_used_capacity'} * $pool->{$tier}->{'raw_fmt_raid_capacity_koef'};
    $pool->{'-'}->{'raw_snapshot_used_capacity'} += $pool->{$tier}->{'raw_snapshot_used_capacity'};
    $pool->{$tier}->{'raw_replicated_used_capacity'} = $pool->{$tier}->{'fmt_replicated_used_capacity'} * $pool->{$tier}->{'raw_fmt_raid_capacity_koef'};
    $pool->{'-'}->{'raw_replicated_used_capacity'} += $pool->{$tier}->{'raw_replicated_used_capacity'};
  }
  else {
    $pool->{$tier}->{'compress_savings'} += &PoolCompressSavings( $pool, $cfg, $tier );
  }
  $pool->{$tier}->{'thinprov_savings'} = &PoolThinprovSavings( $pool, $cfg, $tier );
  $pool->{$tier}->{'dedup_savings'}    = &PoolDedupSavings( $pool, $cfg, $tier );
  #
  $pool->{$tier}->{'raw_system_used_capacity'} = $pool->{$tier}->{'raw_data_used_capacity'} - $pool->{$tier}->{'raw_snapshot_used_capacity'} - $pool->{$tier}->{'raw_replicated_used_capacity'};
}

sub CapCalculateVolumeData {

  # Calulates Pools Snapshot and Replication Used Data Capacity
  my $volume  = $cfg_ref->{'svc_vdsk'};
  my $pool    = $capdata_ref->{'svc_pool'};
  my @reclist = keys( %{$volume} );
  foreach my $rec ( sort(@reclist) ) {
    my $volume_copy_list = &VolumeCopyList( $volume->{$rec} );
    for my $volume_copy ( @{$volume_copy_list} ) {
      my $tier_list = VolumeTierList( $volume->{$rec}, $volume_copy );
      for my $tier_level ( @{$tier_list} ) {
        my $pool_rec                     = $volume->{$rec}->{ "copy_" . $volume_copy . "_mdisk_grp_name" };
        my $fmt_snapshot_used_capacity   = &VolumeFmtSnapshotUsedCapacity( $volume->{$rec}, $volume_copy, $tier_level );
        my $fmt_replicated_used_capacity = &VolumeFmtReplicatedUsedCapacity( $volume->{$rec}, $volume_copy, $tier_level );
        $pool->{$pool_rec}->{$tier_level}->{'fmt_data_used_capacity'} += &VolumeFmtDataUsedCapacity( $volume->{$rec}, $volume_copy, $tier_level );
        if ( $fmt_snapshot_used_capacity != $fmt_replicated_used_capacity ) {
          $pool->{$pool_rec}->{$tier_level}->{'fmt_snapshot_used_capacity'} += $fmt_snapshot_used_capacity;
        }
        $pool->{$pool_rec}->{$tier_level}->{'fmt_replicated_used_capacity'} += $fmt_replicated_used_capacity;
      }
    }
  }

}

# -------------------------------------------------------------------------------------------------
sub CapPrintData {
  #
  if ($debug) { &message("Output file: ${out_cap_file}.tmp.") }
  open( CAPOUT, ">${out_cap_file}.tmp" ) or die("Print data to file ${out_cap_file}.tmp failed.");
  &CapPrintSystemData();
  &CapPrintPoolData();
  &CapPrintVolumeData();

  close(CAPOUT);
  if ($debug) { &message("Rename output file to: ${out_cap_file}.") }
  #my $ret = system( "$cmd_mv " . ${out_cap_file} . ".tmp " . ${out_cap_file} );
  my $ret = system( "$cmd_mv '${out_cap_file}.tmp' '${out_cap_file}'" );
  if ($ret) { error( "Couldn't rename file " . ${out_cap_file} . ".tmp ." ); return; }
}

sub ConvSizeUnits {
  my ( $unit, $size ) = @_;
  if ( !defined $size ) { $size = 0; }
  if ( $unit eq "B" )   { return ($size) }
  if ( $unit eq "KB" )  { return ( $size / 1024 ) }
  if ( $unit eq "MB" )  { return ( $size / 1048576 ) }
  if ( $unit eq "GB" )  { return ( $size / 1073741824 ) }
  if ( $unit eq "TB" )  { return ( $size / 1099511627776 ) }
}

# -------------------------------------------------------------------------------------------------

sub CapPrintSystemData {
  print CAPOUT "\nULOZISTE\n";
  print CAPOUT "------------------\n";
  print CAPOUT "\tMachine Name:                    ", $cfg_ref->{svc_system}->{name},        "\n";
  print CAPOUT "\tMachine Type-Model:              ", &MachineTypeModel,                     "\n";
  print CAPOUT "\tMachine Serial:                  ", &MachineSerial,                        "\n";
  print CAPOUT "\tCode Level:                      ", $cfg_ref->{svc_system}->{code_level},, "\n";
  if ( defined $cfg_ref->{svc_system}->{'product_name'} ) {
    print CAPOUT "\tProduct Name:                    ", $cfg_ref->{svc_system}->{'product_name'}, "\n";
  }
  print CAPOUT "\n";
  print CAPOUT "\tDate:           ", $cfg_ref->{date}->{'cfg_date'}, "\n";
  print CAPOUT "\n";

  my $unit      = "GB";
  my $tier_list = &StorageTierList;

  my $hdrline = "Jmeno uloziste,Kapacitni vrstva,Hruba kapacita,Hruba pouzita kapacita,Spare kapacita,";
  $hdrline .= "Hruba RAID kapacita,Hruba RAID volna kapacita,Hruba pouzita datova kapacita,";
  $hdrline .= "Hruba pouzita system kapacita,Hruba pouzita snapshot kapacita,Hruba pouzita replikovana kapacita,";
  $hdrline .= "Formatovana RAID kapacita,Formatovana RAID volna kapacita,Formatovana Pouzita datova kapacita,";
  $hdrline .= "Formatovana pouzita system kapacita,Formatovana pouzita snapshot kapacita,Formatovana pouzita replikovana kapacita,";
  $hdrline .= "Uspora thin-provisioning,Uspora deduplikace,Uspora komprese,Celkova efektivita,";
  print CAPOUT "\nKapacita uloziste\n";
  print CAPOUT "------------------------\n";
  print CAPOUT $hdrline, "\n";

  for my $tier_level ( sort { $b cmp $a } ( @{$tier_list} ) ) {
    my $raw_capacity      = &StorageRawCapacity($tier_level);
    my $raw_used_capacity = &StorageRawUsedCapacity($tier_level);
    my $spare_capacity    = &StorageSpareCapacity($tier_level);
    #
    my $raw_raid_capacity      = &StorageRawRaidCapacity($tier_level);
    my $raw_raid_free_capacity = &StorageRawRaidFreeCapacity($tier_level);
    my $raw_data_used_capacity = &StorageRawDataUsedCapacity($tier_level);
    #
    my $raw_snapshot_used_capacity   = &StorageRawSnapshotUsedCapacity($tier_level);
    my $raw_replicated_used_capacity = &StorageRawReplicatedUsedCapacity($tier_level);
    #
    my $raw_system_used_capacity = $raw_data_used_capacity - $raw_snapshot_used_capacity - $raw_replicated_used_capacity;
    #
    my $fmt_raid_capacity      = &StorageFmtRaidCapacity($tier_level);
    my $fmt_raid_free_capacity = &StorageFmtRaidFreeCapacity($tier_level);
    my $fmt_data_used_capacity = &StorageFmtDataUsedCapacity($tier_level);
    #
    my $fmt_snapshot_used_capacity   = &StorageFmtSnapshotUsedCapacity($tier_level);
    my $fmt_replicated_used_capacity = &StorageFmtReplicatedUsedCapacity($tier_level);
    #
    my $fmt_system_used_capacity = $fmt_data_used_capacity - $fmt_snapshot_used_capacity - $fmt_replicated_used_capacity;
    #
    my $thinprov_savings = &StorageThinprovSavings($tier_level);
    my $dedup_savings    = &StorageDedupSavings($tier_level);
    my $compress_savings = &StorageCompressSavings($tier_level);
    my $savings          = $thinprov_savings + $dedup_savings + $compress_savings;

    # Jmeno uloziste,Kapacitni vrstva,
    printf CAPOUT ( "%s,%s,", $cfg_ref->{svc_system}->{name}, $tier_level );

    # Hruba kapacita,Hruba pouzita kapacita,Spare kapacita,
    printf CAPOUT ( "%.3f,%.3f,%.3f,", ConvSizeUnits( $unit, $raw_capacity ), ConvSizeUnits( $unit, $raw_used_capacity ), ConvSizeUnits( $unit, $spare_capacity ) );

    # Hruba RAID kapacita,Hruba RAID volna kapacita,Hruba Pouzita datova kapacita,
    printf CAPOUT ( "%.3f,%.3f,%.3f,", ConvSizeUnits( $unit, $raw_raid_capacity ), ConvSizeUnits( $unit, $raw_raid_free_capacity ), ConvSizeUnits( $unit, $raw_data_used_capacity ) );

    # Hruba Pouzita system kapacita,Hruba Pouzita snapshot kapacita,Hruba Pouzita replikovana kapacita"
    printf CAPOUT ( "%.3f,%.3f,%.3f,", ConvSizeUnits( $unit, $raw_system_used_capacity ), ConvSizeUnits( $unit, $raw_snapshot_used_capacity ), ConvSizeUnits( $unit, $raw_replicated_used_capacity ) );

    # Formatovana RAID kapacita,Formatovana RAID volna kapacita,Formatovana Pouzita datova kapacita,
    printf CAPOUT ( "%.3f,%.3f,%.3f,", ConvSizeUnits( $unit, $fmt_raid_capacity ), ConvSizeUnits( $unit, $fmt_raid_free_capacity ), ConvSizeUnits( $unit, $fmt_data_used_capacity ) );

    # Formatovana Pouzita system kapacita,Formatovana Pouzita snapshot kapacita,Formatovana Pouzita replikovana kapacita,
    printf CAPOUT ( "%.3f,%.3f,%.3f,", ConvSizeUnits( $unit, $fmt_system_used_capacity ), ConvSizeUnits( $unit, $fmt_snapshot_used_capacity ), ConvSizeUnits( $unit, $fmt_replicated_used_capacity ) );

    # Uspora thin-provisioning,Uspora deduplikace,Uspora komprese,Celkova efektivita
    printf CAPOUT ( "%.3f,%.3f,", ConvSizeUnits( $unit, $thinprov_savings ), ConvSizeUnits( $unit, $dedup_savings ) );
    printf CAPOUT ( "%.3f,%.3f,", ConvSizeUnits( $unit, $compress_savings ), ConvSizeUnits( $unit, $savings ) );
    printf CAPOUT ("\n");
  }
  print CAPOUT "\n";
}

sub StorageTierList {
  my %h;
  my @a;

  #my @reclist = keys( %{ $cfg_ref->{svc_drive} } );
  #foreach my $rec ( sort(@reclist) ) {
  #  $h{$cfg_ref->{svc_drive}->{$rec}->{'tech_type'}} = 1;
  #}
  #@a = keys(%h);
  #if ( scalar @a == 0 ) {
  my @reclist = keys( %{ $capdata_ref->{'svc_pool'} } );
  foreach my $rec ( sort(@reclist) ) {
    my @tier_list = keys( %{ $capdata_ref->{'svc_pool'}->{$rec} } );
    for my $tier_level ( sort(@tier_list) ) {
      if ( $tier_level ne "-" ) {
        $h{$tier_level} = 1;
      }
    }
  }
  @a = keys(%h);

  #}
  push( @a, "-" );
  return ( \@a );
}

sub StorageRawCapacity {

  # Returns Raw Capacity of Storage Drives
  my $tier    = shift;
  my $cap     = 0;
  my @reclist = keys( %{ $cfg_ref->{svc_drive} } );
  foreach my $rec ( sort(@reclist) ) {
    if ( $tier ne "-" ) {
      if ( $tier eq "ssd" && $cfg_ref->{svc_drive}->{$rec}->{'tech_type'} eq "sas_ssd" ) {
        if ( defined $cfg_ref->{svc_drive}->{$rec}->{'physical_capacity'} ) {
          $cap += $cfg_ref->{svc_drive}->{$rec}->{'physical_capacity'};
        }
        else {
          $cap += $cfg_ref->{svc_drive}->{$rec}->{'capacity'};
        }
      }
      elsif ( $tier eq "enterprise" && $cfg_ref->{svc_drive}->{$rec}->{'tech_type'} eq "sas_hdd" ) {
        if ( defined $cfg_ref->{svc_drive}->{$rec}->{'physical_capacity'} ) {
          $cap += $cfg_ref->{svc_drive}->{$rec}->{'physical_capacity'};
        }
        else {
          $cap += $cfg_ref->{svc_drive}->{$rec}->{'capacity'};
        }
      }
      elsif ( $tier eq $cfg_ref->{svc_drive}->{$rec}->{'tech_type'} ) {
        if ( defined $cfg_ref->{svc_drive}->{$rec}->{'physical_capacity'} ) {
          $cap += $cfg_ref->{svc_drive}->{$rec}->{'physical_capacity'};
        }
        else {
          $cap += $cfg_ref->{svc_drive}->{$rec}->{'capacity'};
        }
      }
    }
    else {
      if ( defined $cfg_ref->{svc_drive}->{$rec}->{'physical_capacity'} ) {
        $cap += $cfg_ref->{svc_drive}->{$rec}->{'physical_capacity'};
      }
      else {
        $cap += $cfg_ref->{svc_drive}->{$rec}->{'capacity'};
      }
    }
  }
  return ($cap);
}

sub StorageRawUsedCapacity {

  # Returns Raw Used Capacity of Storage Drives
  # Raw Used Capacity
  # - capacity drives in arrays (member + spare)
  my $tier    = shift;
  my $cap     = 0;
  my @reclist = keys( %{ $cfg_ref->{svc_drive} } );
  foreach my $rec ( sort(@reclist) ) {
    if ( $cfg_ref->{svc_drive}->{$rec}->{'use'} eq "member" or $cfg_ref->{svc_drive}->{$rec}->{'use'} eq "spare" ) {
      if ( $tier ne "-" ) {
        if ( $tier eq "ssd" && $cfg_ref->{svc_drive}->{$rec}->{'tech_type'} eq "sas_ssd" ) {
          if ( defined $cfg_ref->{svc_drive}->{$rec}->{'physical_capacity'} ) {
            $cap += $cfg_ref->{svc_drive}->{$rec}->{'physical_capacity'};
          }
          else {
            $cap += $cfg_ref->{svc_drive}->{$rec}->{'capacity'};
          }
        }
        elsif ( $tier eq "enterprise" && $cfg_ref->{svc_drive}->{$rec}->{'tech_type'} eq "sas_hdd" ) {
          if ( defined $cfg_ref->{svc_drive}->{$rec}->{'physical_capacity'} ) {
            $cap += $cfg_ref->{svc_drive}->{$rec}->{'physical_capacity'};
          }
          else {
            $cap += $cfg_ref->{svc_drive}->{$rec}->{'capacity'};
          }
        }
        elsif ( $tier eq $cfg_ref->{svc_drive}->{$rec}->{'tech_type'} ) {
          if ( defined $cfg_ref->{svc_drive}->{$rec}->{'physical_capacity'} ) {
            $cap += $cfg_ref->{svc_drive}->{$rec}->{'physical_capacity'};
          }
          else {
            $cap += $cfg_ref->{svc_drive}->{$rec}->{'capacity'};
          }
        }
      }
      else {
        if ( defined $cfg_ref->{svc_drive}->{$rec}->{'physical_capacity'} ) {
          $cap += $cfg_ref->{svc_drive}->{$rec}->{'physical_capacity'};
        }
        else {
          $cap += $cfg_ref->{svc_drive}->{$rec}->{'capacity'};
        }
      }
    }
  }
  return ($cap);
}

sub StorageSpareCapacity {

  # Returns Capacity of Spare Storage Drives
  my $tier    = shift;
  my $cap     = 0;
  my @reclist = keys( %{ $cfg_ref->{svc_drive} } );
  foreach my $rec ( sort(@reclist) ) {
    if ( $cfg_ref->{svc_drive}->{$rec}->{'use'} eq "spare" ) {
      if ( $tier ne "-" ) {
        if ( $tier eq "ssd" && $cfg_ref->{svc_drive}->{$rec}->{'tech_type'} eq "sas_ssd" ) {
          if ( defined $cfg_ref->{svc_drive}->{$rec}->{'physical_capacity'} ) {
            $cap += $cfg_ref->{svc_drive}->{$rec}->{'physical_capacity'};
          }
          else {
            $cap += $cfg_ref->{svc_drive}->{$rec}->{'capacity'};
          }
        }
        elsif ( $tier eq "enterprise" && $cfg_ref->{svc_drive}->{$rec}->{'tech_type'} eq "sas_hdd" ) {
          if ( defined $cfg_ref->{svc_drive}->{$rec}->{'physical_capacity'} ) {
            $cap += $cfg_ref->{svc_drive}->{$rec}->{'physical_capacity'};
          }
          else {
            $cap += $cfg_ref->{svc_drive}->{$rec}->{'capacity'};
          }
        }
        elsif ( $tier eq $cfg_ref->{svc_drive}->{$rec}->{'tech_type'} ) {
          if ( defined $cfg_ref->{svc_drive}->{$rec}->{'physical_capacity'} ) {
            $cap += $cfg_ref->{svc_drive}->{$rec}->{'physical_capacity'};
          }
          else {
            $cap += $cfg_ref->{svc_drive}->{$rec}->{'capacity'};
          }
        }
      }
      else {
        if ( defined $cfg_ref->{svc_drive}->{$rec}->{'physical_capacity'} ) {
          $cap += $cfg_ref->{svc_drive}->{$rec}->{'physical_capacity'};
        }
        else {
          $cap += $cfg_ref->{svc_drive}->{$rec}->{'capacity'};
        }
      }
    }
  }
  return ($cap);
}

sub StorageRawRaidCapacity {
  my $tier    = shift;
  my $cap     = 0;
  my $pool    = $capdata_ref->{'svc_pool'};
  my @reclist = keys( %{$pool} );
  foreach my $rec ( sort(@reclist) ) {
    if ( defined $pool->{$rec}->{$tier} ) {
      $cap += $pool->{$rec}->{$tier}->{'raw_raid_capacity'};
    }
  }
  return ($cap);
}

sub StorageRawRaidFreeCapacity {
  my $tier    = shift;
  my $cap     = 0;
  my $pool    = $capdata_ref->{'svc_pool'};
  my @reclist = keys( %{$pool} );
  foreach my $rec ( sort(@reclist) ) {
    if ( defined $pool->{$rec}->{$tier} ) {
      $cap += $pool->{$rec}->{$tier}->{'raw_raid_free_capacity'};
    }
  }
  return ($cap);
}

sub StorageRawDataUsedCapacity {

  # Returns Raw Used Capacity of Storage Drives for data (system + snapshot + replicated)
  # Data Used Capacity
  # - Compute it from pool capacity
  # - HW compressed drives (NVMe) - in svc_drive data (lsdrive)
  # - Non compressed drives - compute from pool data
  my $tier    = shift;
  my $cap     = 0;
  my $pool    = $capdata_ref->{'svc_pool'};
  my @reclist = keys( %{$pool} );
  foreach my $rec ( sort(@reclist) ) {
    if ( defined $pool->{$rec}->{$tier} ) {
      $cap += $pool->{$rec}->{$tier}->{'raw_data_used_capacity'};
    }
  }
  return ($cap);
}

sub StorageRawSnapshotUsedCapacity {

  # Returns Raw Used Capacity for Snapshot Volumes
  # - compute from volumes data
  my $tier    = shift;
  my $cap     = 0;
  my $pool    = $capdata_ref->{'svc_pool'};
  my @reclist = keys( %{$pool} );
  foreach my $rec ( sort(@reclist) ) {
    if ( defined $pool->{$rec}->{$tier} ) {
      $cap += $pool->{$rec}->{$tier}->{'raw_snapshot_used_capacity'};
    }
  }
  return ($cap);
}

sub StorageRawReplicatedUsedCapacity {

  # Returns Raw Used Capacity for Replicated Volumes
  # - compute from volumes data
  my $tier    = shift;
  my $cap     = 0;
  my $pool    = $capdata_ref->{'svc_pool'};
  my @reclist = keys( %{$pool} );
  foreach my $rec ( sort(@reclist) ) {
    if ( defined $pool->{$rec}->{$tier} ) {
      $cap += $pool->{$rec}->{$tier}->{'raw_replicated_used_capacity'};
    }
  }
  return ($cap);
}

sub StorageFmtRaidCapacity {
  my $tier    = shift;
  my $cap     = 0;
  my $pool    = $capdata_ref->{'svc_pool'};
  my @reclist = keys( %{$pool} );

  foreach my $rec ( sort(@reclist) ) {
    if ( defined $pool->{$rec}->{$tier} && $pool->{$rec}->{$tier}->{'type'} eq "parent" ) {

      # Ignore child pool capacity
      $cap += $pool->{$rec}->{$tier}->{'fmt_raid_capacity'};
    }
  }
  return ($cap);
}

sub StorageFmtRaidFreeCapacity {
  my $tier    = shift;
  my $cap     = 0;
  my $pool    = $capdata_ref->{'svc_pool'};
  my @reclist = keys( %{$pool} );
  foreach my $rec ( sort(@reclist) ) {
    if ( defined $pool->{$rec}->{$tier} ) {
      $cap += $pool->{$rec}->{$tier}->{'fmt_raid_free_capacity'};
    }
  }
  return ($cap);
}

sub StorageFmtDataUsedCapacity {
  my $tier    = shift;
  my $cap     = 0;
  my $pool    = $capdata_ref->{'svc_pool'};
  my @reclist = keys( %{$pool} );
  foreach my $rec ( sort(@reclist) ) {
    if ( defined $pool->{$rec}->{$tier} ) {
      $cap += $pool->{$rec}->{$tier}->{'fmt_data_used_capacity'};
    }
  }
  return ($cap);
}
#
sub StorageFmtSnapshotUsedCapacity {
  my $tier    = shift;
  my $cap     = 0;
  my $pool    = $capdata_ref->{'svc_pool'};
  my @reclist = keys( %{$pool} );
  foreach my $rec ( sort(@reclist) ) {
    if ( defined $pool->{$rec}->{$tier} ) {
      $cap += $pool->{$rec}->{$tier}->{'fmt_snapshot_used_capacity'};
    }
  }
  return ($cap);
}

sub StorageFmtReplicatedUsedCapacity {
  my $tier    = shift;
  my $cap     = 0;
  my $pool    = $capdata_ref->{'svc_pool'};
  my @reclist = keys( %{$pool} );
  foreach my $rec ( sort(@reclist) ) {
    if ( defined $pool->{$rec}->{$tier} ) {
      $cap += $pool->{$rec}->{$tier}->{'fmt_replicated_used_capacity'};
    }
  }
  return ($cap);
}

sub StorageThinprovSavings {
  my $tier    = shift;
  my $cap     = 0;
  my $pool    = $capdata_ref->{'svc_pool'};
  my @reclist = keys( %{$pool} );
  foreach my $rec ( sort(@reclist) ) {
    if ( defined $pool->{$rec}->{$tier} ) {
      $cap += $pool->{$rec}->{$tier}->{'thinprov_savings'};
    }
  }
  return ($cap);
}

sub StorageCompressSavings {
  my $tier    = shift;
  my $cap     = 0;
  my $pool    = $capdata_ref->{'svc_pool'};
  my @reclist = keys( %{$pool} );
  foreach my $rec ( sort(@reclist) ) {
    if ( defined $pool->{$rec}->{$tier} ) {
      $cap += $pool->{$rec}->{$tier}->{'compress_savings'};
    }
  }
  return ($cap);
}

sub StorageDedupSavings {
  my $tier    = shift;
  my $cap     = 0;
  my $pool    = $capdata_ref->{'svc_pool'};
  my @reclist = keys( %{$pool} );
  foreach my $rec ( sort(@reclist) ) {
    if ( defined $pool->{$rec}->{$tier} ) {
      $cap += $pool->{$rec}->{$tier}->{'dedup_savings'};
    }
  }
  return ($cap);
}

# -------------------------------------------------------------------------------------------------
sub CapPrintPoolData {

  my $unit = "GB";

  my $hdrline = "Cislo poolu,Jmeno poolu,Kapacitni vrstva,RAID,";
  $hdrline .= "Hruba kapacita,Hruba pouzita kapacita,Spare kapacita,";
  $hdrline .= "Hruba RAID kapacita,Hruba RAID volna kapacita,Hruba pouzita datova kapacita,";
  $hdrline .= "Hruba pouzita system kapacita,Hruba pouzita snapshot kapacita,Hruba pouzita replikovana kapacita,";
  $hdrline .= "Formatovana RAID kapacita,Formatovana RAID volna kapacita,Formatovana Pouzita datova kapacita,";
  $hdrline .= "Formatovana pouzita system kapacita,Formatovana pouzita snapshot kapacita,Formatovana pouzita replikovana kapacita,";
  $hdrline .= "Uspora thin-provisioning,Uspora deduplikace,Uspora komprese,Celkova efektivita,";
  print CAPOUT "\nKapacita data pool\n";
  print CAPOUT "------------------------\n";
  print CAPOUT $hdrline, "\n";
  my @reclist = keys( %{ $capdata_ref->{'svc_pool'} } );

  foreach my $rec ( sort(@reclist) ) {

    my @tier_list = keys( %{ $capdata_ref->{'svc_pool'}->{$rec} } );
    for my $tier_level ( sort { $b cmp $a } (@tier_list) ) {

      my $raid_type = $capdata_ref->{'svc_pool'}->{$rec}->{$tier_level}->{'raid_type'};
      #
      my $raw_capacity      = $capdata_ref->{'svc_pool'}->{$rec}->{$tier_level}->{'raw_capacity'};
      my $raw_used_capacity = $capdata_ref->{'svc_pool'}->{$rec}->{$tier_level}->{'raw_used_capacity'};
      my $spare_capacity    = $capdata_ref->{'svc_pool'}->{$rec}->{$tier_level}->{'spare_capacity'};
      #
      my $raw_raid_capacity      = $capdata_ref->{'svc_pool'}->{$rec}->{$tier_level}->{'raw_raid_capacity'};
      my $raw_raid_free_capacity = $capdata_ref->{'svc_pool'}->{$rec}->{$tier_level}->{'raw_raid_free_capacity'};
      my $raw_data_used_capacity = $capdata_ref->{'svc_pool'}->{$rec}->{$tier_level}->{'raw_data_used_capacity'};
      #
      my $raw_snapshot_used_capacity   = $capdata_ref->{'svc_pool'}->{$rec}->{$tier_level}->{'raw_snapshot_used_capacity'};
      my $raw_replicated_used_capacity = $capdata_ref->{'svc_pool'}->{$rec}->{$tier_level}->{'raw_replicated_used_capacity'};
      #
      my $raw_system_used_capacity = $raw_data_used_capacity - $raw_snapshot_used_capacity - $raw_replicated_used_capacity;
      #
      my $fmt_raid_capacity      = $capdata_ref->{'svc_pool'}->{$rec}->{$tier_level}->{'fmt_raid_capacity'};
      my $fmt_raid_free_capacity = $capdata_ref->{'svc_pool'}->{$rec}->{$tier_level}->{'fmt_raid_free_capacity'};
      my $fmt_data_used_capacity = $capdata_ref->{'svc_pool'}->{$rec}->{$tier_level}->{'fmt_data_used_capacity'};
      #
      my $fmt_snapshot_used_capacity   = $capdata_ref->{'svc_pool'}->{$rec}->{$tier_level}->{'fmt_snapshot_used_capacity'};
      my $fmt_replicated_used_capacity = $capdata_ref->{'svc_pool'}->{$rec}->{$tier_level}->{'fmt_replicated_used_capacity'};
      #
      my $fmt_system_used_capacity = $fmt_data_used_capacity - $fmt_snapshot_used_capacity - $fmt_replicated_used_capacity;
      #
      my $thinprov_savings = $capdata_ref->{'svc_pool'}->{$rec}->{$tier_level}->{'thinprov_savings'};
      my $dedup_savings    = $capdata_ref->{'svc_pool'}->{$rec}->{$tier_level}->{'dedup_savings'};
      my $compress_savings = $capdata_ref->{'svc_pool'}->{$rec}->{$tier_level}->{'compress_savings'};
      my $savings          = $thinprov_savings + $dedup_savings + $compress_savings;

      # Cislo poolu,Jmeno poolu,Kapacitni vrstva,RAID,
      printf CAPOUT ( "%d,%s,%s,%s,", $cfg_ref->{'svc_pool'}->{$rec}->{id}, $cfg_ref->{'svc_pool'}->{$rec}->{name}, $tier_level, $raid_type );

      # Hruba kapacita,Hruba pouzita kapacita,Spare kapacita,
      printf CAPOUT ( "%.3f,%.3f,%.3f,", ConvSizeUnits( $unit, $raw_capacity ), ConvSizeUnits( $unit, $raw_used_capacity ), ConvSizeUnits( $unit, $spare_capacity ) );

      # Hruba RAID kapacita,Hruba RAID volna kapacita,Hruba Pouzita datova kapacita,
      printf CAPOUT ( "%.3f,%.3f,%.3f,", ConvSizeUnits( $unit, $raw_raid_capacity ), ConvSizeUnits( $unit, $raw_raid_free_capacity ), ConvSizeUnits( $unit, $raw_data_used_capacity ) );

      # Hruba Pouzita system kapacita,Hruba Pouzita snapshot kapacita,Hruba Pouzita replikovana kapacita"
      printf CAPOUT ( "%.3f,%.3f,%.3f,", ConvSizeUnits( $unit, $raw_system_used_capacity ), ConvSizeUnits( $unit, $raw_snapshot_used_capacity ), ConvSizeUnits( $unit, $raw_replicated_used_capacity ) );

      # Formatovana RAID kapacita,Formatovana RAID volna kapacita,Formatovana Pouzita datova kapacita,
      printf CAPOUT ( "%.3f,%.3f,%.3f,", ConvSizeUnits( $unit, $fmt_raid_capacity ), ConvSizeUnits( $unit, $fmt_raid_free_capacity ), ConvSizeUnits( $unit, $fmt_data_used_capacity ) );

      # Formatovana Pouzita system kapacita,Formatovana Pouzita snapshot kapacita,Formatovana Pouzita replikovana kapacita,
      printf CAPOUT ( "%.3f,%.3f,%.3f,", ConvSizeUnits( $unit, $fmt_system_used_capacity ), ConvSizeUnits( $unit, $fmt_snapshot_used_capacity ), ConvSizeUnits( $unit, $fmt_replicated_used_capacity ) );

      # Uspora thin-provisioning,Uspora deduplikace,Uspora komprese,Celkova efektivita
      printf CAPOUT ( "%.3f,%.3f,", ConvSizeUnits( $unit, $thinprov_savings ), ConvSizeUnits( $unit, $dedup_savings ) );
      printf CAPOUT ( "%.3f,%.3f,", ConvSizeUnits( $unit, $compress_savings ), ConvSizeUnits( $unit, $savings ) );
      printf CAPOUT ("\n");
    }
  }
  print CAPOUT "\n";
}

sub PoolTierList {
  my $data = shift;
  my @a;
  push( @a, $data->{'tier_scm'} ) if ( defined $data->{'tier_scm_mdisk_count'} && $data->{'tier_scm_mdisk_count'} > 0 );
  push( @a, $data->{'tier_0'} )   if ( defined $data->{'tier_0_mdisk_count'}   && $data->{'tier_0_mdisk_count'} > 0 );
  push( @a, $data->{'tier_0n'} )  if ( defined $data->{'tier_0n_mdisk_count'}  && $data->{'tier_0n_mdisk_count'} > 0 );
  push( @a, $data->{'tier_1'} )   if ( defined $data->{'tier_1_mdisk_count'}   && $data->{'tier_1_mdisk_count'} > 0 );
  push( @a, $data->{'tier_2'} )   if ( defined $data->{'tier_2_mdisk_count'}   && $data->{'tier_2_mdisk_count'} > 0 );
  push( @a, "-" );
  return ( \@a );
}

sub PoolRaidType {
  my $data    = shift;
  my $tier    = shift;
  my $pool_id = $data->{id};
  my %h;
  my @reclist = keys( %{ $cfg_ref->{svc_mdsk} } );
  foreach my $rec ( sort(@reclist) ) {
    if ( $cfg_ref->{svc_mdsk}->{$rec}->{mdisk_grp_id} eq $pool_id && $cfg_ref->{svc_mdsk}->{$rec}->{mode} eq "array" ) {
      if ( $tier eq "-" ) {
        $h{ $cfg_ref->{svc_mdsk}->{$rec}->{raid_level} } = 1;
      }
      elsif ( $cfg_ref->{svc_mdsk}->{$rec}->{'tier'} eq $tier ) {
        $h{ $cfg_ref->{svc_mdsk}->{$rec}->{raid_level} } = 1;
      }
    }
  }
  return ( join( ':', sort( keys(%h) ) ) );
}

sub PoolRawCapacity {

  # Suma kapacity disku pouzitych pro pool
  my $data    = shift;
  my $tier    = shift;
  my $pool_id = $data->{id};
  my @reclist = keys( %{ $cfg_ref->{svc_mdsk} } );
  my $cap     = 0;
  foreach my $rec ( sort(@reclist) ) {
    if ( $cfg_ref->{svc_mdsk}->{$rec}->{mdisk_grp_id} eq $pool_id && $cfg_ref->{svc_mdsk}->{$rec}->{mode} eq "array" ) {
      my $mdsk_id       = $cfg_ref->{svc_mdsk}->{$rec}->{id};
      my @drive_reclist = keys( %{ $cfg_ref->{svc_drive} } );
      foreach my $drive_rec ( sort(@drive_reclist) ) {
        my $drive_capacity = 0;
        if ( defined $cfg_ref->{svc_drive}->{$drive_rec}->{'physical_capacity'} ) {
          $drive_capacity = $cfg_ref->{svc_drive}->{$drive_rec}->{'physical_capacity'};
        }
        else {
          $drive_capacity = $cfg_ref->{svc_drive}->{$drive_rec}->{'capacity'};
        }
        if ( $cfg_ref->{svc_drive}->{$drive_rec}->{'use'} eq "member" && $cfg_ref->{svc_drive}->{$drive_rec}->{'mdisk_id'} == $mdsk_id ) {
          if ( $tier ne "-" ) {
            if ( $tier eq $cfg_ref->{svc_drive}->{$drive_rec}->{'tech_type'} ) {
              $cap += $drive_capacity;
            }
          }
          else {
            $cap += $drive_capacity;
          }
        }
      }
    }
  }
  return ($cap);
}

sub PoolRawUsedCapacity {

  # Suma kapacity disku pouzitych pro pool
  #   - bez HW komprese - cela kapacita
  #   - s HW kompresi - pouzita fyzicka kapacita
  my $data    = shift;
  my $tier    = shift;
  my $pool_id = $data->{id};
  my @reclist = keys( %{ $cfg_ref->{svc_mdsk} } );
  my $cap     = 0;
  foreach my $rec ( sort(@reclist) ) {
    if ( $cfg_ref->{svc_mdsk}->{$rec}->{mdisk_grp_id} eq $pool_id && $cfg_ref->{svc_mdsk}->{$rec}->{mode} eq "array" ) {
      my $mdsk_id       = $cfg_ref->{svc_mdsk}->{$rec}->{id};
      my @drive_reclist = keys( %{ $cfg_ref->{svc_drive} } );
      foreach my $drive_rec ( sort(@drive_reclist) ) {
        if ( $cfg_ref->{svc_drive}->{$drive_rec}->{'use'} eq "member" && $cfg_ref->{svc_drive}->{$drive_rec}->{'mdisk_id'} == $mdsk_id ) {
          if ( defined $cfg_ref->{svc_drive}->{$drive_rec}->{'compressed'} && $cfg_ref->{svc_drive}->{$drive_rec}->{'compressed'} eq "yes" ) {

            # HW Compression
            if ( $tier ne "-" ) {
              if ( $tier eq $cfg_ref->{svc_drive}->{$drive_rec}->{'tech_type'} ) {
                $cap += $cfg_ref->{svc_drive}->{$drive_rec}->{'physical_capacity'};
              }
            }
            else {
              $cap += $cfg_ref->{svc_drive}->{$drive_rec}->{'physical_capacity'};
            }
          }
          else {
            if ( $tier ne "-" ) {
              if ( $tier eq $cfg_ref->{svc_drive}->{$drive_rec}->{'tech_type'} ) {
                $cap += $cfg_ref->{svc_drive}->{$drive_rec}->{'capacity'};
              }
            }
            else {
              $cap += $cfg_ref->{svc_drive}->{$drive_rec}->{'capacity'};
            }
          }
        }
      }
    }
  }
  return ($cap);
}

sub PoolSpareCapacity {

  # Spare disk nema prislusnost k urcitemu poolu dokud neni pouzit.
  # Ponechano pro pripad novych zjisteni.
  return (0);
}
#
sub PoolRawRaidCapacity {
  my $pool    = shift;
  my $data    = shift;
  my $tier    = shift;
  my $pool_id = $data->{id};
  my $cap     = 0;
  if ( $tier eq "-" ) {
    my @reclist = keys( %{$pool} );
    foreach my $rec ( sort(@reclist) ) {
      $cap += $pool->{$rec}->{'raw_raid_capacity'};
    }
  }
  else {
    my @reclist = keys( %{ $cfg_ref->{'svc_mdsk'} } );
    foreach my $rec ( sort(@reclist) ) {
      if ( $cfg_ref->{'svc_mdsk'}->{$rec}->{'mdisk_grp_id'} eq $pool_id && $cfg_ref->{'svc_mdsk'}->{$rec}->{'mode'} eq "array" ) {
        $cap += &MdiskRawCapacity( $cfg_ref->{'svc_mdsk'}->{$rec}->{'id'}, $tier );
      }
    }
  }
  return ($cap);
}

sub MdiskRawCapacity {
  my $mdisk_id      = shift;
  my $tier          = shift;
  my $cap           = 0;
  my @drive_reclist = keys( %{ $cfg_ref->{svc_drive} } );
  foreach my $drive_rec ( sort(@drive_reclist) ) {
    if ( $cfg_ref->{svc_drive}->{$drive_rec}->{'use'} eq "member" && $cfg_ref->{svc_drive}->{$drive_rec}->{'mdisk_id'} == $mdisk_id ) {
      if ( defined $cfg_ref->{svc_drive}->{$drive_rec}->{'compressed'} && $cfg_ref->{svc_drive}->{$drive_rec}->{'compressed'} eq "yes" ) {

        # HW Compression
        if ( $tier ne "-" ) {
          if ( $tier eq $cfg_ref->{svc_drive}->{$drive_rec}->{'tech_type'} ) {
            $cap += $cfg_ref->{svc_drive}->{$drive_rec}->{'physical_used_capacity'};
          }
        }
        else {
          $cap += $cfg_ref->{svc_drive}->{$drive_rec}->{'physical_used_capacity'};
        }
      }
      else {
        if ( $tier ne "-" ) {
          if ( $tier eq "ssd" && $cfg_ref->{svc_drive}->{$drive_rec}->{'tech_type'} eq "sas_ssd" ) {
            $cap += $cfg_ref->{svc_drive}->{$drive_rec}->{'capacity'};
          }
          elsif ( $tier eq "enterprise" && $cfg_ref->{svc_drive}->{$drive_rec}->{'tech_type'} eq "sas_hdd" ) {
            $cap += $cfg_ref->{svc_drive}->{$drive_rec}->{'capacity'};
          }
          elsif ( $tier eq $cfg_ref->{svc_drive}->{$drive_rec}->{'tech_type'} ) {
            $cap += $cfg_ref->{svc_drive}->{$drive_rec}->{'capacity'};
          }
        }
        else {
          $cap += $cfg_ref->{svc_drive}->{$drive_rec}->{'capacity'};
        }
      }
    }
  }
  return ($cap);
}

sub PoolRawRaidCapacityHwc {
  my $data    = shift;
  my $tier    = shift;
  my $pool_id = $data->{id};
  my @reclist = keys( %{ $cfg_ref->{'svc_mdsk'} } );
  my $cap     = 0;
  foreach my $rec ( sort(@reclist) ) {
    if ( $cfg_ref->{'svc_mdsk'}->{$rec}->{'mdisk_grp_id'} eq $pool_id && $cfg_ref->{'svc_mdsk'}->{$rec}->{'mode'} eq "array" ) {
      $cap += $cfg_ref->{'svc_mdsk'}->{$rec}->{'physical_capacity'};
    }
  }
  return ($cap);
}

sub MdiskHwCompression {
  my $mdisk_id      = shift;
  my $tier          = shift;
  my @drive_reclist = keys( %{ $cfg_ref->{svc_drive} } );
  foreach my $drive_rec ( sort(@drive_reclist) ) {
    if ( $cfg_ref->{svc_drive}->{$drive_rec}->{'use'} eq "member" && $cfg_ref->{svc_drive}->{$drive_rec}->{'mdisk_id'} == $mdisk_id ) {
      if ( defined $cfg_ref->{svc_drive}->{$drive_rec}->{'compressed'} && $cfg_ref->{svc_drive}->{$drive_rec}->{'compressed'} eq "yes" ) {
        if ( $tier ne "-" ) {
          if ( $tier eq $cfg_ref->{svc_drive}->{$drive_rec}->{'tech_type'} ) {
            return (1);
          }
          elsif ( "sas_" . $tier eq $cfg_ref->{svc_drive}->{$drive_rec}->{'tech_type'} ) {
            return (1);
          }
        }
      }
    }
  }
  return (0);
}

sub PoolRawRaidFreeCapacity {
  my $data    = shift;
  my $tier    = shift;
  my $pool_id = $data->{id};
  my @reclist = keys( %{ $cfg_ref->{'svc_mdsk'} } );
  my $cap     = 0;
  foreach my $rec ( sort(@reclist) ) {
    if ( defined $cfg_ref->{'svc_mdsk'}->{$rec}->{'physical_free_capacity'}
      && $cfg_ref->{'svc_mdsk'}->{$rec}->{'mdisk_grp_id'} eq $pool_id
      && $cfg_ref->{'svc_mdsk'}->{$rec}->{'mode'} eq "array" )
    {
      if ( $tier eq "-" ) {
        $cap += $cfg_ref->{'svc_mdsk'}->{$rec}->{'physical_free_capacity'};
      }
      elsif ( $cfg_ref->{'svc_mdsk'}->{$rec}->{'tier'} eq $tier ) {
        $cap += $cfg_ref->{'svc_mdsk'}->{$rec}->{'physical_free_capacity'};
      }
    }
  }
  return ($cap);
}

sub PoolRawDataUsedCapacity {
  my $pool          = shift;
  my $data          = shift;
  my $tier          = shift;
  my @tier_key_list = ( "scm", "0", "0n", "1", "2" );
  if ( $tier eq "-" ) {
    return ( $data->{'used_capacity'} * $pool->{$tier}->{'raw_fmt_raid_capacity_koef'} );
  }
  else {
    for my $key (@tier_key_list) {
      if ( defined $data->{ 'tier_' . $key } && $data->{ 'tier_' . $key } eq $tier ) {
        return ( ( $data->{ 'tier_' . $key . '_capacity' } - $data->{ 'tier_' . $key . '_free_capacity' } ) * $pool->{$tier}->{'raw_fmt_raid_capacity_koef'} );
      }
    }
  }
  return (undef);
}

sub PoolRawDataUsedCapacityHwc {
  my $data    = shift;
  my $tier    = shift;
  my $pool_id = $data->{id};
  my @reclist = keys( %{ $cfg_ref->{'svc_mdsk'} } );
  my $cap     = 0;
  foreach my $rec ( sort(@reclist) ) {
    if ( defined $cfg_ref->{'svc_mdsk'}->{$rec}->{'physical_free_capacity'} && $cfg_ref->{'svc_mdsk'}->{$rec}->{'mdisk_grp_id'} eq $pool_id && $cfg_ref->{'svc_mdsk'}->{$rec}->{'mode'} eq "array" ) {
      if ( $cfg_ref->{'svc_mdsk'}->{$rec}->{'tier'} && $cfg_ref->{'svc_mdsk'}->{$rec}->{'tier'} eq $tier ) {
        $cap += $cfg_ref->{'svc_mdsk'}->{$rec}->{'physical_capacity'} - $cfg_ref->{'svc_mdsk'}->{$rec}->{'physical_free_capacity'};
      }
    }
  }
  return ($cap);
}
#
sub PoolFmtRaidCapacity {
  my $data          = shift;
  my $tier          = shift;
  my @tier_key_list = ( "scm", "0", "0n", "1", "2" );
  if ( $tier eq "-" ) {
    return ( $data->{'capacity'} );
  }
  else {
    for my $key (@tier_key_list) {
      return ( $data->{ 'tier_' . $key . '_capacity' } ) if ( defined $data->{ 'tier_' . $key } && $data->{ 'tier_' . $key } eq $tier );
    }
  }
  return (undef);
}

sub PoolFmtRaidFreeCapacity {
  my $data          = shift;
  my $tier          = shift;
  my @tier_key_list = ( "scm", "0", "0n", "1", "2" );
  if ( $tier eq "-" ) {
    return ( $data->{'free_capacity'} );
  }
  else {
    for my $key (@tier_key_list) {
      return ( $data->{ 'tier_' . $key . '_free_capacity' } ) if ( defined $data->{ 'tier_' . $key } && $data->{ 'tier_' . $key } eq $tier );
    }
  }
  return (undef);
}

# -------------------------------------------------------------------------------------------------
sub CapPrintVolumeData {

  my $unit = "GB";

  my $hdrline = "Cislo volumu,Jmeno volumu,Kopie volumu,Kapacitni vrstva,RAID,Cislo poolu,Jmeno poolu,";
  $hdrline .= "Hruba kapacita,Hruba pouzita kapacita,Spare kapacita,";
  $hdrline .= "Hruba RAID kapacita,Hruba RAID volna kapacita,Hruba pouzita datova kapacita,";
  $hdrline .= "Hruba pouzita system kapacita,Hruba pouzita snapshot kapacita,Hruba pouzita replikovana kapacita,";
  $hdrline .= "Formatovana RAID kapacita,Formatovana RAID volna kapacita,Formatovana Pouzita datova kapacita,";
  $hdrline .= "Formatovana pouzita system kapacita,Formatovana pouzita snapshot kapacita,Formatovana pouzita replikovana kapacita,";
  $hdrline .= "Uspora thin-provisioning,Uspora deduplikace,Uspora komprese,Celkova efektivita,";
  print CAPOUT "\nKapacita Volumu\n";
  print CAPOUT "------------------------\n";
  print CAPOUT $hdrline, "\n";

  my $volume  = $cfg_ref->{'svc_vdsk'};
  my @reclist = keys( %{$volume} );
  foreach my $rec ( sort(@reclist) ) {
    my $volume_copy_list = &VolumeCopyList( $volume->{$rec} );

    for my $volume_copy ( sort @{$volume_copy_list} ) {
      my $tier_list = VolumeTierList( $volume->{$rec}, $volume_copy );

      for my $tier_level ( sort { $b cmp $a } @{$tier_list} ) {

        my $raid_type = &VolumeRaidType( $volume->{$rec}, $volume_copy, $tier_level );
        #
        my $raw_capacity                 = 0;                                                                                       #  = &VolumeRawCapacity($volume->{$rec},$volume_copy,$tier_level);
        my $raw_used_capacity            = 0;                                                                                       #  = &VolumeRawUsedCapacity($volume->{$rec},$volume_copy,$tier_level);
        my $spare_capacity               = 0;                                                                                       #  = &VolumeSpareCapacity($volume->{$rec},$volume_copy,$tier_level);
                                                                                                                                    #
        my $raw_raid_capacity            = 0;                                                                                       #  = &VolumeRawRaidCapacity($volume->{$rec},$volume_copy,$tier_level);
        my $raw_raid_free_capacity       = 0;                                                                                       #  = &VolumeRawRaidFreeCapacity($volume->{$rec},$volume_copy,$tier_level);
        my $raw_data_used_capacity       = 0;                                                                                       #  = &VolumeRawDataUsedCapacity($volume->{$rec},$volume_copy,$tier_level);
                                                                                                                                    #
        my $raw_snapshot_used_capacity   = 0;                                                                                       #  = &VolumeRawSnapshotUsedCapacity($volume->{$rec},$volume_copy,$tier_level);
        my $raw_replicated_used_capacity = 0;                                                                                       #  = &VolumeRawReplicatedUsedCapacity($volume->{$rec},$volume_copy,$tier_level);
                                                                                                                                    #
        my $raw_system_used_capacity     = $raw_data_used_capacity - $raw_snapshot_used_capacity - $raw_replicated_used_capacity;
        #
        my $fmt_raid_capacity      = &VolumeFmtRaidCapacity( $volume->{$rec}, $volume_copy, $tier_level );
        my $fmt_raid_free_capacity = 0;                                                                                             #  = &VolumeFmtRaidFreeCapacity($volume->{$rec},$volume_copy,$tier_level);
        my $fmt_data_used_capacity = &VolumeFmtDataUsedCapacity( $volume->{$rec}, $volume_copy, $tier_level );
        #
        my $fmt_snapshot_used_capacity   = &VolumeFmtSnapshotUsedCapacity( $volume->{$rec}, $volume_copy, $tier_level );
        my $fmt_replicated_used_capacity = &VolumeFmtReplicatedUsedCapacity( $volume->{$rec}, $volume_copy, $tier_level );
        #
        my $fmt_system_used_capacity;
        if ( $fmt_snapshot_used_capacity == $fmt_replicated_used_capacity ) {

          # Volume is FC Target Volume and RC Volume -> use only one capacity
          $fmt_snapshot_used_capacity = 0;
          $fmt_system_used_capacity   = $fmt_data_used_capacity - $fmt_replicated_used_capacity;
        }
        else {
          $fmt_system_used_capacity = $fmt_data_used_capacity - $fmt_snapshot_used_capacity - $fmt_replicated_used_capacity;
        }

        # Cislo volumu,Jmeno volumu,Kopie volumu,Kapacitni vrstva,
        printf CAPOUT ( "%d,%s,%s,%s,", $volume->{$rec}->{id}, $volume->{$rec}->{name}, $volume_copy, $tier_level );

        # RAID,Cislo poolu,Jmeno poolu,
        printf CAPOUT ( "%s,%s,%s,", $raid_type, $volume->{$rec}->{ "copy_" . $volume_copy . "_mdisk_grp_id" }, $volume->{$rec}->{ "copy_" . $volume_copy . "_mdisk_grp_name" } );

        # Hruba kapacita,Hruba pouzita kapacita,Spare kapacita,
        printf CAPOUT ( "%.3f,%.3f,%.3f,", ConvSizeUnits( $unit, $raw_capacity ), ConvSizeUnits( $unit, $raw_used_capacity ), ConvSizeUnits( $unit, $spare_capacity ) );

        # Hruba RAID kapacita,Hruba RAID volna kapacita,Hruba Pouzita datova kapacita,
        printf CAPOUT ( "%.3f,%.3f,%.3f,", ConvSizeUnits( $unit, $raw_raid_capacity ), ConvSizeUnits( $unit, $raw_raid_free_capacity ), ConvSizeUnits( $unit, $raw_data_used_capacity ) );

        # Hruba Pouzita system kapacita,Hruba Pouzita snapshot kapacita,Hruba Pouzita replikovana kapacita"
        printf CAPOUT ( "%.3f,%.3f,%.3f,", ConvSizeUnits( $unit, $raw_system_used_capacity ), ConvSizeUnits( $unit, $raw_snapshot_used_capacity ), ConvSizeUnits( $unit, $raw_replicated_used_capacity ) );

        # Formatovana RAID kapacita,Formatovana RAID volna kapacita,Formatovana Pouzita datova kapacita,
        printf CAPOUT ( "%.3f,%.3f,%.3f,", ConvSizeUnits( $unit, $fmt_raid_capacity ), ConvSizeUnits( $unit, $fmt_raid_free_capacity ), ConvSizeUnits( $unit, $fmt_data_used_capacity ) );

        # Formatovana Pouzita system kapacita,Formatovana Pouzita snapshot kapacita,Formatovana Pouzita replikovana kapacita,
        printf CAPOUT ( "%.3f,%.3f,%.3f,", ConvSizeUnits( $unit, $fmt_system_used_capacity ), ConvSizeUnits( $unit, $fmt_snapshot_used_capacity ), ConvSizeUnits( $unit, $fmt_replicated_used_capacity ) );

        # Uspora thin-provisioning,Uspora deduplikace,Uspora komprese,Celkova efektivita
        printf CAPOUT ( "%.3f,%.3f,%.3f,%.3f,", 0, 0, 0, 0 );
        printf CAPOUT ("\n");
      }
    }
  }
  print CAPOUT "\n";
}

sub VolumeCopyList {
  my $data = shift;
  my @a;
  push( @a, 0 ) if ( defined $data->{copy_0_mdisk_grp_id} );
  push( @a, 1 ) if ( defined $data->{copy_1_mdisk_grp_id} );
  return ( \@a );
}

sub VolumeTierList {
  my $data         = shift;
  my $copy_id      = shift;
  my $pool_rec_key = $data->{ "copy_" . $copy_id . "_mdisk_grp_name" };
  my $pool_data    = $cfg_ref->{'svc_pool'}->{$pool_rec_key};

  my @a;
  push( @a, $pool_data->{'tier_scm'} ) if ( defined $pool_data->{'tier_scm_mdisk_count'} && $pool_data->{'tier_scm_mdisk_count'} > 0 );
  push( @a, $pool_data->{'tier_0'} )   if ( defined $pool_data->{'tier_0_mdisk_count'}   && $pool_data->{'tier_0_mdisk_count'} > 0 );
  push( @a, $pool_data->{'tier_0n'} )  if ( defined $pool_data->{'tier_0n_mdisk_count'}  && $pool_data->{'tier_0n_mdisk_count'} > 0 );
  push( @a, $pool_data->{'tier_1'} )   if ( defined $pool_data->{'tier_1_mdisk_count'}   && $pool_data->{'tier_1_mdisk_count'} > 0 );
  push( @a, $pool_data->{'tier_2'} )   if ( defined $pool_data->{'tier_2_mdisk_count'}   && $pool_data->{'tier_2_mdisk_count'} > 0 );
  push( @a, "-" );
  return ( \@a );
}

sub VolumeRaidType {
  my $data       = shift;
  my $copy       = shift;
  my $tier_level = shift;
  if ( $tier_level eq "-" ) {
    return ("");
  }
  else {
    if ( defined $data->{ "copy_" . $copy . "_tier_scm" } && $data->{ "copy_" . $copy . "_tier_scm" } eq $tier_level ) { return (""); }
    if ( defined $data->{ "copy_" . $copy . "_tier_0" }   && $data->{ "copy_" . $copy . "_tier_0" } eq $tier_level )   { return (""); }
    if ( defined $data->{ "copy_" . $copy . "_tier_0n" }  && $data->{ "copy_" . $copy . "_tier_0n" } eq $tier_level )  { return (""); }
    if ( defined $data->{ "copy_" . $copy . "_tier_1" }   && $data->{ "copy_" . $copy . "_tier_1" } eq $tier_level )   { return (""); }
    if ( defined $data->{ "copy_" . $copy . "_tier_2" }   && $data->{ "copy_" . $copy . "_tier_2" } eq $tier_level )   { return (""); }
  }
}

sub VolumeFmtRaidCapacity {
  my $data       = shift;
  my $copy       = shift;
  my $tier_level = shift;
  if ( $tier_level eq "-" ) {
    return (0);
  }
  else {
    if ( defined $data->{ "copy_" . $copy . "_tier_scm" } && $data->{ "copy_" . $copy . "_tier_scm" } eq $tier_level ) { return (0); }
    if ( defined $data->{ "copy_" . $copy . "_tier_0" }   && $data->{ "copy_" . $copy . "_tier_0" } eq $tier_level )   { return (0); }
    if ( defined $data->{ "copy_" . $copy . "_tier_0n" }  && $data->{ "copy_" . $copy . "_tier_0n" } eq $tier_level )  { return (0); }
    if ( defined $data->{ "copy_" . $copy . "_tier_1" }   && $data->{ "copy_" . $copy . "_tier_1" } eq $tier_level )   { return (0); }
    if ( defined $data->{ "copy_" . $copy . "_tier_2" }   && $data->{ "copy_" . $copy . "_tier_2" } eq $tier_level )   { return (0); }
  }
}

sub VolumeFmtDataUsedCapacity {

  # Return volume used capacity
  my $data          = shift;
  my $copy          = shift;
  my $tier_level    = shift;
  my @tier_key_list = ( "scm", "0", "0n", "1", "2" );

  if ( $tier_level eq "-" ) {
    if ( defined $data->{'deduplicated_copy'} && $data->{'deduplicated_copy'} eq "yes" ) {
      return ( $data->{"used_capacity_before_reduction"} );
    }
    elsif ( defined $data->{ "copy_" . $copy . "_real_capacity" } && $data->{ "copy_" . $copy . "_real_capacity" } ne "" ) {
      return ( $data->{ "copy_" . $copy . "_real_capacity" } );
    }
    else {
      return (0);
    }
  }
  else {
    foreach my $tier (@tier_key_list) {
      if ( defined $data->{ "copy_" . $copy . "_tier_" . $tier } && $data->{ "copy_" . $copy . "_tier_" . $tier } eq $tier_level ) {
        return ( $data->{ "copy_" . $copy . "_tier_" . $tier . "_capacity" } );
      }
    }
    return (0);
  }
}

sub VolumeFmtSnapshotUsedCapacity {

  # Return volume used capacity if volume is FC Target Volume
  my $data          = shift;
  my $copy          = shift;
  my $tier_level    = shift;
  my @tier_key_list = ( "scm", "0", "0n", "1", "2" );
  if ( defined $data->{'FC_id'} && $data->{'FC_id'} ) {
    my $fctarget_id;
    if ( $data->{'FC_id'} eq "many" ) {
      $fctarget_id = &FcTagetVolumeByVolumeId( $data->{id} );
      if ( defined $fctarget_id ) {

        # FC Target Volume
        $fctarget_id = sprintf( "%04d", $fctarget_id );
      }
      else {
        # FC Source Volume
        return (0);
      }
    }
    else {
      my $fcmap_id = sprintf( "%04d", $data->{'FC_id'} );
      $fctarget_id = $cfg_ref->{svc_fcmap}->{$fcmap_id}->{'target_vdisk_id'};
    }
    if ( defined $fctarget_id && $data->{id} == $fctarget_id ) {

      # Volume is FC Target Volume -> return real capacity
      if ( $tier_level eq "-" ) {
        if ( defined $data->{'deduplicated_copy'} && $data->{'deduplicated_copy'} eq "yes" ) {
          return ( $data->{"used_capacity_before_reduction"} );
        }
        elsif ( defined $data->{ "copy_" . $copy . "_real_capacity" } && $data->{ "copy_" . $copy . "_real_capacity" } ne "" ) {
          return ( $data->{ "copy_" . $copy . "_real_capacity" } );
        }
        else {
          return (0);
        }
      }
      else {
        foreach my $tier (@tier_key_list) {
          return ( $data->{ "copy_" . $copy . "_tier_" . $tier . "_capacity" } ) if ( defined $data->{ "copy_" . $copy . "_tier_" . $tier } && $data->{ "copy_" . $copy . "_tier_" . $tier } eq $tier_level );
        }
        return (0);
      }
    }
  }
  return (0);
}

sub FcTagetVolumeByVolumeId {
  my $id      = shift;
  my @reclist = keys( %{ $cfg_ref->{'svc_fcmap'} } );
  foreach my $rec ( sort(@reclist) ) {
    if ( defined $cfg_ref->{'svc_fcmap'}->{$rec}->{'target_vdisk_id'} && $id == $cfg_ref->{'svc_fcmap'}->{$rec}->{'target_vdisk_id'} ) {
      return ( $cfg_ref->{'svc_fcmap'}->{$rec}->{'target_vdisk_id'} );
    }
  }
  return (undef);
}

sub VolumeFmtReplicatedUsedCapacity {

  # Return volume real capacity if volume is RC Volume
  my $data          = shift;
  my $copy          = shift;
  my $tier_level    = shift;
  my @tier_key_list = ( "scm", "0", "0n", "1", "2" );
  if ( defined $data->{'RC_id'} && $data->{'RC_id'} ne "" ) {
    my $rcrel_id = sprintf( "%04d", $data->{'RC_id'} );

    # RC: copy_type state
    if ( defined $cfg_ref->{svc_rcrel}->{$rcrel_id}->{'copy_type'} ) {

      # Volume is RC Volume -> return real capacity
      if ( $data->{'function'} eq "master" || $data->{'function'} eq "aux" ) {

        # Ignore master_change or aux_change volumes
        if ( $tier_level eq "-" ) {
          if ( defined $data->{'deduplicated_copy'} && $data->{'deduplicated_copy'} eq "yes" ) {
            return ( $data->{"used_capacity_before_reduction"} );
          }
          elsif ( defined $data->{ "copy_" . $copy . "_real_capacity" } && $data->{ "copy_" . $copy . "_real_capacity" } ne "" ) {
            return ( $data->{ "copy_" . $copy . "_real_capacity" } );
          }
          else {
            return (0);
          }
        }
        else {
          foreach my $tier (@tier_key_list) {
            return ( $data->{ "copy_" . $copy . "_tier_" . $tier . "_capacity" } ) if ( defined $data->{ "copy_" . $copy . "_tier_" . $tier } && $data->{ "copy_" . $copy . "_tier_" . $tier } eq $tier_level );
          }
          return (0);
        }
      }
    }
  }
  return (0);
}

sub CapStoreData {
  #
  nstore( $capdata_ref, $tmp_cap_file ) or die("svcconfig.pl: Store data to file $tmp_cap_file failed.");
}

# -------------------------------------------------------------------------------------------------
